// chunkff.cpp

#include <string.h>
#include "chunkff.h"
#include <stdio.h>
#include <vector>
#include "file.h"

// globals
static char tb[256];

// Chunk class implementation
void Chunk::Resize(unsigned long size)
{
	if (size == this->size) return;
	
	void* buf = (void*) new char[size];

	// copy current contents
	if (data)
	{
		unsigned long bytes = size < this->size ? size : this->size;
		memcpy(buf, data, bytes);
		delete [] (char*)data;
	}

	data = buf;
	this->size = size;
}

void Chunk::Expand(unsigned long bytes)
{
	Resize(size + bytes);
}

void Chunk::Shrink(unsigned long bytes)
{
	if (bytes >= size)
	{
		Resize(0);
	}
	else
	{
		Resize(size - bytes);
	}
}

Chunk::Chunk()
{
	id = 0;
	size = 0;
	data = 0;
	parent = 0;
	children = ChunkList();
}

Chunk::Chunk(VFS_FILE* fp)
{
	id = 0;
	size = 0;
	data = 0;
	parent = 0;
	children = ChunkList();

	Load(fp);
}

Chunk::~Chunk()
{
	Destroy();
}

unsigned int Chunk::GetID() const
{
	return id;
}

void Chunk::SetID(unsigned long id)
{
	this->id = id;
}

void Chunk::Destroy()
{
	// free chunk's data
	if (data)
	{
		delete [] (char*)data;
		data = 0;
	}

	// delete all children
	for (unsigned long i=0; i<children.size(); i++)
	{
		if (children[i])
		{
			delete children[i];
			children[i] = 0;
		}
	}

	parent = 0;
	children = ChunkList();
}

#define CHUNK_FAIL {Destroy();return false;}

// this destroys the chunk !!!
bool Chunk::Load(VFS_FILE* fp)
{
	Destroy();

	if (!fp) CHUNK_FAIL;

	// read id
	if (!Read<unsigned long>(&id, 1, fp)) CHUNK_FAIL;
	
	// read size
	if (!Read<unsigned long>(&size, 1, fp)) CHUNK_FAIL;

	// allocate memory for this chunk
	data = (void*) new char[size];

	// read data
	if (!Read<char>((char*)data, size, fp)) CHUNK_FAIL;

#ifndef LITTLE_ENDIAN
	FlipEndian();
#endif

	// read number of children
	unsigned long num_children = 0;
	if (!Read<unsigned long>(&num_children, 1, fp)) CHUNK_FAIL;

	for (unsigned long i=0; i<num_children; i++)
	{
		Mount(new Chunk(fp));
	}

	// SUCCESS
	return true;
}

bool Chunk::LoadFile(const char *fname) {
	VFS_FILE *fp = vfs_open(fname, "rb");
	if(!fp) return false;

	if(!Load(fp)) {
		vfs_close(fp);
		return false;
	}

	vfs_close(fp);
	return true;
}

bool Chunk::Save(VFS_FILE* fp) const
{
	if (!fp) return false;

	// write id
	unsigned long ID = GetID();
	Write<unsigned long>(&ID, 1, fp);
	
	// write size
	ID = size; // compiler error ?????????
	Write<unsigned long>(&ID, 1, fp);

	// write data
#ifndef LITTLE_ENDIAN
	FlipEndian();
#endif
	Write<char>((char*)data, size, fp);
#ifndef LITTLE_ENDIAN
	FlipEndian();
#endif
	// write children count
	unsigned long num_children = (unsigned long) children.size();
	Write<unsigned long>(&num_children, 1, fp);
	for (unsigned long i=0; i<num_children; i++)
	{
		children[i]->Save(fp);
	}

	return true;
}

bool Chunk::SaveFile(const char *fname) const {
	VFS_FILE *fp = vfs_open(fname, "wb");
	if(!fp) return false;

	if(!Save(fp)) {
		vfs_close(fp);
		return false;
	}

	vfs_close(fp);
	return true;
}

// flips the endianness of the datablock
void Chunk::FlipEndian()
{
	unsigned int *p = (unsigned int*)GetData();
	for (unsigned int i=0; i<GetSize() / 4; i++)
	{
		::FlipEndian<unsigned int>(p + i);
	}
}

unsigned long Chunk::GetSize() const
{
	return size;
}

void* Chunk::GetData() const
{
	return data;
}

// this performs a copy
void Chunk::SetData(unsigned long size, void* data)
{
	if (this->size != size)
	{
		if (this->data)
		{
			delete [] (char*)this->data;
			this->data = 0;
		}
	}

	this->size = size;
	this->data = (void*) new char[size];

	memcpy(this->data, data, size);
}

Chunk* Chunk::GetParent() const
{
	return parent;
}

ChunkList Chunk::GetChildren() const
{
	return children;
}

ChunkList Chunk::GetIDChildren(unsigned long id) const
{
	ChunkList ret;
	for (unsigned long i=0; i<children.size(); i++)
	{
		if (children[i]->GetID() == id)
		{
			ret.push_back(children[i]);
		}
	}

	return ret;
}

// returns the first child with specified ID (or zero)
Chunk* Chunk::GetIDChild(unsigned long id) const
{
	for (unsigned long i=0; i<children.size(); i++)
	{
		if (children[i]->GetID() == id)
		{
			return children[i];
		}
	}

	return 0;
}

// returns all the subchunks with the id specified recursively
ChunkList Chunk::GetIDChildrenRec(unsigned long id) const
{
	ChunkList ret;
	ret = GetIDChildren(id);

	for (size_t i=0; i<children.size(); i++)
	{
		ChunkList l = children[i]->GetIDChildrenRec(id);
		for (size_t j=0; j<l.size(); j++)
		{
			ret.push_back(l[j]);
		}
	}
	return ret;
}

// returns false if the chunk already existed
bool Chunk::Mount(Chunk* child)
{
	// avoid a duplicate mount
	for (unsigned long i=0; i<children.size(); i++)
	{
		if (children[i] == child) return false;
	}

	// For God's shake no!
	if (this == child) return false;

	child->parent = this;
	children.push_back(child);
	return true;
}

bool Chunk::Unmount(Chunk* child)
{
	for (unsigned long i=0; i<children.size(); i++)
	{
		if (children[i] == child)
		{
			children.erase(children.begin() + i, children.begin() + i + 1);
			return true;
		}
	}

	return false;
}

bool Chunk::Delete(Chunk *child)
{
	bool success = Unmount(child);
	if (success)
	{
		delete child;
	}
	return success;
}

// info functions
std::string Chunk::GetInfo()
{
	std::string ret;
	sprintf(tb, "Generic chunk description\n=========================\n");
	ret += tb;
	sprintf(tb, "Chunk ID is : 0x%x\n", GetID());
	ret += tb;
	sprintf(tb, "Chunk size is %d bytes\n", (int)GetSize());
	ret += tb;
	sprintf(tb, "Number of children: %d\n", (int)GetChildren().size());
	ret += tb;
	return ret;
}

std::string Chunk::GetDetails() const
{
	std::string ret;
	ret += "No specific information is available for this chunk\n";
	return ret;
}

// stringChunk class implementation
stringChunk::stringChunk()
{

}
stringChunk::stringChunk(std::string str)
{
	Setstring(str);
}
stringChunk::~stringChunk()
{

}

std::string stringChunk::Getstring() const
{
	std::string ret;
	char* str_buf = (char*) GetData();
	for (unsigned long i=0; i<GetSize(); i++)
	{
		ret += str_buf[i];
	}

	return ret;
}
void stringChunk::Setstring(const std::string& str)
{
	Resize((unsigned long)str.size());

	char* str_buf = (char*) GetData();
	for (unsigned long i=0; i<str.size(); i++)
	{
		str_buf[i] = str[i];
	}
}

// info functions
std::string stringChunk::GetDetails() const
{
	std::string ret;
	ret += "String Chunk\n";
	ret += "Contents:\n";
	ret += Getstring();
	ret += "\n";
	return ret;
}

// NamedChunk class implementation
std::string NameChunk::GetDetails() const
{
	std::string ret;
	ret += "NameChunk\n";
	ret += "Contents:\n";
	ret += this->Getstring();
	return ret;
}

// NamedChunk class implementation
std::string NamedChunk::GetName() const
{
	// search for the first stringChunk
	ChunkList strs = GetIDChildren(NAME_CHUNK);
	if (!strs.size()) return "";
	return ((NameChunk*)strs[0])->Getstring();
}
void NamedChunk::SetName(const std::string& str)
{
	ChunkList strs = GetIDChildren(NAME_CHUNK);

	// remove and delete all the "name" subchunks
	for (unsigned long i=0; i<strs.size(); i++)
	{
		Unmount(strs[i]);
		delete strs[i];
	}

	NameChunk *name_chunk = new NameChunk();
	name_chunk->Setstring(str);
	Mount(name_chunk);
}

Chunk* NamedChunk::GetChildByName(const std::string& name) const
{
	ChunkList ch = GetChildren();
	for (unsigned long i=0; i<ch.size(); i++)
	{
		if (dynamic_cast<NamedChunk*>(ch[i]))
		{
			if (((NamedChunk*)ch[i])->GetName() == name)
			{
				return ch[i];
			}
		}
	}

	return 0;
}
Chunk* NamedChunk::GetChildByNameRecursive(const std::string& name) const
{
	ChunkList ch = GetChildren();
	for (unsigned long i=0; i<ch.size(); i++)
	{
		if (dynamic_cast<NamedChunk*>(ch[i]))
		{
			if (((NamedChunk*)ch[i])->GetName() == name)
			{
				return ch[i];
			}

			Chunk* tmp = ((NamedChunk*)ch[i])->GetChildByNameRecursive(name);
			if (tmp)
			{
				return tmp;
			}
		}
	}

	return 0;
}

// info functions
std::string NamedChunk::GetDetails()  const
{
	std::string ret;
	ret += "Named Chunk\n";
	ret += "Contents:\n";
	ret += GetName();
	return ret;
}














