// sce.h
// scene file format exporter

#ifndef _SCE_HEADER_
#define _SCE_HEADER_

#include "chunkff.h"
#include "chunk_types.h"
#include <math3d.h>
#include "prs.h"

enum LightDecay
{
	LD_NONE,
	LD_INVERSE,
	LD_SQUARE,

	LD_FORCE_32BIT = 0xFFFFFFFF
};

enum SpotLightType
{
	SLT_RECT,
	SLT_CIRCLE,

	SLT_FORCE_32BIT = 0xFFFFFFFF
};

enum LightChunkType
{
	LT_DIRECTIONAL,
	LT_OMNI,
	LT_SPOT,

	LT_FORCE_32BIT = 0x7fffffff
};

// keep this in sync with "bt_size" table in sce.cpp
enum SceBasicType
{
	SCE_BT_INT8,
	SCE_BT_INT16,
	SCE_BT_INT32,
	SCE_BT_FLOAT,
	SCE_BT_DOUBLE,

	SCE_FORCE_32BIT = 0x7fffffff
};

enum GeometryArrayType
{
	POSITION_ARRAY = 0,
	NORMAL_ARRAY,
	TANGENT_ARRAY,
	BINORMAL_ARRAY,
	COLOR_ARRAY,
	INDEX_ARRAY,
	TEXCOORD0_ARRAY,
	TEXCOORD1_ARRAY,
	TEXCOORD2_ARRAY,
	TEXCOORD3_ARRAY,
	BONE_WEIGHT_ARRAY,
	BONE_INDEX_ARRAY,
	BONE_MAP_ARRAY,
	SMOOTHING_GROUP_ARRAY,
	TRIANGLE_PROPERTIES_ARRAY0,
	TRIANGLE_PROPERTIES_ARRAY1,

	MAX_GEOMETRY_ARRAYS,

	GEOM_ARRAY_TYPE_FORCE_32BIT = 0x7fffffff
};

enum SceTextureType
{
	STT_DIFFUSE,
	STT_LIGHT,
	STT_OPACITY,
	STT_NORMAL,

	STT_UNSUPPORTED,

	STT_FORCE_32BIT = 0x7fffffff
};

enum SceneChunk
{
	SCE_MAIN = 0x30454353, // ascii "SCE0" (little endian)
	SCE_NODE,
	SCE_MESH,
	SCE_GEOMETRY_ARRAY,
	SCE_MATERIAL,
	SCE_LIGHT,
	SCE_CAMERA,
	SCE_TEXTURE
};

class CameraChunk : public Chunk
{
private:
	struct CameraData
	{
		BOOL32 target;
		float fov;
		float near_clip;
		float far_clip;
		unsigned int target_node_id;
	};

public:
	CameraChunk();
	~CameraChunk();

	unsigned int GetID() const
	{
		return SCE_CAMERA;
	}

	bool IsTarget() const;
	void SetTarget(bool target);

	float GetFOV() const;
	void SetFOV(float fov);

	float GetNearClip() const;
	void SetNearClip(float near_clip);

	float GetFarClip() const;
	void SetFarClip(float far_clip);

	unsigned long GetTargetNodeID() const;
	void SetTargetNodeID(unsigned long target_node_id);
}; // end class CameraChunk

class LightChunk : public Chunk
{
private:
	struct LightData
	{
		LightChunkType type;
		Vector4 color;
		float multiplier;
		float aspect_ratio;
		float falloff;
		float hot_spot;
		unsigned int target_node_id;
		LightDecay decay;
		SpotLightType spot_light_type;
		float atten_start, atten_end;
	};

public:
	LightChunk();
	~LightChunk();

	unsigned int GetID() const
	{
		return SCE_LIGHT;
	}

	LightChunkType GetType() const;
	void SetType(LightChunkType type);

	Vector4 GetColor() const;
	void SetColor(const Vector4& color);

	float GetMultiplier() const;
	void SetMultiplier(float multiplier);

	float GetAspectRatio() const;
	void SetAspectRatio(float aspect_ratio);

	float GetFallOff() const;
	void SetFallOff(float falloff);

	float GetHotSpot() const;
	void SetHotSpot(float hot_spot);

	unsigned long GetTargetNodeID() const;
	void SetTargetNodeID(unsigned long target_node_id);

	LightDecay GetLightDecay() const;
	void SetLightDecay(LightDecay decay);

	SpotLightType GetSpotLightType() const;
	void SetSpotLightType(SpotLightType slt);
	
	float GetAttenStart() const;
	void SetAttenStart(float start);

	float GetAttenEnd() const;
	void SetAttenEnd(float end);
}; // end class LightChunk

class TextureChunk : public NamedChunk
{
private:
	struct TextureData
	{
		SceTextureType type;
	};
public:
	TextureChunk();
	~TextureChunk();

	unsigned int GetID() const
	{
		return SCE_TEXTURE;
	}

	SceTextureType GetType() const;
	void SetType(SceTextureType type);
}; // end class TextureChunk

class MaterialChunk : public NamedChunk
{
private:
	struct MaterialData
	{
		Vector4 ambient;
		Vector4 diffuse;
		Vector4 emissive;
		Vector4 specular;
	};
public:
	MaterialChunk();
	~MaterialChunk();

	unsigned int GetID() const
	{
		return SCE_MATERIAL;
	}

	Vector4 GetAmbient() const;
	void SetAmbient(const Vector4& ambient);

	Vector4 GetDiffuse() const;
	void SetDiffuse(const Vector4& diffuse);

	Vector4 GetEmissive() const;
	void SetEmissive(const Vector4& emissive);

	Vector4 GetSpecular() const;
	void SetSpecular(const Vector4& specular);

	// high level
	TextureChunk* GetTexture(SceTextureType type);
}; // end class MaterialChunk

class GeometryArray : public Chunk
{
private:
	struct GeometryArrayData
	{
		GeometryArrayType type;
		SceBasicType basic_type;
		unsigned int basic_type_count;
		unsigned int element_count;
	};

public:
	GeometryArray(GeometryArrayType type, SceBasicType basic_type,
		unsigned long basic_type_count, unsigned long element_count);
	~GeometryArray();

	unsigned int GetID() const
	{
		return SCE_GEOMETRY_ARRAY;
	}

	GeometryArrayType GetType() const;
	SceBasicType GetBasicType() const;
	unsigned long GetBasicTypeCount() const;
	unsigned long GetElementCount() const;

	void* GetArrayData();
}; // end class Geometry Array 

class MeshChunk : public NamedChunk
{
private:
	struct MeshData
	{
		BOOL32 proxy, pvs, mesh, link;
		unsigned int material_id;
	};

public:
	MeshChunk();
	~MeshChunk();

	unsigned int GetID() const
	{
		return SCE_MESH;
	}

	bool IsProxy() const;
	void SetProxy(bool proxy);

	bool IsPVS() const;
	void SetPVS(bool pvs);

	bool IsMesh() const;
	void SetMesh(bool mesh);

	bool IsLink() const;
	void SetLink(bool link) const;

	bool IsIndexed() const;
	unsigned long GetMaterialID() const;
	void SetMaterialID(unsigned long material_id);

	GeometryArray* GetArray(GeometryArrayType type) const;

	bool CalculateNormals();
	bool CalculateTangents();

	// high level
	unsigned long GetTriangleCount() const;
	void GetTriangle(unsigned int idx, unsigned int &a,
		unsigned int &b, unsigned int &c) const;
	void Transform(const Matrix4x4 &mat);
}; // end class MeshChunk

typedef std::vector<MeshChunk*> MeshList;

class NodeChunk : public NamedChunk
{
private:
	struct NodeData
	{
		unsigned int id;
		BOOL32 dynamic;
		PRS world, local, object, pivot;
	};

public:
	NodeChunk();
	~NodeChunk();

	unsigned int GetID() const
	{
		return SCE_NODE;
	}

	Matrix4x4 GetTransform() const;
	
	PRS GetWorldPRS() const;
	void SetWorldPRS(const PRS& prs);
	
	PRS GetLocalPRS() const;
	void SetLocalPRS(const PRS &prs);

	PRS GetObjectPRS() const;
	void SetObjectPRS(const PRS &prs);

	PRS GetPivotPRS() const;
	void SetPivotPRS(const PRS &prs);

	bool GetDynamic() const;
	void SetDynamic(bool dynamic);

	unsigned long GetNodeID() const;
	void SetNodeID(unsigned long node_id);

	// high-level
	unsigned long GetSubNodeCount() const;
	NodeChunk* GetSubNode(unsigned long id) const;
	unsigned long GetMeshCount() const;
	MeshChunk* GetMesh(unsigned long id) const;
	unsigned long GetProxyCount() const;
	MeshChunk* GetProxy(unsigned long id) const;
	unsigned long GetPVSCount() const;
	MeshChunk* GetPVS(unsigned long id) const;
	unsigned long GetLinkCount() const;
	MeshChunk* GetLink(unsigned long id) const;
	unsigned long GetLightCount() const;
	LightChunk* GetLight(unsigned long i) const;
	unsigned long GetCameraCount() const;
	CameraChunk* GetCamera(unsigned long i) const;

	void BatchMeshes();

	NodeChunk* GetNodeByIDRecursive(unsigned long id) const;
	NodeChunk* GetNodeByIDTree(unsigned long id) const;
}; // end class NodeChunk

class SceneFile : public NamedChunk
{
private:
	struct SceneHeader
	{
		unsigned int version;
	};
public:
	SceneFile();
	~SceneFile();

	unsigned int GetID() const
	{
		return SCE_MAIN;
	}

	unsigned long GetVersion() const;
	void SetVersion(unsigned long version);

	std::string GetMessage() const;
	void SetMessage(const std::string& message);

	// high-level
	unsigned long GetMaterialCount() const;
	MaterialChunk* GetMaterial(unsigned long id) const;
	unsigned long GetRootNodeCount() const;
	NodeChunk* GetRootNode(unsigned long id) const;
	NodeChunk* GetNodeByIDRecursive(unsigned long id) const;
	MeshList GetPVSMeshes() const;
	MeshList GetLinkMeshes() const;
}; // end class SceneFile

// helpers
MeshChunk *ConcatenateMeshChunks(MeshList &chunks);


#endif // ndef _SCE_HEADER_
