// collision.cpp
#include "collision.h"
#include <errlog.h>

// CollisionObject class implementation
// -------------------------------------
CollisionObject::CollisionObject()
{
	transform = Matrix4x4();
	last_collision_obj = 0;
	my_type = CONTAINER;
}

CollisionObject::~CollisionObject()
{
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		printf("colobj: %d\n", collision_objects.size());
		if (collision_objects[i]) delete collision_objects[i];
		collision_objects[i] = 0;
	}
	collision_objects.clear();
}


void CollisionObject::AddCollisionObject(CollisionObject* obj)
{
	if (!obj)
	{
		Error("CollisionObject::AddObject failed: null pointer\n");
		return;
	}
	collision_objects.push_back(obj);

	// update bounding volume
	std::vector<Vector3> bb_verts;
	for (unsigned int i=0; i<8; i++)
	{
		bb_verts.push_back(bounding_box.GetVertex(i));
		bb_verts.push_back(obj->bounding_box.GetVertex(i));
	}
	
	// TODO: this box can be oriented
	bounding_box = Box(bb_verts);
}


void CollisionObject::SetTransform(const Matrix4x4& mat)
{
	transform = mat;
}

Matrix4x4 CollisionObject::GetTransform() const
{
	return transform;
}


bool CollisionObject::Collision(const CollisionMesh& obj)
{
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		if (collision_objects[i]->Collision(obj))
		{
			last_collision_obj = collision_objects[i];
			return true;
		}
	}
	return false;
}

bool CollisionObject::Collision(const Ray& ray, bool closest, float maxl)
{
	float closest_dist = 100000.0f;
	bool ret = false;
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		/*
		// bounding box chack
		Box bb = collision_objects[i]->bounding_box;
		bb.Transform(collision_objects[i]->transform);

		if (!Intersect(ray, bb)) continue;
		*/
		if (collision_objects[i]->Collision(ray, closest))
		{
			ret = true;
			if (!closest)
			{
				// return first collision
				last_collision_obj = collision_objects[i];
				return true;
			}
			else
			{
				// get the distance of this result
				Vector3 point = collision_objects[i]->GetCollisionPoint(true);
				point -= ray.GetOrigin();
				float dist = DotProduct(point, ray.GetDirection());
				if (dist < closest_dist)
				{
					last_collision_obj = collision_objects[i];
					closest_dist = dist;
				}
			}
		}
	}
	return ret;
}

bool CollisionObject::Collision(const Line& line, bool closest)
{
	float closest_dist = 100000.0f;
	bool ret = false;
	Ray ray = line.GetRay();
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		if (collision_objects[i]->Collision(line, closest))
		{
			ret = true;
			if (!closest)
			{
				// return first collision
				last_collision_obj = collision_objects[i];
				return true;
			}
			else
			{
				// get the distance of this result
				Vector3 point = collision_objects[i]->GetCollisionPoint(true);
				point -= ray.GetOrigin();
				float dist = DotProduct(point, ray.GetDirection());
				if (dist < closest_dist)
				{
					last_collision_obj = collision_objects[i];
					closest_dist = dist;
				}
			}
		}
	}
	return ret;
}

bool CollisionObject::Collision(const Sphere& sph)
{
	bool ret = false;
	collision_tris.erase(collision_tris.begin(), collision_tris.end());
	col_props.erase(col_props.begin(), col_props.end());
	col_meshes.erase(col_meshes.begin(), col_meshes.end());
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		if (collision_objects[i]->Collision(sph))
		{
			ret = true;
			last_collision_obj = collision_objects[i];
			for (unsigned int j=0; j<last_collision_obj->GetCollisionTriangleCount(); j++)
			{
				collision_tris.push_back(
					last_collision_obj->GetCollisionTriangleList(j, true));
				col_props.push_back(last_collision_obj->GetCollisionTrianglePropertiesList(j));
				col_meshes.push_back(last_collision_obj->GetCollisionMeshList(j));
			}
		}
	}
	return ret;
}


// these are valid only after a collision
CollisionObject *CollisionObject::GetCollisionObject()
{
	return last_collision_obj;
}

CollisionMesh *CollisionObject::GetCollisionMesh()
{
	if (my_type == MESH)
	{
		return (CollisionMesh*) this;
	}
	if (!last_collision_obj) return 0;

	/*
	if (dynamic_cast<CollisionMesh*>(last_collision_obj))
	{
		return (CollisionMesh*)last_collision_obj;
	}
	*/

	return last_collision_obj->GetCollisionMesh();
}

Triangle CollisionObject::GetCollisionTriangle(CollisionTriangle which, bool world)
{
	if (!last_collision_obj) return Triangle();
	return GetCollisionMesh()->GetCollisionTriangle(which, world);
}

TriangleProperties CollisionObject::GetCollisionTriangleProperties()
{
	if (!last_collision_obj) return TriangleProperties();
	return GetCollisionMesh()->GetCollisionTriangleProperties();
}

unsigned int CollisionObject::GetCollisionTriangleIndex(CollisionTriangle which)
{
	if (!last_collision_obj) return 0;
	return GetCollisionMesh()->GetCollisionTriangleIndex(which);
}

unsigned int CollisionObject::GetCollisionTriangleCount()
{
	return (unsigned int)collision_tris.size();
}

Triangle CollisionObject::GetCollisionTriangleList(unsigned int id, bool world)
{
	if (id >= collision_tris.size())
	{
		return Triangle();
	}
	return collision_tris[id];
}

unsigned int CollisionObject::GetCollisionTriangleIndexList(unsigned int id)
{
	// TODO: implement this
	return 0;
}

TriangleProperties CollisionObject::GetCollisionTrianglePropertiesList(unsigned int id)
{
	return col_props[id];
}

CollisionMesh *CollisionObject::GetCollisionMeshList(unsigned int id)
{
	return col_meshes[id];
}

Vector3 CollisionObject::GetCollisionPoint(bool world)
{
	if (!last_collision_obj) return Vector3();
	return GetCollisionMesh()->GetCollisionPoint(world);
}

Vector3 CollisionObject::GetCollisionNormal(bool world)
{
	if (!last_collision_obj) return Vector3();
	return GetCollisionMesh()->GetCollisionNormal(world);
}

std::vector<CollisionMesh*> CollisionObject::GetMeshes() const
{
	std::vector<CollisionMesh*> ret;
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		std::vector<CollisionMesh*> v = collision_objects[i]->GetMeshes();
		for (unsigned int j=0; j<v.size(); j++)
		{
			ret.push_back(v[j]);
		}
	}
	return ret;
}

Box CollisionObject::GetBoundingBox()
{
	return bounding_box;
}

// CollisionMesh class implementation
// -----------------------------------
CollisionMesh::CollisionMesh()
{
#ifdef USE_COLDET
	obj = newCollisionModel3D();
#endif
	my_type = MESH;
}

CollisionMesh::~CollisionMesh()
{
#ifdef USE_COLDET
	delete obj;
#endif
}

void CollisionMesh::AddTriangle(const Triangle& tri)
{

	Vector3 v1 = tri.GetVertex(0);
	Vector3 v2 = tri.GetVertex(1);
	Vector3 v3 = tri.GetVertex(2);
#ifdef USE_COLDET
	obj->addTriangle((float*)&v1, (float*)&v2, (float*)&v3);
#else
	
#endif

	tris.push_back(tri);
	tri_normals.push_back(tri.GetNormal());
	TriangleProperties p;
	props.push_back(p);

	// update bounding volume
	std::vector<Vector3> bb_verts;
	bb_verts.push_back(v1);
	bb_verts.push_back(v2);
	bb_verts.push_back(v3);

	for (unsigned int i=0; i<8; i++)
	{
		bb_verts.push_back(bounding_box.GetVertex(i));
	}

	// TODO: this box can also be oriented
	bounding_box = Box(bb_verts);
}

Triangle CollisionMesh::GetTriangle(unsigned int i)
{
	if (i >= tris.size()) return Triangle();
	return tris[i];
}

unsigned int CollisionMesh::GetTriangleCount()
{
	return (unsigned int)tris.size();
}

void CollisionMesh::SetTriangleProperties(unsigned int id, const TriangleProperties &prop)
{
	props[id] = prop;
}

TriangleProperties CollisionMesh::GetTriangleProperties(unsigned int id) const
{
	return props[id];
}

void CollisionMesh::Finalize()
{
#ifdef USE_COLDET
	obj->finalize();
#else
	// do mesh precalculations here
#endif
}

void CollisionMesh::SetTransform(const Matrix4x4& mat)
{
	transform = mat;
#ifdef USE_COLDET
	Matrix4x4 transp = mat.Transposed();
	obj->setTransform((float*) &transp);
#endif
}

bool CollisionMesh::Collision(const CollisionMesh& obj)
{
#ifdef USE_COLDET
	return this->obj->collision(obj.obj);
#else
	Matrix4x4 rel_xform = obj.GetTransform() * transform.Inverse();
	for (unsigned int i=0; i<tris.size(); i++)
	{
		for (unsigned int j=0; j<obj.tris.size(); j++)
		{
			Triangle obj_tri = obj.tris[j];
			obj_tri.Transform(rel_xform);
			if (Intersect(tris[i], obj_tri))
			{
				that_xform = obj.GetTransform();
				this_tri = tris[i];
				that_tri = obj.tris[j];
				this_tri_idx = i;
				that_tri_idx = j;
				return true;
			}
		}
	}

	return false;
#endif
}

bool CollisionMesh::Collision(const Ray& ray, bool closest, float maxl)
{
	Vector3 orig = ray.GetOrigin();
	Vector3 dir = ray.GetDirection();
#ifdef USE_COLDET
	bool success = obj->rayCollision((float*)&orig, (float*)&dir, closest);
	if (success)
	{
//?		this_tri_idx = 
	}
	return success;
#else
	// transform ray into model space
	Matrix4x4 inv = transform.Inverse();
	orig.Transform(inv);
	dir.Transform((Matrix3x3) inv);
	Ray xray(orig, dir);
	float min_dist = 1000000.0f;
	bool ret = false;
	unsigned int idx = 0;
	for (unsigned int i=0; i<tris.size(); i++)
	{
		if (Intersect(xray, tris[i]))
		{
			Vector3 pt = PointOfIntersection(xray, tris[i]);
			float dist = ray.GetDist(pt.Transformed(transform));
			if (!closest && dist <= maxl)
			{
				isect_point = pt;
				this_tri = tris[i];
				this_tri_idx = i;
				return true;
			}
			else
			{
				if (dist >= 0 && dist <= maxl && dist < min_dist)
				{
					ret = true;
					isect_point = pt;
					this_tri = tris[i];
					this_tri_idx = i;
					min_dist = dist;
				}
			}
		}
	}
	return ret;
#endif
}

bool CollisionMesh::Collision(const Line& line, bool closest)
{
	Ray ray = line.GetRay();
	float maxl = line.GetLength();
#ifdef USE_COLDET
	Vector3 orig = ray.GetOrigin();
	Vector3 dir = ray.GetDirection();
	return obj->rayCollision((float*)&orig, (float*)&dir, closest, 0, maxl);
#else
	return Collision(ray, closest, maxl);
#endif
}

bool CollisionMesh::Collision(const Sphere& sph)
{
#ifdef USE_COLDET
	Vector3 center = sph.GetCenter();
	bool success = obj->sphereCollision((float*)&center, sph.GetRadius(), true);
	col_props.erase(col_props.begin(), col_props.end());
	for (unsigned int i=0; i<GetCollisionTriangleCount(); i++)
	{
		col_props.push_back(GetTriangleProperties(GetCollisionTriangleIndexList(i)));
	}
	return success;
#else
	col_tris.erase(col_tris.begin(), col_tris.end());
	col_tris_idx.erase(col_tris_idx.begin(), col_tris_idx.end());
	// transform the sphere to our model space
	Matrix4x4 inv = transform.Inverse();
	Vector3 center = sph.GetCenter().Transformed(inv);
	// for the radius use the average scaling in x,y and z
	float sx = ((Vector3)inv.GetColumnVector(0)).Length();
	float sy = ((Vector3)inv.GetColumnVector(0)).Length();
	float sz = ((Vector3)inv.GetColumnVector(0)).Length();
	float sc = (sx + sy + sz) / 3.0f;
	float rad = sph.GetRadius() * sc;
	Sphere xsphere(center, rad);
	bool ret = false;
	for (unsigned int i=0; i<tris.size(); i++)
	{
		if (Intersect(xsphere, tris[i]))
		{
			this_tri = tris[i];
			col_tris.push_back(this_tri);
			this_tri_idx = i;
			col_tris_idx.push_back(this_tri_idx);
			ret =  true;
		}
	}
	
	return ret;
#endif
}

// these are valid only after a collision
Triangle CollisionMesh::GetCollisionTriangle(CollisionTriangle which, bool world)
{
#ifdef USE_COLDET
	Triangle ret;
	if (which == CT_THIS)
	{
		obj->getCollidingTriangles((float*)&ret, 0, !world);
	}
	else
	{
		obj->getCollidingTriangles(0, (float*)&ret, !world);
	}

	return ret;
#else
	Triangle ret;
	if (which == CT_THIS)
	{
		ret = this_tri;
		if (world) ret.Transform(transform);
	}
	else
	{
		ret = that_tri;
		if (world) ret.Transform(that_xform);
	}
	return ret;
#endif
}

TriangleProperties CollisionMesh::GetCollisionTriangleProperties()
{
	return GetTriangleProperties(GetCollisionTriangleIndex(CT_THIS));
}

unsigned int CollisionMesh::GetCollisionTriangleIndex(CollisionTriangle which)
{
#ifdef USE_COLDET
	int ret, tmp;
	if (which == CT_THIS)
	{
		obj->getCollidingTriangles(ret, tmp);
	}
	else
	{
		obj->getCollidingTriangles(tmp, ret);
	}
	if (ret >= 0) return ret;
	return 0;
#else
	if (which == CT_THIS) return this_tri_idx;
	return that_tri_idx;
#endif
}

unsigned int CollisionMesh::GetCollisionTriangleCount()
{
#ifdef USE_COLDET
	return obj->getCollidingTriangleCount();
#else
	return (unsigned int) col_tris.size();
#endif
}

Triangle CollisionMesh::GetCollisionTriangleList(unsigned int id, bool world)
{
	Triangle ret;
#ifdef USE_COLDET
	obj->getCollidingTriangle(id, (float*)&ret, !world);
#else
	if (id < col_tris.size()) ret = col_tris[id];
	if (world) ret.Transform(transform);
#endif
	return ret;
}

unsigned int CollisionMesh::GetCollisionTriangleIndexList(unsigned int id)
{
#ifdef USE_COLDET
	return obj->getCollidingTriangleIndex(id);
#else
	return (unsigned int) col_tris_idx.size();
#endif
}

TriangleProperties CollisionMesh::GetCollisionTrianglePropertiesList(unsigned int id)
{
	return col_props[id];
}

CollisionMesh *CollisionMesh::GetCollisionMeshList(unsigned int id)
{
	return this;
}

Vector3 CollisionMesh::GetCollisionPoint(bool world)
{
#ifdef USE_COLDET
	Vector3 ret;
	obj->getCollisionPoint((float*)&ret, !world);
	return ret;
#else
	Vector3 ret = isect_point;
	if (world) ret.Transform(transform);
	return ret;
#endif
}
	
Vector3 CollisionMesh::GetTriangleNormal(unsigned int idx, bool world)
{
	Vector3 ret = tri_normals[idx];
	if (world) ret.Transform(GetTransform());
	return ret;
}

Vector3 CollisionMesh::GetCollisionNormal(bool world)
{
	Triangle tri = GetCollisionTriangle(CT_THIS, world);
	return tri.GetNormal();
}

std::vector<CollisionMesh*> CollisionMesh::GetMeshes() const
{
	std::vector<CollisionMesh*> ret;
	ret.push_back((CollisionMesh*)this);
	return ret;
}
