// collision.h
#define USE_COLDET

#ifdef USE_COLDET
#include <coldet.h>
#endif
#include <geom.h>

#ifndef _COLLISION_HEADER_
#define _COLLISION_HEADER_

enum CollisionTriangle
{
	CT_THIS,
	CT_THAT
};

struct TriangleProperties
{
	unsigned int prop0, prop1;
	TriangleProperties() {prop0 = prop1 = 0;};
};

enum CollisionObjectType
{
	CONTAINER,
	MESH
};

// fwd declaration
class CollisionMesh;

// This class represents a generic collision object. All collision objects 
// are derived from it. It also acts as a collision object container.
// TODO: optimize massive collision tests (octrees?)
class CollisionObject
{
protected:
	Matrix4x4 transform;
	Box bounding_box;
	std::vector<CollisionObject*> collision_objects;
	CollisionObject *last_collision_obj;

	std::vector<Triangle> collision_tris;
	std::vector<TriangleProperties> col_props;
	std::vector<CollisionMesh*> col_meshes;
	std::vector<Vector3> tri_normals;

	CollisionObjectType my_type;
public:
	CollisionObject();
	virtual ~CollisionObject();

	void AddCollisionObject(CollisionObject* obj);

	virtual void SetTransform(const Matrix4x4& mat);
	virtual Matrix4x4 GetTransform() const;

	virtual bool Collision(const CollisionMesh& obj);
	virtual bool Collision(const Ray& ray, bool closest = false, float maxl=3.4e+38F);	// <- XXX what is this value?
	virtual bool Collision(const Line& line, bool closest = false);
	virtual bool Collision(const Sphere& sph);

	// these are valid only after a collision
	CollisionObject *GetCollisionObject();
	CollisionMesh *GetCollisionMesh();
	virtual Triangle GetCollisionTriangle(CollisionTriangle which, bool world = false);
	virtual TriangleProperties GetCollisionTriangleProperties();
	virtual unsigned int GetCollisionTriangleIndex(CollisionTriangle which);
	
	virtual unsigned int GetCollisionTriangleCount();
	virtual Triangle GetCollisionTriangleList(unsigned int id, bool world);
	virtual unsigned int GetCollisionTriangleIndexList(unsigned int id);
	virtual TriangleProperties GetCollisionTrianglePropertiesList(unsigned int id);
	virtual CollisionMesh *GetCollisionMeshList(unsigned int id);

	virtual Vector3 GetCollisionPoint(bool world = false);
	virtual Vector3 GetCollisionNormal(bool world = false);

	virtual std::vector<CollisionMesh*> GetMeshes() const;


	Box GetBoundingBox();
};

class CollisionMesh : public CollisionObject
{
protected:
	std::vector<TriangleProperties> props;
	std::vector<Triangle> tris;
#ifdef USE_COLDET
	CollisionModel3D* obj;
#else

	std::vector<Triangle> col_tris;
	std::vector<unsigned int> col_tris_idx;
	Triangle this_tri, that_tri;
	unsigned int this_tri_idx, that_tri_idx;
	Matrix4x4 that_xform;
	Vector3 isect_point;
#endif
public:
	// constructor
	CollisionMesh();
	virtual ~CollisionMesh();

	void AddTriangle(const Triangle& tri);
	Triangle GetTriangle(unsigned int i);
	unsigned int GetTriangleCount();
	void SetTriangleProperties(unsigned int id, const TriangleProperties &prop);
	TriangleProperties GetTriangleProperties(unsigned int id) const;
	void Finalize();

	void SetTransform(const Matrix4x4& mat);

	bool Collision(const CollisionMesh& obj);
	bool Collision(const Ray& ray, bool closest = false, float maxl=3.4e+38F);
	bool Collision(const Line& line, bool closest = false);
	bool Collision(const Sphere& sph);

	// these are valid only afXter a collision
	Triangle GetCollisionTriangle(CollisionTriangle which, bool world = false);
	TriangleProperties GetCollisionTriangleProperties();
	unsigned int GetCollisionTriangleIndex(CollisionTriangle which);
	
	unsigned int GetCollisionTriangleCount();
	Triangle GetCollisionTriangleList(unsigned int id, bool world);
	unsigned int GetCollisionTriangleIndexList(unsigned int id);
	TriangleProperties GetCollisionTrianglePropertiesList(unsigned int id);
	CollisionMesh *GetCollisionMeshList(unsigned int id);

	Vector3 GetCollisionPoint(bool world = false);
	Vector3 GetCollisionNormal(bool world = false);
	
	Vector3 GetTriangleNormal(unsigned int idx, bool world = false);

	virtual std::vector<CollisionMesh*> GetMeshes() const;
};

#endif // ndef _COLLISION_HEADER_
