#ifndef __BENCH_H__
#define __BENCH_H__

#ifdef USE_OPENMP
#include <omp.h>
#include <cassert>
#include <string>
#include <map>

/*

This is a benchmark class which can support multiple threads.

Usage:
======

Mark sections of code you want to profile with BM_START() and BM_STOP() macros.
When you run the program, real-time statistics will appear on the monitoring
client.


BM_START( region_name );

// code

BM_STOP( region_name );


Features:
=========

OpenMP: For sections that lay inside parallel regions, statistics will be kept
   for each thread and presented accordingly. No action is required from the
   user.


Limitations:
- It is not allowed for two or more sections to have the same name.

- The implementation uses a threadprivate bm_thread_number variable which is
meant to be initialised only once, when the program begins (no action is
required from the user). OpenMP guarantees that local copies of threadprivate
data persists between consecutive parallel regions but sets the following
restrictions:
a. No nested parallel regions.
b. No change in the number of threads (i.e. don't call omp_set_num_threads,
   either inside or outside parallel regions).
c. No dynamic threads (i.e. don't call omp_set_dynamic(1)).

- OpenMP standard v2.5 is not very detailed on how static class objects (non
POD) are constructed when declared inside the scope of a parallel region, and
this results in different interpretations among compiler vendors.
Although it's also not absolutely clear to me what happens with POD types, it
seems that all the compilers I tried would behave in the same way, except for
Microsoft's VC8 which won't compile at all.
The OpenMP ARB aims to clarify these ambiguities in OpenMP v3.0.
This implementation uses a static POD bool (which is initialised only once) to
check whether the object has been previously initialised and if not, proceeds
with explicit initialisation. Unfortunately this involves a small overhead for
the branch, which should always be predicted nontheless.

*/

//This refers to whether the number of times a section will open-close will be
//counted. There is a very small overhead to that (increasing a counter) which
//is very unlikely to become an issue.
//Comment out the following line if you want to swich it off.
#define BM_COUNT

//Set this to anything bigger than the size of the CPU's cache line (in bytes).
//This will be used to avoid false sharing. For the time being, I will leave
//it to 512, as I am not aware of any CPU using a cache line bigger than that.
#define BM_CACHE_LINE 512


#ifdef NDEBUG
#define BM_DEBUG_ST(a)
#else
#define BM_DEBUG_ST(a) a
#endif

#ifdef BM_COUNT
#define BM_COUNT_ST(a)
#else
#define BM_COUNT_ST(a) a
#endif


enum BM_ERROR
{
	BM_ERR_01, // Attempted to close a serial benchmark section inside a parallel region.
	BM_ERR_02, // Attempted to close a parallel benchmark section inside a serial region.
	BM_ERR_03, // Section not found. This may have happen because either 
			   //  - the section was not opened, 
			   //  - the benchmark section was was opened in a parallel region with N
			   //    threads and attempted to be closed in another parallel region with M
			   //    threads, where M > N.
	BM_ERR_04  // A section with this name already exists.
};


class BMSection;
class Benchmark;
extern Benchmark bm;


extern int bm_thread_num;
#pragma omp threadprivate(bm_thread_num)
extern int bm_num_threads;
#pragma omp threadprivate(bm_num_threads)


struct BMSection
{
private:
	char padding[BM_CACHE_LINE - 1];
public:
	//Time counting variables
	double temp;
	double time;

	//The name of the section, assigned by user.
	std::string name;

	//File and line of BMSection instance.
	std::string file;
	int line;

	//Times a section has been opened.
	int count;

	//DEBUG ONLY!
	//If the section is started with BM_START, this is true.
	bool running;

	//OpenMP binding of the section.
	int threadNum;
	bool inParallel;
};



void bm_init();
void bm_reset();
void bm_section_init(BMSection **bms, std::string section_name, std::string file_name, int line);
void bm_section_reset(BMSection *bms);
void bm_report(int iter);


struct Benchmark
{
	std::map<std::string, BMSection *> smap;

	Benchmark();

	BMSection *findSection(std::string name);

	//For sections with multiple threads, multiple BMSection objects are
	//allocated from the heap contiguously. Since they all share the same name
	//only the pointer of the first object in the block is inserted in smap.
	void submit(BMSection *section);
	void checkout();
};


#define BM_START( section_name )\
	do {\
		static bool first_time = true;\
		static BMSection *bms;\
		if (first_time) {\
			bm_section_init(&bms, section_name, __FILE__, __LINE__);\
			first_time = false;\
		}\
		BM_DEBUG_ST(assert(bms[bm_thread_num].running == false););\
		BM_DEBUG_ST(bms[bm_thread_num].running = true);\
		bms[bm_thread_num].temp = omp_get_wtime();\
	} while(false)
	

#define BM_STOP( section_name )\
	do {\
		static bool first_time = true;\
		static BMSection *bms;\
		if (first_time) {\
			bms = bm.findSection(section_name);\
			first_time = false;\
		}\
		BM_DEBUG_ST(assert(bms[bm_thread_num].running == true););\
		BM_DEBUG_ST(bms[bm_thread_num].running = false);\
		BM_COUNT_ST(bms[bm_thread_num].count++;);\
		bms[bm_thread_num].time += omp_get_wtime() - bms[bm_thread_num].temp;\
	} while(false)

//#define BM_SPLIT( section_name )\
//	do {\
//		static bool first_time = true;\
//		static BMSection *bms;\
//		if (first_time) {\
//			bms = bm.findSection(section_name);\
//			bm_section_split(&bms, __FILE__, __LINE__);
//			first_time = false;\
//		}\
//		BM_DEBUG_ST(assert(bms[bm_thread_num].running == false););\
//		BM_DEBUG_ST(bms[bm_thread_num].running = true);\
//		bms[bm_thread_num].temp = omp_get_wtime();\
//	} while(false)

inline int bm_get_checkout_iter(int a) {return a-1;}
inline int bm_get_checkout_iter() {return 0;}

#define BM_CHECKOUT( iter )\
	do {\
		static int bmc = bm_get_checkout_iter(iter);\
		BM_DEBUG_ST(assert(omp_in_parallel() == false););\
		BM_DEBUG_ST(assert(bmc >= 0););\
		if (!bmc--) {\
			bm_report(bm_get_checkout_iter(iter)+1);\
			bm_reset();\
			bmc = bm_get_checkout_iter(iter);\
		}\
	} while(false)

#endif

#endif
