#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "errlog.h"

#define MAX_FNAME_SZ	512
#define MSG_BUFFER_SZ	2048

static char msg_buf[MSG_BUFFER_SZ];

struct LogOut {
	enum {OUT_FILE, OUT_STREAM, OUT_CALLBACK} type;
	union {
		char fname[MAX_FNAME_SZ];
		FILE *stream;
		void (*callback)(const char*);
	} out;
	unsigned int msg_mask;

	struct LogOut *next;
};

#ifdef _MSC_VER
#define vsnprintf	_vsnprintf
#else
int vsnprintf(char *str, size_t size, const char *format, va_list ap);
#endif

static void CleanUp(void);

struct LogOut *out_list = 0;

void SetLogFile(const char *fname, unsigned int mtypes) {
	struct LogOut *node;
	if(!out_list) {
		out_list = malloc(sizeof *out_list);
		out_list->next = 0;
		atexit(CleanUp);
	}

	if((node = malloc(sizeof *node))) {
		node->type = OUT_FILE;
		strncpy(node->out.fname, fname, MAX_FNAME_SZ);
		node->msg_mask = mtypes;
		node->next = out_list->next;
		out_list->next = node;
		remove(fname);
	}
}

void SetLogStream(FILE *fp, unsigned int mtypes) {
	struct LogOut *node;
	if(!out_list) {
		out_list = malloc(sizeof *out_list);
		out_list->next = 0;
		atexit(CleanUp);
	}

	if((node = malloc(sizeof *node))) {
		node->type = OUT_STREAM;
		node->out.stream = fp;
		node->msg_mask = mtypes;
		node->next = out_list->next;
		out_list->next = node;
	}
}

void SetLogCallback(void (*func)(const char*), unsigned int mtypes) {
	struct LogOut *node;

	if(!out_list) {
		out_list = malloc(sizeof *out_list);
		out_list->next = 0;
		atexit(CleanUp);
	}

	if((node = malloc(sizeof *node))) {
		node->type = OUT_CALLBACK;
		node->out.callback = func;
		node->msg_mask = mtypes;
		node->next = out_list->next;
		out_list->next = node;
	}
}

static void Log(const char *str, unsigned int type) {
	struct LogOut *node;
	if(!out_list) return;
	node = out_list->next;

	while(node) {
		if(type & node->msg_mask) {
			switch(node->type) {
			case OUT_FILE:
				{
					FILE *fp = fopen(node->out.fname, "a");
					if(fp) {
						fputs(str, fp);
						fclose(fp);
					}
				}
				break;

			case OUT_STREAM:
				fputs(str, node->out.stream);
				break;

			case OUT_CALLBACK:
				node->out.callback(str);
				break;

			default:
				break;
			}
		}
		node = node->next;
	}
}

void Info(const char *str, ...) {
	va_list arg_list;
	strcpy(msg_buf, "I: ");

	va_start(arg_list, str);
	vsnprintf(msg_buf + 3, MSG_BUFFER_SZ - 4, str, arg_list);
	va_end(arg_list);

	Log(msg_buf, LOG_INFO);
}

void Warning(const char *str, ...) {
	va_list arg_list;
	strcpy(msg_buf, "W: ");

	va_start(arg_list, str);
	vsnprintf(msg_buf + 3, MSG_BUFFER_SZ - 4, str, arg_list);
	va_end(arg_list);

	Log(msg_buf, LOG_WARNING);
}

void Error(const char *str, ...) {
	va_list arg_list;
	strcpy(msg_buf, "E: ");

	va_start(arg_list, str);
	vsnprintf(msg_buf + 3, MSG_BUFFER_SZ - 4, str, arg_list);
	va_end(arg_list);

	Log(msg_buf, LOG_ERROR);
}

static void CleanUp(void) {
	while(out_list) {
		struct LogOut *tmp = out_list;
		out_list = out_list->next;
		free(tmp);
	}
	out_list = 0;
}
