// character.cpp
#include "character.h"
#include <errlog.h>

#define COLLISION_PUSH 0.5f
#define MAX_ITERATIONS 1

char *char_node_names[] =
{
	"Bip01",
	"Bip01 L Hand",
	"Bip01 R Hand",
	"Bip01 L Calf",
	"Bip01 R Calf",
	"Bip01 L Foot",
	"Bip01 R Foot",
	"Bip01 Head"
};

Character::Character()
{
	state[0] = state[1] = STAY;
	state0_msec = state1_msec = 0;

	parent_scene = 0;

	origin_node = 0;
	
	og = 0;

	t0 = t1 = 0;
}

Character::~Character()
{

}

bool Character::SetOG(ObjectGroup *og)
{
	if (!og)
	{
		Error("Character og invalid\n");
		return false;
	}
	this->og = og;

	// get interesting nodes
	for (unsigned int i=0; i<MAX_CHAR_NODES; i++)
	{
		OGNode *n = og->GetNode(char_node_names[i]);
		if (!n)
		{
			Error("Character OG has no %s node\n", char_node_names[i]);
			return false;
		}
		nodes[i] = n;
	}

	// character metrics
	leg_length = 8.5f;//(GetNodePos(ROOT) - GetNodePos(L_FOOT)).Length();
	height = 2.0f * leg_length;
	// approximately a leg is 1 mt
	float one_mm = leg_length / 1000.0f;
	node_radius[ROOT] = leg_length / 2.0f;
	
	node_radius[L_HAND] = 0.1f * (GetNodePos(ROOT) - GetNodePos(L_HAND)).Length();
	node_radius[R_HAND] = 0.1f * (GetNodePos(ROOT) - GetNodePos(R_HAND)).Length();
	node_radius[L_KNEE] = leg_length / 4.0f;
	node_radius[R_KNEE] = leg_length / 4.0f;
	node_radius[L_FOOT] = 0.5f * GetNodePos(L_FOOT).y;// - 5.0f * one_mm;
	node_radius[R_FOOT] = 0.5f * GetNodePos(R_FOOT).y;// - 5.0f * one_mm;
	node_radius[HEAD] = 0.2f * (GetNodePos(ROOT) - GetNodePos(HEAD)).Length();

	return true;
}

bool Character::SetAnimation(CharState which, unsigned int ground, Animation *anim)
{
	if (!anim)
	{
		Error("Invalid animation\n");
		return false;
	}

	CharAnimation added_anim;
	added_anim.anim = anim;
	added_anim.ground = ground;
	
	anim->DeleteLastKey();

	if (ground & (PX | PY | PZ))
	{
		bool gx = false, gy = false, gz = false;
		if (!(ground & PX)) gx = true;
		if (!(ground & PY)) gy = true;
		if (!(ground & PZ)) gz = true;

		// process animation translation
		Timeline tm = anim->GetKeys(nodes[ROOT]->GetID());

		Timeline vel_tm;
		float anim_spf = 1.0f / anim->GetFPS();
		Vector3 prev_vel = Vector3(0, 0, 0);
		for (unsigned int i=0; i<tm.keys.size() - 1; i++)
		{
			Vector3 curr, next;
			curr = tm.keys[i].GetPosition();
		
			next = tm.keys[i + 1].GetPosition();
			Vector3 vel = (next - curr) / anim_spf;

			if (gx) vel.x = 0;
			if (gy) vel.y = 0;
			if (gz) vel.z = 0;

			if (vel.Length() > 1000.0f)
			{
				vel = prev_vel;
			}

			prev_vel = vel;

			vel_tm.keys.push_back(PRS(vel, Quaternion(), Vector3(1, 1, 1)));
		}

		// last frame of velocity is interpolated first and last velocity
		Vector3 last_vel = 0.5f * (vel_tm.keys[0].GetPosition()
			+ vel_tm.keys[vel_tm.keys.size() - 1].GetPosition());
		
		vel_tm.keys.push_back(PRS(last_vel, Quaternion(), Vector3(1, 1, 1)));
		
		added_anim.velocity = vel_tm;
	}

	animations[which] = added_anim;

	if (which == STAY)
	{
		og->SetAnimation(added_anim.anim);
		og->SetAnimationBlend(false);
		og->ProcessAnimation();
		// approximately a leg is 1 mt
		float one_mm = leg_length / 1000.0f;
		node_radius[L_FOOT] = GetNodePos(L_FOOT).y - one_mm;
		node_radius[R_FOOT] = GetNodePos(R_FOOT).y - one_mm;
	}

	return true;
}

void Character::Update(const Control &control, unsigned int msec)
{
	state0_msec += msec;
	state1_msec += msec;
	transition_msec += msec;
	t0 = state0_msec / 1000.0f;
	t1 = state1_msec / 1000.0f;
	tr = transition_msec / 1000.0f;
	dt = msec / 1000.0f;

	ctrl = control;

	state[1] = STAY;
	if (control.pad1.y > 0.01f)
	{
		state[1] = WALK;
	}
	if (control.pad1.y > 0.5f)
	{
		state[1] = RUN;
	}
	if (control.pad2.x > 0.1f)
	{
		state[1] = SIDESTEP_RIGHT;
	}
	if (control.pad2.x < -0.1f)
	{
		state[1] = SIDESTEP_LEFT;
	}
	if (control.pad1.y < -0.1f)
	{
		state[1] = WALK_BACK;
	}

	float ang_vel = Pi;
	rel_prs.SetRotation(rel_prs.GetRotation() * Quaternion(Vector3(0, -1, 0), 
		dt * control.pad1.x * ang_vel));

	if (state[0] == state[1])
	{
		transition_msec = state1_msec = 0;
		ProcessGenericState();
	}
	else
	{
		ProcessGenericTransition();
	}

	ProcessCollisions();
	og->SetModelMatrix(GetGlobalTransform());
	og->ProcessAnimation();
}

Vector3 Character::GetNodePos(CharNode which)
{
	Vector3 ret = Vector3(0, 0, 0);
	Matrix4x4 me = rel_prs.GetTransform();
	Matrix4x4 anim = nodes[which]->GetModelMatrix();
	return ret.Transformed(GetGlobalTransform() * anim);
}

Sphere Character::GetNodeSphere(CharNode which)
{
	return Sphere(GetNodePos(which), node_radius[which]);
}

void Character::ProcessGenericState()
{
	// animaiton
	og->SetAnimation(animations[state[0]].anim, 0);
	animations[state[0]].anim->SetLoopMode(ALM_LOOP);
	animations[state[0]].anim->SetSampleMode(ASM_LINEAR);
	og->SetAnimationTime(t0, 0);
	og->SetAnimationBlend(false, 0);

	// translation
	if (animations[state[0]].ground & (PX | PY | PZ))
	{
		Vector3 curr_vel = Sample(animations[state[0]].velocity, t0 *
			animations[state[0]].anim->GetFPS()).GetPosition();

		// rotate velocity
		curr_vel.Transform(rel_prs.GetRotation().GetRotationMatrix());

		rel_prs.SetPosition(rel_prs.GetPosition() + dt * curr_vel);
	}

	// ground animation components
	nodes[ROOT]->SetGround(animations[state[0]].ground);
}

void Character::ProcessGenericTransition()
{
	// animaiton
	og->SetAnimation(animations[state[0]].anim, 0);
	animations[state[0]].anim->SetLoopMode(ALM_LOOP);
	animations[state[0]].anim->SetSampleMode(ASM_LINEAR);
	og->SetAnimationTime(t0, 0);

	og->SetAnimation(animations[state[1]].anim, 1);
	animations[state[1]].anim->SetLoopMode(ALM_LOOP);
	animations[state[1]].anim->SetSampleMode(ASM_LINEAR);
	og->SetAnimationTime(t1, 1);

	// blending
	float trans_secs = 0.2f;
	float blend_factor = tr / trans_secs;
	if (blend_factor > 1.0f)
	{
		blend_factor = 1.0f;
		state[0] = state[1];
		state0_msec = state1_msec;
		state1_msec = 0;
	}
	og->SetAnimationBlend(true, blend_factor);

	// translation
	Vector3 vel0 = Vector3(0, 0, 0);
	if (animations[state[0]].ground & (PX | PY | PZ))
	{
		vel0 = Sample(animations[state[0]].velocity, t0 *
			animations[state[0]].anim->GetFPS()).GetPosition();
	}
	Vector3 vel1 = Vector3(0, 0, 0);
	if (animations[state[1]].ground & (PX | PY | PZ))
	{
		vel1 = Sample(animations[state[1]].velocity, t1 *
			animations[state[1]].anim->GetFPS()).GetPosition();
	}

	// interpolate velocity
	vel0 = Lerp(vel0, vel1, blend_factor);
	vel0.Transform(rel_prs.GetRotation().GetRotationMatrix());
	rel_prs.SetPosition(rel_prs.GetPosition() + dt * vel0);

	// ground animation components
	nodes[ROOT]->SetGround0(animations[state[0]].ground);
	nodes[ROOT]->SetGround1(animations[state[1]].ground);
}

Matrix4x4 Character::GetOriginTransform()
{
	Matrix4x4 m0;
	if (origin_node)
	{
		m0 = origin_node->GetModelMatrix();
	}
	Matrix4x4 origin = m0 * origin2_matrix;

	// make sure that origin coordinate system is orthogonal,
	// normalized and that up vector always looks up
	Vector3 xvec, yvec(0, 1, 0), zvec;
	xvec = origin.GetColumnVector(0);	
	zvec = CrossProduct(xvec, yvec).Normalized();
	xvec = CrossProduct(yvec, zvec).Normalized();
	origin.SetColumnVector(xvec, 0);
	origin.SetColumnVector(yvec, 1);
	origin.SetColumnVector(zvec, 2);
	origin.SetRowVector(Vector4(0, 0, 0, 1), 3);

	return origin;
}

Matrix4x4 Character::GetGlobalTransform()
{
	return GetOriginTransform() * rel_prs.GetTransform();
}

void Character::SetAbsolutePosition(const Vector3 &pos)
{
	rel_prs.SetPosition(pos.Transformed(GetOriginTransform().Inverse()));
}

Vector3 Character::GetAbsolutePosition()
{
	return rel_prs.GetPosition().Transformed(GetOriginTransform());
}

void Character::ProcessCollisions()
{
	Vector3 char_root = GetNodePos(ROOT);
	Ray r(char_root, Vector3(0, -1, 0));
	Vector3 floor_pos;
	
	if (parent_scene->Collision(r, true))
	{
		floor_pos = parent_scene->GetCollisionPoint(true);
		Vector3 correct_pos = GetAbsolutePosition();
		correct_pos.y = floor_pos.y;

		// extract collision node
		ObjectGroup *cog = (ObjectGroup*)parent_scene->GetCollisionObject();
		CollisionMesh *cmesh = parent_scene->GetCollisionMesh();
		OGNode *cnode = cog->GetProxyNode(cmesh);

		// reparent character, if needed
		if (cnode != origin_node)
		{
			Matrix4x4 abs_transform = GetGlobalTransform();//GetOriginTransform() * rel_prs.GetTransform();
			Matrix4x4 new_origin_transform = cnode->GetModelMatrix();
			Matrix4x4 intermediate = new_origin_transform.Inverse() * abs_transform;
			origin2_matrix = intermediate;
			origin_node = cnode;
			rel_prs = PRS();
		}

		SetAbsolutePosition(correct_pos);
	}

	unsigned int iterations = 0;
	bool isect = true;
	while (isect && iterations < MAX_ITERATIONS)
	{
		iterations++;
		// character sphere collision
		Vector3 max_isect = Vector3(0, 0, 0);
		unsigned int max_isect_node;
		for (unsigned int i=0; i<MAX_CHAR_NODES; i++)
		{
			col_results[i].collision = false;
			Sphere s = GetNodeSphere((CharNode)i);
			if (parent_scene->Collision(s))
			{
				isect = true;

				unsigned int tri_count = parent_scene->GetCollisionTriangleCount();
				for (unsigned int i=0; i<tri_count; i++)
				{
					Triangle tri = parent_scene->GetCollisionTriangleList(i, true);
					Plane pl = tri.GetPlane();
					if (UnderPlane(s.GetCenter(), pl)) continue;
					// see how deep was the intersection
					float dist = DistanceFromPlane(s.GetCenter(), pl);
					float pen = s.GetRadius() - dist;
					Vector3 correction = pen * pl.GetNormal();
					SetAbsolutePosition(GetAbsolutePosition() + correction);
				}

				/*
				// fill collision result struct
				col_results[i].collision = true;
				col_results[i].mesh = parent_scene->GetCollisionMesh();
				col_results[i].triangle_id = 
					col_results[i].mesh->GetCollisionTriangleIndex(CT_THIS);

				Triangle tri = col_results[i].mesh->GetCollisionTriangle(CT_THIS, true);
				Plane pl = tri.GetPlane();
				// see how deep was the intersection
				float dist = DistanceFromPlane(s.GetCenter(), pl);
				float pen = s.GetRadius() - dist;
				//printf("%f\n", pen);
				if (pen > max_isect.Length())
				{
					max_isect = pen * pl.GetNormal();
					max_isect_node = i;
				}
				*/
			}
		}

		/*
		// push the character back to prevent collision
		if (isect)
		{
			//printf("%f, %f, %f\n", max_isect.x, max_isect.y, max_isect.z);
			SetAbsolutePosition(GetAbsolutePosition() + max_isect);
		}
		*/
	}
}
