#include <string.h>
#include <errno.h>
#include <expat.h>
#include "level.h"
#include <errlog.h>

Level::Level()
{
	scene = new Scene;
	time_step = 33;
	player_cam = new TargetCamera;
}

Level::~Level()
{
	delete scene;
	delete player_cam;
}

enum XmlTag {
	TAG_LEVEL,
	TAG_OBJGROUP,
	TAG_CHAR,
	TAG_ANIM,

	TAG_COUNT
};
#define TAG_UNKNOWN	TAG_COUNT

struct XmlSymb {
	char *name;
	enum XmlTag tag;
} symb_table[] = {
	{"level",		TAG_LEVEL},
	{"objgroup",	TAG_OBJGROUP},
	{"character",	TAG_CHAR},
	{"animation",	TAG_ANIM}
};


struct LoadTreeNode {
	XmlTag tag;
	char *name, *fname;
	LoadTreeNode *parent;
	std::vector<LoadTreeNode*> children;

	GameObjGroup *gobj;
};

struct ParseState {
	Level *lvl;
	LoadTreeNode *ldtree;
	LoadTreeNode *cur_node;
	int prog, load_count;
};

static void StartHandler(void *udata, const XML_Char *name, const XML_Char **atts);
static void EndHandler(void *udata, const XML_Char *name);
static void DataHandler(void *udata, const XML_Char *str, int len);

static enum XmlTag GetTag(const char *name);
static const char *GetAttrStr(const XML_Char **atts, const char *aname);

static void ExecuteLoadTree(LoadTreeNode *node, ParseState *pst, void (*prog_func)(float));
static void DeleteLoadTree(LoadTreeNode *node);

bool Level::Load(const char *fname, void (*progr_func)(float))
{
	VFS_FILE *fp;
	if(!(fp = vfs_open(fname, "r"))) {
		Error("failed to open level file %s: %s\n", fname, strerror(errno));
		return false;
	}

	XML_Parser xml;
	if(!(xml = XML_ParserCreate("US-ASCII"))) {
		Error("level loading failed to create XML parser\n");
		vfs_close(fp);
		return false;
	}

	ParseState pst;
	pst.lvl = this;
	pst.ldtree = new LoadTreeNode;
	pst.cur_node = pst.ldtree;
	pst.ldtree->tag = TAG_UNKNOWN;
	pst.ldtree->name = pst.ldtree->fname = 0;
	pst.ldtree->parent = 0;
	pst.prog = pst.load_count = 0;

	XML_SetElementHandler(xml, StartHandler, EndHandler);
	XML_SetCharacterDataHandler(xml, DataHandler);
	XML_SetUserData(xml, &pst);
	XML_SetParamEntityParsing(xml, XML_PARAM_ENTITY_PARSING_ALWAYS);

	int done = 0;
	while(!done) {
		char *buf = (char*)XML_GetBuffer(xml, 1024);
		if(!vfs_gets(buf, 1024, fp) || strstr(buf, "</level>")) {
			done = 1;
		}

		if(!XML_ParseBuffer(xml, (int)strlen(buf), done)) {
			Error("level loading failed due to XML error: %s\n", XML_ErrorString(XML_GetErrorCode(xml)));
			vfs_close(fp);
			XML_ParserFree(xml);
			return false;
		}
	}

	XML_ParserFree(xml);
	vfs_close(fp);

	ExecuteLoadTree(pst.ldtree, &pst, progr_func);
	DeleteLoadTree(pst.ldtree);
	return true;
}

static void StartHandler(void *udata, const XML_Char *name, const XML_Char **atts)
{
	ParseState *pst = (ParseState*)udata;
	enum XmlTag tag = GetTag(name);
	const char *str;

	LoadTreeNode *node = new LoadTreeNode;
	node->tag = tag;
	node->name = node->fname = 0;
	node->parent = pst->cur_node;
	pst->cur_node->children.push_back(node);

	switch(tag) {
	case TAG_LEVEL:
		break;

	case TAG_CHAR:
	case TAG_OBJGROUP:
	case TAG_ANIM:
		if((str = GetAttrStr(atts, "name"))) {
			node->name = new char[strlen(str) + 1];
			strcpy(node->name, str);
		} else if((str = GetAttrStr(atts, "file"))) {
			node->fname = new char[strlen(str) + 1];
			strcpy(node->fname, str);
		}
		pst->load_count++;
		break;

	default:
		Error("XML parsing error, unknown tag: <%s>\n", name);
	}
}

static void EndHandler(void *udata, const XML_Char *name)
{
	ParseState *pst = (ParseState*)udata;
	pst->cur_node = pst->cur_node->parent;
}

static void DataHandler(void *udata, const XML_Char *str, int len) {}

static enum XmlTag GetTag(const char *name)
{
	for(int i=0; i<TAG_COUNT; i++) {
		if(strcmp(symb_table[i].name, name) == 0) {
			return symb_table[i].tag;
		}
	}
	return TAG_UNKNOWN;
}

static const char *GetAttrStr(const XML_Char **atts, const char *aname)
{
	while(*atts) {
		if(!strcmp(*atts, aname)) {
			return *(atts + 1);
		}
		atts++;
	}
	return 0;
}

static void ExecuteLoadTree(LoadTreeNode *node, ParseState *pst, void (*prog_func)(float))
{
	bool advance_prog = false;

	switch(node->tag) {
	case TAG_OBJGROUP:
		node->gobj = new GameObjGroup;
		node->gobj->SetName(node->name);
		if(!node->gobj->Load(node->fname)) {
			Error("failed to load GameObjGroup: %s\n", node->fname);
			delete node->gobj;
		}
		advance_prog = true;
		break;

	case TAG_CHAR:
		{
			Character *chr = new Character;
			chr->SetName(node->name);
			if(!chr->Load(node->fname)) {
				Error("failed to load character: %s\n", node->fname);
				delete chr;
			}
			node->gobj = chr;
		}
		advance_prog = true;
		break;

	case TAG_ANIM:
		{
			if(!node->parent || (node->parent->tag != TAG_OBJGROUP && node->parent->tag != TAG_CHAR)) {
				Error("<anim>'s parent must be <objectgroup> or <character>\n");
			}
			if(!node->parent->gobj->LoadAnimation(node->fname, node->name)) {
				Error("failed to load animation: %s\n", node->fname);
			}
		}
		advance_prog = true;
		break;

	default:
		break;
	}

	if(advance_prog && prog_func) {
		prog_func((float)++pst->prog / (float)pst->load_count);
	}

	for(size_t i=0; i<node->children.size(); i++) {
		ExecuteLoadTree(node->children[i], pst, prog_func);
	}
}

static void DeleteLoadTree(LoadTreeNode *node)
{
	for(size_t i=0; i<node->children.size(); i++) {
		DeleteLoadTree(node->children[i]);
	}

	delete node->name;
	delete node->fname;
	delete node;
}
