#define LIBGC_SOURCE_FILE
#include <math3d.h>
#include <libsys.h>
#include "gc.h"
#include "gc_impl.h"
#include "immed.h"
#include "statecache.h"
#include "uniforms.h"
#include "textures.h"
#include <console.h>
#include <errlog.h>
#include <devhelp.h>


static GCInitParams init_params;

static Matrix4x4 projection;
static Matrix4x4 model;
static Matrix4x4 view, inv_view;
static Matrix4x4 modelview;
static Matrix4x4 modelviewprojection;
static Matrix4x4 texture_matrix[MAX_TEXTURE_UNITS];


GCInitParams::GCInitParams(int w, int h, bool fullscr)
{
	dmode.width = w;
	dmode.height = h;
	dmode.bpp = 32;
	dmode.frequency = 0;
	depth_bits = 24;
	stencil_bits = 8;
	fullscreen = fullscr;
	win_ext = 0;
}

void InitializeGCStates()
{
	SetClearColor(Color(0.0f, 0.0f, 0.0f, 1.0f));
	SetClearDepth(1.0f);
	SetClearStencil(0);

	SetColorMask(true);

	SetDepthMask(true);
	SetDepthTest(true);
	SetDepthFunc(CMP_LESS);

	SetBlending(false);
	SetBlendFunc(BLEND_ONE, BLEND_ONE);

	SetAlphaTest(false);
	SetAlphaRef(0.0f);
	SetAlphaFunc(CMP_LEQUAL);

	SetCulling(CULL_CCW);
	SetProjectionMatrix(PerspectiveProj(70.0f, (float) SysGetWindowWidth() / (float) SysGetWindowHeight(), 1.0f, 2500.0f));

	Matrix4x4 m;

	SetViewMatrix(m);
	SetModelMatrix(m);

	for(int i=0; i<8; i++) {
        SetTextureMatrix(i, m);

		if(i < 4) {
			SetSamplerState(i, ST_MAGFILTER, TF_LINEAR);
			SetSamplerState(i, ST_MINFILTER, TF_LINEAR);
			SetSamplerState(i, ST_MIPFILTER, TF_LINEAR);
		}
	}
}

bool InitSubsys()
{
	// Bad Sector: gia fullscreen modes o mouse cursor prepei na einai invisible
	// alla gia windowed modes prepei na fainetai.
	if (init_params.fullscreen)
		SysShowCursor(false);
	else
		SysShowCursor(true);

	if(!InitializeGCUniforms()) {
		Error("failed to initialize the shader uniforms\n");
		return false;
	}
	if(!InitializeGPUPrograms()) {
		Error("failed to initialize the shader subsystem\n");
		return false;
	}

	InitializeGCStates();
	//InitGCResources();  // TODO : LAALLALAALALLA
	
	MountDevData();
		
	if(!ImInit()) {
		Error("failed to initialize immediate mode rendering\n");
		return false;
	}
	if(!ConInit()) {
		Error("failed to initialize the debugging console\n");
		//return false;
	}
	ConHide();
	SetLogCallback(ConPrintStr, LOG_ALL);
	// stderr and logfile logging is already enabled

	GCInitParams params = GetInitParams();
	SetHalfPixelSize(params.dmode.width, params.dmode.height);
	return true;
}

void CloseSubsys()
{
	ImDestroy();	// destroy immediate mode emulation

	//ReleaseGCResources();	// TODO : ALALALALALAL
	CloseGPUPrograms();
	ReleaseGCUniforms();
}

void SetInitParams(GCInitParams iparam)
{
	init_params = iparam;
}

GCInitParams GetInitParams()
{
	return init_params;
}

void SetProjectionMatrix(const Matrix4x4 &m)
{
	projection = m;
	modelviewprojection = projection * modelview;

	Matrix4x4 tproj = m.Transposed();
	Matrix4x4 tmvp = modelviewprojection.Transposed();

	APISetProjectionMatrix((float*)&tproj);
	
	SetUniform(STATE_PROJECTION_MATRIX, (float*)&tproj);
	SetUniform(STATE_MODELVIEWPROJECTION_MATRIX, (float*)&tmvp);
}

void SetViewMatrix(const Matrix4x4 &m)
{
	view = m;
	modelview = view * model;
	modelviewprojection = projection * modelview;
	inv_view = m.Inverse();

	Matrix4x4 tview = view.Transposed();
	Matrix4x4 tmv = modelview.Transposed();
	Matrix4x4 tmvp = modelviewprojection.Transposed();

	APISetViewMatrix((float*)&tview);
	APISetModelViewMatrix((float*)&tmv);

	SetUniform(STATE_VIEW_MATRIX, (float*)&tview);
	SetUniform(STATE_MODELVIEW_MATRIX, (float*)&tmv);
	SetUniform(STATE_MODELVIEWPROJECTION_MATRIX, (float*)&tmvp);
}

void SetModelMatrix(const Matrix4x4 &m)
{
	model = m;
	modelview = view * model;
	modelviewprojection = projection * modelview;

	Matrix4x4 tmodel = model.Transposed();
	Matrix4x4 tmv = modelview.Transposed();
	Matrix4x4 tmvp = modelviewprojection.Transposed();

	APISetModelMatrix((float*)&tmodel);
	APISetModelViewMatrix((float*)&tmv);

	SetUniform(STATE_MODEL_MATRIX, (float*)&tmodel);
	SetUniform(STATE_MODELVIEW_MATRIX, (float*)&tmv);
	SetUniform(STATE_MODELVIEWPROJECTION_MATRIX, (float*)&tmvp);
}

void SetTextureMatrix(unsigned int tid, const Matrix4x4 &m)
{
	unsigned int cstate = 0;
	switch(tid) {
	case 0:
		cstate = STATE_TEXTURE0_MATRIX; break;
	case 1:
		cstate = STATE_TEXTURE1_MATRIX; break;
	case 2:
		cstate = STATE_TEXTURE2_MATRIX; break;
	case 3:
		cstate = STATE_TEXTURE3_MATRIX; break;
	case 4:
		cstate = STATE_TEXTURE4_MATRIX; break;
	case 5:
		cstate = STATE_TEXTURE5_MATRIX; break;
	case 6:
		cstate = STATE_TEXTURE6_MATRIX; break;
	case 7:
		cstate = STATE_TEXTURE7_MATRIX; break;
	default:
		return;
	}

	texture_matrix[tid] = m;

	Matrix4x4 ttex = m.Transposed();

	APISetTextureMatrix(tid, (float*)&ttex);
	SetUniform(cstate, (float*)&ttex);
}

// the get functions used the commented-out code that retrieves the matrix from the
// state cache. however we already store each matrix to a static variable here, so
// why do the extra effort to retranspose the matrix?
Matrix4x4 GetModelMatrix()
{
	/*Matrix4x4 m;
	GetUniform(STATE_MODEL_MATRIX, (float*)&m);
	return m.Transposed();*/
	return model;
}

Matrix4x4 GetViewMatrix()
{
	/*Matrix4x4 m;
	GetUniform(STATE_VIEW_MATRIX, (float*)&m);
	return m.Transposed();*/
	return view;
}

Matrix4x4 GetInvViewMatrix()
{
	return inv_view;
}

Matrix4x4 GetProjectionMatrix()
{
	/*Matrix4x4 m;
	GetUniform(STATE_PROJECTION_MATRIX, (float*)&m);
	return m.Transposed();*/
	return projection;
}

Matrix4x4 GetTextureMatrix(int unit)
{
	return texture_matrix[unit];
}

Matrix4x4 GetModelViewMatrix()
{
	return modelview;
}

Box GetFrustum()
{
	return CalculateViewFrustum(GetViewMatrix(), GetProjectionMatrix());
}

Ray GetPickingRay(Vector2 mpos)
{
	Box fr = GetFrustum();

	Vector3 start = fr.GetVertex(0);
	start += mpos.x * (fr.GetVertex(3) - fr.GetVertex(0));
	start += mpos.y * (fr.GetVertex(1) - fr.GetVertex(0));

	Vector3 end = fr.GetVertex(4);
	end += mpos.x * (fr.GetVertex(7) - fr.GetVertex(4));
	end += mpos.y * (fr.GetVertex(5) - fr.GetVertex(4));
	return Line(start, end).GetRay();
}

void SetBoneMatrix(unsigned int which, const Matrix4x4 &mat)
{
	if (which >= MAX_BONES) return;
	Matrix4x4 m = mat;
	//m.Transpose();
	//m.Rotate(Vector3(-Pi/2, 0, 0));
	//m = m * mat;
	float *uniform = (float*) GetUniformDataPtr(STATE_BONE_MATRIX);

	memcpy(uniform + 12 * which, &m, 12 * sizeof(float));
}

void SetHalfPixelSize(int width, int height)
{
	float hp_width, hp_height;
	hp_width = 0.5f / (float) width;
	hp_height = 0.5f / (float) height;
	SetUniform(STATE_HALF_PIXEL_WIDTH, &hp_width);
	SetUniform(STATE_HALF_PIXEL_HEIGHT, &hp_height);
}
