#define LIBGC_SOURCE_FILE
#include <builddefs.h>

#ifdef BUILD_GL
#include <libsys.h>
#include <sysgc.h>
#include <errlog.h>
#include "gc.h"
#include "gc_stats.h"

bool InitSubsys();
void CloseSubsys();
static void LoadExtensions();

static GCInitParams init_params;

static float cache_alpha_ref;
static CompareFunc cache_alpha_func = CMP_ALWAYS;
TextureAddress cache_tex_addr[8];
static TextureFilter cache_tex_min_filter[8];
static TextureFilter cache_tex_mag_filter[8];
static TextureFilter cache_tex_mip_filter[8];

static int stencil_ref;
static CompareFunc stencil_func;
static StencilOp stencil_fail = STENCIL_KEEP;
static StencilOp stencil_zfail = STENCIL_KEEP;
static StencilOp stencil_pass = STENCIL_KEEP;

// X Window System (GLX) specific part of the OpenGL gc stuff, too small to merit its own file
#ifdef WSYS_X11
#include <X11/Xlib.h>
#include <GL/glx.h>

// XXX: this must be the same as in libsys/libsys_x11.cpp
struct InitData {
	Display *dpy;
	GLXContext ctx;
	XVisualInfo *vis;
	Window win;
	bool fullscreen;
};

static Display *dpy;
static GLXContext ctx;
static Window win;

bool InitGC(const GCInitParams &params)
{
#ifndef NDEBUG
	SetLogStream(stderr, LOG_ALL);
	SetLogFile("t7engine.log", LOG_ALL);
#endif

	if(!(dpy = XOpenDisplay(0))) {
		Error("failed to connect to the X server\n");
		return false;
	}
	int scr = DefaultScreen(dpy);

	int rbits = params.dmode.bpp >= 24 ? 8 : 5;
	int gbits = params.dmode.bpp >= 24 ? 8 : 6;
	int bbits = params.dmode.bpp >= 24 ? 8 : 5;

	int attr[] = {
		GLX_RGBA, GLX_DOUBLEBUFFER,
		GLX_RED_SIZE, rbits,
		GLX_GREEN_SIZE, gbits,
		GLX_BLUE_SIZE, bbits,
		GLX_DEPTH_SIZE, params.depth_bits,
		GLX_STENCIL_SIZE, params.stencil_bits,
		None
	};

	XVisualInfo *vis;
	if(!(vis = glXChooseVisual(dpy, scr, attr))) {
		Error("failed to get a suitable GLX visual\n");
		XCloseDisplay(dpy);
		return false;
	}

	if(!(ctx = glXCreateContext(dpy, vis, 0, True))) {
		Error("failed to create a GLX context\n");
		XFree(vis);
		XCloseDisplay(dpy);
		return false;
	}

	InitData xinit;
	xinit.dpy = dpy;
	xinit.ctx = ctx;
	xinit.vis = vis;
	xinit.fullscreen = params.fullscreen;

	if(!SysCreateWindow(params.dmode.width, params.dmode.height, false, &xinit)) {
		XFree(vis);
		XCloseDisplay(dpy);
		return false;
	}
	win = xinit.win;

	if(glXMakeCurrent(dpy, win, ctx) == False) { 
		Error("glXMakeCurrent failed\n");
		glXDestroyContext(dpy, ctx);
		SysCloseWindow();
		XCloseDisplay(dpy);
		return false;
	}

	LoadExtensions();
	return InitSubsys();
}

void CloseGC()
{
	CloseSubsys();
	glXDestroyContext(dpy, ctx);
	SysCloseWindow();
	XCloseDisplay(dpy);
}

void Flip()
{
	glXSwapBuffers(dpy, win);
	StatsReset();
}

#endif	// WSYS_X11

// Win32 (WGL) specific part of the OpenGL gc stuff, too small to merit its own file
#ifdef WSYS_WIN32
#include <windows.h>

HWND win;
HDC hdc;
HGLRC ctx;

bool InitGC(const GCInitParams &params)
{
	if(params.fullscreen) {
		DEVMODE dmode;
		memset(&dmode, 0, sizeof dmode);
		dmode.dmSize = sizeof dmode;
		dmode.dmBitsPerPel = params.dmode.bpp;
		dmode.dmPelsWidth = params.dmode.width;
		dmode.dmPelsHeight = params.dmode.height;
		dmode.dmDisplayFrequency = params.dmode.frequency;
		dmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;

		if(ChangeDisplaySettings(&dmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL) {
			Error("InitGC failed to change display mode\n");
			CloseGC();
			return false;
		}
	}

	if(!SysCreateWindow(params.dmode.width, params.dmode.height, !params.fullscreen, params.win_ext)) {
		Error("InitGC failed to create graphics window\n");
		CloseGC();
		return false;
	}
	win = (HWND)SysGetWindow();
	hdc = GetDC(win);

	PIXELFORMATDESCRIPTOR pf;
	memset(&pf, 0, sizeof pf);
	pf.nSize = sizeof pf;
	pf.nVersion = 1;
	pf.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pf.cColorBits = PFD_TYPE_RGBA;
	pf.cDepthBits = params.depth_bits;
	pf.cStencilBits = params.stencil_bits;

	int pixel_fmt = ChoosePixelFormat(hdc, &pf);
	if(!pixel_fmt) {
		Error("InitGC failed to select pixel format\n");
		CloseGC();
		return false;
	}

	if(!SetPixelFormat(hdc, pixel_fmt, &pf)) {
		Error("InitGC failed to set pixel format\n");
		CloseGC();
		return false;
	}

	ctx = wglCreateContext(hdc);
	if(!ctx) {
		Error("InitGC failed to create OpenGL context\n");
		CloseGC();
		return false;
	}

	if(!wglMakeCurrent(hdc, ctx)) {
		Error("InitGC failed to make the GL context current\n");
		CloseGC();
		return false;
	}

	LoadExtensions();
	return InitSubsys();
}

void CloseGC()
{
	CloseSubsys();

	wglDeleteContext(ctx);
	ReleaseDC(win, hdc);
	DestroyWindow(win);
    
	// TODO: implement (close WGL context blah)
	SysCloseWindow();
}

void Flip()
{
	SwapBuffers(hdc);
}
#endif	// WSYS_WIN32

void BeginScene() {}
void EndScene() {}

void SetClearColor(const Color &color)
{
	glClearColor(color.r, color.g, color.b, color.a);
}

void SetClearDepth(float depth)
{
	glClearDepth(depth);
}

void SetClearStencil(unsigned int s)
{
	glClearStencil(s);
}

void Clear()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void Clear(unsigned int mask)
{
	glClear(mask);
}

void SetColorMask(bool red, bool green, bool blue, bool alpha)
{
	glColorMask((GLboolean)red, (GLboolean)green, (GLboolean)blue, (GLboolean)alpha);
}

void SetColorMask(bool enabled)
{
	GLboolean glfoo = (GLboolean)enabled;
	glColorMask(glfoo, glfoo, glfoo, glfoo);
}

void SetDepthMask(bool enabled)
{
	glDepthMask((GLboolean)enabled);
}

bool GetDepthMask()
{
	GLboolean foo;
	glGetBooleanv(GL_DEPTH_WRITEMASK, &foo);
	return foo;
}

void SetDepthTest(bool enabled)
{
	if(enabled) {
		glEnable(GL_DEPTH_TEST);
	} else {
		glDisable(GL_DEPTH_TEST);
	}
}

bool GetDepthTest()
{
	return (bool)glIsEnabled(GL_DEPTH_TEST);
}

void SetDepthFunc(CompareFunc func)
{
	glDepthFunc(func);
}

void SetStencil(bool enabled)
{
	if(enabled) {
		glEnable(GL_STENCIL_TEST);
	} else {
		glDisable(GL_STENCIL_TEST);
	}
}

bool GetStencil()
{
	return (bool)glIsEnabled(GL_STENCIL_TEST);
}

void SetStencilFunc(CompareFunc func)
{
	stencil_func = func;
	glStencilFunc(stencil_func, stencil_ref, 0xffffffff);
}

void SetStencilPass(StencilOp op)
{
	stencil_pass = op;
	glStencilOp(stencil_fail, stencil_zfail, stencil_pass);
}

void SetStencilFail(StencilOp op)
{
	stencil_fail = op;
	glStencilOp(stencil_fail, stencil_zfail, stencil_pass);
}

void SetStencilZFail(StencilOp op)
{
	stencil_zfail = op;
	glStencilOp(stencil_fail, stencil_zfail, stencil_pass);
}

void SetStencilRef(int ref)
{
	stencil_ref = ref;
	glStencilFunc(stencil_func, stencil_ref, 0xffffffff);
}

void SetBlending(bool enabled)
{
	if(enabled) {
		glEnable(GL_BLEND);
	} else {
		glDisable(GL_BLEND);
	}
}

bool GetBlending()
{
	return (bool)glIsEnabled(GL_BLEND);
}

void SetBlendFunc(BlendFunc src, BlendFunc dst)
{
	glBlendFunc(src, dst);
}

BlendFunc GetBlendSrc()
{
	int ret;
	glGetIntegerv(GL_BLEND_SRC, &ret);
	return (BlendFunc)ret;
}

BlendFunc GetBlendDst()
{
	int ret;
	glGetIntegerv(GL_BLEND_DST, &ret);
	return (BlendFunc)ret;
}

void SetAlphaTest(bool enabled)
{
	if(enabled) {
		glEnable(GL_ALPHA_TEST);
	} else {
		glDisable(GL_ALPHA_TEST);
	}
}

void SetAlphaRef(float ref)
{
	cache_alpha_ref = ref;
	glAlphaFunc((GLenum)cache_alpha_func, ref);
}

void SetAlphaFunc(CompareFunc func)
{
	cache_alpha_func = func;
	glAlphaFunc((GLenum)func, cache_alpha_ref);
}

void SetCulling(CullFace face)
{
	if(face == CULL_NONE) {
		glDisable(GL_CULL_FACE);
	} else {
		glEnable(GL_CULL_FACE);
		glFrontFace(face);		// remember, "face" is defined as the opposite in gctypes.h
	}
}

CullFace GetCulling()
{
	if(!glIsEnabled(GL_CULL_FACE)) {
		return CULL_NONE;
	}

	int ret;
	glGetIntegerv(GL_CULL_FACE_MODE, &ret);
	return (CullFace)ret;
}

void APISetProjectionMatrix(float *m)
{
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(m);
}

// these two do nothing, APISetModelViewMatrix() is also called anyway
void APISetViewMatrix(float *m) {}
void APISetModelMatrix(float *m) {}

void APISetModelViewMatrix(float *m)
{
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(m);
}

void APISetTextureMatrix(unsigned int tid, float *m)
{
	GLenum tunit = GL_TEXTURE0 + (GLenum)tid;
	glMatrixMode(tunit);
	glLoadMatrixf(m);
}

void SetTexTransform(bool enable)
{
	// nothing needs to be done here in OpenGL
}

bool GetTexTransform()
{
	return true;
}


void SetViewport(int x, int y, int width, int height)
{
	glViewport(x, y, width, height);
}

Vector2 GetViewport(int *x, int *y, int *width, int *height)
{
	int vp[4];

	glGetIntegerv(GL_VIEWPORT, vp);
	if(x) *x = vp[0];
	if(y) *y = vp[1];
	if(width) *width = vp[2];
	if(height) *height = vp[3];

	return Vector2((float)vp[2], (float)vp[3]);
}

void SetLighting(bool enable)
{
	if(enable) {
		glEnable(GL_LIGHTING);
	} else {
		glDisable(GL_LIGHTING);
	}
}

bool GetLighting()
{
	return (bool)glIsEnabled(GL_LIGHTING);
}

/* XXX: the addressing states are per-texture in OpenGL. We just keep our own
 * state for them, and rely on BindTexture() in textures.cpp to actually set
 * them after calling glBindTexture().
 */
void SetTextureAddressing(int unit, TextureAddress addr)
{
	cache_tex_addr[unit] = addr;
}

void SetSamplerState(unsigned int sampler_id, SamplerStateType type, unsigned int value)
{
	// TODO: implement
}

void SetTextureFiltering(int unit, TextureFilter minf, TextureFilter magf, TextureFilter mipf)
{
	cache_tex_min_filter[unit] = minf;
	cache_tex_mag_filter[unit] = magf;
	cache_tex_mip_filter[unit] = mipf;
}

TextureFilter GetTextureFilteringMin(int unit)
{
	return cache_tex_min_filter[unit];
}

TextureFilter GetTextureFilteringMag(int unit)
{
	return cache_tex_mag_filter[unit];
}

TextureFilter GetTextureFilteringMip(int unit)
{
	return cache_tex_mip_filter[unit];
}

void SetFillMode(FillMode mode)
{
	glPolygonMode(GL_FRONT_AND_BACK, (GLenum)mode);
}

void SetScissorTest(bool enable)
{
	if(enable) {
		glEnable(GL_SCISSOR_TEST);
	} else {
		glDisable(GL_SCISSOR_TEST);
	}
}

bool GetScissorTest()
{
	return (bool)glIsEnabled(GL_SCISSOR_TEST);
}

void SetScissorRect(int x1, int y1, int x2, int y2)
{
	glScissor(x1, y1, x2 - x1, y2 - y1);
}

static void LoadExtensions()
{
	glewInit();
}

#endif	/* BUILD_GL */
