#include <errno.h>
#include <vector>
#include "sysgc.h"
#include <Cg/cg.h>
#include "idman.h"
#include "gpuprog.h"
#include "statecache.h"
#include "uniforms.h"
#include <errlog.h>
#include <libvfs.h>

#ifdef BUILD_DX9
#include <Cg/cgD3D9.h>
extern IDirect3DDevice9* d3d_device;
#elif defined(BUILD_GL)
#include <Cg/cgGL.h>
#endif

static CGcontext cg_context = NULL;

static CGprofile vs_prof, ps_prof;

static unsigned int current_gpu_uid = 0;
static unsigned int current_gpu_id = 0;

struct GPUProgram
{
	CGprogram vs;
	CGprogram ps;

	std::vector<unsigned int> vs_cached_uniforms;
	std::vector<CGparameter> vs_cached_uniform_params;
	std::vector<unsigned int> ps_cached_uniforms;
	std::vector<CGparameter> ps_cached_uniform_params;
};

static void DeleteHWGPUProg(void *prog);
static void AddCachedUniforms(GPUProgram *prog);


IdMan gpu_progs (DeleteHWGPUProg);


bool InitializeGPUPrograms ()
{
	if(!(cg_context = cgCreateContext()))
	{
		return false;
	}
#ifdef BUILD_DX9
	if(FAILED(cgD3D9SetDevice(d3d_device)))
	{
		return false;
	}
	
	vs_prof = cgD3D9GetLatestVertexProfile();
	ps_prof = cgD3D9GetLatestPixelProfile();
#elif defined(BUILD_GL)
	//vs_prof = cgGLGetLatestProfile(CG_GL_VERTEX);
	//ps_prof = cgGLGetLatestProfile(CG_GL_PIXEL);
	vs_prof = CG_PROFILE_ARBVP1;
	ps_prof = CG_PROFILE_ARBFP1;
#endif

	return true;
}

void CloseGPUPrograms ()
{
#ifdef BUILD_DX9
	cgD3D9SetDevice(0);
#endif
	cgDestroyContext(cg_context);
}

static void DeleteHWGPUProg(void* prog)
{
	GPUProgram *gp = (GPUProgram*)prog;
	if(gp->vs) cgDestroyProgram(gp->vs);
	if(gp->ps) cgDestroyProgram(gp->ps);
	delete gp;
}


// searches the shaders of a gpu program to find any cached-uniforms and adds them
static void AddCachedUniforms(GPUProgram *prog) {
	// search vertex shader

	unsigned int i = FirstUniform ();
	while (i != 0)
	{
		CGparameter p;
		if((p = cgGetNamedParameter(prog->vs, GetUniformName(i))))
		{
			prog->vs_cached_uniforms.push_back(i);
			prog->vs_cached_uniform_params.push_back(p);
		}

		i = NextUniform (i);
	}

	i = FirstUniform ();
	while (i != 0)
	{
		CGparameter p;
		if((p = cgGetNamedParameter(prog->ps, GetUniformName(i))))
		{
			prog->ps_cached_uniforms.push_back(i);
			prog->ps_cached_uniform_params.push_back(p);
		}

		i = NextUniform (i);
	}
}


/* ActualCreateGPUProgram()
 * Creates a GPUProgram from source (vssrc / pssrc) adds it to the gpu_progs
 * manager and returns the id.
 * vsfile and psfile are optional. If they are not NULL, it uses them to
 * produce better error messages. (TODO)
 */
static unsigned int ActualCreateGPUProgram(const char *vssrc, const char *pssrc, const char *vsfile, const char *psfile)
{
	if(!vssrc || !pssrc) {
		return 0;
	}

	GPUProgram *prog = new GPUProgram;
	prog->vs = prog->ps = 0;

#ifdef BUILD_DX9
	//TODO: in the future, check if cgD3D9GetOptimalOptions works (as of 23/04/2007
	//it does NOT) and use it.

	prog->vs = cgCreateProgram(cg_context, CG_SOURCE, vssrc, vs_prof, 0, 0);
	if(!prog->vs || cgD3D9LoadProgram(prog->vs, false, 0) != D3D_OK) {
		Error("error while compiling vertex shader: %s\n%s\n", vsfile, cgGetLastListing(cg_context));
		DeleteHWGPUProg(prog);
		return 0;
	}
#elif defined(BUILD_GL)
	cgGLSetOptimalOptions(vs_prof);
	prog->vs = cgCreateProgram(cg_context, CG_SOURCE, vssrc, vs_prof, 0, 0);
	if(!prog->vs) {
		cgGLLoadProgram(prog->vs);

		CGerror err;
		if((err = cgGetError()) != CG_NO_ERROR) {
			Error("error while compiling vertex shader: %s\n%s\n", vsfile, cgGetLastListing(cg_context));
			DeleteHWGPUProg(prog);
			return 0;
		}
	}
#endif

#ifdef BUILD_DX9
	//TODO: in the future, check if cgD3D9GetOptimalOptions works (as of 23/04/2007
	//it does NOT) and use it.

	prog->ps = cgCreateProgram(cg_context, CG_SOURCE, pssrc, ps_prof, 0, 0);
	if(!prog->ps || cgD3D9LoadProgram(prog->ps, false, 0) != D3D_OK) {
		Error("error while compiling pixel shader: %s\n%s\n", psfile, cgGetLastListing(cg_context));
		DeleteHWGPUProg(prog);
		return 0;
	}
#elif defined(BUILD_GL)
	prog->ps = cgCreateProgram(cg_context, CG_SOURCE, pssrc, ps_prof, 0, 0); //<-- if you have doubts on that, please check the reference
	if(!prog->ps) {
		cgGLLoadProgram(prog->ps);

		GLenum err;
		if((err = glGetError()) != GL_NO_ERROR) {
			Error("error while compiling pixel shader: %s\n%s\n", psfile, cgGetLastListing(cg_context));
			DeleteHWGPUProg(prog);
			return 0;
		}
	}
#endif

	AddCachedUniforms(prog);

	unsigned int id = gpu_progs.Add(prog);
	if(!id) {
		DeleteHWGPUProg(prog);
	}
	return id;
}

// Create GPU program from strings in memory
unsigned int CreateGPUProgram(const char *vssrc, const char *pssrc)
{
	return ActualCreateGPUProgram(vssrc, pssrc, 0, 0);
}

// Create GPU program from files
unsigned int CreateGPUProgramFromFile(const char *vsfile, const char *psfile)
{
	unsigned int prog;
	char *vsrc, *psrc;
	VFS_FILE *fp;
	long fsize;

	if(!(fp = vfs_open(vsfile, "rb"))) {
		Error("failed to open shader file %s: %s\n", vsfile, strerror(errno));
		return 0;
	}
	vfs_seek(fp, 0, VFS_SEEK_END);
	fsize = vfs_tell(fp);
	vfs_seek(fp, 0, VFS_SEEK_SET);

	if(!(vsrc = (char*)malloc(fsize + 1))) {
		Error("failed to allocate %ld bytes: %s\n", fsize + 1, strerror(errno));
		vfs_close(fp);
		return 0;
	}
	if(!vfs_read(vsrc, 1, fsize, fp)) {
		Error("failed to load shader file %s: %s\n", vsfile, strerror(errno));
		vfs_close(fp);
		free(vsrc);
		return 0;
	}
	vfs_close(fp);

	vsrc[fsize] = 0;

	if(!(fp = vfs_open(psfile, "rb"))) {
		Error("failed to open shader file %s: %s\n", psfile, strerror(errno));
		free(vsrc);
		return 0;
	}
	vfs_seek(fp, 0, VFS_SEEK_END);
	fsize = vfs_tell(fp);
	vfs_seek(fp, 0, VFS_SEEK_SET);

	if(!(psrc = (char*)malloc(fsize + 1))) {
		Error("failed to allocate %ld bytes: %s\n", fsize + 1, strerror(errno));
		vfs_close(fp);
		free(vsrc);
		return 0;
	}
	if(!vfs_read(psrc, 1, fsize, fp)) {
		Error("failed to load shader file %s: %s\n", psfile, strerror(errno));
		vfs_close(fp);
		free(vsrc);
		free(psrc);
		return 0;
	}
	vfs_close(fp);

	psrc[fsize] = 0;

	prog = ActualCreateGPUProgram(vsrc, psrc, vsfile, psfile);

	free(vsrc);
	free(psrc);
	return prog;
}


void ReleaseGPUProgram (unsigned int id)
{
	gpu_progs.Remove (id);
	if (current_gpu_id == id)
	{
		current_gpu_id = 0;
		current_gpu_uid = 0;
	}
}

bool BindGPUProgram(unsigned int id)
{
	if(id == 0) {
		current_gpu_uid = 0;
		current_gpu_id = 0;
#ifdef BUILD_DX9
		cgD3D9BindProgram(0);
		d3d_device->SetPixelShader(0);
		d3d_device->SetVertexShader(0);
#elif defined(BUILD_GL)
		cgGLBindProgram(0);
		cgGLDisableProfile(vs_prof);
		cgGLDisableProfile(ps_prof);
#endif
		return true;
	}

	if (!gpu_progs.IsHandle (id))
	{
		return false;
	}

	if (current_gpu_uid != gpu_progs.GetUID(id))
	{
		current_gpu_uid = gpu_progs.GetUID(id);
		current_gpu_id = id;
		GPUProgram *prog = (GPUProgram*)gpu_progs.Get(id);
#ifdef BUILD_DX9
		if(cgD3D9BindProgram(prog->vs) != D3D_OK || cgD3D9BindProgram(prog->ps) != D3D_OK) {
			return false;
		}
#elif defined(BUILD_GL)
		cgGLBindProgram(prog->vs);
		cgGLBindProgram(prog->ps);

		cgGLEnableProfile(vs_prof);
		cgGLEnableProfile(ps_prof);
#endif
	}

	return BindUniforms ();
}

static void SetCgUniform(CGparameter param, unsigned int uid)
{
	float *data = (float*)GetUniformDataPtr(uid);
#ifdef BUILD_DX9
	cgD3D9SetUniform(param, data);
#elif defined(BUILD_GL)
	CGGL_API void CGGLENTRY (*set_float_uniform[])(CGparameter, const float*) = {
		0,
		cgGLSetParameter1fv,
		cgGLSetParameter2fv,
		cgGLSetParameter3fv,
		cgGLSetParameter4fv
	};

	unsigned int float_num = GetUniformFloats(uid);
	if(float_num <= 4) {
		set_float_uniform[float_num](param, data);
	} else {
		// XXX: I thing though, that we mostly use 3x4 matrices, does this expect 4x4? test it.
		cgGLSetMatrixParameterfc(param, data);
	}
#endif
}


bool BindUniforms ()
{
	if (!gpu_progs.IsHandle (current_gpu_id))
	{
		return false;
	}
	GPUProgram *prog = (GPUProgram*)gpu_progs.Get(current_gpu_id);

	for(size_t i = 0; i < prog->vs_cached_uniforms.size(); i++) {
		SetCgUniform(prog->vs_cached_uniform_params[i], prog->vs_cached_uniforms[i]);
	}
	for(size_t i = 0; i < prog->ps_cached_uniforms.size(); i++) {
		SetCgUniform(prog->ps_cached_uniform_params[i], prog->ps_cached_uniforms[i]);
	}
	return true;
}

void DumpUniforms(const char *filename)
{
#ifdef BUILD_DX9
	FILE *fp = fopen(filename, "w+");
	if (!fp) return;
	if (!gpu_progs.IsHandle (current_gpu_id))
	{
		return;
	}
	GPUProgram *prog = (GPUProgram *) gpu_progs.Get (current_gpu_id);

	fprintf(fp, "Vertex shader cached uniforms\n---------------\n");
	for (unsigned int i = 0; i < prog->vs_cached_uniforms.size(); i++)
	{
		fprintf(fp, "%s :", GetUniformName (prog->vs_cached_uniforms[i]));

		for (unsigned int j=0; j<GetUniformFloats(prog->vs_cached_uniforms[i]); j++)
		{
			if (!(j%4)) fprintf(fp, "\n");
			float x = ((float*)GetUniformDataPtr(prog->vs_cached_uniforms[i]))[j];
			fprintf(fp, "%f ", x);
		}
		fprintf(fp, "\n\n");
		cgD3D9SetUniform(prog->vs_cached_uniform_params[i],
			GetUniformDataPtr (prog->vs_cached_uniforms[i]));
	}
	for (unsigned int i = 0; i < prog->ps_cached_uniforms.size(); i++)
	{
		cgD3D9SetUniform(prog->ps_cached_uniform_params[i],
			GetUniformDataPtr (prog->ps_cached_uniforms[i]));
	}
	fclose(fp);
#endif
}
