#include <string.h>
#include "idman.h"

#define INIT_RESOURCES_COUNT 32


IdMan::IdMan (void (*destructor) (void*))
{
	uid = 1;
	this->destructor = destructor;
	resources.resize (INIT_RESOURCES_COUNT);
	uids.resize (INIT_RESOURCES_COUNT);
	for (unsigned int i = 0; i < resources.size(); i++)
	{
		resources[i] = 0;
	}
}

IdMan::~IdMan ()
{
}

// the second argument specifies a desired id for this resource...
// if it's unavailable returns 0, if it's 0 then we don't care for a specific id.
unsigned int IdMan::Add(void *resource, unsigned int id)
{
	if(id) {	// specific id requested
		size_t cur_size = resources.size();
		if(id >= cur_size) {
			try {
				resources.resize(id + 1);
				uids.resize(id + 1);
			}
			catch(...) {
				return 0;
			}
			// mark the new slots unused
			memset(&resources[cur_size], 0, (resources.size() - cur_size) * sizeof(void*));
		}

		if(resources[id] == 0) {
			resources[id] = resource;
			uids[id] = uid++;
			return id;
		} else {
			return 0;
		}
	}

	for(size_t i=1; i<resources.size(); i++) {
		if(resources[i] == 0) {
			resources[i] = resource;
			uids[i] = uid++;
			return (unsigned int)i;
		}
	}
	resources.push_back(resource);
	uids.push_back(uid++);
	return (unsigned int)resources.size() - 1;
}

void IdMan::Remove (unsigned int id)
{
	if (IsHandle (id))
	{
		if (destructor)
		{
			destructor(resources[id]);
		}
		resources[id] = 0;
	}
}

void *IdMan::Get (unsigned int id) const
{
	if (IsHandle (id))
	{
		return resources[id];
	}
	return 0;
}

bool IdMan::IsHandle (unsigned int id) const
{
	if ((id >= resources.size()) || (resources[id] == 0))
	{
		return false;
	}
	return true;
}

void IdMan::Clear ()
{
	for (unsigned int i = 1; i < resources.size(); i++)
	{
		Remove (i);
	}
	uids.resize (INIT_RESOURCES_COUNT);
	resources.resize (INIT_RESOURCES_COUNT);
	//uid = 1;
}

unsigned int IdMan::GetUID (unsigned int id) const
{
	if (IsHandle(id))
	{
		return uids[id];
	}
	return 0;
}

unsigned int IdMan::GetResCount () const
{
	return (unsigned int) uids.size();
}


IdManRef::IdManRef (void (*destructor) (void*)) : IdMan (destructor)
{
	refs.resize (INIT_RESOURCES_COUNT);
}

IdManRef::~IdManRef ()
{
}

unsigned int IdManRef::Add(void *resource, unsigned int id)
{
	unsigned int idx = IdMan::Add(resource, id);
	if (idx >= refs.size())
	{
		refs.push_back(1);
	}
	else
	{
		refs[idx]++;
	}
	return idx;
}

void IdManRef::Remove (unsigned int id)
{
	if (IsHandle(id))
	{
		if (--refs[id] == 0)
		{
			IdMan::Remove (id);
		}
	}
}

void IdManRef::Clear ()
{
	IdMan::Clear();
	refs.resize (INIT_RESOURCES_COUNT);
}


void IdManRef::AddRef (unsigned int id)
{
	if (IsHandle (id))
	{
		refs[id]++;
	}
}

