// D3D OpenGL-ification, hurray!
// aka. how to make Defect3D usable.

#include <stack>
#include "gctypes.h"
#include "gc.h"
#include "textures.h"
#include "gpuprog.h"
#include "immed.h"
#include "geometry.h"
#include "vbuffer.h"

using namespace std;

static void SetFixedShader();


static stack<Matrix4x4> mat_stack[IM_MATRIX_MODE_COUNT];
static int mat_mode;
static bool set_xform = true;


/* Possible optimization : when ImBegin,
 * do not start from the begining of the array,
 * but from current position, so that the lock on the 
 * vertex buffer won't occur on old data(possibly being rendered)
 * AND, maybe increase IMM_VERTEX_COUNT
 */

#define IMM_VERTEX_COUNT   5000

static Geometry geometry;				// the geometry for vb update
static unsigned int v_buffer = 0;		// the immed vertex buffer
static int ff_v_buffer = 0;				// the fixed function vertex buffer

static VertexArray *va = 0;				// vertex array of geometry
static NormalArray *na = 0;				// normal array of geometry
static TexCoordArray *ta[3] = {0,0,0};	// texcoord arrays of geometry
static ColorArray *ca = 0;				// color array of geometry
static TangentArray *tang = 0;			// tangent array of geometry

static Vector3 *va_ptr = 0;				// vertex ptr of vertex array
static Vector3 *na_ptr = 0;				// normal ptr of normal array
static Vector2 *ta_ptr[3] = {0, 0, 0};	// texcoord ptrs of texcoord arrays
static unsigned int *ca_ptr = 0;		// color ptr of color array
static Vector3 *tang_ptr = 0;			// tangent ptr of tangent array

static  int imm_vertex_count = 0;		// current vertex count 
static ImPrimitiveType current_pt;		// current primitive type

										// these values dont change between end / begin
static Vector3 current_normal;			// current normal
static Vector3 current_tangent;			// current tangent
static unsigned int current_color;		// current color
static Vector2 current_tex_coord[3];	// current tex coords

static float curr_billboard_size = 1.0f;
static Matrix4x4 curr_billboard_matrix;

static bool fixed_function = false;


bool ImInit()
{
	for(int i=0; i<IM_MATRIX_MODE_COUNT; i++) {
		mat_stack[i].push(Matrix4x4::identity);
	}
	mat_mode = IM_MODEL;

	va = geometry.CreateVertexArray();
	na = geometry.CreateNormalArray();
	ta[0] = geometry.CreateTexCoordArray(0);
	ta[1] = geometry.CreateTexCoordArray(1);
	ta[2] = geometry.CreateTexCoordArray(2);
	ca = geometry.CreateColorArray();
	tang = geometry.CreateTangentArray();

	va_ptr = (Vector3*)va->Create(IMM_VERTEX_COUNT);
	na_ptr = (Vector3*)na->Create(IMM_VERTEX_COUNT);
	ta_ptr[0] = (Vector2*)ta[0]->Create(IMM_VERTEX_COUNT);
	ta_ptr[1] = (Vector2*)ta[1]->Create(IMM_VERTEX_COUNT);
	ta_ptr[2] = (Vector2*)ta[2]->Create(IMM_VERTEX_COUNT);
	ca_ptr = (unsigned int*)ca->Create(IMM_VERTEX_COUNT);
	tang_ptr = (Vector3*)tang->Create(IMM_VERTEX_COUNT);

	v_buffer = CreateVertexBuffer(geometry.GetVertexFormat(), IMM_VERTEX_COUNT, true);

	if(v_buffer == 0) {
		return false;
	}

	VertexFormat fvf = geometry.GetVertexFormat();
	fvf.tangent = false;
	ff_v_buffer = CreateVertexBuffer(fvf, IMM_VERTEX_COUNT, true);

	if(ff_v_buffer == 0) {
		return false;
	}

	current_normal = Vector3(0.0f, 1.0f, 0.0f);
	current_tangent = Vector3(1.0f, 0.0f, 0.0f);
	current_color = 0xffffffff;
	current_tex_coord[0] = Vector2(0.0f, 0.0f);
	current_tex_coord[1] = Vector2(0.0f, 0.0f);
	current_tex_coord[2] = Vector2(0.0f, 0.0f);

	ImFixedFunction ();

	return true;
}

void ImDestroy()
{
	if(v_buffer) {
		DeleteVertexBuffer(v_buffer);
		v_buffer = 0;
	}
	geometry.Release();
}

void ImFixedFunction (bool ff)
{
	fixed_function = ff;
}

bool ImFixedFunctionUsed()
{
	return fixed_function;
}

PrimitiveType ImGetPrimitiveType()
{
	PrimitiveType primitive_type = PT_TRIANGLE_LIST;

	switch(current_pt) {
	case IM_POINTS:
		primitive_type = PT_POINT_LIST;
		break;
	
	case IM_LINES:
		primitive_type = PT_LINE_LIST;
		break;

	case IM_LINE_STRIP:
		primitive_type = PT_LINE_STRIP;
		break;

	case IM_TRIANGLES:
		primitive_type = PT_TRIANGLE_LIST;
		break;

	case IM_TRIANGLE_STRIP:
		primitive_type = PT_TRIANGLE_STRIP;
		break;

	case IM_TRIANGLE_FAN:
		primitive_type = PT_TRIANGLE_FAN;
		break;

	case IM_BILLBOARDS:
	case IM_QUADS:
		primitive_type = PT_TRIANGLE_LIST;
		break;
	}
	return primitive_type;
}

unsigned int ImGetPrimitiveCount()
{
	unsigned int primitive_count = 0;
	switch(current_pt) {
	case IM_POINTS:
		primitive_count = imm_vertex_count;
		break;

	case IM_LINES:
		primitive_count = imm_vertex_count / 2;
		break;

	case IM_LINE_STRIP:
		primitive_count = imm_vertex_count == 0 ? 0 : imm_vertex_count - 1;
		break;

	case IM_TRIANGLES:
		primitive_count = imm_vertex_count / 3;
		break;

	case IM_TRIANGLE_STRIP:
		primitive_count = imm_vertex_count < 3 ? 0 : imm_vertex_count - 2;
		break;

	case IM_TRIANGLE_FAN:
		primitive_count = imm_vertex_count < 3 ? 0 : imm_vertex_count - 2;
		break;

	case IM_BILLBOARDS:
	case IM_QUADS:
		primitive_count = imm_vertex_count < 6 ? 0 :(imm_vertex_count / 6) * 2;
		break;
	}
	return primitive_count;
}


void ImRender()
{
	unsigned int primitive_count = ImGetPrimitiveCount();
	PrimitiveType primitive_type = ImGetPrimitiveType();

	if(primitive_count > 0)
	{
		if (fixed_function)
		{
			if(UpdateVertexBuffer(ff_v_buffer, 0, &geometry, imm_vertex_count))
			{
				BindVertexBuffer(ff_v_buffer);
			}
		}
		else
		{
			if(UpdateVertexBuffer(v_buffer, 0, &geometry, imm_vertex_count))
			{
				BindVertexBuffer(v_buffer);
			}
		}
		RenderPrimitives(primitive_type, 0, primitive_count);
	}
}

void ImBegin(ImPrimitiveType ptype)
{
	if(fixed_function) {
		BindGPUProgram(0);
	}

	if (current_pt == IM_BILLBOARDS)
	{
		curr_billboard_matrix = GetInvViewMatrix();
		curr_billboard_matrix.SetRowVector(Vector3(0, 0, 0), 3);
	}

	if (set_xform)
	{
		ImApplyXForm();
	}

	current_pt = ptype;
	imm_vertex_count = 0;
}

void ImEnd()
{
	ImRender();
	imm_vertex_count = 0; // to prevent multiple ImEnd();
}

void ImCopyVertex(unsigned int dst, unsigned int src)
{
	va_ptr[dst] = va_ptr [src];
	na_ptr[dst] = na_ptr [src];
	ca_ptr[dst] = ca_ptr [src];
	ta_ptr[0][dst] = ta_ptr[0][src];
	ta_ptr[1][dst] = ta_ptr[1][src];
	ta_ptr[2][dst] = ta_ptr[2][src];
	tang_ptr[dst] = tang_ptr[src];
}

void ImReset()
{
	switch(current_pt)
	{
	case IM_POINTS:
		imm_vertex_count = 0; 
		break;

	case IM_LINES:
		if(imm_vertex_count % 2 != 0) {
			ImCopyVertex(0, imm_vertex_count-1); 
			imm_vertex_count = 1;
		} else {
			imm_vertex_count = 0;
		}
		break;

	case IM_LINE_STRIP:
		ImCopyVertex(0, imm_vertex_count-1); 
		imm_vertex_count = 1;
		break;

	case IM_TRIANGLES:
		if(imm_vertex_count % 3 == 0) {
			imm_vertex_count = 0;
		}
		if(imm_vertex_count % 3 == 1) {
			ImCopyVertex(0, imm_vertex_count-1);
			imm_vertex_count = 1;
		}
		if(imm_vertex_count % 3 == 2) {
			ImCopyVertex(0, imm_vertex_count-2);
			ImCopyVertex(1, imm_vertex_count-1);
			imm_vertex_count = 2;
		}
		break;

	case IM_TRIANGLE_STRIP:
		ImCopyVertex(0, imm_vertex_count - 2);
		ImCopyVertex(1, imm_vertex_count - 1);
		imm_vertex_count = 2;
		break;

	case IM_TRIANGLE_FAN:
		 // vertex0 remains the same...
		ImCopyVertex(1, imm_vertex_count - 1);
		imm_vertex_count = 2;
		break;

	case IM_BILLBOARDS:
	case IM_QUADS:
		if(imm_vertex_count % 6 == 0) {
			imm_vertex_count = 0;
		}
		if(imm_vertex_count % 6 == 1) {
			ImCopyVertex(0, imm_vertex_count-1);
			imm_vertex_count = 1;
		}
		if(imm_vertex_count % 6 == 2) {
			ImCopyVertex(0, imm_vertex_count-2);
			ImCopyVertex(1, imm_vertex_count-1);
			imm_vertex_count = 2;
		}
		if(imm_vertex_count % 6 == 3) {
			ImCopyVertex(0, imm_vertex_count-3);
			ImCopyVertex(1, imm_vertex_count-2);
			ImCopyVertex(2, imm_vertex_count-1);
			imm_vertex_count = 3;
		}
		if(imm_vertex_count % 6 == 4) {
			ImCopyVertex(0, imm_vertex_count-4);
			ImCopyVertex(1, imm_vertex_count-3);
			ImCopyVertex(2, imm_vertex_count-2);
			ImCopyVertex(3, imm_vertex_count-1);
			imm_vertex_count = 4;
		}
		if(imm_vertex_count % 6 == 5) {
			ImCopyVertex(0, imm_vertex_count-5);
			ImCopyVertex(1, imm_vertex_count-4);
			ImCopyVertex(2, imm_vertex_count-3);
			ImCopyVertex(3, imm_vertex_count-2);
			ImCopyVertex(4, imm_vertex_count-1);
			imm_vertex_count = 5;
		}
		break;
	}
}

static Vector3 bill_verts[4] = {
	Vector3(-0.5f, 0.5f, 0.0f),
	Vector3(0.5f, 0.5f, 0.0f),
	Vector3(0.5f, -0.5f, 0.0f),
	Vector3(-0.5f, -0.5f, 0.0f)
};

void ImVertex(float x, float y, float z)
{
	if (current_pt == IM_BILLBOARDS)
	{
		current_pt = IM_QUADS;

		curr_billboard_matrix.SetColumnVector(Vector3(x, y, z), 3);

		ImTexCoord(0, 0.0f, 0.0f);
		ImVertex((curr_billboard_size * bill_verts[0]).Transformed(curr_billboard_matrix));

		ImTexCoord(0, 1.0f, 0.0f);
		ImVertex((curr_billboard_size * bill_verts[1]).Transformed(curr_billboard_matrix));

		ImTexCoord(0, 1.0f, 1.0f);
		ImVertex((curr_billboard_size * bill_verts[2]).Transformed(curr_billboard_matrix));

		ImTexCoord(0, 0.0f, 1.0f);
		ImVertex((curr_billboard_size * bill_verts[3]).Transformed(curr_billboard_matrix));

		current_pt = IM_BILLBOARDS;
		return;
	}

	va_ptr[imm_vertex_count].x = x;
	va_ptr[imm_vertex_count].y = y;
	va_ptr[imm_vertex_count].z = z;

	na_ptr[imm_vertex_count] = current_normal;
	ca_ptr[imm_vertex_count] = current_color;
	ta_ptr[0][imm_vertex_count] = current_tex_coord[0];
	ta_ptr[1][imm_vertex_count] = current_tex_coord[1];
	ta_ptr[2][imm_vertex_count] = current_tex_coord[2];
	tang_ptr[imm_vertex_count] = current_tangent;
	
	imm_vertex_count++;

	if(current_pt == IM_QUADS)
	{
		if((imm_vertex_count % 3 == 0) &&(imm_vertex_count % 6 != 0))
		{
			ImCopyVertex(imm_vertex_count, imm_vertex_count - 3);
			ImCopyVertex(imm_vertex_count+1, imm_vertex_count-1);
			imm_vertex_count += 2;
		}
	}

	if(imm_vertex_count >= IMM_VERTEX_COUNT - 6)
	{
		ImRender();
		ImReset();
	}
}

void ImVertex(const Vector3 &v)
{
	ImVertex(v.x, v.y, v.z);
}

void ImNormal(float x, float y, float z)
{
	current_normal.x = x;
	current_normal.y = y;
	current_normal.z = z;
}

void ImNormal(const Vector3 &v)
{
	current_normal.x = v.x;
	current_normal.y = v.y;
	current_normal.z = v.z;
}

void ImColor(float r, float g, float b, float a)
{
	current_color = PACKF_RGBA(r,g,b,a);
}

void ImColor(const Color &c)
{
	current_color = PACKF_RGBA(c.r, c.g, c.b, c.a);
}

void ImTexCoord(unsigned int tunit, float u, float v)
{
	if(tunit < 3)
	{
		current_tex_coord[tunit].x = u;
		current_tex_coord[tunit].y = v;
	}
}

void ImTexCoord(float u, float v)
{
	ImTexCoord (0, u, v);
}

void ImTexCoord(unsigned int tunit, const Vector2 &v)
{					  
	if(tunit < 3)
	{
		current_tex_coord[tunit] = v;
	}
}

void ImTangent(float x, float y, float z)
{
	current_tangent.x = x;
	current_tangent.y = y;
	current_tangent.z = z;
}

void ImTangent(const Vector3 &tangent)
{
	current_tangent = tangent;
}

void ImBillboardSize(float s)
{
	curr_billboard_size = s;
}

void ImMatrixMode(MatrixMode mode)
{
	mat_mode = mode;
}

void ImLoadIdentity()
{
	mat_stack[mat_mode].top() = Matrix4x4::identity;
}

void ImLoadMatrix(const Matrix4x4 &m)
{
	mat_stack[mat_mode].top() = m;
}

void ImMultMatrix(const Matrix4x4 &m)
{
	mat_stack[mat_mode].top() *= m;
}

void ImPushMatrix()
{
	mat_stack[mat_mode].push(mat_stack[mat_mode].top());
}

void ImPopMatrix()
{
	mat_stack[mat_mode].pop();
}

void ImTranslate(float x, float y, float z)
{
	Matrix4x4 m;
	m.SetTranslation(Vector3(x, y, z));
	ImMultMatrix(m);
}

void ImRotate(float angle, float x, float y, float z)
{
	Matrix4x4 m;
	m.SetRotation(Vector3(x, y, z), DEG_TO_RAD(angle));
	ImMultMatrix(m);
}

void ImScale(float x, float y, float z)
{
	Matrix4x4 m;
	m.SetScaling(Vector3(x, y, z));
	ImMultMatrix(m);
}

void ImOverlay(const Vector2 &v1, const Vector2 &v2, const Color &col, unsigned int tex, unsigned int gfxprog)
{
	/*
	float l = v1.x * 2.0f - 1.0f;
	float r = v2.x * 2.0f - 1.0f;
	float u = -v1.y * 2.0f + 1.0f;
	float d = -v2.y * 2.0f + 1.0f;
	*/

	Matrix4x4 proj_mat = GetProjectionMatrix();
	Matrix4x4 model_mat = GetModelMatrix();
	Matrix4x4 view_mat = GetViewMatrix();
	
	// setup orthographic projection
	ImMatrixMode(IM_PROJECTION);
	ImPushMatrix();
	ImLoadMatrix(OrthoProj(0, 1, 0, 1, 0, 10));
	ImMatrixMode(IM_VIEW);
	ImPushMatrix();
	ImLoadIdentity();
	ImMatrixMode(IM_MODEL);
	ImPushMatrix();
	ImLoadIdentity();

	if(tex != IM_TEX_CUR) {
		BindTexture(0, tex);
	}
	if(fixed_function) {
		gfxprog = 0;
	}

	ImMatrixMode(IM_TEXTURE);
	ImPushMatrix();
	ImLoadMatrix(GetTextureMatrix());

	if(gfxprog != IM_SDR_CUR) {
		BindGPUProgram(gfxprog);
		BindUniforms();
	}

	bool prev_tex_xform = GetTexTransform();
	//if(!gfxprog) {
		SetTexTransform(true);
	//}

	bool prev_depth_test = GetDepthTest();
	CullFace prev_cull_mode = GetCulling();
	bool prev_lighting = GetLighting();
	bool blending = GetBlending();

	SetDepthTest(false);
	SetCulling(CULL_NONE);
	SetLighting(false);

	if(col.a < 1.0 - SmallNumber) {
		SetBlending(true);
		SetBlendFunc(BLEND_SRC_ALPHA, BLEND_ONE_MINUS_SRC_ALPHA);
	}

	// draw quad
	ImBegin(IM_QUADS);
	{
		ImColor(col);
		
		ImTexCoord(0, 0);
		ImVertex(v1.x, v1.y, 1);
		
		ImTexCoord(1, 0);
		ImVertex(v2.x, v1.y, 1);
		
		ImTexCoord(1, 1);
		ImVertex(v2.x, v2.y, 1);
		
		ImTexCoord(0, 1);
		ImVertex(v1.x, v2.y, 1);
	}
	ImEnd();

	if(col.a < 1.0 - SmallNumber) {
		SetBlending(blending);
	}

	if(tex != IM_TEX_CUR) {
		BindTexture(0, 0);
	}

	SetLighting(prev_lighting);
	SetCulling(prev_cull_mode);
	SetDepthTest(prev_depth_test);

	ImMatrixMode(IM_TEXTURE);
	ImPopMatrix();

	//if(!gfxprog) {
		SetTexTransform(prev_tex_xform);
	//}

	ImMatrixMode(IM_PROJECTION);
	ImPopMatrix();
	ImMatrixMode(IM_VIEW);
	ImPopMatrix();
	ImMatrixMode(IM_MODEL);
	ImPopMatrix();

	SetProjectionMatrix(proj_mat);
	SetViewMatrix(view_mat);
	SetModelMatrix(model_mat);
}

void ImSetApplyXForm(bool enable)
{
	set_xform = enable;
}

void ImApplyXForm()
{
	SetModelMatrix(mat_stack[IM_MODEL].top());
	SetViewMatrix(mat_stack[IM_VIEW].top());
	SetProjectionMatrix(mat_stack[IM_PROJECTION].top());
	SetTextureMatrix(0, mat_stack[IM_TEXTURE].top());
	BindUniforms();
}

void ImWireBox(const Box &box)
{
	for (unsigned int i=0; i<6; i++)
	{
		Quad q = box.GetQuad(i);
		for (unsigned int j=0; j<4; j++)
		{
			Line l = q.GetLine(j);
			ImVertex(l.GetVertex(0));
			ImVertex(l.GetVertex(1));
		}
	}
}

void ImQuad(const Quad &quad)
{
	for (unsigned int i=0; i<4; i++)
	{
		ImVertex(quad.GetVertex(i));
	}
}


/* ------- random helpers ------- */

void ImGrid(float spacing, float sz)
{
	float hsz = sz * 0.5;
	int lines = (int)(hsz / spacing) + 1;

	bool prev_lighting = GetLighting();
	
	SetLighting(false);

	ImBegin(IM_LINES);
	ImColor(Color(0.4, 0.4, 0.4));
	for(int i=1; i<lines; i++) {
		ImVertex(i * spacing, 0, -hsz);
		ImVertex(i * spacing, 0, hsz);
		ImVertex(-i * spacing, 0, -hsz);
		ImVertex(-i * spacing, 0, hsz);
		ImVertex(-hsz, 0, i * spacing);
		ImVertex(hsz, 0, i * spacing);
		ImVertex(-hsz, 0, -i * spacing);
		ImVertex(hsz, 0, -i * spacing);
	}
	ImEnd();

	ImBegin(IM_LINES);
	ImColor(Color(0.6, 0.6, 0.6));
	ImVertex(-hsz, 0, 0);
	ImVertex(hsz, 0, 0);
	ImVertex(0, 0, -hsz);
	ImVertex(0, 0, hsz);
	ImEnd();

	SetLighting(prev_lighting);
}
