#include "lighting.h"
#include "statecache.h"
#include "uniforms.h"
#include "gc.h"

void SetLightAmbient (unsigned int lid, const Color &color)
{
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_AMBIENT, (float*) &color); break;
	case 1 : SetUniform (STATE_LIGHT1_AMBIENT, (float*) &color); break;
	case 2 : SetUniform (STATE_LIGHT2_AMBIENT, (float*) &color); break;
	case 3 : SetUniform (STATE_LIGHT3_AMBIENT, (float*) &color); break;
	}
}

void SetLightDiffuse (unsigned int lid, const Color &color)
{
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_DIFFUSE, (float*) &color); break;
	case 1 : SetUniform (STATE_LIGHT1_DIFFUSE, (float*) &color); break;
	case 2 : SetUniform (STATE_LIGHT2_DIFFUSE, (float*) &color); break;
	case 3 : SetUniform (STATE_LIGHT3_DIFFUSE, (float*) &color); break;
	}
}

void SetLightSpecular (unsigned int lid, const Color &color)
{
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_SPECULAR, (float*) &color); break;
	case 1 : SetUniform (STATE_LIGHT1_SPECULAR, (float*) &color); break;
	case 2 : SetUniform (STATE_LIGHT2_SPECULAR, (float*) &color); break;
	case 3 : SetUniform (STATE_LIGHT3_SPECULAR, (float*) &color); break;
	}
}


void SetLightPos (unsigned int lid, const Vector3 &pos)
{
	Vector3 vpos = pos;
	vpos.Transform (GetViewMatrix());
	switch (lid)
	{
	case 0 : 
		SetUniform (STATE_LIGHT0_POS, (float*) &vpos); 
		SetUniform (STATE_LIGHT0_WPOS, (float*) &pos);
		break;

	case 1 : 
		SetUniform (STATE_LIGHT1_POS, (float*) &vpos); 
		SetUniform (STATE_LIGHT1_WPOS, (float*) &pos);
		break;

	case 2 : 
		SetUniform (STATE_LIGHT2_POS, (float*) &vpos); 
		SetUniform (STATE_LIGHT2_WPOS, (float*) &pos);
		break;

	case 3 : 
		SetUniform (STATE_LIGHT3_POS, (float*) &vpos); 
		SetUniform (STATE_LIGHT3_WPOS, (float*) &pos);
		break;
	}
}

void SetLightDir (unsigned int lid, const Vector3 &dir)
{
	Vector3 vpos = dir;
	vpos.Transform ((Matrix3x3)GetViewMatrix());
	vpos.Normalize();
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_DIR, (float*) &vpos); break;
	case 1 : SetUniform (STATE_LIGHT1_DIR, (float*) &vpos); break;
	case 2 : SetUniform (STATE_LIGHT2_DIR, (float*) &vpos); break;
	case 3 : SetUniform (STATE_LIGHT3_DIR, (float*) &vpos); break;
	}
}

void SetLightInner (unsigned int lid, float inner)
{
	float inner_cos = cos (DEG_TO_RAD(inner/2.0f));
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_INNER, (float*) &inner_cos); break;
	case 1 : SetUniform (STATE_LIGHT1_INNER, (float*) &inner_cos); break;
	case 2 : SetUniform (STATE_LIGHT2_INNER, (float*) &inner_cos); break;
	case 3 : SetUniform (STATE_LIGHT3_INNER, (float*) &inner_cos); break;
	}
}

void SetLightOuter (unsigned int lid, float outer)
{
	float outer_cos = cos (DEG_TO_RAD(outer/2.0f));
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_OUTER, (float*) &outer_cos); break;
	case 1 : SetUniform (STATE_LIGHT1_OUTER, (float*) &outer_cos); break;
	case 2 : SetUniform (STATE_LIGHT2_OUTER, (float*) &outer_cos); break;
	case 3 : SetUniform (STATE_LIGHT3_OUTER, (float*) &outer_cos); break;
	}
}

void SetLightAttenStart (unsigned int lid, float start)
{
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_ATTEN_START, (float*) &start); break;
	case 1 : SetUniform (STATE_LIGHT1_ATTEN_START, (float*) &start); break;
	case 2 : SetUniform (STATE_LIGHT2_ATTEN_START, (float*) &start); break;
	case 3 : SetUniform (STATE_LIGHT3_ATTEN_START, (float*) &start); break;
	}
}

void SetLightAttenEnd (unsigned int lid, float end)
{
	switch (lid)
	{
	case 0 : SetUniform (STATE_LIGHT0_ATTEN_END, (float*) &end); break;
	case 1 : SetUniform (STATE_LIGHT1_ATTEN_END, (float*) &end); break;
	case 2 : SetUniform (STATE_LIGHT2_ATTEN_END, (float*) &end); break;
	case 3 : SetUniform (STATE_LIGHT3_ATTEN_END, (float*) &end); break;
	}
}

void SetMatAmbient (const Color &color)
{
	SetUniform (STATE_MATERIAL_AMBIENT, (float*) &color);
}

void SetMatDiffuse (const Color &color)
{
	SetUniform (STATE_MATERIAL_DIFFUSE, (float*) &color);
}

void SetMatSpecular (const Color &color)
{
	SetUniform (STATE_MATERIAL_SPECULAR, (float*) &color);
}

void SetMatEmissive (const Color &color)
{
	SetUniform (STATE_MATERIAL_EMISSIVE, (float*) &color);
}

void SetMatPower (float power)
{
	SetUniform (STATE_MATERIAL_POWER, &power);
}
		


Material::Material ()
{
	ambient = Color (0.2f, 0.2f, 0.2f, 1.0f);
	diffuse = Color (0.5f, 0.5f, 0.5f, 1.0f);
	specular = Color (0.9f, 0.9f, 0.9f, 1.0f);
	emissive = Color (0.0f, 0.0f, 0.0f, 0.0f);
	power = 128.0f;
}

bool Material::SaveToFP (VFS_FILE *fp)
{
	size_t bytes_write;

	try
	{
		// write the AMBIENT components...
		bytes_write = vfs_write (&ambient.r, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&ambient.g, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&ambient.b, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&ambient.a, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		// write the DIFFUSE components...
		bytes_write = vfs_write (&diffuse.r, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&diffuse.g, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&diffuse.b, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&diffuse.a, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		// write the SPECULAR components...
		bytes_write = vfs_write (&specular.r, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&specular.g, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&specular.b, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&specular.a, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		// write the EMISSIVE components...
		bytes_write = vfs_write (&emissive.r, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&emissive.g, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&emissive.b, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_write = vfs_write (&emissive.a, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}

		// write the POWER component...
		bytes_write = vfs_write (&power, 1, sizeof (float), fp);
		if (bytes_write != sizeof (float))
		{
			throw (int) 0;
		}
	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool Material::LoadFromFP (VFS_FILE *fp)
{
	size_t bytes_read;

	try
	{
		// write the AMBIENT components...
		bytes_read = vfs_read (&ambient.r, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&ambient.g, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&ambient.b, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&ambient.a, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		// write the DIFFUSE components...
		bytes_read = vfs_read (&diffuse.r, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&diffuse.g, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&diffuse.b, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&diffuse.a, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		// write the SPECULAR components...
		bytes_read = vfs_read (&specular.r, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&specular.g, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&specular.b, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&specular.a, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		// write the EMISSIVE components...
		bytes_read = vfs_read (&emissive.r, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&emissive.g, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&emissive.b, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
		bytes_read = vfs_read (&emissive.a, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}

		// write the POWER component...
		bytes_read = vfs_read (&power, 1, sizeof (float), fp);
		if (bytes_read != sizeof (float))
		{
			throw (int) 0;
		}
	}
	catch (...)
	{
		return false;
	}
	return true;
}


void SetMaterial (const Material &mat)
{
	SetMatAmbient (mat.ambient);
	SetMatDiffuse (mat.diffuse);
	SetMatSpecular (mat.specular);
	SetMatEmissive (mat.emissive);
	SetMatPower (mat.power);
}





Light::Light()
{
	lt = LT_POINT_LIGHT;
	ambient = Color (1.0f, 1.0f, 1.0f, 1.0f);
	diffuse = Color (1.0f, 1.0f, 1.0f, 1.0f);
	specular = Color (1.0f, 1.0f, 1.0f, 1.0f);
	pos = Vector3 (0.0f, 0.0f, 0.0f);
	dir = Vector3 (0.0f, 0.0f, 1.0f);
	inner = 0.0f;
	outer = 0.0f;
	atten_start = 200.0f;
	atten_end = 250.0f;
}

void SetLight (unsigned int lid, const Light &light)
{
	SetLightAmbient (lid, light.ambient);
	SetLightDiffuse (lid, light.diffuse);
	SetLightSpecular (lid, light.specular);
	SetLightPos (lid, light.pos);
	SetLightDir (lid, light.dir);
	SetLightInner (lid, light.inner);
	SetLightOuter (lid, light.outer);
	SetLightAttenStart (lid, light.atten_start);
	SetLightAttenEnd (lid, light.atten_end);
}

