#include <string.h>
#include <errno.h>
#include <assert.h>
#include <ft2build.h>
#include FT_FREETYPE_H
#include "gc.h"
#include "immed.h"
#include "gpuprog.h"
#include "textures.h"
#include "text.h"
#include "idman.h"
#include <errlog.h>
#include <libvfs.h>


#define MAX_CHARS		128
#define MAX_IMG_WIDTH	1024
#define SIZE_PIXELS(x)	((x) / 64)

struct FontStruct {
	unsigned int tex_id;
	float scale;	// this compensates if a higher res font is loaded in its stead, with a new CreateFont
	float line_adv;	// vertical advance to go to the next line
	struct {
		Vector2 pos, size;		// glyph position (from origin) and size in normalized coords [0, 1]
		float advance;			// advance in normalized coords
		Vector2 tc_pos, tc_sz;	// tex coord box pos/size
	} glyphs[MAX_CHARS];
};

static void BlitFontGlyph(FontStruct *fnt, int x, int y, FT_GlyphSlot glyph, unsigned int *img, int xsz, int ysz);
static void CleanUp();
static void DeleteInternalFont(void *ptr);

static FT_Library ft;
static Vector2 text_pos;
static float text_size = 1.0;
static Color text_color;
static FontStruct *act_fnt;

IdMan font_list(DeleteInternalFont);

unsigned int CreateFont(const char *fname, int font_size)
{

	if(!ft) {
		if(FT_Init_FreeType(&ft) != 0) {
			Error("failed to initialize freetype\n");
			return 0;
		}
		atexit(CleanUp);
	}

	struct vfs_stat st;
	if(vfs_stat(fname, &st) == -1) {
		Error("failed to stat font file %s: %s\n", fname, strerror(errno));
		return 0;
	}

	size_t font_mem_size = st.st_size;
	void *font_mem;
	if(!(font_mem = malloc(font_mem_size))) {
		Error("failed to allocate %u bytes: %s\n", (unsigned int)font_mem_size, strerror(errno));
		return 0;
	}

	VFS_FILE *fp;
	if(!(fp = vfs_open(fname, "rb"))) {
		Error("failed to load font %s: %s\n", fname, strerror(errno));
		free(font_mem);
		return 0;
	}
	if(vfs_read(font_mem, 1, font_mem_size, fp) != font_mem_size) {
		Warning("unexpected size of bytes read from font: %s\n", fname);
		free(font_mem);
		vfs_close(fp);
		return 0;
	}
	vfs_close(fp);

	FT_Face face;
	if(FT_New_Memory_Face(ft, (unsigned char*)font_mem, (FT_Long)font_mem_size, 0, &face) != 0) {
		Error("failed to load font: %s\n", fname);
		free(font_mem);
		return 0;
	}

	FT_Set_Pixel_Sizes(face, 0, font_size);

	FontStruct *fnt = new FontStruct;
	int max_width = MAX_CHARS * SIZE_PIXELS(face->bbox.xMax - face->bbox.xMin);
	int xsz = MAX_IMG_WIDTH;
	int ysz = SIZE_PIXELS(face->bbox.yMax - face->bbox.yMin) * max_width / xsz;

	int tex_xsz = NextPow2(xsz);
	int tex_ysz = NextPow2(ysz);

	unsigned int *img;
	try {
        img = new unsigned int[tex_xsz * tex_ysz];
	}
	catch(...) {
		Error("mem alloc failed in CreateFontStruct (font texture image: %dx%d)\n", tex_xsz, tex_ysz);
		return 0;
	}
	memset(img, 0xff, tex_xsz * tex_ysz * sizeof *img);	// TODO: after debugging change this to 0

	int vport_xsz, vport_ysz;
	GetViewport(0, 0, &vport_xsz, &vport_ysz);

	int max_glyph_y = 0;
	int max_glyph_x = 0;
	for(int i=0; i<MAX_CHARS; i++) {
		FT_Load_Char(face, i, 0);
		int width = SIZE_PIXELS(face->glyph->metrics.width);
		int height = SIZE_PIXELS(face->glyph->metrics.height);
		
		if(height > max_glyph_y) {
			max_glyph_y = height;
		}

		if(width > max_glyph_x) {
			max_glyph_x = width;
		}
	}

	int gx = 0, gy = 0;
	for(int i=0; i<MAX_CHARS; i++) {
		FT_Load_Char(face, i, FT_LOAD_RENDER);
		FT_GlyphSlot g = face->glyph;

		int gwidth = SIZE_PIXELS(g->metrics.width);
		int gheight = SIZE_PIXELS(g->metrics.height);

		if(gx > MAX_IMG_WIDTH - gwidth) {
			gx = 0;
			gy += max_glyph_y;
		}

		BlitFontGlyph(fnt, gx, gy, g, img, tex_xsz, tex_ysz);
		fnt->scale = 1.0;
		fnt->line_adv = (float)SIZE_PIXELS(g->metrics.vertAdvance) / (float)vport_ysz;
		fnt->glyphs[i].tc_pos.x = (float)gx / (float)tex_xsz;
		fnt->glyphs[i].tc_pos.y = (float)gy / (float)tex_ysz;
		fnt->glyphs[i].tc_sz.x = (float)gwidth / (float)tex_xsz;
		fnt->glyphs[i].tc_sz.y = (float)gheight / (float)tex_ysz;
		fnt->glyphs[i].size.x = (float)gwidth / (float)vport_xsz;
		fnt->glyphs[i].size.y = (float)gheight / (float)vport_ysz;
		fnt->glyphs[i].pos.x = (float)SIZE_PIXELS(g->metrics.horiBearingX) / (float)vport_xsz;
		fnt->glyphs[i].pos.y = -(float)SIZE_PIXELS(g->metrics.horiBearingY) / (float)vport_ysz;
		fnt->glyphs[i].advance = (float)SIZE_PIXELS(g->metrics.horiAdvance) / (float)vport_xsz;

		gx += gwidth;
	}

	FT_Done_Face(face);
	free(font_mem);

	if(!(fnt->tex_id = CreateTexture(tex_xsz, tex_ysz, TEX_2D, TFMT_32, false))) {
		delete [] img;
		delete fnt;
		return 0;
	}
	SetTexImage(fnt->tex_id, img, tex_xsz, tex_ysz, TFMT_32);
	delete [] img;

	unsigned int id = font_list.Add(fnt);	// this used fname XXXXXXX
	if(!id)
	{
		DeleteTexture(fnt->tex_id);
		delete fnt;
		return 0;
	}

	if(!act_fnt) act_fnt = fnt;
	return id;
}

void DeleteFont(unsigned int fid)
{
	font_list.Remove(fid);
}

unsigned int GetFont(const char *fname, int sz)
{
	unsigned int id = 0;//font_list.GetIndex(fname);
	if(!id) {
		return CreateFont(fname, sz);
	}

	// XXX: if we change the font list to ref-counted, add an AddRef call here.
	return id;
}

bool BindFont(unsigned int fid)
{
	void *tmp = font_list.Get(fid);
	if(!tmp) return false;

	act_fnt = (FontStruct*)tmp;
	return true;
}

void SetTextPos(const Vector2 &pos)
{
	text_pos = pos;
}

Vector2 GetTextPos()
{
	return text_pos;
}

void TextLineAdvance(int adv)
{
	text_pos.y += (float)adv * act_fnt->line_adv;
}

void TextCRet()
{
	text_pos.x = 0.0;
}

void SetTextSize(float sz)
{
	text_size = sz;
}

float GetTextSize()
{
	return text_size;
}

void SetTextColor(const Color &col)
{
	text_color = col;
}

Color GetTextColor()
{
	return text_color;
}

float PrintChar(char c)
{
	// get texture coordinates for the glyph, and construct the texture matrix
	float tx = act_fnt->glyphs[(int)c].tc_pos.x;
	float ty = act_fnt->glyphs[(int)c].tc_pos.y;
	float sx = act_fnt->glyphs[(int)c].tc_sz.x;
	float sy = act_fnt->glyphs[(int)c].tc_sz.y;

	Matrix4x4 mat(	sx, 0, tx, 0,
					0, sy, ty, 0,
					0, 0, 1, 0,
					0, 0, 0, 1);
	
	ImMatrixMode(IM_TEXTURE);
	ImPushMatrix();
	ImLoadMatrix(mat);

	SetTextureFiltering(0, TF_POINT);
	SetBlending(true);
	SetBlendFunc(BLEND_SRC_ALPHA, BLEND_ONE_MINUS_SRC_ALPHA);
		
	Vector2 pos = text_pos + act_fnt->glyphs[(int)c].pos * act_fnt->scale;
	ImOverlay(pos, pos + act_fnt->glyphs[(int)c].size * act_fnt->scale, text_color, act_fnt->tex_id);

	SetBlending(false);
	SetTextureFiltering(0, TF_LINEAR);

	ImMatrixMode(IM_TEXTURE);
	ImPopMatrix();

	return act_fnt->glyphs[(int)c].advance * act_fnt->scale;
}

static TextureFilter prev_min_filt, prev_mag_filt, prev_mip_filt;
static bool prev_blending;
static bool prev_lighting;
static CullFace prev_culling;
static bool prev_depth_test;

// this function contains the preamble of all block text drawing functions (i.e. not printchar above)
static void PreDraw()
{
	ImMatrixMode(IM_PROJECTION);
	ImPushMatrix();
	ImLoadMatrix(OrthoProj(0, 1, 0, 1, 0, 10));
	ImMatrixMode(IM_VIEW);
	ImPushMatrix();
	ImLoadIdentity();
	ImMatrixMode(IM_MODEL);
	ImPushMatrix();
	ImLoadIdentity();
	ImMatrixMode(IM_TEXTURE);
	ImPushMatrix();
	ImLoadIdentity();

	BindTexture(0, act_fnt->tex_id);
	BindGPUProgram(0);

	prev_min_filt = GetTextureFilteringMin(0);
	prev_mag_filt = GetTextureFilteringMag(0);
	prev_mip_filt = GetTextureFilteringMip(0);
	prev_blending = GetBlending();
	prev_depth_test = GetDepthTest();
	prev_culling = GetCulling();
	prev_lighting = GetLighting();

	SetTextureFiltering(0, TF_POINT);
	SetBlending(true);
	SetBlendFunc(BLEND_SRC_ALPHA, BLEND_ONE_MINUS_SRC_ALPHA);
	SetDepthTest(false);
	SetCulling(CULL_NONE);
	SetLighting(false);

	ImBegin(IM_QUADS);
	ImColor(text_color);
}

static void PostDraw()
{
	ImEnd();

	SetLighting(prev_lighting);
	SetCulling(prev_culling);
	SetDepthTest(prev_depth_test);
	SetBlending(prev_blending);
	SetTextureFiltering(0, prev_min_filt, prev_mag_filt, prev_mip_filt);

	ImMatrixMode(IM_TEXTURE);
	ImPopMatrix();
	ImMatrixMode(IM_MODEL);
	ImPopMatrix();
	ImMatrixMode(IM_VIEW);
	ImPopMatrix();
	ImMatrixMode(IM_PROJECTION);
	ImPopMatrix();
}

float PrintString(const char *str, bool standalone)
{
	if(standalone) PreDraw();
	
	float start_x = text_pos.x;
	while(*str) {
		float tx = act_fnt->glyphs[(int)*str].tc_pos.x;
		float ty = act_fnt->glyphs[(int)*str].tc_pos.y;
		float sx = act_fnt->glyphs[(int)*str].tc_sz.x;
		float sy = act_fnt->glyphs[(int)*str].tc_sz.y;

		Vector2 tc1 = Vector2(tx, ty);
		Vector2 tc2 = Vector2(tx + sx, ty + sy);

		Vector2 v1 = text_pos + (act_fnt->glyphs[(int)*str].pos * act_fnt->scale);
		Vector2 v2 = v1 + (act_fnt->glyphs[(int)*str].size * act_fnt->scale);
		float l = v1.x;// * 2.0f - 1.0f;
		float r = v2.x;// * 2.0f - 1.0f;
		float u = v1.y;// * 2.0f + 1.0f;
		float d = v2.y;// * 2.0f + 1.0f;

		ImTexCoord(tc1.x, tc1.y);
		ImVertex(l, u, 1);
		ImTexCoord(tc2.x, tc1.y);
		ImVertex(r, u, 1);
		ImTexCoord(tc2.x, tc2.y);
		ImVertex(r, d, 1);
		ImTexCoord(tc1.x, tc2.y);
		ImVertex(l, d, 1);

		text_pos.x += act_fnt->glyphs[(int)*str++].advance * act_fnt->scale;
	}

	if(standalone) PostDraw();
	return text_pos.x - start_x;
}

void PrintStringLines(const char **str, int lines)
{
	PreDraw();

	while(lines-- > 0) {
		PrintString(*str++, false);
		TextLineAdvance();
		TextCRet();
	}

	PostDraw();
}

static void BlitFontGlyph(FontStruct *fnt, int x, int y, FT_GlyphSlot glyph, unsigned int *img, int xsz, int ysz)
{
	if(glyph->bitmap.pixel_mode != FT_PIXEL_MODE_GRAY) {
		Error("BlitFontGlyph: unsupported pixel mode: %d\n", glyph->bitmap.pixel_mode);
		return;
	}

	unsigned int *dest = img + y * xsz + x;
	unsigned char *src = glyph->bitmap.buffer;

	for(int j=0; j<glyph->bitmap.rows; j++) {
		for(int i=0; i<glyph->bitmap.width; i++) {
			dest[i] = 0x00ffffff | ((unsigned int)src[i] << ASHIFT);
		}
		dest += xsz;
		src += glyph->bitmap.pitch;
	}
}

static void CleanUp()
{
	FT_Done_FreeType(ft);
}


static void DeleteInternalFont(void *ptr)
{
	FontStruct *fnt = (FontStruct*)ptr;
	delete fnt;
}


float GetMaxDescent()
{
	FontStruct *fnt = act_fnt;

	float max_descent = 0.0f;
	
	for(int i=0; i<MAX_CHARS; i++) {
		float des = fnt->glyphs[i].pos.y + fnt->glyphs[i].size.y;
		if(isprint(i) && des > max_descent) {
			max_descent = des;
		}
	}

	return max_descent;
}

float GetLineAdvance()
{
	return act_fnt->line_adv;
}

float GetTextWidth(const char *str)
{
	float width = 0;
	while(*str) {
		width += act_fnt->glyphs[(int)*str++].advance * act_fnt->scale * text_size;
	}
	return width;
}
