#include <stack>
#include <limits.h>
#include <builddefs.h>
#include <math3d.h>
#include <convimg.h>
#include <errlog.h>
#include "textures.h"
#include "idman.h"
#include "gc.h"
#include "caps.h"
#include "sysgc.h"

//TODO : otan kanoume bind ena texture, na tsekaroume an einai hdh bound

#ifdef BUILD_DX9
//typedef IDirect3DTexture9* HwTexture;
typedef IDirect3DBaseTexture9* HwTexture;
#elif defined(BUILD_GL)
typedef unsigned int HwTexture;
#endif

struct Texture {
	int width, height;
	int rt_width, rt_height;	// this is valid for render targets, to keep the "real" (viewport) size
	TextureFormat fmt;
	TextureType type;
	int mip_levels;
	HwTexture hwtex;
	bool need_texmat;
	int rt_cube_face_bound;

#ifdef BUILD_DX9
	IDirect3DSurface9 *ds_surf;	// this should only be non-null if this texture is a render target
#elif defined(BUILD_GL)
	int ds_surf;	// same as above, but only used as a boolean flag
	unsigned int depth_buffer, stencil_buffer, fbo;
#endif
};

static const char *GetFormatName(TextureFormat fmt); 
static void IdManDelete(void *tex);

static Matrix4x4 CalcTexMat(Texture *tex);

static unsigned int current_uid[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static TextureType current_ttype[8] = {TEX_INVALID, TEX_INVALID, TEX_INVALID, TEX_INVALID, TEX_INVALID, TEX_INVALID, TEX_INVALID, TEX_INVALID};
IdMan tex_list(IdManDelete);

std::stack<unsigned int> rtstack[4];


unsigned int AddHWTexture (int w, int h, TextureType type, TextureFormat fmt, bool mip, void *hw_tex)
{
	Texture *tex = new Texture;
	memset (tex, 0, sizeof *tex);

	tex->width = w;
	tex->height = h;
	tex->fmt = fmt;
	tex->type = type;
	tex->mip_levels = mip ? (int)log2((float)MIN(w, h)) + 1 : 1;
#ifdef BUILD_DX9
	tex->hwtex = (HwTexture)hw_tex;
#elif defined(BUILD_GL)
	tex->hwtex = *(unsigned int*)&hw_tex;
#endif
	
	unsigned int id = tex_list.Add(tex);
	return id;
}

// creates an actual texture through the graphics API
static HwTexture CreateHwTexture(int xsz, int ysz, TextureType type, TextureFormat fmt, int mip_levels)
{
#ifdef BUILD_DX9
	HwTexture hwtex;
	if(type == TEX_CUBE) {
		if(d3d_device->CreateCubeTexture(xsz, mip_levels, 0, (D3DFORMAT)fmt, D3DPOOL_MANAGED, (IDirect3DCubeTexture9**)&hwtex, 0) != D3D_OK) {
			return 0;
		}
	} else {
		if(d3d_device->CreateTexture(xsz, ysz, mip_levels, 0, (D3DFORMAT)fmt, D3DPOOL_MANAGED, (IDirect3DTexture9**)&hwtex, 0) != D3D_OK) {
			return 0;
		}
	}
	return hwtex;
#elif defined(BUILD_GL)
	unsigned int tex;
	glGenTextures(1, &tex);
	glBindTexture(type, tex);

	glTexParameteri(type, GL_TEXTURE_MIN_FILTER, mip_levels > 1 ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri(type, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(type, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(type, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexImage2D(type, 0, fmt, xsz, ysz, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	return tex;
#endif
}

// destroyes the underlying API texture
static void DeleteHwTexture(HwTexture tex)
{
#ifdef BUILD_DX9
	tex->Release();
#elif defined(BUILD_GL)
	glDeleteTextures(1, &tex);
#endif
}

// this function is called by the ID manager to delete the texture
static void IdManDelete(void *texptr)
{
	Texture *tex = (Texture*)texptr;

	DeleteHwTexture(tex->hwtex);
	if(tex->ds_surf) {
#ifdef BUILD_DX9
		tex->ds_surf->Release();
#elif defined(BUILD_GL)
		glDeleteRenderbuffersEXT(1, &tex->depth_buffer);
		glDeleteRenderbuffersEXT(1, &tex->stencil_buffer);
		glDeleteFramebuffersEXT(1, &tex->fbo);
#endif
	}
	delete tex;
}

unsigned int CreateTexture(int w, int h, TextureType type, TextureFormat fmt, bool mip, unsigned int req_id)
{
	int xsz = w;
	int ysz = h;

	Texture *tex;
	try {
		tex = new Texture;
		memset(tex, 0, sizeof *tex);
	}
	catch(...) {
		return 0;
	}

	bool ispow2 = IsPow2(w) && IsPow2(h);
	tex->need_texmat = (!GetGcCaps()->npot_textures && !ispow2) ? true : false;
	
	if(tex->need_texmat) {		
		xsz = NextPow2(w);
		ysz = NextPow2(h);
	}

	tex->width = xsz;
	tex->height = ysz;
	tex->rt_width = w;
	tex->rt_height = h;
	tex->fmt = fmt;
	tex->type = type;
	tex->mip_levels = mip ? (int)log2((float)MIN(w, h)) + 1 : 1;
	
	if(!(tex->hwtex = CreateHwTexture(xsz, ysz, type, fmt, tex->mip_levels))) {
		Error("CreateTexture failed to create the texture\n");
		delete tex;
		return 0;
	}
	
	unsigned int id = tex_list.Add(tex, req_id);
	if(!id && req_id) {
		tex_list.Remove(req_id);
		id = tex_list.Add(tex, req_id);
	}

	if(!id) {
		Error("CreateTexture failed to add to the tex_list\n");
		DeleteHwTexture(tex->hwtex);
		delete tex;
	}
	return id;
}

bool SetTexImage(unsigned int tid, void *pixels, int w, int h, TextureFormat tfmt, int face)
{
	bool result = false;
	void *pixels_rgba, *pixels_tmp, *pixels_dest = 0;

	int mip_w = w;
	int mip_h = h;

	Texture *tex = (Texture*)tex_list.Get(tid);
	if(!tex) {
		Error("Invalid texture id (%u) specified to SetTexImage\n", tid);
		return false;
	}

	// create the 32bit pixel data if it's in another format
	if(!(pixels_rgba = malloc(img_size(w, h, TFMT_32)))) {
		Error("memory allocation failed\n");
		return false;
	}
	if(!convert(pixels, pixels_rgba, w, h, tfmt, TFMT_32)) {
		Error("unsupported image conversion: %s -> %s\n", GetFormatName(tfmt), GetFormatName(TFMT_32));
		goto cleanup;
	}

	// create a scratch buffer (32bit) for the resamplings
	if(!(pixels_tmp = malloc(img_size(w, h, TFMT_32)))) {
		Error("memory allocation failed\n");
		goto cleanup;
	}

#ifdef BUILD_DX9
	// this buffer will hold the dest-format pixel data for each level (needed only for d3d)
	if(!(pixels_dest = malloc(img_size(w, h, tex->fmt)))) {
		Error("memory allocation failed\n");
		goto cleanup;
	}
#endif

	for(int i=0; i<tex->mip_levels; i++) {
#ifdef BUILD_DX9
		// create the dest-format image for this level
		if(!convert(pixels_rgba, pixels_dest, mip_w, mip_h, TFMT_32, tex->fmt)) {
			Error("unsupported image conversion: %s -> %s\n", GetFormatName(TFMT_32), GetFormatName(tex->fmt));
			goto cleanup;
		}

		D3DLOCKED_RECT lock_rect;
		D3DCUBEMAP_FACES cbface = (D3DCUBEMAP_FACES)(face - (int)CUBE_FACE_PX);
		
		if(tex->type == TEX_CUBE) {
			if(((IDirect3DCubeTexture9*)tex->hwtex)->LockRect(cbface, i, &lock_rect, 0, 0) != D3D_OK) {
				Error("failed to lock cube texture %d, face %d, level %d\n", tid, face, i);
				goto cleanup;
			}
		} else {
			if(((IDirect3DTexture9*)tex->hwtex)->LockRect(i, &lock_rect, 0, 0) != D3D_OK) {
				Error("failed to lock texture %d, level %d\n", tid, i);
				goto cleanup;
			}
		}

		// XXX: assumes no padding between scanlines
		memcpy(lock_rect.pBits, pixels_dest, img_size(mip_w, mip_h, tex->fmt));
		
		if(tex->type == TEX_CUBE) {
			((IDirect3DCubeTexture9*)tex->hwtex)->UnlockRect(cbface, i);
		} else {
			((IDirect3DTexture9*)tex->hwtex)->UnlockRect(i);
		}
#elif defined(BUILD_GL)
		glBindTexture(tex->type, tex->hwtex);
		if(tex->type == TEX_2D) {
			glTexSubImage2D(GL_TEXTURE_2D, i, 0, 0, mip_w, mip_h, GL_BGRA, GL_UNSIGNED_BYTE, pixels_rgba);
		}
#endif

		memcpy(pixels_tmp, pixels_rgba, img_size(mip_w, mip_h, TFMT_32));
		resample((unsigned int*)pixels_tmp, (unsigned int*)pixels_rgba, mip_w, mip_h, mip_w / 2, mip_h / 2);
		mip_w /= 2;
		mip_h /= 2;
	}

	result = true;
cleanup:
	free(pixels_rgba);
	free(pixels_tmp);
	free(pixels_dest);
	return result;
}

bool GetTexImage(unsigned int tid, void *pixels, TextureFormat fmt, int face)
{
	bool result = false;
	void *pixels_tex = 0;
	void *pixels_rgba = 0;

	Texture *tex = (Texture*)tex_list.Get(tid);
	if(!tex) {
		Error("Invalid texture id (%u) specified to SetTexImage\n", tid);
		return false;
	}

	if(!(pixels_tex = malloc(img_size(tex->width, tex->height, tex->fmt)))) {
		Error("failed to allocate memory\n");
		goto cleanup;
	}
	if(!(pixels_rgba = malloc(img_size(tex->width, tex->height, TFMT_32)))) {
		Error("failed to allocate memory\n");
		goto cleanup;
	}

#ifdef BUILD_DX9
	D3DLOCKED_RECT lock_rect;
	D3DCUBEMAP_FACES cbface;
	cbface = (D3DCUBEMAP_FACES)(face - (int)CUBE_FACE_PX);

	if(tex->type == TEX_CUBE) {
		if(((IDirect3DCubeTexture9*)tex->hwtex)->LockRect(cbface, 0, &lock_rect, 0, 0)) {
			Error("failed to lock cube texture %d, face %d, level 0\n", tid, face);
			goto cleanup;
		}
	} else {
        if(((IDirect3DTexture9*)tex->hwtex)->LockRect(0, &lock_rect, 0, 0) != D3D_OK) {
			Error("failed to lock texture %d, level 0\n", tid);
			goto cleanup;
		}
	}
	// XXX: assumes no padding between scanlines
	memcpy(pixels_tex, lock_rect.pBits, img_size(tex->width, tex->height, tex->fmt));

	if(tex->type == TEX_CUBE) {
		((IDirect3DCubeTexture9*)tex->hwtex)->UnlockRect(cbface, 0);
	} else {
        ((IDirect3DTexture9*)tex->hwtex)->UnlockRect(0);
	}

	if(!convert(pixels_tex, pixels_rgba, tex->width, tex->height, tex->fmt, TFMT_32)) {
		Error("unupported image conversion: %s -> %s\n", GetFormatName(tex->fmt), GetFormatName(TFMT_32));
		goto cleanup;
	}
#elif defined(BUILD_GL)
	glGetTexImage(GL_TEXTURE_2D, 0, GL_BGRA, GL_UNSIGNED_BYTE, pixels_rgba);
	if(glGetError() != GL_NO_ERROR) {
		Error("failed to get texture %d, level 0\n", tid);
		goto cleanup;
	}
#endif

	if(!convert(pixels_rgba, pixels, tex->width, tex->height, TFMT_32, tex->fmt)) {
		Error("unupported image conversion: %s -> %s\n", GetFormatName(TFMT_32), GetFormatName(fmt));
		goto cleanup;
	}

	result = true;
cleanup:
	free(pixels_rgba);
	free(pixels_tex);
	return result;
}

unsigned int CreateRenderTarget(int w, int h, TextureFormat fmt)
{
	int xsz = w;
	int ysz = h;

	Texture *tex;
	try {
		tex = new Texture;
	}
	catch(...) {
		return 0;
	}

	bool ispow2 = IsPow2(w) && IsPow2(h);
	tex->need_texmat = (!GetGcCaps()->npot_textures && !ispow2) ? true : false;
	
	if(tex->need_texmat) {		
		xsz = NextPow2(w);
		ysz = NextPow2(h);
	}
	
	tex->width = xsz;
	tex->height = ysz;
	tex->rt_width = w;
	tex->rt_height = h;
	tex->fmt = fmt;

#ifdef BUILD_DX9
	if(d3d_device->CreateTexture(xsz, ysz, 1, D3DUSAGE_RENDERTARGET, (D3DFORMAT)fmt, D3DPOOL_DEFAULT, (IDirect3DTexture9**)&tex->hwtex, 0) != D3D_OK) {
		Error("CreateRenderTarget(%d, %d, %d) failed\n", w, h, (int)fmt);
		delete tex;
		return 0;
	}
	if(d3d_device->CreateDepthStencilSurface(xsz, ysz, D3DFMT_D24S8, D3DMULTISAMPLE_NONE, 0, TRUE, &tex->ds_surf, 0) != D3D_OK) {
		Error("CreateRenderTarget failed in depth/stencil creation\n");
		tex->hwtex->Release();
		delete tex;
		return 0;
	}
#elif defined(BUILD_GL)
	// create the underlying color-buffer texture
	if(!(tex->hwtex = CreateHwTexture(xsz, ysz, TEX_2D, TFMT_32, 1))) {
		Error("CreateRenderTarget(%d, %d, %d) failed\n", w, h, (int)fmt);
		delete tex;
		return 0;
	}

	// create the depth buffer
	glGenRenderbuffersEXT(1, &tex->depth_buffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, tex->depth_buffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, xsz, ysz);

	// create the stencil buffer
	glGenRenderbuffersEXT(1, &tex->stencil_buffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, tex->stencil_buffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_STENCIL_INDEX, xsz, ysz);

	// create the framebuffer objct and attach the color/depth buffers
	glGenFramebuffersEXT(1, &tex->fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, tex->fbo);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex->hwtex, 0);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, tex->depth_buffer);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, tex->stencil_buffer);

	GLenum fbstat = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if(fbstat != GL_FRAMEBUFFER_COMPLETE_EXT) {
		Error("CreateRenderTarget failed, framebuffer not complete ...\n");
		delete tex;
		return 0;
	}

	tex->ds_surf = 1;
#endif

	unsigned int id = tex_list.Add(tex);
	if(!id) 
	{
		Error("CreateRenderTarget failed to add to the tex_list\n");
#ifdef BUILD_DX9
		tex->ds_surf->Release();
		tex->hwtex->Release();
#elif defined(BUILD_GL)
		DeleteHwTexture(tex->hwtex);
		glDeleteRenderbuffersEXT(1, &tex->depth_buffer);
		glDeleteRenderbuffersEXT(1, &tex->stencil_buffer);
		glDeleteFramebuffersEXT(1, &tex->fbo);
#endif
		delete tex;
		return 0;
	}

	return id;
}

unsigned int CreateCubeRenderTarget(int w, int h, TextureFormat fmt)
{
	int xsz = w;
	int ysz = h;

	Texture *tex;
	try {
		tex = new Texture;
	}
	catch(...) {
		return 0;
	}

	if (xsz != ysz)
	{
		Error("CreateCubeRenderTarget failed: dimensions %d, %d are not equal\n", xsz, ysz);
	}

	/*
	bool ispow2 = IsPow2(w) && IsPow2(h);
	tex->need_texmat = (!GetGcCaps()->npot_textures && !ispow2) ? true : false;

	if(tex->need_texmat) {		
		xsz = NextPow2(w);
		ysz = NextPow2(h);
	}
	*/
	
	tex->width = xsz;
	tex->height = ysz;
	tex->rt_width = w;
	tex->rt_height = h;
	tex->fmt = fmt;

#ifdef BUILD_DX9
	if(d3d_device->CreateCubeTexture(xsz, 1, D3DUSAGE_RENDERTARGET, (D3DFORMAT)fmt, D3DPOOL_DEFAULT, (IDirect3DCubeTexture9**)&tex->hwtex, 0) != D3D_OK) {
		Error("CreateCubeRenderTarget(%d, %d, %d) failed\n", w, h, (int)fmt);
		delete tex;
		return 0;
	}
	if(d3d_device->CreateDepthStencilSurface(xsz, ysz, D3DFMT_D24S8, D3DMULTISAMPLE_NONE, 0, TRUE, &tex->ds_surf, 0) != D3D_OK) {
		Error("CreateCubeRenderTarget failed in depth/stencil creation\n");
		(IDirect3DCubeTexture9*)tex->hwtex->Release();
		delete tex;
		return 0;
	}
#elif defined(BUILD_GL)
// TODO: implement opengl cube rendertarget creation
#endif

	unsigned int id = tex_list.Add(tex);
	if(!id) 
	{
		Error("CreateCubeRenderTarget failed to add to the tex_list\n");
#ifdef BUILD_DX9
		tex->ds_surf->Release();
		(IDirect3DCubeTexture9*)tex->hwtex->Release();
#elif defined(BUILD_GL)
		// TODO: implement this
#endif
		delete tex;
		return 0;
	}

	return id;
}

void DeleteTexture(unsigned int tid) 
{
	tex_list.Remove(tid);
}

bool BindTexture(unsigned int tex_unit, unsigned int tid) 
{
	// TODO : check tex_unit for overflow
	if(!tid || !tex_list.IsHandle(tid)) {
		current_uid[tex_unit] = 0;
		current_ttype[tex_unit] = TEX_INVALID;
#ifdef BUILD_DX9
		d3d_device->SetTexture(tex_unit, 0);
#elif defined(BUILD_GL)
		glBindTexture(current_ttype[tex_unit], 0);
		glDisable(current_ttype[tex_unit]);
#endif
		SetTextureMatrix(tex_unit, Matrix4x4::identity);
		return tid ? false : true;
	}

	Texture *tex = (Texture*)tex_list.Get(tid);
	if(tex_list.GetUID(tid) != current_uid[tex_unit]) {
		current_uid[tex_unit] = tex_list.GetUID(tid);
#ifdef BUILD_DX9
		if(FAILED(d3d_device->SetTexture(tex_unit, tex->hwtex))) {
			BindTexture(tex_unit, 0);
			return false;
		}
#elif defined(BUILD_GL)
		if(current_ttype[tex_unit] != tex->type) {
			BindTexture(tex_unit, 0);
		}

		glBindTexture(tex->type, tex->hwtex);
		if(glGetError() != GL_NO_ERROR) {
			BindTexture(tex_unit, 0);
			return false;
		}
#endif
		current_ttype[tex_unit] = tex->type;
		if(tex->need_texmat) {
			SetTextureMatrix(tex_unit, CalcTexMat(tex));
		}
	}
	return true;
}

bool PushRenderTarget(int mrt_id)
{
	if(mrt_id == MRT_ALL) {
		bool ret[4];
		for(int i=0; i<4; i++) {
			ret[i] = PushRenderTarget(i);
		}
		return ret[0] && ret[1] && ret[2] && ret[3];
	}

	if(mrt_id < 0 || mrt_id > 3) {
		return false;
	}

	unsigned int top = rtstack[mrt_id].empty() ? 0 : rtstack[mrt_id].top();
	rtstack[mrt_id].push(top);

	return true;
}

bool PopRenderTarget(int mrt_id)
{
	if(mrt_id == MRT_ALL) {
		bool ret[4];
		for(int i=0; i<4; i++) {
			ret[i] = PopRenderTarget(i);
		}
		return ret[0] && ret[1] && ret[2] && ret[3];
	}

	if(mrt_id < 0 || mrt_id > 3) {
		return false;
	}

	if(!rtstack[mrt_id].empty()) {
		rtstack[mrt_id].pop();
	}

	unsigned int rt = rtstack[mrt_id].empty() ? 0 : rtstack[mrt_id].top();

	if (!rt)
	{
		Error("PopRenderTarget returned zero as previews rt\n");
		return false;
	}
	bool success;
	Texture *tex = (Texture*)tex_list.Get(rt);
	if (tex->rt_cube_face_bound == -1)
	{
		// regular texture
		success = SetRenderTarget(rt, mrt_id);
	}
	else
	{
		success = SetRenderTarget(rt, mrt_id, tex->rt_cube_face_bound);
	}
	return success;
}

/* if tid == 0, it restores the original back-buffer render target */
// TODO: implement MRT for the OpenGL version
bool SetRenderTarget(unsigned int tid, unsigned int mrt_id, unsigned int cube_face)
{
#ifdef BUILD_DX9
	static IDirect3DSurface9 *back_surf, *back_ds, *prev_surf;
	IDirect3DSurface9 *surf;

	if(!back_surf) {	// if this is the first call, save the original backbuffer/depth/stencil
		if(d3d_device->GetRenderTarget(0, &back_surf) != D3D_OK) {
			back_surf = 0;
			return false;
		}

		if(d3d_device->GetDepthStencilSurface(&back_ds) != D3D_OK) {
			back_surf->Release();
			back_surf = back_ds = 0;
			return false;
		}
		prev_surf = back_surf;
	}
#endif

	if(tid) {
		Texture *tex = (Texture*)tex_list.Get(tid);
		if(!tex->ds_surf) {
			return false;	// not a render target
		}
		
#ifdef BUILD_DX9
		if (cube_face > 5)
		{
			if(((IDirect3DTexture9*)tex->hwtex)->GetSurfaceLevel(0, &surf) != D3D_OK) {
				return false;
			}
			// regular texture
			tex->rt_cube_face_bound = -1;
		}
		else
		{
			if(((IDirect3DCubeTexture9*)tex->hwtex)->GetCubeMapSurface((D3DCUBEMAP_FACES)cube_face, 0, &surf) != D3D_OK) {
				return false;
			}
			// cube texture
			tex->rt_cube_face_bound = cube_face;
		}
	
		if(d3d_device->SetRenderTarget(mrt_id, surf) != D3D_OK) {
			surf->Release();
			return false;
		}

		if(d3d_device->SetDepthStencilSurface(tex->ds_surf) != D3D_OK) {
			d3d_device->SetRenderTarget(mrt_id, prev_surf);
			surf->Release();
			return false;
		}
#elif defined(BUILD_GL)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, tex->fbo);
		if(glGetError() != GL_NO_ERROR) {
			return false;
		}
#endif

		if(mrt_id == 0) {
            SetViewport(0, 0, tex->rt_width, tex->rt_height);
		}
	} else {
#ifdef BUILD_DX9
		surf = back_surf;
		if(d3d_device->SetRenderTarget(mrt_id, mrt_id ? 0 : back_surf) != D3D_OK) {
			return false;
		}

		if(d3d_device->SetDepthStencilSurface(mrt_id ? 0 : back_ds) != D3D_OK) {
			d3d_device->SetRenderTarget(mrt_id, prev_surf);
			return false;
		}

		if(mrt_id == 0) {
            D3DSURFACE_DESC sdesc;
			back_surf->GetDesc(&sdesc);
			SetViewport(0, 0, sdesc.Width, sdesc.Height);
		}
#elif defined(BUILD_GL)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
#endif
	}

#ifdef BUILD_DX9
	if(prev_surf != back_surf) {
		prev_surf->Release();	// we got the previous with GetSurfaceLevel, so it must be released (d3d sucks).
	}
	prev_surf = surf;
#endif

	if(!rtstack[mrt_id].empty()) {
		rtstack[mrt_id].top() = tid;
	}

	return true;
}

Matrix4x4 GetRenderTargetTexMat(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return CalcTexMat(tex);
}

unsigned int GetTextureWidth(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex ? tex->width : 0;
}

unsigned int GetTextureHeight(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex ? tex->height : 0;
}

TextureType GetTextureType(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex ? tex->type : (TextureType)0;
}

TextureFormat GetTextureFormat(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex ? tex->fmt : TFMT_INVALID;
}

bool IsRenderTarget(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex && tex->ds_surf;
}

unsigned int GetMipLevels(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	return tex ? tex->mip_levels : 0;
}

size_t GetTextureMemUsage(unsigned int tid)
{
	Texture *tex = (Texture*)tex_list.Get(tid);
	if(!tex) {
		return 0;
	}
	size_t debug = (size_t)((float)img_size(tex->width, tex->height, tex->fmt) * (1.0f - pow(0.25f, tex->mip_levels + 1)) / 0.75f);
	return debug;
}

static const char *GetFormatName(TextureFormat fmt)
{
	switch(fmt) {
	case TFMT_8:
		return "TFMT_8";

	case TFMT_16:
		return "TFMT_16";

	case TFMT_24:
		return "TFMT_24";

	case TFMT_32:
		return "TFMT_32";

	case TFMT_COMPRESSED_24:
		return "TFMT_COMPRESSED_24";

	case TFMT_COMPRESSED_32:
		return "TFMT_COMPRESSED_32";

	default:
		break;
	}
	return "TFMT_UNKNOWN";
}

static Matrix4x4 CalcTexMat(Texture *tex)
{
	if(!tex->need_texmat) {
		return Matrix4x4::identity;
	}

	Matrix4x4 mat;
	float sx = (float)tex->rt_width / (float)tex->width;
	float sy = (float)tex->rt_height / (float)tex->height;
	mat.SetScaling(Vector4(sx, sy, 1.0f, 1.0f));
	return mat;
}
