#include <builddefs.h>

#ifdef BUILD_DX9
#include <assert.h>
#include "sysgc.h"
#include "vbuffer.h"
#include "idman.h"
#include "gc_stats.h"

extern IDirect3DDevice9  *d3d_device;

static const unsigned int index_element_size = INDEX_SIZE;

struct VertexDeclaration
{
	VertexFormat vertex_format;
	IDirect3DVertexDeclaration9 *vdecl;
};

struct VertexBuffer
{
	bool					is_dynamic;
	VertexFormat			vertex_format;
	IDirect3DVertexBuffer9 *vertex_buffer;
	unsigned int			vertex_format_size;
	unsigned int			vertex_declaration_id;
	unsigned int			vertex_count;
};

struct IndexBuffer
{
	bool is_dynamic;
	IDirect3DIndexBuffer9 *index_buffer;
	unsigned int index_count;
};

static void DeleteHWVDecl (void *vdecl);
static void DeleteHWVBuffer (void *vb);
static void DeleteHWIndex (void* ib);


IdManRef vdecl_list(DeleteHWVDecl);
IdMan vbuffer_list (DeleteHWVBuffer);
IdMan ibuffer_list (DeleteHWIndex);

unsigned int current_vdecl_bound_id = 0;
unsigned int current_vbuffer_bound_id = 0;
unsigned int current_ibuffer_bound_id = 0;
unsigned int current_vbuffer_stride = 0;

static IDirect3DVertexDeclaration9 *CreateHWVDecl (const VertexFormat &vf)
{
	D3DVERTEXELEMENT9  vertex_element [20];
	D3DVERTEXELEMENT9  vertex_end = D3DDECL_END ();
	unsigned int current_element = 0;

	if (vf.vertex)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetVertexOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_FLOAT3;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_POSITION;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	if (vf.normal)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetNormalOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_FLOAT3;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_NORMAL;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	if (vf.color)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetColorOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_D3DCOLOR;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_COLOR;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	if (vf.bone_weight)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetBoneWeightOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_FLOAT3;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_BLENDWEIGHT;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	if (vf.bone_index)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetBoneIndexOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_UBYTE4;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_BLENDINDICES;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	if (vf.tangent)
	{
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetTangentOffset ();
		vertex_element [current_element].Type = D3DDECLTYPE_FLOAT3;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_TANGENT;
		vertex_element [current_element].UsageIndex = 0;

		current_element++;
	}
	for (unsigned int i = 0; i < vf.tc_count; i++)
	{
		unsigned char type = D3DDECLTYPE_FLOAT2;
		switch (vf.tc_stream_count [i])
		{
		case (1): type = D3DDECLTYPE_FLOAT1; break;
		case (2): type = D3DDECLTYPE_FLOAT2; break;
		case (3): type = D3DDECLTYPE_FLOAT3; break;
		case (4): type = D3DDECLTYPE_FLOAT4; break;
		}
		vertex_element [current_element].Stream = 0;
		vertex_element [current_element].Offset = vf.GetTexCoordOffset(i);
		vertex_element [current_element].Type = type;
		vertex_element [current_element].Method = D3DDECLMETHOD_DEFAULT;
		vertex_element [current_element].Usage = D3DDECLUSAGE_TEXCOORD;
		vertex_element [current_element].UsageIndex = i;

		current_element++;
	}

	memcpy (&vertex_element [current_element], &vertex_end, sizeof (D3DVERTEXELEMENT9));

	IDirect3DVertexDeclaration9  *d3d_declaration = NULL;

	d3d_device->CreateVertexDeclaration (vertex_element, &d3d_declaration);

	return d3d_declaration;
}


static unsigned int CreateVertexDeclaration (const VertexFormat &vf)
{
	unsigned int count = vdecl_list.GetResCount();
	for (unsigned int i = 1; i < count; i++)
	{	
		if (vdecl_list.IsHandle(i))
		{
			VertexDeclaration *vdecl = (VertexDeclaration *) vdecl_list.Get(i);
			if (vdecl->vertex_format == vf)
			{
				vdecl_list.AddRef(i);
				return i;
			}
		}
	}
	IDirect3DVertexDeclaration9  *d3d_declaration = CreateHWVDecl (vf);
	if (d3d_declaration == NULL)
	{
		return 0;
	}

	VertexDeclaration *vd = new VertexDeclaration;
	vd->vdecl = d3d_declaration;
	vd->vertex_format = vf;

	unsigned int id = vdecl_list.Add (vd);
	if (id == 0)
	{
		d3d_declaration->Release();
		delete vd;
		return 0;
	}

	return id;
}

static void DeleteVertexDeclaration (unsigned int vdecl)
{
	vdecl_list.Remove (vdecl);
}

static void DeleteHWVDecl (void *vdecl)//VertexDeclaration *vdecl)
{
	//TODO : check if bound
	((VertexDeclaration *) vdecl)->vdecl->Release();
	delete (VertexDeclaration *) vdecl;
}

static bool BindVertexDeclaration (unsigned int id)
{
	if (!vdecl_list.IsHandle (id))
	{
		d3d_device->SetVertexDeclaration (NULL);
		return false;
	}

	if (vdecl_list.GetUID (id) != current_vdecl_bound_id)
	{
		VertexDeclaration *vdecl = (VertexDeclaration *) vdecl_list.Get(id);

		HRESULT res = d3d_device->SetVertexDeclaration (vdecl->vdecl);
		if (FAILED (res))
		{
			return false;
		}
		current_vdecl_bound_id = vdecl_list.GetUID (id);
	}
	return true;
}

unsigned int CreateVertexBuffer (const VertexFormat &vf, unsigned int elements_count, bool dynamic)
{
	unsigned int vdecl = CreateVertexDeclaration (vf);
	if (vdecl == 0)
	{
		return 0;
	}

	unsigned int usage = D3DUSAGE_WRITEONLY;
	if(dynamic) {
		usage |= D3DUSAGE_DYNAMIC;
	}

	unsigned int pool = dynamic ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED;

	IDirect3DVertexBuffer9 *hw_vb = NULL;

	HRESULT hRes = d3d_device->CreateVertexBuffer (vf.GetSize() * elements_count, usage, 0, (D3DPOOL) pool, &hw_vb, NULL);
	if (FAILED(hRes))
	{
		DeleteVertexDeclaration (vdecl);
		return 0;
	}

	VertexBuffer *vb = new VertexBuffer;
	vb->vertex_format = vf;
	vb->vertex_buffer = hw_vb;
	vb->vertex_declaration_id = vdecl;
	vb->vertex_format_size = vb->vertex_format.GetSize();
	vb->is_dynamic = dynamic;
	vb->vertex_count = elements_count;

	unsigned int the_id = vbuffer_list.Add (vb);
	if (the_id == 0)
	{
		hw_vb->Release ();
		delete vb;
		DeleteVertexDeclaration (vdecl);
		return 0;
	}

	return the_id;
}


void DeleteVertexBuffer (unsigned int id)
{
	vbuffer_list.Remove(id);
}


static void DeleteHWVBuffer (void *vb) //VertexBuffer *vb)
{
	//TODO : check if bound
	((VertexBuffer *)vb)->vertex_buffer->Release();
	DeleteVertexDeclaration (((VertexBuffer *)vb)->vertex_declaration_id);
	delete (VertexBuffer *)vb;
}


bool BindVertexBuffer (unsigned int id)
{
	// TODO : if it's already bound, then dont do it...
	if (!vbuffer_list.IsHandle(id))
	{
		d3d_device->SetStreamSource (0, NULL, 0, 0);
		current_vbuffer_bound_id = 0;
		current_vbuffer_stride = 0;
		return false;
	}

	if (vbuffer_list.GetUID (id) != current_vbuffer_bound_id)
	{
		VertexBuffer *vb = (VertexBuffer *) vbuffer_list.Get (id);
		if (BindVertexDeclaration (vb->vertex_declaration_id) == false)
		{
			current_vbuffer_bound_id = 0;
			current_vbuffer_stride = 0;
			return false;
		}

		HRESULT res = d3d_device->SetStreamSource (0, vb->vertex_buffer, 0, vb->vertex_format_size); 
		if (FAILED (res))
		{
			current_vbuffer_bound_id = 0;
			current_vbuffer_stride = 0;
			return false;
		}
		current_vbuffer_bound_id = vbuffer_list.GetUID (id);
		current_vbuffer_stride = vb->vertex_format_size;
	}

	return true;
}

void *LockVertexBuffer (unsigned int id, unsigned int start_vertex, unsigned int vertex_count)
{
	if (vbuffer_list.IsHandle (id))
	{
		VertexBuffer *vb = (VertexBuffer *) vbuffer_list.Get (id);
		unsigned int stride = vb->vertex_format_size;

		HRESULT res;
		void *the_ptr = NULL;

		if ((start_vertex == 0) && (vertex_count == 0))
		{
			res = vb->vertex_buffer->Lock (0, 0, &the_ptr, vb->is_dynamic? D3DLOCK_DISCARD : 0);
		}
		else
		{
			res = vb->vertex_buffer->Lock (start_vertex*stride, vertex_count*stride, &the_ptr, vb->is_dynamic? D3DLOCK_DISCARD : 0);
		}
		if (FAILED(res))
		{
			return NULL;
		}

		return the_ptr;
	}
	return NULL;
}

void UnlockVertexBuffer (unsigned int id)
{
	if (vbuffer_list.IsHandle (id))
	{
		VertexBuffer *vb = (VertexBuffer *) vbuffer_list.Get (id);
		vb->vertex_buffer->Unlock ();
	}
}

VertexFormat GetVertexFormat (unsigned int id)
{
	VertexFormat vf;

	if (vbuffer_list.IsHandle (id))
	{
		VertexBuffer *vb = (VertexBuffer *) vbuffer_list.Get (id);
		vf = vb->vertex_format;
	}

	return vf;
}



//--------------- INDEX BUFFER -------------------------------------------------------



unsigned int CreateIndexBuffer (unsigned int indices_count, bool dynamic)
{
	IDirect3DIndexBuffer9 *ib = NULL;

	unsigned int usage = D3DUSAGE_WRITEONLY | (dynamic ? D3DUSAGE_DYNAMIC : 0);
	D3DFORMAT fmt = index_element_size == 4 ? D3DFMT_INDEX32 : D3DFMT_INDEX16;
	D3DPOOL pool = dynamic ? D3DPOOL_DEFAULT : D3DPOOL_MANAGED;

	HRESULT res = d3d_device->CreateIndexBuffer (indices_count * index_element_size, usage, fmt, pool, &ib, NULL);
	if (FAILED (res))
	{
		return false;
	}

	IndexBuffer *index_buffer;
	try
	{
		index_buffer = new IndexBuffer;
	}
	catch (...)
	{
		ib->Release();
		return false;
	}

	index_buffer->is_dynamic = dynamic;
	index_buffer->index_buffer = ib;
	index_buffer->index_count = indices_count;
	unsigned int id = ibuffer_list.Add (index_buffer);
	if (id == 0)
	{
		ib->Release();
		delete index_buffer;
	}
	return id;
}

void DeleteIndexBuffer (unsigned int id)
{
	ibuffer_list.Remove (id);
}

static void DeleteHWIndex (void *ib) //(IndexBuffer *ib)
{
	//TODO : check if bound
	((IndexBuffer *) ib)->index_buffer->Release();
	delete (IndexBuffer *) ib;
}


bool BindIndexBuffer (unsigned int id)
{
	if (!ibuffer_list.IsHandle (id))
	{
		d3d_device->SetIndices (NULL);
		return false;
	}
	if (current_ibuffer_bound_id != ibuffer_list.GetUID(id))
	{
		IndexBuffer *ib = (IndexBuffer *) ibuffer_list.Get (id);
		HRESULT res = d3d_device->SetIndices (ib->index_buffer);
		if (FAILED (res))
		{
			return false;
		}
		current_ibuffer_bound_id = ibuffer_list.GetUID(id);
	}
	return true;
}

void *LockIndexBuffer (unsigned int id, unsigned int start, unsigned int count)
{
	if (!ibuffer_list.IsHandle (id))
	{
		return NULL;
	}
	IndexBuffer *ib = (IndexBuffer *) ibuffer_list.Get (id);
	void *the_ptr = NULL;

	ib->index_buffer->Lock (start * index_element_size, count * index_element_size, &the_ptr, ib->is_dynamic ? D3DLOCK_DISCARD : 0);
	return the_ptr;
}

void UnlockIndexBuffer (unsigned int id)
{
	if (ibuffer_list.IsHandle (id))
	{
		IndexBuffer *ib = (IndexBuffer *) ibuffer_list.Get (id);
		ib->index_buffer->Unlock();
	}
}

bool RenderPrimitives (PrimitiveType pt, unsigned int start_vertex, unsigned int primitive_count)
{
	if (current_vbuffer_bound_id == 0)
	{
		return false;
	}

	StatsAddBatch(primitive_count);

	HRESULT res = d3d_device->DrawPrimitive ((D3DPRIMITIVETYPE)pt, start_vertex, primitive_count);
	if (FAILED (res))
	{
		return false;
	}
	return true;
}


bool RenderIndexedPrimitives (PrimitiveType pt, unsigned int start_index, unsigned int primitive_count, unsigned int start_vertex, unsigned int vertices_count)
{
	if (current_vbuffer_bound_id == 0)
	{
		return false;
	}
	if (current_ibuffer_bound_id == 0)
	{
		return false;
	}

	StatsAddBatch(primitive_count);

	HRESULT res = d3d_device->DrawIndexedPrimitive ((D3DPRIMITIVETYPE)pt, 0, start_vertex, vertices_count, start_index, primitive_count);

	if (FAILED (res))
	{
		return false;
	}
	return true;
}

#endif	// BUILD_DX9
