#include <builddefs.h>

#ifdef BUILD_GL
#include <vector>
#include <assert.h>
#include "sysgc.h"
#include "vbuffer.h"
#include "gc_stats.h"
//#include "idman.h"
#include <errlog.h>

static std::vector<VertexFormat> vfmt(256);
static int act_vbo, act_ibo;

unsigned int CreateVertexBuffer(const VertexFormat &vf, unsigned int elements_count, bool dynamic)
{
	unsigned int vbo;

	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER_ARB, vbo);
	glBufferData(GL_ARRAY_BUFFER_ARB, elements_count * vf.GetSize(), 0, dynamic ? GL_DYNAMIC_DRAW : GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER_ARB, act_vbo);

	if(vbo >= vfmt.size()) {
		size_t new_sz = vfmt.size() * 2;
		Info("resizing vfmt array to: %lu\n", new_sz);
		try {
			vfmt.resize(new_sz);
		}
		catch(...) {
			Error("memory allocation failed!");
			glDeleteBuffers(1, &vbo);
		}
	}
	vfmt[vbo] = vf;

	return vbo;
}

void DeleteVertexBuffer(unsigned int id)
{
	glDeleteBuffers(1, &id);
}

bool BindVertexBuffer(unsigned int id)
{
	act_vbo = id;
	glBindBuffer(GL_ARRAY_BUFFER_ARB, id);
	return glGetError() == GL_NO_ERROR;
}


/* XXX: I assume we never actually read from these buffers, so semantically,
 * this "locking" interface is broken.
 * Based on this assumption, for performance reasons I choose to implement it
 * with glBufferSubData() instead of glMapBuffer()
 */

static struct {
	void *data;
	unsigned int start, count;
	unsigned int vbo;
	VertexFormat vf;
} vdata;

void *LockVertexBuffer(unsigned int id, unsigned int start_vertex, unsigned int vertex_count)
{
	assert(!vdata.data);	// XXX if you ever hit this, put vdata in a stack

	vdata.vf = GetVertexFormat(id);
	if(!(vdata.data = malloc(vdata.vf.GetSize() * vertex_count))) {
		Error("failed to allocate buffer\n");
		return 0;
	}

	vdata.vbo = id;
	vdata.start = start_vertex;
	vdata.count = vertex_count;

	return vdata.data;
}

void UnlockVertexBuffer(unsigned int id)
{
	assert(vdata.data && vdata.vbo == id);

	unsigned int offs = vdata.start * vdata.vf.GetSize();
	unsigned int size = vdata.count * vdata.vf.GetSize();

	BindVertexBuffer(id);
	glBufferSubData(GL_ARRAY_BUFFER_ARB, offs, size, vdata.data);

	free(vdata.data);
	vdata.data = 0;
}

VertexFormat GetVertexFormat(unsigned int id)
{
	if(id >= vfmt.size()) {
		id = 0;
	}
	return vfmt[id];
}


unsigned int CreateIndexBuffer(unsigned int indices_count, bool dynamic)
{
	unsigned int ibo;

	glGenBuffers(1, &ibo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER_ARB, ibo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER_ARB, indices_count * INDEX_SIZE, 0, dynamic ? GL_DYNAMIC_DRAW_ARB : GL_STATIC_DRAW_ARB);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER_ARB, act_ibo);

	return ibo;
}

void DeleteIndexBuffer(unsigned int id)
{
	glDeleteBuffers(1, &id);
}

bool BindIndexBuffer(unsigned int id)
{
	act_ibo = id;
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER_ARB, id);
	return glGetError() == GL_NO_ERROR;
}

static struct {
	void *data;
	unsigned int start, count;
	unsigned int ibo;
} idata;

void *LockIndexBuffer(unsigned int id, unsigned int start, unsigned int count)
{
	assert(!idata.data);	// XXX if you ever hit this, put vdata in a stack

	if(!(idata.data = malloc(count * INDEX_SIZE))) {
		Error("failed to allocate buffer\n");
		return 0;
	}

	idata.ibo = id;
	idata.start = start;
	idata.count = count;

	return idata.data;
}

void UnlockIndexBuffer(unsigned int id)
{
	assert(idata.data && idata.ibo == id);

	unsigned int offs = idata.start * INDEX_SIZE;
	unsigned int size = idata.count * INDEX_SIZE;

	BindIndexBuffer(id);
	glBufferSubData(GL_ARRAY_BUFFER_ARB, offs, size, idata.data);

	free(idata.data);
	idata.data = 0;
}

static void EnableVertexArrays(VertexFormat *vf)
{
	
	if(vf->vertex) {
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, vf->GetSize(), (void*)vf->GetVertexOffset());
	}

	if(vf->normal) {
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, vf->GetSize(), (void*)vf->GetNormalOffset());
	}

	if(vf->color) {
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_FLOAT, vf->GetSize(), (void*)vf->GetColorOffset());
	}

	/*if(vf->bone_index) {
		glEnableClientState(GL_MATRIX_INDEX_ARRAY_ARB);
		glMatrixIndexPointerARB(4, GL_UNSIGNED_BYTE, vf->GetSize(), (void*)vf->GetBoneIndexOffset());
	}

	if(vf->bone_weight) {
		glEnableClientState(GL_WEIGHT_ARRAY_ARB);
		glWeightPointerARB(4, GL_FLOAT, vf->GetSize(), (void*)vf->GetBoneWeightOffset());
	}*/

	for(int i=0; i<(int)vf->tc_count; i++) {
		glClientActiveTexture(GL_TEXTURE0_ARB + (GLenum)i);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(vf->tc_stream_count[i], GL_FLOAT, vf->GetSize(), (void*)vf->GetTexCoordOffset(i));
	}
}

static void DisableVertexArrays(VertexFormat *vf)
{
	if(vf->vertex) {
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	if(vf->normal) {
		glDisableClientState(GL_NORMAL_ARRAY);
	}
	if(vf->color) {
		glDisableClientState(GL_COLOR_ARRAY);
	}
	if(vf->bone_index) {
		glDisableClientState(GL_MATRIX_INDEX_ARRAY_ARB);
	}
	if(vf->bone_weight) {
		glDisableClientState(GL_WEIGHT_ARRAY_ARB);
	}
	for(int i=0; i<(int)vf->tc_count; i++) {
		glClientActiveTexture(GL_TEXTURE0_ARB + (GLenum)i);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
}

static int PrimVerts(PrimitiveType pt)
{
	switch(pt) {
	case PT_TRIANGLE_LIST:
		return 3;
	case PT_POINT_LIST:
		return 1;
	case PT_LINE_LIST:
		return 2;
	case PT_QUAD_LIST:
		return 4;
	default:
		return 0;	// TODO: support the rest ptypes somehow
	}
	return 0;
}

bool RenderPrimitives(PrimitiveType pt, unsigned int start_vertex, unsigned int primitive_count)
{
	GLenum err;

	if(!act_vbo) {
		return false;
	}
	VertexFormat *vf = &vfmt[act_vbo];

	EnableVertexArrays(vf);
	glDrawArrays((GLenum)pt, start_vertex, PrimVerts(pt) * primitive_count);
	err = glGetError();
	DisableVertexArrays(vf);

	StatsAddBatch(primitive_count);
	return err == GL_NO_ERROR;
}

bool RenderIndexedPrimitives(PrimitiveType pt, unsigned int start_index, unsigned int primitive_count, unsigned int start_vertex, unsigned int vertices_count)
{
	GLenum err;

	if(!act_vbo || !act_ibo) {
		return false;
	}
	VertexFormat *vf = &vfmt[act_vbo];

	EnableVertexArrays(vf);
	glDrawElements((GLenum)pt, PrimVerts(pt) * primitive_count, GL_UNSIGNED_INT, (void*)start_index);
	err = glGetError();
	DisableVertexArrays(vf);

	StatsAddBatch(primitive_count);
	return err == GL_NO_ERROR;
}


#endif	// BUILD_GL
