#include "vertexformat.h"


#define VF_TEXCOORD_0     0
#define VF_TEXCOORD_1     1
#define VF_TEXCOORD_2     2
#define VF_TEXCOORD_3     3
#define VF_TEXCOORD_4     4
#define VF_TEXCOORD_5     5
#define VF_TEXCOORD_6     6
#define VF_TEXCOORD_7     7
#define VF_TEXCOORD_8     8

#define VF_VERTEX       16
#define VF_NORMAL       32
#define VF_TANGENT      64
#define VF_COLOR       128
#define VF_BONE_W      256
#define VF_BONE_I      512

unsigned int GetTexCoordOrdinal (const VertexFormat &vf)
{
	unsigned int ordinal = 0;
	for (unsigned int i = 0; i < vf.tc_count; i++)
	{
		unsigned int current_tc = vf.tc_stream_count[i];
		switch (current_tc)
		{
		case 1: current_tc = 0; break;
		case 2: current_tc = 1; break;
		case 3: current_tc = 2; break;
		case 4: current_tc = 3; break;
		default: current_tc = 0; break;
		}
		current_tc <<= (i*2) + 16;
		ordinal += current_tc;
	}
	return ordinal;
}


VertexFormat::VertexFormat ()
{
    vertex = true;
    normal = false;
    tangent = false;
    color = false;
    bone_weight = false;
    bone_index = false;
    tc_count = 0;
    for (int i = 0; i < MAX_TEX_COORDS; i ++)
    {
        tc_stream_count[i] = 2;
    }
}


bool VertexFormat::IsEqual (const VertexFormat &vf) const
{
	return *this == vf;
}

bool VertexFormat::operator!=(const VertexFormat &vf) const
{
	return !(*this == vf);
}


bool VertexFormat::operator==(const VertexFormat &vf) const
{
    if (vf.vertex != vertex)
    {
        return false;
    }
    if (vf.normal != normal)
    {
        return false;
    }
    if (vf.tangent != tangent)
    {
        return false;
    }
    if (vf.color != color)
    {
        return false;
    }
    if (vf.bone_weight != bone_weight)
    {
        return false;
    }
    if (vf.bone_index != bone_index)
    {
        return false;
    }
    if (vf.tc_count != tc_count)
    {
        return false;
    }
    for (unsigned int i = 0; i < tc_count; i++)
    {
        if (vf.tc_stream_count[i] != tc_stream_count[i])
        {
            return false;
        }
    }

    return true;
}



bool VertexFormat::ContainsFormat (const VertexFormat &vf) const
{
    if (vf.vertex)
    {
        if (vertex == false)
        {
            return false;
        }
    }

    if (vf.normal)
    {
        if (normal == false)
        {
            return false;
        }
    }

    if (vf.color)
    {
        if (color == false)
        {
            return false;
        }
    }

    if (vf.tangent)
    {
        if (tangent == false)
        {
            return false;
        }
    }

    if (vf.bone_index)
    {
        if (bone_index == false)
        {
            return false;
        }
    }

    if (vf.bone_weight)
    {
        if (bone_weight == false)
        {
            return false;
        }
    }

    if (vf.tc_count > tc_count)
    {
        return false;
    }

    for (unsigned int i = 0; i < vf.tc_count; i++)
    {
        if (vf.tc_stream_count [i] != tc_stream_count [i])
        {
            return false;
        }
    }

    return true;
}



unsigned int VertexFormat::GetSize () const
{
    unsigned int the_size = 0;
    if (vertex)
    {
        the_size += sizeof (float) * 3;
    }
    if (normal)
    {
        the_size += sizeof (float) * 3;
    }
    if (tangent)
    {
        the_size += sizeof (float) * 3;
    }
    if (color)
    {
        the_size += 4;
    }
    if (bone_weight)
    {
        the_size += sizeof (float) * 3;
    }
    if (bone_index)
    {
        the_size += 4;
    }
    for (unsigned int i = 0; i < tc_count; i ++)
    {
        the_size += sizeof (float) * tc_stream_count [i];
    }

    return the_size;
}



unsigned int VertexFormat::GetVertexOffset () const
{
    return 0;
}

unsigned int VertexFormat::GetNormalOffset () const
{
    return GetVertexOffset () + GetVertexSize ();
}

unsigned int VertexFormat::GetColorOffset () const
{
    return GetNormalOffset () + GetNormalSize ();
}

unsigned int VertexFormat::GetBoneWeightOffset () const
{
    return GetColorOffset() + GetColorSize ();
}

unsigned int VertexFormat::GetBoneIndexOffset () const
{
    return GetBoneWeightOffset () + GetBoneWeightSize ();
}

unsigned int VertexFormat::GetTangentOffset () const
{
    return GetBoneIndexOffset () + GetBoneIndexSize ();
}

unsigned int VertexFormat::GetTexCoordOffset (unsigned int idx) const
{
    unsigned int start = GetTangentOffset () + GetTangentSize ();
    for (unsigned int i = 0; i < idx; i ++)
    {
        start += GetTexCoordSize (idx);
    }
    return start;
}



unsigned int VertexFormat::GetVertexSize () const
{
    if (vertex)
    {
        return sizeof(float) * 3;
    }
    return 0;
}

unsigned int VertexFormat::GetNormalSize () const
{
    if (normal)
    {
        return sizeof(float) * 3;
    }
    return 0;
}

unsigned int VertexFormat::GetColorSize () const
{
    if (color)
    {
        return 4;
    }
    return 0;
}

unsigned int VertexFormat::GetBoneWeightSize () const
{
    if (bone_weight)
    {
        return sizeof (float) * 3;
    }
    return 0;
}

unsigned int VertexFormat::GetBoneIndexSize () const
{
    if (bone_index)
    {
        return 4;
    }
    return 0;
}

unsigned int VertexFormat::GetTangentSize () const
{
    if (tangent)
    {
        return sizeof(float) * 3;
    }
    return 0;
}

unsigned int VertexFormat::GetTexCoordSize (unsigned int idx) const
{
    if (idx < tc_count)
    {
        return sizeof(float) * tc_stream_count[idx];
    }
    return 0;
}

bool VertexFormat::WriteToFP (VFS_FILE *fp)
{
	size_t bytes_write;
	try
	{
		unsigned int bool_int;
		
		bool_int = 0;
		if (vertex)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bool_int = 0;
		if (normal)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bool_int = 0;
		if (tangent)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bool_int = 0;
		if (color)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bool_int = 0;
		if (bone_weight)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bool_int = 0;
		if (bone_index)
		{
			bool_int = 1;
		}
		bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}

		bytes_write = vfs_write (&tc_count, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		
		bytes_write = vfs_write (tc_stream_count, 1, sizeof (unsigned int) * MAX_TEX_COORDS, fp);
		if (bytes_write != sizeof (unsigned int) * MAX_TEX_COORDS)
		{
			throw (int) 0;
		}
	}
	catch (...)
	{
		return false;
	}

	return true;
}

bool VertexFormat::LoadFromFP (VFS_FILE *fp)
{
	size_t bytes_read;
	try
	{
		unsigned int bool_int;
		
		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		vertex = false;
		if (bool_int == 1)
		{
			vertex = true;
		}

		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		normal = false;
		if (bool_int == 1)
		{
			normal = true;
		}

		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		tangent = false;
		if (bool_int == 1)
		{
			tangent = true;
		}

		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		color = false;
		if (bool_int == 1)
		{
			color = true;
		}

		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		bone_weight = false;
		if (bool_int == 1)
		{
			bone_weight = true;
		}

		bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		bone_index = false;
		if (bool_int == 1)
		{
			bone_index = true;
		}


		bytes_read = vfs_read (&tc_count, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 0;
		}
		
		bytes_read = vfs_read (tc_stream_count, 1, sizeof (unsigned int) * MAX_TEX_COORDS, fp);
		if (bytes_read != sizeof (unsigned int) * MAX_TEX_COORDS)
		{
			throw (int) 0;
		}
	}
	catch (...)
	{
		return false;
	}

	return true;
}


unsigned int VertexFormat::GetOrdinal () const
{
	unsigned int ordinal = 0;

	if (vertex) ordinal |= VF_VERTEX;
    if (normal) ordinal |= VF_NORMAL;
    if (tangent) ordinal |= VF_TANGENT;
    if (color) ordinal |= VF_COLOR;
    if (bone_weight) ordinal |= VF_BONE_W;
    if (bone_index) ordinal |= VF_BONE_I;
    switch (tc_count)
	{
	case (0): ordinal |= VF_TEXCOORD_0; break;
	case (1): ordinal |= VF_TEXCOORD_1; break;
	case (2): ordinal |= VF_TEXCOORD_2; break;
	case (3): ordinal |= VF_TEXCOORD_3; break;
	case (4): ordinal |= VF_TEXCOORD_4; break;
	case (5): ordinal |= VF_TEXCOORD_5; break;
	case (6): ordinal |= VF_TEXCOORD_6; break;
	case (7): ordinal |= VF_TEXCOORD_7; break;
	case (8): ordinal |= VF_TEXCOORD_8; break;
	}
	ordinal |= GetTexCoordOrdinal (*this);

	return ordinal;
}
