#include <builddefs.h>
#include <errlog.h>
#include <uniforms.h>
#include <statecache.h>
#include "texman.h"
#include "fullmat.h"


FullMaterial::FullMaterial()
{
	gc_mat.ambient = Color(0.1f, 0.1f, 0.1f, 1.0f);
	gc_mat.diffuse = Color(0.5f, 0.5f, 0.5f, 1.0f);
	gc_mat.specular = Color(1.0f, 1.0f, 1.0f, 1.0f);
	gc_mat.emissive = Color(0.0f, 0.0f, 0.0f, 0.0f);
	gc_mat.power = 60.0f;

	memset(tex, 0, sizeof *tex * FMTT_COUNT);

	SetTexFilename(FMTT_DIFFUSE, DEV_DATA_PATH "/defmap_diffuse.png");
	SetTexFilename(FMTT_SPECULAR, DEV_DATA_PATH "/defmap_specular.png");
	SetTexFilename(FMTT_LIGHTMAP, DEV_DATA_PATH "/defmap_lightmap.png");
	SetTexFilename(FMTT_EMISSIVE, DEV_DATA_PATH "/defmap_emissive.png");
	SetTexFilename(FMTT_NORMAL, DEV_DATA_PATH "/defmap_normal.png");

	CreateTextures();
}

FullMaterial::~FullMaterial()
{
	ReleaseTextures();
}

void FullMaterial::ReleaseTextures()
{
	for(int i = 0; i < FMTT_COUNT; i++) {
		SetTexFilename(i, 0);
		if(tex[i].tex_id) {
			ReleaseTexture(tex[i].tex_id);
			tex[i].tex_id = 0;
		}
	}
}


void FullMaterial::SetTexFilename(unsigned int tex_type, const char *name)
{
	if(tex_type >= FMTT_COUNT) {
		return;
	}
	if(tex[tex_type].fname) {
		free(tex[tex_type].fname);
		tex[tex_type].fname = 0;
	}

	if(name && *name) {
		size_t name_len = strlen(name);
		
		tex[tex_type].fname = (char*)malloc(name_len + 1);
		if(!tex[tex_type].fname) {
			Error("SetTexFilename failed to allocate memory\n");
			return;
		}
		strcpy(tex[tex_type].fname, name);
	}
}

//#define DEFMAP_PATH		DEV_DATA_PATH "/textures/"
#define DEFMAP_PATH	 "/devdata/textures/"
static const char *defmaps[] = {
	DEFMAP_PATH "defmap_diffuse.png",
	DEFMAP_PATH "defmap_specular.png",
	DEFMAP_PATH "defmap_lightmap.png",
	DEFMAP_PATH "defmap_emissive.png",
	DEFMAP_PATH "defmap_normal.png"
};

bool FullMaterial::CreateTextures()
{
	for(int i=0; i<FMTT_COUNT; i++) {
		if(tex[i].fname) {
			tex[i].tex_id = GetTexture(tex[i].fname);
		}
		if(!tex[i].tex_id) {
			if(!(tex[i].tex_id = GetTexture(defmaps[i]))) {
				Error("FullMaterial::CreateTextures(): failed to load default map\n");
				return false;
			}
		}
	}
	return true;
}

void FullMaterial::Bind() const
{
	SetUniform(STATE_MATERIAL_AMBIENT, (void*)&gc_mat.ambient);
	SetUniform(STATE_MATERIAL_DIFFUSE, (void*)&gc_mat.diffuse);
	SetUniform(STATE_MATERIAL_EMISSIVE, (void*)&gc_mat.emissive);
	SetUniform(STATE_MATERIAL_SPECULAR, (void*)&gc_mat.specular);
	SetUniform(STATE_MATERIAL_POWER, (void*)&gc_mat.power);

	BindTexture(0, tex[FMTT_DIFFUSE].tex_id);
	BindTexture(1, tex[FMTT_NORMAL].tex_id);
	BindTexture(2, tex[FMTT_SPECULAR].tex_id);
	BindTexture(3, tex[FMTT_LIGHTMAP].tex_id);

	SetTextureMatrix(0, Matrix4x4::identity);
	SetTextureMatrix(1, Matrix4x4::identity);
	SetTextureMatrix(2, Matrix4x4::identity);
	SetTextureMatrix(3, Matrix4x4::identity);

	SetTextureAddressing(0, TA_WRAP);
	SetTextureAddressing(1, TA_WRAP);
	SetTextureAddressing(2, TA_WRAP);
	SetTextureAddressing(3, TA_WRAP);
}

FullMaterial *GetDefaultMaterial()
{
	static FullMaterial def_mat;
	return &def_mat;
}


