#include <assert.h>
#include <vbuffer.h>
#include "geometry.h"
#include "gctypes.h"

// TODO: remove this... this is for the obscure D3DX cache-optimization trick below
//       measure the performance with and without it. If it is useful (I doubt it)
//       move in libgc, otherwise get rid of it asap.
#ifdef BUILD_DX9
#include <d3dx9.h>
#endif

// TODO : check Transform if anything else needs to be transformed

Geometry::Geometry ()
{
    vertex_array = NULL;
    normal_array = NULL;
    for (int i = 0; i < MAX_TEX_COORDS; i++)
    {
        tex_coord_array [i] = NULL;
    }
    tangent_array = NULL;
    color_array = NULL;
    index_array = NULL;
	bone_index_array = NULL;
	bone_weight_array = NULL;

    SetPrimitiveType (PT_TRIANGLE_LIST);
}


Geometry::~Geometry ()
{
    Release ();
}


// The following functions creates the arrays.
// if the array is already created, then the function deletes it first.
VertexArray *Geometry::CreateVertexArray ()
{
    ReleaseVertexArray ();
	try
	{
		vertex_array = new VertexArray ();
	}
	catch (...)
	{
		vertex_array = NULL;
	}
    return vertex_array;
}

NormalArray *Geometry::CreateNormalArray ()
{
    ReleaseNormalArray ();
	try
	{
		normal_array = new NormalArray();
	}
	catch (...)
	{
		normal_array = NULL;
	}
    return normal_array;
}

TangentArray *Geometry::CreateTangentArray ()
{
    ReleaseTangentArray ();
	try
	{
		tangent_array = new TangentArray ();
	}
	catch (...)
	{
		tangent_array = NULL;
	}
    return tangent_array;
}

ColorArray *Geometry::CreateColorArray ()
{
    ReleaseColorArray ();
	try
	{
		color_array = new ColorArray ();
	}
	catch (...)
	{
		color_array = NULL;
	}
    return color_array;
}

TexCoordArray *Geometry::CreateTexCoordArray (unsigned int idx, unsigned int tex_coord_count)
{
    if (idx >= MAX_TEX_COORDS)
    {
        return NULL;
    }
    ReleaseTexCoordArray (idx);
	try
	{
		tex_coord_array[idx] = new TexCoordArray (tex_coord_count);
	}
	catch (...)
	{
		tex_coord_array[idx] = NULL;
	}
    return tex_coord_array [idx];
}

IndexArray *Geometry::CreateIndexArray ()
{
    ReleaseIndexArray ();
    index_array = new IndexArray ();
    return index_array;

}

BoneIndexArray *Geometry::CreateBoneIndexArray()
{
    ReleaseBoneIndexArray ();
	try
	{
		bone_index_array = new BoneIndexArray ();
	}
	catch (...)
	{
		bone_index_array = NULL;
	}
    return bone_index_array;
}

BoneWeightArray *Geometry::CreateBoneWeightArray()
{
    ReleaseBoneWeightArray ();
	try
	{
		bone_weight_array = new BoneWeightArray ();
	}
	catch (...)
	{
		bone_weight_array = NULL;
	}
    return bone_weight_array;
}

// delete some array...
void Geometry::ReleaseVertexArray ()
{
    delete vertex_array;
    vertex_array = NULL;
}

void Geometry::ReleaseNormalArray ()
{
    delete normal_array;
    normal_array = NULL;
}

void Geometry::ReleaseTangentArray ()
{
    delete tangent_array;
    tangent_array = NULL;
}

void Geometry::ReleaseColorArray ()
{
    delete color_array;
    color_array = NULL;
}

void Geometry::ReleaseTexCoordArray (unsigned int idx)
{
    if (idx < MAX_TEX_COORDS)
    {
        delete tex_coord_array[idx];
        tex_coord_array[idx] = NULL;
    }
}

void Geometry::ReleaseIndexArray ()
{
    delete index_array;
    index_array = NULL;
}

void Geometry::ReleaseBoneIndexArray()
{
    delete bone_index_array;
    bone_index_array = NULL;
}

void Geometry::ReleaseBoneWeightArray()
{
    delete bone_weight_array;
    bone_weight_array = NULL;
}


// Get a pointer to some array, or null if array is not created.
VertexArray *Geometry::GetVertexArray () const
{
    return vertex_array;
}

NormalArray *Geometry::GetNormalArray () const
{
    return normal_array;
}

TangentArray *Geometry::GetTangentArray () const
{
    return tangent_array;
}

ColorArray *Geometry::GetColorArray () const
{
    return color_array;
}

TexCoordArray *Geometry::GetTexCoordArray (unsigned int idx) const
{
    if (idx < MAX_TEX_COORDS)
    {
        return tex_coord_array[idx];
    }
    return NULL;
}

IndexArray *Geometry::GetIndexArray () const
{
    return index_array;
}

BoneIndexArray *Geometry::GetBoneIndexArray() const
{
    return bone_index_array;
}

BoneWeightArray *Geometry::GetBoneWeightArray() const
{
    return bone_weight_array;
}


void Geometry::SetPrimitiveType (PrimitiveType pt)
{
    primitive_type = pt;
}

PrimitiveType Geometry::GetPrimitiveType () const
{
    return primitive_type;
}


void Geometry::Release ()
{
    ReleaseVertexArray ();
    ReleaseNormalArray ();
    ReleaseTangentArray ();
    ReleaseColorArray ();
    for (unsigned int i = 0; i < MAX_TEX_COORDS; i ++)
    {
        ReleaseTexCoordArray (i);
    }
    ReleaseIndexArray ();
	ReleaseBoneIndexArray();
	ReleaseBoneWeightArray();
}


VertexFormat Geometry::GetVertexFormat () const
{
    VertexFormat vf;

    if (vertex_array != NULL)
    {
        vf.vertex = true;
    }
    if (normal_array != NULL)
    {
        vf.normal = true;
    }
    if (tangent_array != NULL)
    {
        vf.tangent = true;
    }
    if (color_array != NULL)
    {
        vf.color = true;
    }
	if (bone_index_array != NULL)
    {
        vf.bone_index = true;
    }
	if (bone_weight_array != NULL)
    {
        vf.bone_weight = true;
    }
    vf.tc_count = 0;
    for (int i = 0; i < MAX_TEX_COORDS; i++)
    {
        if (tex_coord_array [i] != 0)
        {
            vf.tc_count = i + 1;
            vf.tc_stream_count [i] = tex_coord_array[i]->GetBasicTypeCount();
        }
        else
        {
            break;
        }
    }
    return vf;
}


bool Geometry::MapGeometry (void *dest, const VertexFormat &vf, unsigned int vertices_to_update) const
{
    VertexFormat current_vf = GetVertexFormat ();

    if (current_vf.ContainsFormat (vf) == false)
    {
        return false;
    }

    char *destination_ptr = (char*) dest;
    char *destination_ptr_index;

    unsigned int the_stride = vf.GetSize ();

    if (vf.vertex)
    {
        destination_ptr_index = &destination_ptr [vf.GetVertexOffset()];
        vertex_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    if (vf.normal)
    {
        destination_ptr_index = &destination_ptr [vf.GetNormalOffset()];
        normal_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    if (vf.color)
    {
        destination_ptr_index = &destination_ptr [vf.GetColorOffset()];
        color_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    if (vf.tangent)
    {
        destination_ptr_index = &destination_ptr [vf.GetTangentOffset()];
        tangent_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    if (vf.bone_weight)
    {
        destination_ptr_index = &destination_ptr [vf.GetBoneWeightOffset()];
        bone_weight_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    if (vf.bone_index)
    {
        destination_ptr_index = &destination_ptr [vf.GetBoneIndexOffset()];
        bone_index_array->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }
    for (unsigned int i = 0; i < vf.tc_count; i ++)
    {
        destination_ptr_index = &destination_ptr [vf.GetTexCoordOffset(i)];
        tex_coord_array[i]->MapWithStride (destination_ptr_index, the_stride, vertices_to_update);
    }

    return true;
}

bool Geometry::MapIndex (void *dest, unsigned int base) const
{
	if (index_array == NULL)
	{
		return false;
	}

	unsigned int *destination = (unsigned int *) dest;
	unsigned int *source = (unsigned int*) index_array->GetData ();

	for (unsigned int i = 0; i < index_array->GetElementsCount(); i++)
	{
		destination[i] = source[i] + base;
	}
	return true;
}

unsigned int Geometry::GetVertexCount () const
{
    if (vertex_array)
    {
        return vertex_array->GetElementsCount();
    }
    return 0;
}

unsigned int Geometry::GetIndexCount () const
{
	if (index_array)
	{
		return index_array->GetElementsCount();
	}
	return 0;
}


unsigned int Geometry::GetPrimitiveCount () const
{
	unsigned int pc = 0;
	if (index_array)
	{
		switch (primitive_type)
		{
		case PT_POINT_LIST: pc = index_array->GetElementsCount(); break;
	    case PT_LINE_LIST:  pc = index_array->GetElementsCount() / 2; break;
	    case PT_LINE_STRIP:  pc = index_array->GetElementsCount() - 1; break;
	    case PT_TRIANGLE_LIST:  pc = index_array->GetElementsCount() / 3; break;
	    case PT_TRIANGLE_STRIP:  pc = index_array->GetElementsCount() - 2; break;
	    case PT_TRIANGLE_FAN:  pc = index_array->GetElementsCount() - 2; break;
		default:
			assert(0);
		}
	}
	else
	{
		switch (primitive_type)
		{
		case PT_POINT_LIST: pc = vertex_array->GetElementsCount(); break;
	    case PT_LINE_LIST:  pc = vertex_array->GetElementsCount() / 2; break;
	    case PT_LINE_STRIP:  pc = vertex_array->GetElementsCount() - 1; break;
	    case PT_TRIANGLE_LIST:  pc = vertex_array->GetElementsCount() / 3; break;
	    case PT_TRIANGLE_STRIP:  pc = vertex_array->GetElementsCount() - 2; break;
	    case PT_TRIANGLE_FAN:  pc = vertex_array->GetElementsCount() - 2; break;
		default:
			assert(0);
		}
	}
	return pc;
}


bool Geometry::WriteToFP (VFS_FILE *fp)
{
	VertexFormat vf = GetVertexFormat ();
	try
	{
		if (vf.WriteToFP (fp) == false)
		{
			throw (int) 0;
		}
		unsigned int bool_int = 0;
		if (index_array)
		{
			bool_int = 1;
		}
		size_t bytes_write = vfs_write (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_write != sizeof (unsigned int))
		{
			throw (int) 1;
		}

		if (vertex_array)
		{
			unsigned int vertex_count = vertex_array->GetElementsCount();
			bytes_write = vfs_write (&vertex_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (vertex_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (normal_array)
		{
			unsigned int normal_count = normal_array->GetElementsCount();
			bytes_write = vfs_write (&normal_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (normal_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (tangent_array)
		{
			unsigned int tangent_count = tangent_array->GetElementsCount();
			bytes_write = vfs_write (&tangent_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (tangent_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (color_array)
		{
			unsigned int color_count = color_array->GetElementsCount();
			bytes_write = vfs_write (&color_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (color_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (bone_weight_array)
		{
			unsigned int bw_count = bone_weight_array->GetElementsCount();
			bytes_write = vfs_write (&bw_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (bone_weight_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (bone_index_array)
		{
			unsigned int bi_count = bone_index_array->GetElementsCount();
			bytes_write = vfs_write (&bi_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (bone_index_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}

		for (unsigned int i = 0; i < vf.tc_count; i ++)
		{
			unsigned int tc_count = tex_coord_array[i]->GetElementsCount();
			bytes_write = vfs_write (&tc_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (tex_coord_array[i]->WriteToFP (fp) == false)
			{
				throw (int) 1;
			}
		}

		if (index_array)
		{
			unsigned int i_count = index_array->GetElementsCount();
			bytes_write = vfs_write (&i_count, 1, sizeof (unsigned int), fp);
			if (bytes_write != sizeof (unsigned int))
			{
				throw (int) 1;
			}

			if (index_array->WriteToFP (fp) == false)
			{
				throw (int) 0;
			}
		}

	}
	catch (...)
	{
		return false;
	}
	return true;
}

bool Geometry::ReadFromFP (VFS_FILE *fp)
{
	VertexFormat vf;
	try
	{
		if (vf.LoadFromFP(fp) == false)
		{
			throw (int) 0;
		}
		unsigned int bool_int = 0;

		size_t bytes_read = vfs_read (&bool_int, 1, sizeof (unsigned int), fp);
		if (bytes_read != sizeof (unsigned int))
		{
			throw (int) 1;
		}


		if (vf.vertex)
		{
			unsigned int vertex_count;
			bytes_read = vfs_read (&vertex_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateVertexArray () == NULL)
			{
				throw (int) 0;
			}
			if (vertex_array->Create (vertex_count) == NULL)
			{
				throw (int) 0;
			}
			if (vertex_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (vf.normal)
		{
			unsigned int normal_count;
			bytes_read = vfs_read (&normal_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateNormalArray () == NULL)
			{
				throw (int) 0;
			}
			if (normal_array->Create (normal_count) == NULL)
			{
				throw (int) 0;
			}
			if (normal_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (vf.tangent)
		{
			unsigned int tangent_count;
			bytes_read = vfs_read (&tangent_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateTangentArray () == NULL)
			{
				throw (int) 0;
			}
			if (tangent_array->Create (tangent_count) == NULL)
			{
				throw (int) 0;
			}
			if (tangent_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (vf.color)
		{
			unsigned int color_count;
			bytes_read = vfs_read (&color_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateColorArray () == NULL)
			{
				throw (int) 0;
			}
			if (color_array->Create (color_count) == NULL)
			{
				throw (int) 0;
			}
			if (color_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (vf.bone_weight)
		{
			unsigned int bw_count;
			bytes_read = vfs_read (&bw_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateBoneWeightArray () == NULL)
			{
				throw (int) 0;
			}
			if (bone_weight_array->Create (bw_count) == NULL)
			{
				throw (int) 0;
			}
			if (bone_weight_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}
		if (vf.bone_index)
		{
			unsigned int bi_count;
			bytes_read = vfs_read (&bi_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateBoneIndexArray () == NULL)
			{
				throw (int) 0;
			}
			if (bone_index_array->Create (bi_count) == NULL)
			{
				throw (int) 0;
			}
			if (bone_index_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}

		for (unsigned int i = 0; i < vf.tc_count; i ++)
		{
			unsigned int tc_count;
			bytes_read = vfs_read (&tc_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int)) {
				throw 0;
			}

			if (CreateTexCoordArray (i, vf.tc_stream_count[i]) == NULL) {
				throw 0;
			}
			if (tex_coord_array[i]->Create (tc_count) == NULL) {
				throw 0;
			}
			if (tex_coord_array[i]->LoadFromFP (fp) == false) {
				throw 1;
			}
		}

		if (bool_int == 1)
		{
			unsigned int i_count;
			bytes_read = vfs_read (&i_count, 1, sizeof (unsigned int), fp);
			if (bytes_read != sizeof (unsigned int))
			{
				throw (int) 0;
			}
			if (CreateIndexArray () == NULL)
			{
				throw (int) 0;
			}
			if (index_array->Create (i_count) == NULL)
			{
				throw (int) 0;
			}
			if (index_array->LoadFromFP (fp) == false)
			{
				throw (int) 0;
			}
		}

	}
	catch (...)
	{
		return false;
	}
	return true;
}

void Geometry::Transform (const Matrix4x4 &xform)
{
	VertexArray *va = GetVertexArray ();
	if (va) {
		char *va_data = (char*) va->GetData();
		char va_elm_size = va->GetSingleElementSize();
		for (unsigned int i = 0; i < va->GetElementsCount(); i++) {
			Vector3 tv;
			memcpy(&tv, &va_data[va_elm_size * i], va_elm_size);
			tv.Transform (xform);
			memcpy(&va_data[va_elm_size * i], &tv, va_elm_size);
		}
	}

	Matrix3x3 rot_matrix = xform;

	NormalArray *na = GetNormalArray ();
	if (na) {
		char *na_data = (char*) na->GetData();
		char na_elm_size = na->GetSingleElementSize();
		for (unsigned int i = 0; i < na->GetElementsCount(); i++) {
			Vector3 tv;
			memcpy(&tv, &na_data[na_elm_size * i], na_elm_size);
			tv.Transform (xform);
			memcpy(&na_data[na_elm_size * i], &tv, na_elm_size);
			//ndata[i].Transform (rot_matrix);
		}
	}

	TangentArray *ta = GetTangentArray ();
	if (ta)
	{
		char *ta_data = (char*)ta->GetData();
		char ta_elm_size = ta->GetSingleElementSize();
		for (unsigned int i = 0; i < ta->GetElementsCount(); i++)
		{
			Vector3 tv;
			memcpy(&tv, &ta_data[ta_elm_size * i], ta_elm_size);
			tv.Transform (xform);
			memcpy(&ta_data[ta_elm_size * i], &tv, ta_elm_size);
			//tdata[i].Transform (rot_matrix);
		}
	}
}




bool Geometry::CalculateBoundingPoints (Vector3 &min_point, Vector3 &max_point)
{
	if (vertex_array == NULL)
	{
		return false;
	}
	Vector3  min_b, max_b;
	char *gdata = (char *)vertex_array->GetData();
	int gdata_elm_size = vertex_array->GetSingleElementSize();
	if (gdata == NULL)
	{
		return false;
	}

	vertex_array->GetElement(&min_b, 0);
	vertex_array->GetElement(&max_b, 0);

	for (unsigned int i = 1; i < vertex_array->GetElementsCount(); i++)
	{
		Vector3 gdata_i;
		vertex_array->GetElement(&gdata_i, i);
		if (min_b.x > gdata_i.x)
		{
			min_b.x = gdata_i.x;
		}
		if (min_b.y > gdata_i.y)
		{
			min_b.y = gdata_i.y;
		}
		if (min_b.z > gdata_i.z)
		{
			min_b.z = gdata_i.z;
		}

		if (max_b.x	< gdata_i.x)
		{
			max_b.x = gdata_i.x;
		}
		if (max_b.y	< gdata_i.y)
		{
			max_b.y = gdata_i.y;
		}
		if (max_b.z	< gdata_i.z)
		{
			max_b.z = gdata_i.z;
		}
	}

	min_point = min_b;
	max_point = max_b;

	return true;
}

struct Face
{
	unsigned int v1, v2, v3;
};

void Geometry::Optimize ()
{
#ifdef BUILD_DX9
	unsigned int faces_count = GetIndexArray()->GetElementsCount() / 3;
	unsigned int vertices_count = GetVertexArray()->GetElementsCount();

	unsigned int *faces_remap;
	Face *faces;
	faces = (Face *) GetIndexArray()->GetData();
	faces_remap = new unsigned int [faces_count];
	for (unsigned int i = 0; i < faces_count; i++)
	{
		faces_remap[i] = i;
	}

	D3DXOptimizeFaces(faces, faces_count, vertices_count, TRUE, (DWORD*) faces_remap);

	Face *face2 = new Face [faces_count];
	for (unsigned int i = 0; i < faces_count; i++)
	{
		face2[i] = faces[faces_remap[i]];
	}
	memcpy (faces, face2, sizeof (unsigned int)*faces_count*3);
	delete [] face2;


	unsigned int *vertex_remap = new unsigned int [vertices_count];
	for (unsigned int i = 0; i < vertices_count; i++)
	{
		vertex_remap[i] = i;
	}

	D3DXOptimizeVertices (faces, faces_count, vertices_count, TRUE, (DWORD*)vertex_remap);

	if (vertex_array)
	{
		vertex_array->Remap (vertex_remap);
	}
	if (normal_array)
	{
		normal_array->Remap (vertex_remap);
	}

	for (unsigned int i = 0; i < MAX_TEX_COORDS; i++)
	{
		if (tex_coord_array[i])
		{
			tex_coord_array[i]->Remap(vertex_remap);
		}
	}
    if (tangent_array)
	{
		tangent_array->Remap (vertex_remap);
	}
    if (color_array)
	{
		color_array->Remap(vertex_remap);
	}
	if (bone_index_array)
	{
		bone_index_array->Remap(vertex_remap);
	}
	if (bone_weight_array)
	{
		bone_weight_array->Remap (vertex_remap);
	}
	for (unsigned int i = 0; i < faces_count; i++)
	{
		faces[i].v1 = vertex_remap[faces[i].v1];
		faces[i].v2 = vertex_remap[faces[i].v2];
		faces[i].v3 = vertex_remap[faces[i].v3];
	}
	delete [] vertex_remap;
#endif	// BUILD_DX9
}


bool UpdateVertexBuffer (unsigned int id, unsigned int vertex_idx, const Geometry *geom, unsigned int total_vertices)
{
    unsigned int vertices_to_update = total_vertices;
    if (vertices_to_update == 0)
    {
        vertices_to_update = geom->GetVertexCount();
    }
    void *ptr = LockVertexBuffer (id, vertex_idx, vertices_to_update);
    if (ptr)
    {
        // den tsekarw an to id exists, giati gia na exei perasei apo to LockVertexBuffer shmainei oti einai ok.
        bool done = geom->MapGeometry (ptr, GetVertexFormat (id), vertices_to_update);
        UnlockVertexBuffer (id);
        return done;
    }
    return false;
}

bool UpdateIndexBuffer (unsigned int id, unsigned int index_start, const Geometry *geom, unsigned int base)
{
	if (geom->GetIndexCount() == 0)
	{
		return false;
	}

	void *the_ptr = LockIndexBuffer (id, index_start, geom->GetIndexCount());
	if (the_ptr)
	{
		bool ok = geom->MapIndex (the_ptr, base);
		UnlockIndexBuffer (id);
		return ok;
	}
	return false;
}


