#ifndef GEOMETRY_H
#define GEOMETRY_H

#include <stdio.h>
#include "elements.h"
#include <gctypes.h>
#include "vertexformat.h"
#include "math3d.h"


class Geometry
{
protected:
    PrimitiveType    primitive_type;
    VertexArray     *vertex_array;
    NormalArray     *normal_array;
    TexCoordArray   *tex_coord_array[MAX_TEX_COORDS];
    TangentArray    *tangent_array;
    ColorArray      *color_array;
    IndexArray      *index_array;
	BoneIndexArray  *bone_index_array;
	BoneWeightArray *bone_weight_array;

public:
    Geometry ();
    ~Geometry ();

    // The following functions creates the arrays.
    // if the array is already created, then the function deletes it first.
    VertexArray *CreateVertexArray ();
    NormalArray *CreateNormalArray ();
    TangentArray *CreateTangentArray ();
    ColorArray *CreateColorArray ();
    TexCoordArray *CreateTexCoordArray (unsigned int idx, unsigned int tex_coord_count = 2);
    IndexArray *CreateIndexArray ();
	BoneIndexArray *CreateBoneIndexArray();
	BoneWeightArray *CreateBoneWeightArray();

    // delete some array...
    void ReleaseVertexArray ();
    void ReleaseNormalArray ();
    void ReleaseTangentArray ();
    void ReleaseColorArray ();
    void ReleaseTexCoordArray (unsigned int idx);
    void ReleaseIndexArray ();
	void ReleaseBoneIndexArray();
	void ReleaseBoneWeightArray();

    // Get a pointer to some array, or null if array is not created.
    VertexArray *GetVertexArray () const;
    NormalArray *GetNormalArray () const;
    TangentArray *GetTangentArray () const;
    ColorArray *GetColorArray () const;
    TexCoordArray *GetTexCoordArray (unsigned int idx) const;
    IndexArray *GetIndexArray () const;
	BoneIndexArray *GetBoneIndexArray() const;
	BoneWeightArray *GetBoneWeightArray() const;

    void SetPrimitiveType (PrimitiveType pt);
    PrimitiveType GetPrimitiveType () const;

	void Transform (const Matrix4x4 &xform);

    void Release ();

    VertexFormat GetVertexFormat () const;

    bool MapGeometry (void *dest, const VertexFormat &vf, unsigned int vertices_to_update = 0) const;
	bool MapIndex (void *dest, unsigned int base) const;

    // returns the number of vertex_array elements
    unsigned int GetVertexCount () const;
	unsigned int GetIndexCount () const;
	unsigned int GetPrimitiveCount () const;

	bool WriteToFP (VFS_FILE *fp);
	bool ReadFromFP (VFS_FILE *fp);

	bool CalculateBoundingPoints (Vector3 &min_point, Vector3 &max_point);

	void Optimize ();
};


// if total_vertices == 0 (default) then copies the whole geometry
bool UpdateVertexBuffer (unsigned int id, unsigned int vertex_idx, const Geometry *geom, unsigned int total_vertices = 0);
bool UpdateIndexBuffer (unsigned int id, unsigned int index_start, const Geometry *geom, unsigned int base);



#endif //GEOMETRY_H

