#include <algorithm>
#include "objectgroup.h"
#include <errlog.h>
#include <progress.h>

// ----------------------
// sampler routines
// ----------------------
// float frame = time * fps;
static inline int Once(int f, unsigned int size)
{
	if (f < 0) return 0;
	if ((unsigned int)f >= size) return size - 1;
	return f;
}

static inline int Loop(int f, unsigned int size)
{
	if (!size) return 0;
	f = f % size;
	return f;
}

static inline int Bounce(int f, unsigned int size)
{
	if (!size) return 0;
	if (f < 0)
	{
		f = -f;
		f += size;
	}
	if (! ((f/size) % 2))
	{
		return f;
	}
	return (size -1) - f;
}

static inline PRS PointSampleOnce(const Timeline &tm, float frame)
{
	int fr = (int) (frame + 0.5f);
	return tm.keys[Once(fr, (unsigned int)tm.keys.size())];
}
static inline PRS PointSampleLoop(const Timeline &tm, float frame)
{
	int fr = (int) (frame + 0.5f);
	return tm.keys[Loop(fr, (unsigned int)tm.keys.size())];
}
static inline PRS PointSampleBounce(const Timeline &tm, float frame)
{
	int fr = (int) (frame + 0.5f);
	return tm.keys[Bounce(fr, (unsigned int)tm.keys.size())];
}


static inline PRS LinearSampleOnce(const Timeline &tm, float frame)
{
	float fr1 = floorf(frame);
	float fr2 = fr1 + 1.0f;
	float p = frame - fr1;
	PRS p1 = PointSampleOnce(tm, fr1);
	PRS p2 = PointSampleOnce(tm, fr2);
	return Slerp(p1, p2, p);
}
static inline PRS LinearSampleLoop(const Timeline &tm, float frame)
{
	float fr1 = floorf(frame);
	float fr2 = fr1 + 1.0f;
	float p = frame - fr1;

	PRS p1 = PointSampleLoop(tm, fr1);
	
	PRS p2 = PointSampleLoop(tm, fr2);
	
	PRS ret =  Slerp(p1, p2, p);	
	
	return ret;
}
static inline PRS LinearSampleBounce(const Timeline &tm, float frame)
{
	float fr1 = floorf(frame);
	float fr2 = fr1 + 1.0f;
	float p = frame - fr1;
	PRS p1 = PointSampleBounce(tm, fr1);
	PRS p2 = PointSampleBounce(tm, fr2);
	return Slerp(p1, p2, p);
}

PRS Sample(const Timeline &tm, float frame, 
		   AnimSampleMode asmp, AnimLoopMode alm)
{
	switch (asmp)
	{
		case ASM_POINT:
			switch (alm)
			{
				case ALM_ONCE:
					return PointSampleOnce(tm, frame);
				case ALM_LOOP:
					return PointSampleLoop(tm, frame);
				case ALM_BOUNCE:
					return PointSampleBounce(tm, frame);
			}
		case ASM_LINEAR:
			switch (alm)
			{
				case ALM_ONCE:
					return LinearSampleOnce(tm, frame);
				case ALM_LOOP:
					return LinearSampleLoop(tm, frame);
				case ALM_BOUNCE:
					return LinearSampleBounce(tm, frame);
			}
	}

	return PRS();
}

// Animation class implementation
Animation::Animation()
{
	fps = 24.0f;

	asmp = ASM_LINEAR;
	alm = ALM_LOOP;
}

Animation::~Animation()
{
	Destroy();
}

void Animation::Destroy()
{
	for (unsigned int i=0; i<nodes.size(); i++)
	{
		nodes[i].keys.clear();
	}
	nodes.clear();
}

bool Animation::Load(const char *filename)
{
	Destroy();

	VFS_FILE *fp = vfs_open(filename, "rb");
	if (!fp) return false;

	AnimationFile *anm = new AnimationFile();
	if (!anm->Load(fp))
	{
		delete anm;
		vfs_close(fp);
		return false;
	}

	fps = (float) anm->GetFPS();

	unsigned int kc = anm->GetKeyframeChunkCount();
	for (unsigned int i=0; i<kc; i++)
	{
		Timeline new_tl;
		KeyframeChunk *kfc = anm->GetKeyframeChunk(i);
		if (!kfc)
		{
			new_tl.is_static = true;
			nodes.push_back(new_tl);
			continue;
		}
		unsigned int key_count = kfc->GetKeyframeCount();
		if (!key_count)
		{
			new_tl.is_static = true;
			new_tl.static_local = kfc->GetDefaultPRS();
			nodes.push_back(new_tl);	
			continue;
		}

		new_tl.is_static = false;
		PRS *keys = kfc->GetKeyframes();
		for (unsigned int j=0; j<key_count; j++)
		{
			new_tl.keys.push_back(keys[j]);
		}

		nodes.push_back(new_tl);
	}

	delete anm;
	vfs_close(fp);
	return true;
}

void Animation::SetSampleMode(AnimSampleMode asmp)
{
	this->asmp = asmp;
}
void Animation::SetLoopMode(AnimLoopMode alm)
{
	this->alm = alm;
}



void Animation::ProcessAnimation(float t)
{
	for (unsigned int i=0; i<nodes.size(); i++)
	{
		// ultra hack
		//if (i!=2) continue;
		if (!nodes[i].is_static)
		{
			PRS prs = Sample(nodes[i], t * fps, asmp, alm);
			nodes[i].curr_local = prs;
		}
		else
		{
			PRS prs = nodes[i].static_local;
			nodes[i].curr_local = prs;
		}
	}
}

// valid only after a call to ProcessAnimation()
PRS Animation::GetNodeLocal(unsigned int node)
{
	if (node >= nodes.size()) return PRS();
	PRS ret = nodes[node].curr_local;
	return ret;
}

void Animation::SetNodeLocal(unsigned int node, const PRS &prs)
{
	if (node >= nodes.size()) return;
	nodes[node].curr_local = prs;
}

Timeline Animation::GetKeys(unsigned int node)
{
	return nodes[node];
}

void Animation::DeleteLastKey()
{
	for (unsigned int i=0; i<nodes.size(); i++)
	{
		unsigned int count = (unsigned int) nodes[i].keys.size();
		if (count > 0)
		{
			nodes[i].keys.erase(nodes[i].keys.begin() + (count - 1),
				nodes[i].keys.begin() + count);
		}
	}
}

float Animation::GetDuration() const
{
	unsigned int frames;
	for (unsigned int i=0; i<nodes.size(); i++)
	{
		frames = (unsigned int) nodes[i].keys.size();
		if (frames) break;
	}
	return frames / fps;
}

float Animation::GetFPS() const
{
	return fps;
}

void Animation::SetFPS(float fps)
{
	this->fps = fps;
}

// ------------------------------------------------------------------------

// OGNode class implementation
OGNode::OGNode()
{
	ground0 = ground1 = 0;
	local_matrix = model_matrix = static_model_matrix = pivot_matrix = Matrix4x4();
}
OGNode::~OGNode()
{
	for (unsigned int i=0; i<children.size(); i++)
	{
		if (children[i])
		{
			delete children[i];
			children[i] = 0;
		}
	}
	children.clear();
}

std::string OGNode::GetName() const
{
	return name;
}

void OGNode::SetName(const std::string &name)
{
	this->name = name;
}

unsigned int OGNode::GetID() const
{
	return id;
}

void OGNode::SetID(unsigned int id)
{
	this->id = id;
}

unsigned int OGNode::GetChildrenCount() const
{
	return (unsigned int)children.size();
}

OGNode *OGNode::GetChild(unsigned int which) const
{
	if (which >= children.size()) return 0;
	return children[which];
}

void OGNode::Mount(OGNode *child)
{
	if (!child) return;
	// avoid duplicate mounts
	Unmount(child);
	children.push_back(child);
	child->parent = this;
}
void OGNode::Unmount(OGNode *child)
{
	for (unsigned int i=0; i<children.size(); i++)
	{
		if (children[i] == child)
		{
			children.erase(children.begin() + i, children.begin() + i + 1);
			return;
		}
	}
}

void OGNode::CalculateStaticModelMatrix()
{
	static_model_matrix = local_matrix * pivot_matrix;
	for (unsigned int i=0; i<children.size(); i++)
	{
		if (!children[i]) continue;
		children[i]->local_matrix = local_matrix * children[i]->local_matrix ;
		children[i]->CalculateStaticModelMatrix();
	}
}

void OGNode::CalculateModelMatrix()
{
	model_matrix = local_matrix * pivot_matrix;
	for (unsigned int i=0; i<children.size(); i++)
	{
		if (!children[i]) continue;
		children[i]->local_matrix = local_matrix * children[i]->local_matrix ;
		//children[i]->local_matrix = children[i]->local_matrix * local_matrix ;
		children[i]->CalculateModelMatrix();
	}
}

void OGNode::SetModelMatrix(const Matrix4x4 &model)
{
	this->model_matrix = model;
}

void OGNode::SetGround(unsigned int ground)
{
	this->ground0 = ground;
}
void OGNode::SetGround0(unsigned int ground)
{
	this->ground0 = ground;
}
void OGNode::SetGround1(unsigned int ground)
{
	this->ground1 = ground;
}
unsigned int OGNode::GetGround() const
{
	return ground0;
}
unsigned int OGNode::GetGround0() const
{
	return ground0;
}
unsigned int OGNode::GetGround1() const
{
	return ground1;
}

void OGNode::AddToAllKeys()
{
	all_keys.push_back(model_matrix);
}

// ObjectGroup class implementation
ObjectGroup::ObjectGroup()
{
	Init ();
	anim_blend = false;
	anim_blend_param = 0.0f;
	anm[0] = NULL;
	anm[1] = NULL;
	anm_time[0] = 0.0f;
	anm_time[1] = 0.0f;
	start[0] = PRS();
	start[1] = PRS();
	model_matrix.ResetIdentity();
	mesh_as_proxy = false;
	random_move = false;
}

ObjectGroup::~ObjectGroup()
{
	Destroy();
}

void ObjectGroup::Init ()
{
	// create root node
	root_node = new OGNode();
	is_static = false;
	force_static = false;
	filename = 0;
	use_model_matrix = false;
}


void ObjectGroup::Destroy()
{
	printf("Destroying OG: %s\n", filename);
	// delete all animations
	std::map<std::string, Animation*>::iterator it;
	for (it = animations.begin(); it != animations.end(); it++)
	{
		if (it->second)
		{
			delete it->second;
			it->second = 0;
		}
	}
	animations.clear();
	anm.clear();
	start.clear();
	anm_time.clear();
	anim_weights.clear();

	// delete nodes
	if (root_node)
	{
		delete root_node;
		root_node = 0;
	}
	nodes.clear();
	node_name_map.clear();

	// delete meshes
	for (unsigned int i=0; i<mesh_list.size(); i++)
	{
		if (mesh_list[i])
		{
			delete mesh_list[i]->geom;
			delete mesh_list[i];
			mesh_list[i] = 0;
		}
	}
	mesh_list.clear();

	// delete materials
	for (unsigned int i=0; i<material_list.size(); i++)
	{
		if (material_list[i])
		{
			delete material_list[i];
			material_list[i] = 0;
		}
	}

	// delete lights
	for (unsigned int i=0; i<light_list.size(); i++)
	{
		if (light_list[i])
		{
			delete light_list[i];
			light_list[i] = 0;
		}
	}

	// delete convex areas
	for (unsigned int i=0; i<convex_areas.size(); i++)
	{
		if (convex_areas[i])
		{
			delete convex_areas[i];
			convex_areas[i] = 0;
		}
	}
	convex_areas.clear();

	// delete cameras
	std::map<std::string, MatrixCamera*>::iterator cit;
	for (cit = cameras.begin(); cit != cameras.end(); cit++)
	{
		if (cit->second)
		{
			delete cit->second;
			cit->second = 0;
		}
	}
	cameras.clear();

	mesh_material_map.clear();
	mesh_node_map.clear();

	mesh_bone_map_map.clear();

	delete [] filename;
}

void ObjectGroup::ProcessGeometry(MeshChunk* mesh)
{
	ComplexMesh *added_mesh = new ComplexMesh(); 

	//added_mesh->random_move = this->random_move;
	added_mesh->geom = new Geometry ();
	//added_mesh->material_id = mesh->GetMaterialID();
	mesh_material_map[added_mesh] = mesh->GetMaterialID();
	NodeChunk *parent_node = (NodeChunk*)mesh->GetParent();
	//added_mesh->node_id = parent_node->GetNodeID();
	mesh_node_map[added_mesh] = parent_node->GetNodeID();

	mesh_list.push_back (added_mesh);
	//added_mesh->model_matrix = &(GetNode(added_mesh)->model_matrix);
	//added_mesh->node = GetNode(added_mesh);
	GeometryArray* mesh_array = mesh->GetArray(POSITION_ARRAY);
	if (mesh_array)
	{
		VertexArray *g_vertex = added_mesh->geom->CreateVertexArray ();
		void *dest_data = g_vertex->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_vertex->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(NORMAL_ARRAY);
	if (mesh_array)
	{
		NormalArray *g_normal = added_mesh->geom->CreateNormalArray ();
		void *dest_data = g_normal->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_normal->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(TANGENT_ARRAY);
	if (mesh_array)
	{
		TangentArray *g_tangent = added_mesh->geom->CreateTangentArray ();
		void *dest_data = g_tangent->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_tangent->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(COLOR_ARRAY);
	if (mesh_array)
	{
		ColorArray *g_color = added_mesh->geom->CreateColorArray ();
		void *dest_data = g_color->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_color->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(INDEX_ARRAY);
	if (mesh_array)
	{
		IndexArray *g_index = added_mesh->geom->CreateIndexArray ();
		void *dest_data = g_index->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_index->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(TEXCOORD0_ARRAY);
	if (mesh_array)
	{
		TexCoordArray *g_texcoord = added_mesh->geom->CreateTexCoordArray (0, mesh_array->GetBasicTypeCount());
		void *dest_data = g_texcoord->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_texcoord->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(TEXCOORD1_ARRAY);
	if (mesh_array)
	{
		TexCoordArray *g_texcoord = added_mesh->geom->CreateTexCoordArray (1, mesh_array->GetBasicTypeCount());
		void *dest_data = g_texcoord->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_texcoord->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(TEXCOORD2_ARRAY);
	if (mesh_array)
	{
		TexCoordArray *g_texcoord = added_mesh->geom->CreateTexCoordArray (2, mesh_array->GetBasicTypeCount());
		void *dest_data = g_texcoord->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_texcoord->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(TEXCOORD3_ARRAY);
	if (mesh_array)
	{
		TexCoordArray *g_texcoord = added_mesh->geom->CreateTexCoordArray (3, mesh_array->GetBasicTypeCount());
		void *dest_data = g_texcoord->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_texcoord->GetSingleElementSize());
	}

	//added_mesh->skin = false;

	mesh_array = mesh->GetArray(BONE_WEIGHT_ARRAY);
	if (mesh_array)
	{
		//added_mesh->skin = true;
		BoneWeightArray *g_bweight = added_mesh->geom->CreateBoneWeightArray ();
		void *dest_data = g_bweight->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_bweight->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(BONE_INDEX_ARRAY);
	if (mesh_array)
	{
		//added_mesh->skin = true;
		BoneIndexArray *g_bindex = added_mesh->geom->CreateBoneIndexArray ();
		void *dest_data = g_bindex->Create (mesh_array->GetElementCount());
		memcpy (dest_data, mesh_array->GetArrayData(), mesh_array->GetElementCount()*g_bindex->GetSingleElementSize());
	}

	mesh_array = mesh->GetArray(BONE_MAP_ARRAY);
	if (mesh_array)
	{
		unsigned int *indices = (unsigned int *)mesh_array->GetArrayData();

		unsigned int count = mesh_array->GetElementCount();
		if (count > MAX_BONES) count = MAX_BONES;
		//added_mesh->bone_count = count;
		mesh_bone_map_map[added_mesh].count = count;
		Info("Bone count: %d\n", count);
		memcpy(mesh_bone_map_map[added_mesh].map, indices, count * sizeof(unsigned int));
	}
	//added_mesh->geom->Optimize();
}

void ObjectGroup::ProcessProxy(MeshChunk *mesh)
{
	if (mesh->GetArray(BONE_WEIGHT_ARRAY))
	{
		return;
	}
	Info("Adding new proxy\n");
	GeometryArray* pos_array = mesh->GetArray(POSITION_ARRAY);
	if (!pos_array) return;
	GeometryArray* index_array = mesh->GetArray(INDEX_ARRAY);
	if (!index_array) return;

	Vector3 *verts = (Vector3*)pos_array->GetArrayData();
	unsigned int *indices = (unsigned int*) index_array->GetArrayData();
	unsigned int triangle_count = index_array->GetElementCount() / 3;

	CollisionMesh *added_mesh = new CollisionMesh();
	
	GeometryArray *prop0_array = mesh->GetArray(TRIANGLE_PROPERTIES_ARRAY0);
	unsigned int *props0 = 0;
	if (prop0_array) props0 = (unsigned int*)prop0_array->GetArrayData();

	GeometryArray *prop1_array = mesh->GetArray(TRIANGLE_PROPERTIES_ARRAY1);
	unsigned int *props1 = 0;
	if (prop1_array) props1 = (unsigned int*)prop1_array->GetArrayData();

	// init collision mesh
	for (unsigned int i=0; i<triangle_count; i++)
	{
		Triangle t;
		t.SetVertex(0, verts[indices[3*i + 0]]);
		t.SetVertex(1, verts[indices[3*i + 1]]);
		t.SetVertex(2, verts[indices[3*i + 2]]);
		added_mesh->AddTriangle(t);
		TriangleProperties p;
		if (props0) p.prop0 = props0[i];
		if (props1) p.prop1 = props1[i];
			
		added_mesh->SetTriangleProperties(i, p);
	}
	added_mesh->Finalize();

	// keep parent node
	NodeChunk *parent_node = (NodeChunk*)mesh->GetParent();
	collision_nodes.push_back(parent_node->GetNodeID());

	// set initial xform
	added_mesh->SetTransform(parent_node->GetTransform());

	// add mesh
	AddCollisionObject(added_mesh);
}

void ObjectGroup::ProcessConvexArea(MeshChunk *mesh, bool link)
{
	if (!mesh) return;
	if (!mesh->GetArray(POSITION_ARRAY)) return;
	GeometryArray *va = mesh->GetArray(POSITION_ARRAY);
	Vector3 *verts = (Vector3*) va->GetArrayData();
	ConvexArea *ca = new ConvexArea();
	for (unsigned int i=0; i<mesh->GetTriangleCount(); i++)
	{
		unsigned int a, b, c;
		mesh->GetTriangle(i, a, b, c);
		ca->AddTriangle(verts[a], verts[b], verts[c]);
	}
	if (link) ca->SetLink(true);
	convex_areas.push_back(ca);
}

void ObjectGroup::PrepareProxy()
{
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		collision_objects[i]->SetTransform(nodes[collision_nodes[i]]->GetModelMatrix());
	}
}

void ObjectGroup::ProcessLight(LightChunk* light)
{
	RLight *rlight = 0;
	PRS prs;
	//FILE *fp;
	Vector3  dir;

	switch (light->GetType())
	{
	case LT_DIRECTIONAL:
		break;

	case LT_OMNI:
		rlight = new RLight ();
		rlight->SetType (RL_OMNI);
		prs = ((NodeChunk*)light->GetParent())->GetWorldPRS ();
		rlight->SetPosition (prs.GetPosition());
		rlight->SetAttenStart (light->GetAttenStart() + 200);
		rlight->SetAttenEnd (light->GetAttenEnd() + 250);

		rlight->SetAmbient (light->GetColor());
		rlight->SetDiffuse (light->GetColor());
		rlight->SetSpecular (light->GetColor());
		break;
	case LT_SPOT:
		rlight = new RLight ();
		rlight->SetType (RL_SPOT);
		prs = ((NodeChunk*)light->GetParent())->GetWorldPRS ();
		rlight->SetPosition (prs.GetPosition());
		rlight->SetAttenStart (light->GetAttenStart());
		rlight->SetAttenEnd (light->GetAttenEnd());
		rlight->SetInnerCone(light->GetHotSpot());
		rlight->SetOuterCone(light->GetFallOff());
		dir = Vector3 (0.0f, -1.0f, 0.0f);
		dir.Transform (prs.GetRotation().GetRotationMatrix());
		rlight->SetDirection(dir);//prs.GetRotation().v.Normalized());

		rlight->SetAmbient (light->GetColor());
		rlight->SetDiffuse (light->GetColor());
		rlight->SetSpecular (light->GetColor());
		rlight->SetOrientation (prs.GetRotation());

		/*fp = fopen ("c:/spot.txt", "wt");
		fprintf (fp, "position : %f %f %f \n", rlight->GetPosition().x, rlight->GetPosition().y, rlight->GetPosition().z);
		fprintf (fp, "inner : %f  \n", rlight->GetInnerCone());
		fprintf (fp, "outer : %f  \n", rlight->GetOuterCone());
		fprintf (fp, "dir : %f %f %f \n", rlight->GetDirection().x, rlight->GetDirection().y, rlight->GetDirection().z);
		fclose (fp);*/
		break;

	default:
		break;
	}

	if (rlight)
	{
		light_list.push_back  (rlight);
	}

}

void ObjectGroup::ProcessCamera(CameraChunk *camera)
{
	NodeChunk *parent = (NodeChunk*) camera->GetParent();
	
	// safe to assume that parent node is already in the list
	MatrixCamera *cam = new MatrixCamera;
	cam->SetAspect(4.0f / 3.0f);
	cam->SetFOV(3.0f / 4.0f * camera->GetFOV() * 180 / Pi);
	cam->SetNearClip(camera->GetNearClip());
	cam->SetFarClip(camera->GetFarClip());
	cam->SetViewMatrix(&nodes[parent->GetNodeID()]->model_matrix);

	cameras[parent->GetName()] = cam;
}

OGNode *ObjectGroup::ProcessNode(NodeChunk *node)
{
	OGNode * new_node = new OGNode();
	nodes.push_back(new_node);
	new_node->SetName(node->GetName());
	new_node->id = node->GetNodeID();
	node_name_map[node->GetName()] = new_node;
	new_node->pivot_matrix = node->GetPivotPRS().GetTransform();
	new_node->local_matrix = node->GetLocalPRS().GetTransform();
	new_node->model_matrix = node->GetWorldPRS().GetTransform() * 
		node->GetPivotPRS().GetTransform();
	new_node->local_matrix = node->GetLocalPRS().GetTransform();

	unsigned int node_meshes = node->GetMeshCount();
	for (unsigned int i = 0; i < node_meshes; i++)
	{
		ProcessGeometry(node->GetMesh(i));

		if (mesh_as_proxy)
			ProcessProxy(node->GetMesh(i));
	}

	unsigned int node_proxys = node->GetProxyCount();
	for (unsigned int i=0; i<node_proxys; i++)
	{
		ProcessProxy(node->GetProxy(i));
	}

	unsigned int node_sets = node->GetPVSCount();
	for (unsigned int i=0; i<node_sets; i++)
	{
		ProcessConvexArea(node->GetPVS(i), false);
	}

	unsigned int node_links = node->GetLinkCount();
	for (unsigned int i=0; i<node_links; i++)
	{
		ProcessConvexArea(node->GetLink(i), true);
	}

	unsigned int node_lights = node->GetLightCount();
	for (unsigned int i = 0; i < node_lights; i++)
	{
		ProcessLight(node->GetLight(i));
	}

	// process cameras
	unsigned int node_cameras = node->GetCameraCount();
	for (unsigned int i = 0; i < node_cameras; i++)
	{
		ProcessCamera(node->GetCamera(i));
	}
	for (unsigned int i = 0; i < node->GetSubNodeCount(); i++)
	{
		new_node->Mount(ProcessNode(node->GetSubNode(i)));
	}

	return new_node;
}


bool ObjectGroup::Load (const char *filename, bool force_static)
{
	VFS_FILE *fp = vfs_open(filename, "rb");
	if (!fp)
	{
		Error("Failed loading scene file %s : no such file or directory\n",
			filename);
		return false;
	}

	bool success = LoadFromFP(fp, force_static);
	vfs_close (fp);
	if (success)
	{
		this->filename = new char[strlen(filename) + 1];
		strcpy(this->filename, filename);
	}

	return success;
}

bool ObjectGroup::LoadFromFP(VFS_FILE *fp, bool force_static)
{
	Destroy();
	Init ();

	this->force_static = force_static;
	
	SceneFile  *sce_file = NULL;

	sce_file = new SceneFile ();

	if (sce_file->Load(fp) == false)
	{
		Error("ObjectGroup::LoadFromFP() failed: bad fp\n");
		return false;
	}

	unsigned int root_node_count;
	root_node_count = sce_file->GetRootNodeCount ();

	for (unsigned int i = 0; i < root_node_count; i++)
	{
		root_node->Mount(ProcessNode(sce_file->GetRootNode(i)));
	}

	// calculate world matrices
	root_node->CalculateStaticModelMatrix();
	PrepareProxy();

	CompleteProgressTask();
	UpdateProgressBar(0);
	for(unsigned int i = 0; i < sce_file->GetMaterialCount(); i++) {
		printf("\n----- trying  material %d\n", i);
		MaterialChunk *mchunk = sce_file->GetMaterial(i);
		printf("material chunk pointer = %p\n", (void*)mchunk);

		FullMaterial *material = new FullMaterial ();
		printf("full material allocation pointer = %p\n", (void*)material);
		material->gc_mat.ambient = Color(mchunk->GetAmbient());
		material->gc_mat.diffuse = Color(mchunk->GetDiffuse());
		//TODO :Add emissive here
		material->gc_mat.specular = Color(mchunk->GetSpecular());
		//TODO :Add specular power here

		material->SetTexFilename(FMTT_DIFFUSE, mchunk->GetTexture(STT_DIFFUSE)->GetName().c_str());
		printf("diffuse: %s\n", mchunk->GetTexture(STT_DIFFUSE)->GetName().c_str());
		//TODO : add specular texture ===> material->SetTexFilename (FMTT_SPECULAR,
		material->SetTexFilename(FMTT_LIGHTMAP, mchunk->GetTexture(STT_LIGHT)->GetName().c_str());
		printf("lightmap: %s\n", mchunk->GetTexture(STT_LIGHT)->GetName().c_str());
		
		//TODO : add emissive texture
		material->SetTexFilename(FMTT_NORMAL, mchunk->GetTexture(STT_NORMAL)->GetName().c_str());
		printf("normal: %s\n", mchunk->GetTexture(STT_NORMAL)->GetName().c_str());
		material_list.push_back(material);

		printf("creating the textures for this material\n");
		if(material->CreateTextures() == false) {
			Error("ObjectGroup::LoadFromFP() : Failed to create textures for material %s\n", mchunk->GetName().c_str());
		}
		printf("textures ok, material is finished\n");

		UpdateProgressBar((float) i / (float)  sce_file->GetMaterialCount());
	}

	FullMaterial *blank_material = new FullMaterial ();
	material_list.push_back (blank_material);
	blank_material->CreateTextures ();

	// 'burn' all vertices in their static position
	if (force_static)
	{
		ForceStatic();
	}

	/*
	// process bone maps
	//printf("Node count: %d\n", nodes.size());
	for (unsigned int i=0; i<mesh_list.size(); i++)
	{
		// TODO: add functions ro provide vertex components etc etc....
		Geometry *geom = mesh_list[i]->geom;
		bool skinned = false;
		if (geom->GetBoneWeightArray() && geom->GetBoneIndexArray())
			skinned = true;
		if (!skinned) continue;
		
		for (unsigned int j=0; j<mesh_list[i]->bone_count; j++)
		{
			mesh_list[i]->skin_nodes[j] = nodes[mesh_list[i]->bone_map[j]];
		}
	}
	*/

	// calculate bounding boxes
	for (unsigned int i=0; i<GetMeshCount(); i++)
	{
		ComplexMesh *mesh = GetMesh(i);
		Vector3 minv, maxv;

		VertexArray *va = mesh->geom->GetVertexArray();
		unsigned int vcount = va->GetElementsCount();
		Vector3 *verts =(Vector3*)  va->GetData();
		
		if (vcount)
		{
			minv = maxv = verts[0];

			for (unsigned int i=1; i<vcount; i++)
			{
				Vector3 tv = verts[i];

				if (tv.x < minv.x) minv.x = tv.x;
				if (tv.y < minv.y) minv.y = tv.y;
				if (tv.z < minv.z) minv.z = tv.z;

				if (tv.x > maxv.x) maxv.x = tv.x;
				if (tv.y > maxv.y) maxv.y = tv.y;
				if (tv.z > maxv.z) maxv.z = tv.z;
			}

			mesh->model_space_bounding_box = Box(minv, maxv);
		}
		mesh->bounding_box = mesh->model_space_bounding_box;
		mesh->bounding_box.Transform(GetNode(mesh)->GetModelMatrix());
	}

	return true;
}

void ObjectGroup::ForceStatic()
{
	/*
	force_static = true;
	ProcessAnimation();
	root_node->CalculateModelMatrix();
	for (unsigned int i=0; i<mesh_list.size(); i++)
	{
		unsigned int node_id = mesh_list[i]->node_id;
		OGNode *node = nodes [node_id];
		Matrix4x4 m = node->model_matrix;

		mesh_list[i]->geom->Transform(m);
		node->model_matrix = Matrix4x4();
	}
	
	
	//TODO :Remove Sorting from here...
	// sort mesh_list
	for (unsigned int i = 0; i < mesh_list.size(); i ++)
	{
		for (unsigned int j = i+1; j < mesh_list.size(); j++)
		{
			if (mesh_list[i]->material_id < mesh_list[j]->material_id)
			{
				ComplexMesh *temp = mesh_list[i];
				mesh_list[i] = mesh_list[j];
				mesh_list[j] = temp;
			}
		}
	}*/
}

unsigned int ObjectGroup::GetNodeCount() const
{
	return (unsigned int)nodes.size();
}

OGNode *ObjectGroup::GetNode(unsigned int which) const
{
	if (which >= nodes.size()) return 0;
	return nodes[which];
}

OGNode *ObjectGroup::GetRootNode() const
{
	return root_node;
}

OGNode *ObjectGroup::GetFirstSplittingNode() const
{
	OGNode *n = root_node;
	if (!n) return 0;
	printf("%s, %s, %s\n", n->GetChild(0)->GetName().c_str(), 
		n->GetChild(1)->GetName().c_str(), 
		n->GetChild(2)->GetName().c_str());
	while (n->GetChildrenCount() == 1)
	{
		OGNode *child = n->GetChild(0);
		if (!child) return n;
		n = child;
	}
	return n;
}

unsigned int ObjectGroup::GetMeshCount () const
{
	return (unsigned int) mesh_list.size ();
}

ComplexMesh *ObjectGroup::GetMesh (unsigned int idx) const
{
	if (idx >= mesh_list.size ())
	{
		return 0;
	}
	return mesh_list[idx];
}

void ObjectGroup::AddMesh(ComplexMesh *mesh)
{
	mesh_list.push_back(mesh);
}

bool ObjectGroup::HasMesh(ComplexMesh *mesh) const
{
	return find(mesh_list.begin(), mesh_list.end(), mesh) != mesh_list.end();
}

const char *ObjectGroup::GetFilename() const
{
	return filename;
}

bool ObjectGroup::IsStatic () const
{
	return force_static ? true : is_static;
}


bool ObjectGroup::LoadAnimation(const char *filename, const char *name)
{
	if (force_static) return false;
	if (!name) name = filename;
	animations[name] = new Animation;
	if (animations[name]->Load(filename))
	{
		is_static = false;
		return true;
	}
	use_model_matrix = false;
	return false;
}

Animation *ObjectGroup::GetAnimation(const char *name)
{
	return animations[name];
}

Animation *ObjectGroup::SetAnimation(const char *name, int dim)
{
	anm[dim] = animations[name];
	use_model_matrix = false;
	return anm[dim];
}

void ObjectGroup::SetAnimation(Animation *anim, int dim)
{
	anm[dim] = anim;
	use_model_matrix = false;
}


void ObjectGroup::SetAnimationTime(float t, int dim)
{
	anm_time[dim] = t;
}


void ObjectGroup::SetAnimationBlend(bool enable, float param)
{
	anim_blend = enable;
	anim_blend_param = param;
}

void ObjectGroup::SetStart(const PRS &s, int dim)
{
	start[dim] = s;
}


void ObjectGroup::ProcessAnimation()
{
	if (!anm[0]) return;
	anm[0]->ProcessAnimation(anm_time[0]);
	// ground specific nodes

	int nodes_size = nodes.size();

	for (int i=0; i<nodes_size; i++)
	{
		PRS node_local = anm[0]->GetNodeLocal(i);
		
		// geiwse thn meta8esh pros to paron
		unsigned int ground = nodes[i]->GetGround0();
		Vector3 pos = node_local.GetPosition();
		if (ground & PX) pos.x = 0;
		if (ground & PY) pos.y = 0;
		if (ground & PZ) pos.z = 0;
		node_local.SetPosition(pos);
		
		anm[0]->SetNodeLocal(i, node_local);
	}
	bool blend = anim_blend;
	if (!anm[1]) blend = false;
	if (blend) 
	{
		anm[1]->ProcessAnimation(anm_time[1]);
		// ground specific nodes

		for (int i=0; i<nodes_size; i++)
		{
			PRS node_local = anm[1]->GetNodeLocal(i);
			
			// geiwse thn meta8esh pros to paron
			unsigned int ground = nodes[i]->GetGround1();
			Vector3 pos = node_local.GetPosition();
			if (ground & PX) pos.x = 0;
			if (ground & PY) pos.y = 0;
			if (ground & PZ) pos.z = 0;
			node_local.SetPosition(pos);
			
			anm[1]->SetNodeLocal(i, node_local);
		}

		// blend
		int ln = (int)anm[0]->nodes.size();
		if ((int)anm[1]->nodes.size() < ln) ln = (unsigned int)anm[1]->nodes.size();


		for (int i=0; i<ln; i++)
		{
			anm[0]->nodes[i].curr_local = Slerp(anm[0]->nodes[i].curr_local + start[0],
				anm[1]->nodes[i].curr_local + start[1], anim_blend_param);
		}
	}
	else
	{
		int anm0_nodes_size = anm[0]->nodes.size();
		for (int i=0; i < anm0_nodes_size; i++)
		{
			anm[0]->nodes[i].curr_local = anm[0]->nodes[i].curr_local + start[0];
		}
	}

	// calculate node matrix
	for (int i=0; i<nodes_size; i++)
	{
		PRS node_local = anm[0]->GetNodeLocal(i);
		nodes[i]->local_matrix = node_local.GetTransform();
	}

	// calculate model matrix of each node
	if (root_node)
	{
		root_node->CalculateModelMatrix();
	}

	// update proxy objects transforms
	PrepareProxy();
}

unsigned int ObjectGroup::GetMaterialCount () const
{
	return (unsigned int) material_list.size();
}

FullMaterial *ObjectGroup::GetMaterial (unsigned int idx) const
{
	if (idx >= material_list.size ())
	{
		return GetDefaultMaterial();
	}
	return material_list[idx];
}

FullMaterial *ObjectGroup::GetMaterial(ComplexMesh *mesh) const
{
	std::map<ComplexMesh*, unsigned int>::const_iterator it;
	it = mesh_material_map.find(mesh);
	if (it == mesh_material_map.end())
	{
		if (material_list.size())
			return material_list[material_list.size() - 1];
		else return 0;
	}
	return GetMaterial(it->second);
}

void ObjectGroup::SetModelMatrix (const Matrix4x4 &m)
{
	use_model_matrix = true;
	model_matrix = m;
}

Matrix4x4 ObjectGroup::GetModelMatrix ()
{
	return model_matrix;
}

unsigned int ObjectGroup::GetLightCount () const
{
	return (unsigned int) light_list.size();
}

RLight *ObjectGroup::GetLight (unsigned int idx) const
{
	if (idx >= light_list.size())
	{
		return 0;
	}
	return light_list[idx];
}

MatrixCamera *ObjectGroup::GetCamera(std::string name) const
{
	std::map<std::string, MatrixCamera*>::const_iterator it;
	it = cameras.find(name);
	if (it == cameras.end())
	{
		return 0;
	}
	return it->second;
}

OGNode *ObjectGroup::GetNode(std::string name) const
{
	std::map<std::string, OGNode*>::const_iterator it;
	it = node_name_map.find(name);
	if (it == node_name_map.end())
	{
		return 0;
	}
	return it->second;
}

OGNode *ObjectGroup::GetNode(ComplexMesh *mesh) const
{
	std::map<ComplexMesh*, unsigned int>::const_iterator it;
	it = mesh_node_map.find(mesh);
	if (it == mesh_node_map.end())
	{
		return root_node;
	}
	return nodes[it->second];
}

// converts a CollisionObject pointer to its OGNode pointer
OGNode *ObjectGroup::GetProxyNode(CollisionObject *obj)
{
	for (unsigned int i=0; i<collision_objects.size(); i++)
	{
		if (collision_objects[i] == obj)
			return nodes[collision_nodes[i]];
	}
	return 0;
}


void ObjectGroup::TreatMeshAsProxy(bool enable)
{
	mesh_as_proxy = enable;
}

// for all animations and for all keyframes it pushes back
// a matrix at each animated node. This list of matrices
// (all_keys) can be used to determine spacetime bb's
void ObjectGroup::CalculateAllKeys()
{
	// clear
	ClearAllKeys();
	std::map<std::string, Animation*>::iterator it;
	for(it = animations.begin(); it != animations.end(); ++it)
	{
		Animation *anim = it->second;
		SetAnimation(anim);
		SetAnimationBlend(false);
		unsigned int frames = (unsigned int)(anim->GetDuration() * anim->GetFPS());
		for (unsigned int i=0; i<frames; i++)
		{
			float t = (float)i / anim->GetFPS();
			SetAnimationTime(t);
			ProcessAnimation();
			for (unsigned int j=0; j<nodes.size(); j++)
			{
				nodes[j]->AddToAllKeys();
			}
		}
	}
}

void ObjectGroup::ClearAllKeys()
{
	for (unsigned int i=0; i<nodes.size(); i++)
	{
		OGNode *n = nodes[i];
		n->all_keys.erase(n->all_keys.begin(), n->all_keys.end());
	}
}

// SmartBox calculation
Box ObjectGroup::CalculateSmartBox() const
{
	std::vector<Vector3> pt;
	for (unsigned int i=0; i<GetNodeCount(); i++)
	{
		OGNode *n = GetNode(i);
		Vector3 pos(0, 0, 0);
		pos.Transform(n->GetModelMatrix());
		if (pos.LengthSq() < 50) continue;
		pt.push_back(Vector3(0, 0, 0).Transformed(
			model_matrix * n->GetModelMatrix()));
	}

	smart_box = SmartBox(pt);
	return smart_box;
}

Box ObjectGroup::GetSmartBox() const
{
	return smart_box;
}

unsigned int ObjectGroup::GetBoneCount(ComplexMesh *mesh) const
{
	std::map<ComplexMesh*, BoneMap>::const_iterator it;
	it = mesh_bone_map_map.find(mesh);
	if (it == mesh_bone_map_map.end())
	{
		return 0;
	}
	return it->second.count;
}

OGNode *ObjectGroup::GetBone(ComplexMesh *mesh, unsigned int idx) const
{
	std::map<ComplexMesh*, BoneMap>::const_iterator it;
	it = mesh_bone_map_map.find(mesh);
	if (it == mesh_bone_map_map.end())
	{
		return 0;
	}
	return GetNode(it->second.map[idx]);
}





