#ifndef OBJECT_GROUP_H
#define OBJECT_GROUP_H

#include <vector>
#include <string>
#include <map>
#include <builddefs.h>

#include <gc.h>
#include <collision.h>
#include <libvfs.h>
#include <libvis.h>
#include <sce.h>
#include <anm.h>
#include <math3d.h>
#include "geometry.h"
#include "meshman.h"
#include "fullmat.h"
#include "rlight.h"
#include <camera.h>
#include "mesh.h"

using namespace std;


// transform type
#define PX 1
#define PY 2
#define PZ 4
#define RX 8
#define RY 16
#define RZ 32
#define SX 64
#define SY 128
#define SZ 256

// fwd declaration
class OGNode;
class ObjectGroup;




enum AnimSampleMode
{
	ASM_POINT,
	ASM_LINEAR
};

enum AnimLoopMode
{
	ALM_ONCE,
	ALM_LOOP,
	ALM_BOUNCE
};

struct Timeline
{
	std::vector<PRS> keys;
	bool is_static;
	PRS static_local;
	PRS curr_local;
};

PRS Sample(const Timeline &tm, float frame, 
		   AnimSampleMode asmp = ASM_LINEAR, AnimLoopMode alm = ALM_LOOP);

class Animation
{
protected:
	AnimSampleMode asmp;
	AnimLoopMode alm;
	std::vector<Timeline> nodes;
	float fps;
public:
	Animation();
	~Animation();

	void Destroy();

	bool Load(const char *filename);
	
	void SetSampleMode(AnimSampleMode asmp);
	void SetLoopMode(AnimLoopMode alm);

	void ProcessAnimation(float t);

	PRS GetNodeLocal(unsigned int node);
	void SetNodeLocal(unsigned int node, const PRS &prs);

	Timeline GetKeys(unsigned int node);

	void DeleteLastKey();

	float GetDuration() const;

	float GetFPS() const;
	void SetFPS(float fps);

	friend class ObjectGroup;
};

class OGNode
{
protected:
	OGNode *parent;
	std::vector<OGNode*> children;
	std::string name;
	unsigned int id;

	Matrix4x4 local_matrix;
	Matrix4x4 model_matrix;
	Matrix4x4 static_model_matrix;
	Matrix4x4 pivot_matrix;

	unsigned int ground0, ground1;

public:

	std::vector<Matrix4x4> all_keys;

	OGNode();
	~OGNode();

	std::string GetName() const;
	void SetName(const std::string &name);

	unsigned int GetID() const;
	void SetID(unsigned int id);

	unsigned int GetChildrenCount() const;
	OGNode *GetChild(unsigned int which) const;

	void Mount(OGNode *child);
	void Unmount(OGNode *child);

	void CalculateStaticModelMatrix();
	void CalculateModelMatrix();
	Matrix4x4 GetPivot() const {return pivot_matrix;}
	Matrix4x4 GetModelMatrix() const {return model_matrix;}
	Matrix4x4 GetStaticModelMatrix() const {return static_model_matrix;}
	
	void SetModelMatrix(const Matrix4x4 &model);
	
	void SetGround(unsigned int ground);
	void SetGround0(unsigned int ground);
	void SetGround1(unsigned int ground);
	unsigned int GetGround() const;
	unsigned int GetGround0() const;
	unsigned int GetGround1() const;
	
	friend class ObjectGroup;

	void AddToAllKeys();
}; // end class OGNode

class ObjectGroup : public CollisionObject
{
protected:

	struct BoneMap
	{
		unsigned int map[MAX_BONES];
		unsigned int count;
		BoneMap() {count=0;}
	};

	Matrix4x4 model_matrix;
	bool use_model_matrix;
	char *filename;
	std::vector<ComplexMesh*> mesh_list;
	std::map<ComplexMesh*, unsigned int> mesh_material_map;
	std::map<ComplexMesh*, unsigned int> mesh_node_map;
	std::vector<ConvexArea*> convex_areas;

	std::map<ComplexMesh*, BoneMap> mesh_bone_map_map;

	std::map<std::string, MatrixCamera*> cameras;

	std::vector<OGNode*> nodes;
	OGNode *root_node;
	std::map<std::string, OGNode*> node_name_map;
	std::map<std::string, Animation*> animations;


	bool anim_blend;
	float anim_blend_param;
	bool is_static, force_static;

	// appplies only in loading time
	bool mesh_as_proxy;

	map<int, float> anim_weights; //this is a replacement for n dimensions

	//Animation *anm0, *anm1;			//////////////////
	//PRS start0, start1;				// obsolete!
	//float anm0_time, anm1_time;		//////////////////

	map<int, Animation*> anm;
	map<int, PRS> start;
	map<int, float> anm_time;

	mutable Box smart_box;

	bool random_move;

	// collision object owner nodes
	std::vector<unsigned int> collision_nodes;

	std::vector<FullMaterial*> material_list;
	std::vector<RLight *> light_list;

	void ProcessGeometry (MeshChunk* mesh);
	void ProcessProxy(MeshChunk *mesh);
	void PrepareProxy();
	void ProcessConvexArea(MeshChunk *mesh, bool link);
	OGNode *ProcessNode (NodeChunk *node);
	void ProcessLight(LightChunk* light);
	void ProcessCamera(CameraChunk *camera);

	void Init ();

public:
	ObjectGroup();
	~ObjectGroup();

	void Destroy();

	bool Load (const char *filename, bool force_static = false);
	bool LoadFromFP(VFS_FILE *fp, bool force_static = false);
	void ForceStatic();

	unsigned int GetNodeCount() const;
	OGNode *GetNode(unsigned int which) const;
	OGNode *GetRootNode() const;
	OGNode *GetFirstSplittingNode() const;

	unsigned int GetMeshCount () const;
	ComplexMesh *GetMesh (unsigned int idx) const;
	void AddMesh(ComplexMesh *mesh);
	bool HasMesh(ComplexMesh *mesh) const;

	const char *GetFilename() const;

	bool IsStatic () const;

	bool LoadAnimation(const char *filename, const char *name);
	Animation *GetAnimation(const char *name);
	Animation *SetAnimation(const char *name, int dim=0);
	void SetAnimation(Animation *anim, int dim=0);
	void SetAnimationTime(float t, int dim=0);
	void SetStart(const PRS &s, int dim=0);
	void SetAnimationBlend(bool enable, float param = 0.0f);
	void ProcessAnimation();

	unsigned int GetMaterialCount () const;
	FullMaterial *GetMaterial (unsigned int idx) const;
	FullMaterial *GetMaterial(ComplexMesh *mesh) const;

	void SetModelMatrix (const Matrix4x4 &m);
	Matrix4x4 GetModelMatrix ();

	unsigned int GetLightCount () const;
	RLight *GetLight (unsigned int idx) const;

	MatrixCamera *GetCamera(std::string name) const;
	OGNode *GetNode(std::string name) const;
	OGNode *GetNode(ComplexMesh *mesh) const;

	inline bool CanUseModelMatrix() { return use_model_matrix; }

	// converts a CollisionObject pointer to its OGNode pointer
	OGNode *GetProxyNode(CollisionObject *obj);

	void TreatMeshAsProxy(bool enable);

	unsigned int GetConvexAreaCount() const {return (unsigned int) convex_areas.size();}
	ConvexArea *GetConvexArea(unsigned int i)
	{
		if (i >= convex_areas.size()) return 0;
		return convex_areas[i];
	}

	// for all animations and for all keyframes it pushes back
	// a matrix at each animated node. This list of matrices
	// (all_keys) can be used to determine spacetime bb's
	void CalculateAllKeys();
	void ClearAllKeys();

	void SetRandomMove(bool enable) {random_move = enable;}
	bool GetRandomMove() {return random_move;}

	// SmartBox calculation
	Box CalculateSmartBox() const;
	Box GetSmartBox() const;

	unsigned int GetBoneCount(ComplexMesh *mesh) const;
	OGNode *GetBone(ComplexMesh *mesh, unsigned int idx) const;
};


#endif //OBJECT_GROUP_H

