#include "renderer_mrt.h"
#include <libgc.h>
#include <libsys.h>
#include <statecache.h>
#include <uniforms.h>
#include <devhelp.h>
#include <errlog.h>


static char *ltmesh_sdr_vertex = {
	"uniform float4x4 STATE_MODELVIEWPROJECTION_MATRIX;\n"
	"float4 main(float4 pos : POSITION) : POSITION\n"
	"{ return mul(pos, STATE_MODELVIEWPROJECTION_MATRIX); }\n"
};

static char *ltmesh_sdr_pixel = {
	"float4 main() : COLOR\n"
	"{ return float4(1.0, 0.0, 0.0, 1.0); }\n"
};

#define SDR_PATH	DEV_DATA_PATH "/shaders/renderer_mrt/"

/* For the create functions, inner means we want a polyhedron approximation
 * around an inner sphere (circumscribed). if false, the sphere is around the
 * approximated polyhdron (the polyhedron is inscribed)
 */

static SimpleMesh *CreateSphere(int sphere_quality = 4, bool inner = true);
static SimpleMesh *CreateCone(int cone_quality = 8, bool inner = true);


RendererMRTDeferred::RendererMRTDeferred()
{
	memset(rt, 0, sizeof rt);
	mrt_sdr = 0;
	memset(light_sdr, 0, sizeof light_sdr);

	SetParam("SHOW_LIGHTS", false);
	SetParam("SHOW_DEBUG_BUF", 0);
	SetParam("LRANGE_SCALE" , 1.0);
	SetParam("DEPTH_FACTOR", 0.1);
	SetParam("USE_LIGHTMAP", true);

	SetParam("RTARG", 0);
}

RendererMRTDeferred::~RendererMRTDeferred()
{
	Release();
}

enum {
	LTSDR_OMNI,
	LTSDR_SPOT,
	LTSDR_POS,
	LTSDR_NORMAL,
	LTSDR_COLOR,
	LTSDR_LMAP,
	LTSDR_HEIGHT,
	LTSDR_SPEC
};

#define SDR(x)	SDR_PATH x
static char *ltsdr_fnames[][2] = {
	{SDR("defer_lomni.v.cg"), SDR("defer_lomni.p.cg")},	// omni deferred light shader
	{SDR("defer_lomni.v.cg"), SDR("defer_lspot.p.cg")},	// spotlight deferred light shader
	{SDR("defer_lomni.v.cg"), SDR("dbg_pos.p.cg")},		// DEBUG: show position
	{SDR("defer_lomni.v.cg"), SDR("dbg_norm.p.cg")},	// DEBUG: show normals
	{SDR("defer_lomni.v.cg"), SDR("dbg_color.p.cg")},	// DEBUG: show colors
	{SDR("defer_lomni.v.cg"), SDR("dbg_lmap.p.cg")},	// DEBUG: show lightmap texels
	{SDR("defer_lomni.v.cg"), SDR("dbg_height.p.cg")},	// DEBUG: show heightfield (from normal/height map)
	{SDR("defer_lomni.v.cg"), SDR("dbg_spec.p.cg")},	// DEBUG: show specular level
	{0, 0}
};


bool RendererMRTDeferred::Init()
{
	// create the render targets (see MRT_<whatever> enumeration in header)
	for(int i=0; i<MRT_ENUM_COUNT; i++) {
		if(!(rt[i] = CreateRenderTarget(SysGetWindowWidth(), SysGetWindowHeight(), TFMT_FLOAT16_RGBA))) {
			Error("RendererMRTDeferred::Init() failed to create render target\n");
			Release();
			return false;
		}
	}

	// load the ambient/lightmap (GI) shader
	if(!(amb_sdr = CreateGPUProgramFromFile(SDR_PATH "defer_lomni.v.cg", SDR_PATH "defer_amb.p.cg"))) {
		Error("RendererMRTDeferred::Init() failed to create ambient shader\n");
		Release();
		return false;
	}

	// load the MRT program (that stores values in the multiple buffers)
	if(!(mrt_sdr = CreateGPUProgramFromFile(SDR_PATH "defer_mrt.v.cg", SDR_PATH "defer_mrt.p.cg"))) {
		Error("RendererMRTDeferred::Init() failed to create mrt shader\n");
		Release();
		return false;
	}
	if(!(mrt_skin_sdr = CreateGPUProgramFromFile(SDR_PATH "defer_skin_mrt.v.cg", SDR_PATH "defer_mrt.p.cg"))) {
		Error("RendererMRTDeferred::Init() failed to create skin mrt shader\n");
		Release();
		return false;
	}

	// create the necessary uniforms for the light shaders
	//uni_light_pos = AddUniform("light_pos", 3);
	//uni_light_rad = AddUniform("light_rad", 1);
	//uni_light_dir = AddUniform("light_dir", 3);
	//uni_light_cone = AddUniform("light_cone", 2);
	//uni_depth_fact = AddUniform("parallax_depth_factor", 1);

	// load the light shader programs for omni and spotlights (plus debugging shaders).
	memset(light_sdr, 0, MAX_LIGHT_SHADERS * sizeof *light_sdr);
	for(int i=0; ltsdr_fnames[i][0]; i++) {
		if(!(light_sdr[i] = CreateGPUProgramFromFile(ltsdr_fnames[i][0], ltsdr_fnames[i][1]))) {
			Error("RendererMRTDeferred::Init() failed to load deferred light shaders: %s, %s\n", ltsdr_fnames[i][0], ltsdr_fnames[i][1]);
			Release();
			return false;
		}
	}

	/* load the shader to be used when drawing the light meshes for debugging
	 * (not the actual light shader).
	 */
	if(!(ltmesh_sdr = CreateGPUProgram(ltmesh_sdr_vertex, ltmesh_sdr_pixel))) {
		Warning("RendererMRTDeferred::Init() failed to load light mesh shader\n");
	}

	// create the light meshes.
	memset(light_mesh, 0, RL_TYPE_COUNT * sizeof *light_mesh);
	if(!(light_mesh[RL_OMNI] = CreateSphere(8, true))) {
		Error("RendererMRTDeferred::Init() failed to create a light sphere\n");
		Release();
		return false;
	}
	if(!(light_mesh[RL_SPOT] = CreateCone())) {
		Error("RendererMRTDeferred::Init() failed to create a light cone\n");
		Release();
		return false;
	}

	return true;
}

void RendererMRTDeferred::Release()
{
	for(int i=0; i<MRT_ENUM_COUNT; i++) {
		if(rt[i]) {
			DeleteTexture(rt[i]);
			rt[i] = 0;
		}
	}

	if(mrt_sdr) {
		ReleaseGPUProgram(mrt_sdr);
		mrt_sdr = 0;
	}

	if(amb_sdr) {
		ReleaseGPUProgram(amb_sdr);
		amb_sdr = 0;
	}

	for(int i=0; i<MAX_LIGHT_SHADERS; i++) {
		if(light_sdr[i]) {
			ReleaseGPUProgram(light_sdr[i]);
			light_sdr[i] = 0;
		}
	}
}

void RendererMRTDeferred::RenderDeferredPass(const Scene *scene) const
{
	for(int i=0; i<MRT_ENUM_COUNT; i++) {
		if(!SetRenderTarget(rt[i], i)) {
			Error("SetRenderTarget(rt[%d], %d) failed\n", i, i);
		}
	}

	//float depth_fact = GetParamFloat("DEPTH_FACTOR");
	//SetUniform(uni_depth_fact, (void*)&depth_fact);

	//BindGPUProgram(mrt_sdr);
	Clear();

	Matrix4x4 model_matrix = GetModelMatrix();

	const Camera *cam = scene->GetActiveCamera();
	std::vector<RObject*> objects = scene->GetVisibleItems(GetViewMatrix(), GetProjectionMatrix()).objects;

	Info("Batch count = %d\n", objects.size());

	SetAlphaTest(true);
	SetAlphaFunc(CMP_GREATER);
	SetAlphaRef(0.95);

	for(size_t i=0; i<objects.size(); i++) {
		RObject *sobj = objects[i];

		SetModelMatrix(sobj->GetNode()->GetModelMatrix());
		if(sobj->IsSkinned()) 
		{
			BindGPUProgram(mrt_skin_sdr);
			for (unsigned int j=0; j<sobj->GetBoneCount(); j++) {
				ComplexMesh *mesh = sobj->GetMesh();
				Matrix4x4 bone;
				bone.ResetIdentity();
			
				bone *= sobj->GetBone(j)->GetModelMatrix();
				bone *= sobj->GetBone(j)->GetStaticModelMatrix().Inverse();
				bone *= sobj->GetNode()->GetModelMatrix();
			
				SetBoneMatrix(j, bone);
			}
			SetModelMatrix(sobj->GetOwner()->GetModelMatrix());
			//DumpUniforms("uniforms.txt");
		} 
		else 
		{
			BindGPUProgram(mrt_sdr);
			if (sobj->GetOwner()->CanUseModelMatrix())
			{
				SetModelMatrix(sobj->GetOwner()->GetModelMatrix());
			}
		}
	
		BindVertexBuffer(sobj->GetMesh()->rg_info.vb_id);
		BindIndexBuffer(sobj->GetMesh()->rg_info.ib_id);
		//SetModelMatrix(*sobj->mesh->model_matrix);

		sobj->GetMaterial()->Bind();
		BindUniforms();

		RGInfo *rginfo = &sobj->GetMesh()->rg_info;
		RenderIndexedPrimitives(PT_TRIANGLE_LIST, rginfo->start_index, rginfo->primitive_count, rginfo->start_vertex, rginfo->vertex_count);

		sobj++;
	}

	SetAlphaTest(false);

	SetModelMatrix(model_matrix);

	for(int i=0; i<MRT_ENUM_COUNT; i++) {
		SetRenderTarget(0, i);
	}
}

void RendererMRTDeferred::RenderLights(const Scene *scene) const
{
	GCInitParams ipar = GetInitParams();
	SetHalfPixelSize(ipar.dmode.width, ipar.dmode.height);

	/* save whatever state is needed */
	Matrix4x4 save_model_mat = GetModelMatrix();
	Matrix4x4 save_view_mat = GetViewMatrix();
	Matrix4x4 save_proj_mat = GetProjectionMatrix();
	bool prev_depth_test = GetDepthTest();
	bool prev_depth_mask = GetDepthMask();
	CullFace prev_cull_mode = GetCulling();
	bool prev_blending = GetBlending();
	

	/* bind the multiple buffers */
	for(int j=0; j<MRT_ENUM_COUNT; j++) {
		BindTexture(j, rt[j]);
		SetTextureFiltering(j, TF_POINT, TF_POINT, TF_NONE);
		SetTextureAddressing(j, TA_CLAMP);
	}

	SetRenderTarget(GetParamInt("RTARG"));
	Clear();

	SetBlending(true);
	SetBlendFunc(BLEND_ONE, BLEND_ONE);
	SetDepthTest(false);
	SetDepthMask(false);
	SetCulling(CULL_CW);

	// make an ambient/GI pass
	if(GetParamBool("USE_LIGHTMAP")) {
        BindGPUProgram(amb_sdr);
		BindUniforms();
		ImFixedFunction(false);
		ImOverlay(Vector2(0, 0), Vector2(1, 1), 1.0, IM_TEX_CUR, IM_SDR_CUR);
	}

	// make the per-light pass
	Box frustum = GetFrustum();

	int dbg_sdr = GetParamInt("SHOW_DEBUG_BUF");
	if(dbg_sdr > 1) {
		BindGPUProgram(light_sdr[dbg_sdr + 1]);
		BindUniforms();

		ImFixedFunction(false);
		ImOverlay(Vector2(0.1, 0.1), Vector2(0.9, 0.9), 1.0, IM_TEX_CUR, IM_SDR_CUR);
	} else {
		// --- regular deferred light pass ---
		int num_lights = scene->GetDynamicLightCount();
		Matrix4x4 view_mat = GetViewMatrix();

		for(int i=0; i<num_lights; i++) {
			float cone_height, cone_range;
			const RLight *light = scene->GetDynamicLight(i);
			int ltype = light->GetType();

			light->Bind(0);
			Matrix4x4 model_matrix;

			float atten_start = light->GetAttenStart() * GetParamFloat("LRANGE_SCALE");
			float atten_end = light->GetAttenEnd() * GetParamFloat("LRANGE_SCALE");
			
			Sphere s = Sphere(light->GetPosition(), atten_end);
			if(!Intersect(s, frustum)) {
				continue;
			}

			BindGPUProgram(light_sdr[ltype]);

			Vector3 light_pos = light->GetPosition().Transformed(view_mat);
			SetUniform(STATE_LIGHT0_POS, &light_pos);
			SetUniform(STATE_LIGHT0_ATTEN_START, &atten_start);
			SetUniform(STATE_LIGHT0_ATTEN_END, &atten_end);

			Color light_color = light->GetDiffuse();
			SetUniform(STATE_LIGHT0_DIFFUSE, &light_color);

			switch(ltype) {
			case RL_OMNI: 
				model_matrix.Translate(light->GetPosition());
				model_matrix.Scale(Vector4(atten_end, atten_end, atten_end, 1.0f));
				break;

			case RL_SPOT: 
				model_matrix.Translate(light->GetPosition());
				model_matrix = model_matrix * ((Matrix4x4)light->GetOrientation().GetRotationMatrix());
				cone_height = atten_end;
				cone_range = light->GetConeRange();
				model_matrix.Scale(Vector4(cone_range, cone_height, cone_range, 1.0f));

				{
					Vector2 cone(light->GetInnerCone(), light->GetOuterCone());
					SetUniform(STATE_LIGHT0_INNER, &cone.x);
					SetUniform(STATE_LIGHT0_OUTER, &cone.y);

					Vector3 dir = light->GetDirection().Transformed(view_mat.Inverse().Transposed());
                    SetUniform(STATE_LIGHT0_DIR, &dir);
				}
				break;

			default:
				Error("RendererMRTDeferred::RenderLights invalid light type\n");
				return;
			}

			SetModelMatrix(model_matrix);

			BindUniforms();

			SetCulling(CULL_CCW);
		
			// render the light meshes (sphere/cone)
			light_mesh[ltype]->DrawIndexed();

			// also render the light meshes if requested
			if(GetParamBool("SHOW_LIGHTS")) {
				BindGPUProgram(ltmesh_sdr);
				bool blending = GetBlending();
				SetBlending(false);
				SetFillMode(FILL_WIREFRAME);
				light_mesh[ltype]->DrawIndexed();
				SetFillMode(FILL_SOLID);
				SetBlending(blending);
			}
		}
	}

	SetRenderTarget(0);

	for(int j=0; j<MRT_ENUM_COUNT; j++) {
		SetTextureFiltering(j, TF_LINEAR, TF_LINEAR, TF_LINEAR);
		BindTexture(j, 0);
	}

	SetCulling(prev_cull_mode);
	SetDepthTest(prev_depth_test);
	SetDepthMask(prev_depth_mask);

	SetModelMatrix(save_model_mat);
	SetViewMatrix(save_view_mat);
	SetProjectionMatrix(save_proj_mat);

	SetBlending(prev_blending);
}

void RendererMRTDeferred::Render(const Scene *scene) const
{
	SetupTransform(scene->GetActiveCamera());

	RenderDeferredPass(scene);
	RenderLights(scene);
}

#define SPHERE_SIZE(q)	(2 * (q) * (q))
static SimpleMesh *CreateSphere(int sphere_quality, bool inner)
{
	float radius = 1.0f;
	if(inner) {
		float n = (float)(2 * sphere_quality);
		radius /= cos(Pi / n);
	}

	unsigned int vcount = SPHERE_SIZE(sphere_quality);
	unsigned int icount = (sphere_quality - 1) * 12 * sphere_quality;

	SimpleVertex *varr, *vptr;
	unsigned int *iarr, *iptr;
	SimpleMesh *mesh;
	try {
		varr = vptr = new SimpleVertex[vcount];
		iarr = iptr = new unsigned int[icount];
		mesh = new SimpleMesh;
	}
	catch(...) {
		return 0;
	}

	int arcs = sphere_quality * 2;
	int arc_pt = sphere_quality;
	float da = TwoPi / (float)arcs;
	float da_pt = Pi / (float)(sphere_quality - 1);

	for(int i=0; i<arcs; i++) {
		for(int j=0; j<arc_pt; j++) {
			Vector3 vec(0, radius, 0);
			vec.Rotate(Vector3(da_pt * (float)j, 0, 0));
			vec.Rotate(Vector3(0, da * (float)i, 0));
			vptr->pos = vptr->norm = vec;
			vptr++;

			if(j < arc_pt - 1) {
                int a = i * arc_pt + j;
				int b = (a + arc_pt) % vcount;
				int c = a + 1;
				int d = b + 1;

				*iptr++ = a;
				*iptr++ = b;
				*iptr++ = d;
				*iptr++ = a;
				*iptr++ = d;
				*iptr++ = c;
			}
		}
	}

	if(!mesh->SetVertexData(varr, vcount) || !mesh->SetIndexData(iarr, icount)) {
		delete [] varr;
		delete [] iarr;
		delete mesh;
		return 0;
	}

	delete [] varr;
	delete [] iarr;
	return mesh;
}

static SimpleMesh *CreateCone(int cone_quality, bool inner)
{
	float radius = 1.0f;
	if(inner) {
		float n = (float)cone_quality;
		radius /= cos(Pi / n);
	}
	int vcount = cone_quality + 2;
	int tcount = cone_quality * 2;
	int icount = tcount * 3;

	SimpleVertex *varr, *vptr;
	unsigned int *iarr, *iptr;
	SimpleMesh *mesh;

	try {
		varr = vptr = new SimpleVertex[vcount];
		//iarr = iptr = new unsigned int[icount];
		iarr = iptr = new (std::nothrow) unsigned int[icount];
		mesh = new SimpleMesh;
	}
	catch(...) {
		return 0;
	}

	float rot = TwoPi / cone_quality;

	// fill vertex buffer
	for(int i=0; i<cone_quality; i++) {
		Vector3 v(radius, 0, 0);
		v.Rotate(Vector3(0, i * rot, 0));
		v.y = -1.0f;
		vptr[i].pos = v;
	}
	vptr[cone_quality].pos = Vector3(0, 0, 0);
	vptr[cone_quality + 1].pos = Vector3(0, -1, 0);

	// fill index buffer
	for(int i=0; i<icount/2; i+=3) {
		int t = i / 3;
		/*
		iptr[i + 0] = t;
		iptr[i + 1] = (t + 1) % cone_quality;
		iptr[i + 2] = cone_quality;
		//t+=tcount/2;

		iptr[icount / 2 + i + 0] = (t + 1) % cone_quality;
		iptr[icount / 2 + i + 1] = t;
		iptr[icount / 2 + i + 2] = cone_quality + 1;
		*/
		iptr[i + 0] = t;
		iptr[i + 1] = cone_quality;
		iptr[i + 2] = (t + 1) % cone_quality;

		iptr[icount / 2 + i + 0] = (t + 1) % cone_quality;
		iptr[icount / 2 + i + 1] = cone_quality + 1;
		iptr[icount / 2 + i + 2] = t;
	}

	if(!mesh->SetVertexData(varr, vcount) || !mesh->SetIndexData(iarr, icount)) {
		delete [] varr;
		delete [] iarr;
		delete mesh;
		return 0;
	}

	delete [] varr;
	delete [] iarr;
	return mesh;
}
