#include <gc.h>
#include <immed.h>
#include <errlog.h>
#include <gpuprog.h>
#include <textures.h>
#include <devhelp.h>
#include <console.h>
#include <statecache.h>
#include <uniforms.h>
#include "sdrgen_simple.h"
#include "renderer_simple.h"

static unsigned int fr_drops = 0;

SimpleRenderer::SimpleRenderer()
{
	SetParam("SHOW_LIGHTS", false);
	rigid_shader = skin_shader = color_shader = 0;
	rigid_shader_shadow = skin_shader_shadow = 0;
	zburn_rigid_shader = zburn_skin_shader = 0;
	ambient_shader = ambient_skin_shader = 0;
	overlay_shader = 0;
	depthsq_shader = depthsq_skin_shader = 0;
	highlight_property = 0;
	highlight_proxy = 0;
	render_only_highlight = false;
	hl_prop_bit = true;
	curr_scene = 0;

	curr_vr = 0;
	
	// settings
	cast_shadows = true;
}

#define VSDR_PATH	DEV_DATA_PATH "/shaders/renderer_simple/vertex"
#define PSDR_PATH	DEV_DATA_PATH "/shaders/renderer_simple/pixel"

#define ZBURN_SDRPATH	DEV_DATA_PATH "/shaders/common"

bool SimpleRenderer::Init()
{
	char tst[] = VSDR_PATH;
	if(!(rigid_shader = CreateGPUProgramFromFile(VSDR_PATH "/vertex.vcg", PSDR_PATH "/pixel_omni.cg"))) {
		Error("SimpleRenderer::Init() : Could not load rigid shader\n");
		Release();
		return false;
	}

	if(!(skin_shader = CreateGPUProgramFromFile("/devdata/shaders/renderer_simple/vertex/skin_vertex.vcg", 
		"/devdata/shaders/renderer_simple/pixel/pixel_omni.cg"))) {
		Error("SimpleRenderer::Init() : Could not load skin shader\n");
		Release();
		return false;
	}

	if(!(rigid_shader_shadow = CreateGPUProgramFromFile(VSDR_PATH "/vertex.vcg", PSDR_PATH "/pixel_omni_shadow.cg"))) {
		Error("SimpleRenderer::Init() : Could not load rigid shader with shadows\n");
		Release();
		return false;
	}

	if(!(skin_shader_shadow = CreateGPUProgramFromFile("/devdata/shaders/renderer_simple/vertex/skin_vertex.vcg", 
		"/devdata/shaders/renderer_simple/pixel/pixel_omni_shadow.cg"))) {
		Error("SimpleRenderer::Init() : Could not load skin shader with shadows\n");
		Release();
		return false;
	}

	if(!(ambient_shader = CreateGPUProgramFromFile(VSDR_PATH "/vertex.vcg", PSDR_PATH "/pixel_ambient.cg"))) {
		Error("SimpleRenderer::Init() : Could not load rigid shader\n");
		Release();
		return false;
	}

	if(!(ambient_skin_shader = CreateGPUProgramFromFile("/devdata/shaders/renderer_simple/vertex/skin_vertex.vcg", 
		"/devdata/shaders/renderer_simple/pixel/pixel_ambient.cg"))) {
		Error("SimpleRenderer::Init() : Could not load skin shader\n");
		Release();
		return false;
	}

	if(!(color_shader = CreateGPUProgramFromFile(VSDR_PATH "/vertex.vcg", PSDR_PATH "/c_pixel.pcg"))) {
		Error("SimpleRenderer::Init() : Could not load color shader\n");
		Release();
		return false;
	}

	if(!(zburn_rigid_shader = CreateGPUProgramFromFile(
			ZBURN_SDRPATH "/zburn_vertex.cg", 
			ZBURN_SDRPATH "/zburn_pixel.cg"))) {
		Error("SimpleRenderer::Init() : Could not load zburn rigid shader\n");
		Release();
		return false;
	}

	if(!(zburn_skin_shader = CreateGPUProgramFromFile(
			ZBURN_SDRPATH "/zburn_skin_vertex.cg", 
			ZBURN_SDRPATH "/zburn_pixel.cg"))) {
		Error("SimpleRenderer::Init() : Could not load zburn skin shader\n");
		Release();
		return false;
	}

	if(!(overlay_shader = CreateGPUProgramFromFile(
			VSDR_PATH "/overlay_vertex.cg", 
			PSDR_PATH "/overlay_pixel.cg"))) {
		Error("SimpleRenderer::Init() : Could not load overlay shader\n");
		Release();
		return false;
	}

	if(!(depthsq_shader = CreateGPUProgramFromFile(
		VSDR_PATH "/vertex.vcg", 
		PSDR_PATH "/pixel_depth_sq.cg"))) {
		Error("SimpleRenderer::Init() : Could not load depthsq rigid shader\n");
		Release();
		return false;
	}

	if(!(depthsq_skin_shader = CreateGPUProgramFromFile(
		VSDR_PATH "/skin_vertex.vcg", 
		PSDR_PATH "/pixel_depth_sq.cg"))) {
		Error("SimpleRenderer::Init() : Could not load depthsq skin shader\n");
		Release();
		return false;
	}

	cube_tex = CreateCubeRenderTarget(2048, 2048, TFMT_FLOAT_32);
	if (!cube_tex) return false;
	
	//SGS_GetShader(0);

	return true;
}

void SimpleRenderer::Release()
{
	if (rigid_shader)
	{
		ReleaseGPUProgram(rigid_shader);
	}
	if (skin_shader)
	{
		ReleaseGPUProgram(skin_shader);
	}
	if (rigid_shader_shadow)
	{
		ReleaseGPUProgram(rigid_shader_shadow);
	}
	if (skin_shader_shadow)
	{
		ReleaseGPUProgram(skin_shader_shadow);
	}
	if (color_shader)
	{
		ReleaseGPUProgram(color_shader);
	}
	if (zburn_skin_shader)
	{
		ReleaseGPUProgram(zburn_skin_shader);
	}
	if (zburn_rigid_shader)
	{
		ReleaseGPUProgram(zburn_rigid_shader);
	}
}


void SimpleRenderer::PumpRObject(const RObject *robj, bool zburn, bool shadow) const
{
	BindVertexBuffer(robj->GetMesh()->rg_info.vb_id);
	BindIndexBuffer(robj->GetMesh()->rg_info.ib_id);
	SetModelMatrix(robj->GetNode()->GetModelMatrix());
	if (shadow && zburn) SetCulling(CULL_CW);

	if (robj->IsSkinned())
	{
		for (unsigned int j=0; j<robj->GetBoneCount(); j++) 
		{
			ComplexMesh *mesh = robj->GetMesh();
			Matrix4x4 bone;
			bone.ResetIdentity();
				
			bone *= robj->GetBone(j)->GetModelMatrix();
			bone *= robj->GetBone(j)->GetStaticModelMatrix().Inverse();
			bone *= robj->GetNode()->GetModelMatrix();//model_matrix;
			
			SetBoneMatrix(j, bone);
		}

		SetModelMatrix(robj->GetOwner()->GetModelMatrix());
	}

	else
	{
		if (robj->GetOwner()->CanUseModelMatrix())
		{
			SetModelMatrix(robj->GetOwner()->GetModelMatrix());
		}
	}
	
	if (!zburn)
	{
		// bind the material (also sets all the textures)
		robj->GetMaterial()->Bind();
		if (shadow)
		{
			BindTexture(2, cube_tex);
			SetTextureFiltering(2, TF_LINEAR, TF_LINEAR);
		}
	}

	BindUniforms();

	RGInfo *rg = &robj->GetMesh()->rg_info;
	int pcount = rg->primitive_count;
	RenderIndexedPrimitives(PT_TRIANGLE_LIST, rg->start_index, pcount, rg->start_vertex, rg->vertex_count);

	if (!zburn && shadow) SetTextureFiltering(2, TF_LINEAR, TF_LINEAR);
	if (shadow && zburn) SetCulling(CULL_CCW);
}

static int stencil_ref = 1;

void SimpleRenderer::Render(const Scene *scene) const
{
	fr_drops = 0;
	//RenderCube(scene, scene->GetActiveCamera(), cube_tex);
	curr_scene = scene;
	//RenderCube(scene, scene->GetActiveCamera(), cube_tex);
	Render(scene, scene->GetActiveCamera());
	//Matrix4x4 cam_rot = (Matrix3x3)scene->GetActiveCamera()->GetViewMatrix();
	//SetUniform(STATE_AUX_MATRIX, &cam_rot);
	//ImFixedFunction(false);
	//ImOverlay(Vector2(0, 0), Vector2(1, 1), Color(1, 1, 1, 1), cube_tex, overlay_shader);
	/*
	Matrix4x4 cam_rot = (Matrix3x3)scene->GetActiveCamera()->GetViewMatrix();
	SetUniform(STATE_AUX_MATRIX, &cam_rot);
	ImFixedFunction(false);
	ImOverlay(Vector2(0, 0), Vector2(1, 1), Color(1, 1, 1, 1), cube_tex, overlay_shader);
	*/

	ConPrintf("Frustum drops : %d\n", fr_drops);
}

void SimpleRenderer::Render(const Scene *scene, const Camera *can, unsigned int cube_face) const
{
	stencil_ref = 1;
	// clear
	//if(scene->GetClearScene()) {
		//SetClearColor(scene->GetBGColor());
	SetClearColor(Color(0, 0, 0, 1)); // needed for additive blending
	SetClearDepth(1.0f);
	Clear();

	SetCulling(CULL_CCW);

	//}
	SetupTransform(scene->GetActiveCamera(), cube_face);
	// DEBUG!!!
	SetUniform(STATE_AUX_MATRIX, &GetViewMatrix().Inverse());
	curr_vr = 0;
	ProcessVisibility();

	// zburn
	SetDepthFunc(CMP_LESS);
	SetColorMask(false);
	ZBurnPass();
	SetColorMask(true);

	// lighting
	SetDepthFunc(CMP_LEQUAL);
	AmbientPass();

	SetBlending(true);
	SetBlendFunc(BLEND_ONE, BLEND_ONE);

	unsigned int active_lt = 0;
	std::map<PVS*, std::vector<RLight*> >::iterator it;
	for (it=vr[0].pvs_specific_lights.begin(); it!=vr[0].pvs_specific_lights.end(); it++)
	{
		PVSPass(it->first);
		/*
		if (it->first->IsOutdoor())
		{
			ConPrintf("OUTDOOR!\n");
		}*/
		for (unsigned int i=0; i<it->second.size(); i++)
		{
			LightPass(it->second[i]);
			active_lt++;
		}
	}
	ConPrintf("active lights = %d\n", active_lt);

	SetStencil(false);
	for (unsigned int i=0; i<vr[0].global_lights.size(); i++)
	{
		LightPass(vr[0].global_lights[i]);
	}

	SetBlending(false);

	// render smart boxes
	//SetModelMatrix(Matrix4x4());
	//ImSetApplyXForm(false);
	//ImBegin(IM_LINES);
	//ImFixedFunction(true);
	//SetLighting(false);
	//ImColor(1, 1, 1);
	//for (unsigned int i=0; i<scene->GetObjectGroupCount(); i++)
	//{
	//	ImWireBox(const_cast<Scene*>(scene)->GetObjectGroup((int)i)->GetSmartBox());
	//}
	//ImEnd();
	//ImSetApplyXForm(true);
}

void SimpleRenderer::RenderCube(const Scene *scene, const Camera *cam, unsigned int rt) const
{
	bool prev_stencil = GetStencil();
	SetStencil(false);
	bool prev_blend = GetBlending();
	SetBlending(false);
	Matrix4x4 v, p;
	v = GetViewMatrix();
	p = GetProjectionMatrix();
	Box v_fr = scene->GetActiveCamera()->GetFrustum();
	std::vector<Box> boxes;
	for (unsigned int i=0; i<6; i++)
	{
		SetRenderTarget(rt, 0, i);
		SetClearDepth(1.0f);
		SetClearColor(Color(10000, 0, 0));
		Clear();

		// see if the current box frustum intersects the
		// view frustum
		Box c_fr = CalculateViewFrustum(cam->GetViewMatrixCube(i), cam->GetProjectionMatrixCube());
		boxes.push_back(c_fr);
		if (!Intersect(v_fr, c_fr)) 
		{
			fr_drops ++;
			continue;
		}

		SetupTransform(cam, i);
		curr_vr = 1;
		ProcessVisibility(true);
		RenderAll(depthsq_shader, depthsq_skin_shader);
		//Render(scene, cam, i);
	}
	SetRenderTarget(0);
	SetProjectionMatrix(p);
	SetViewMatrix(v);

	SetBlending(prev_blend);
	SetStencil(prev_stencil);
}

void SimpleRenderer::ZBurnPass() const
{
	BindGPUProgram(zburn_rigid_shader);
	for (unsigned int i=0; i<vr[curr_vr].rigid_obj.size(); i++)
	{
		RObject *curr = vr[curr_vr].rigid_obj[i];
		PumpRObject(curr, true, false);
	}
	
	BindGPUProgram(zburn_skin_shader);
	for (unsigned int i=0; i<vr[curr_vr].skinned_obj.size(); i++)
	{
		PumpRObject(vr[curr_vr].skinned_obj[i], true, false);
	}
}

void SimpleRenderer::RenderAll(unsigned int rigid, unsigned int skinned) const
{
	BindGPUProgram(rigid);
	for (unsigned int i=0; i<vr[curr_vr].rigid_obj.size(); i++)
	{
		RObject *curr = vr[curr_vr].rigid_obj[i];
		PumpRObject(curr, false, false);
	}
	
	BindGPUProgram(skinned);
	for (unsigned int i=0; i<vr[curr_vr].skinned_obj.size(); i++)
	{
		PumpRObject(vr[curr_vr].skinned_obj[i], false, false);
	}
}

void SimpleRenderer::AmbientPass() const
{
	RenderAll(ambient_shader, ambient_skin_shader);
	/*
	BindGPUProgram(ambient_shader);
	for (unsigned int i=0; i<rigid_obj.size(); i++)
	{
		PumpRObject(rigid_obj[i], false);
	}
	
	BindGPUProgram(ambient_skin_shader);
	for (unsigned int i=0; i<skinned_obj.size(); i++)
	{
		PumpRObject(skinned_obj[i], false);
	}*/
}

void SimpleRenderer::LightPass(RLight *lt) const
{
	if (lt->GetType() != RL_OMNI) return;

	bool shadows = lt->CastShadows();
	if (!cast_shadows) shadows = false;
	if (shadows)
	{
		PRSCamera prscam;
		prscam.SetPosition(lt->GetPosition());
		// prepare shadowmap
		RenderCube(curr_scene, &prscam, cube_tex);
	}

	lt->Bind(0);
	BindGPUProgram(shadows ? rigid_shader_shadow : rigid_shader);
	for (unsigned int i=0; i<vr[0].rigid_obj.size(); i++)
	{
		// see if the light touches the object
		Box bb = vr[0].rigid_obj[i]->GetMesh()->model_space_bounding_box;
		bb.Transform(vr[0].rigid_obj[i]->GetNode()->GetModelMatrix());
		if (!Intersect(lt->GetBoundingSphere(), bb)) continue;

		// if the light is pvs-specific, see if
		// the object belongs to the same set
		if (lt->IsPVSSpecific())
		{
			PVS *pvs = lt->GetPVS();
			Cullable *cull = (Cullable*) vr[0].rigid_obj[i];
			if (!cull->OwnedBy(pvs)) continue;
		}

		PumpRObject(vr[0].rigid_obj[i], false, shadows);
	}
	
	BindGPUProgram(shadows ? skin_shader_shadow : skin_shader);
	for (unsigned int i=0; i<vr[0].skinned_obj.size(); i++)
	{
		// TODO: do the same for skinned meshes
		PumpRObject(vr[0].skinned_obj[i], false, shadows);
	}
}

void SimpleRenderer::RenderProxy(const Scene *scene) const
{
	// clear
	if (scene->GetClearScene())
	{
		SetClearColor(scene->GetBGColor());
		SetClearColor(Color(0, 0, 0, 1));
		SetClearDepth(1.0f);
		Clear();
	}

	SetupTransform(scene->GetActiveCamera());

	SetLighting(true);
	std::vector<CollisionMesh*> proxys = scene->GetMeshes();
	Box frustum = GetFrustum();
	BindGPUProgram(color_shader);
	for (unsigned int i=0; i<proxys.size(); i++)
	{
		if (render_only_highlight)
		{
			if (proxys[i] != highlight_proxy)
				continue;
		}
		Box bb = proxys[i]->GetBoundingBox();
		if (!bb.FrustumCheck(frustum, proxys[i]->GetTransform()))
			continue;
		SetModelMatrix(proxys[i]->GetTransform());
		BindUniforms();
		ImFixedFunction(false);
		ImSetApplyXForm(false);
		ImBegin(IM_TRIANGLES);
		{
			for (unsigned int j=0; j<proxys[i]->GetTriangleCount(); j++)
			{
				// decide color
				ImColor(0.5f, 0.5f, 0.5f, 0.5f);
				if (hl_prop_bit &&
					(proxys[i]->GetTriangleProperties(j).prop0 & highlight_property))
				{
					ImColor(1, 0, 0, 1);
				}
				if (!hl_prop_bit && proxys[i]->GetTriangleProperties(j).prop1 == highlight_property)
				{
					ImColor(1, 0, 0, 1);
				}
				Triangle tri = proxys[i]->GetTriangle(j);
				ImNormal(proxys[i]->GetTriangleNormal(j));
				ImVertex(tri.GetVertex(0));
				ImVertex(tri.GetVertex(1));
				ImVertex(tri.GetVertex(2));
			}
		}
		ImEnd();
		ImSetApplyXForm(true);
	}
}

void SimpleRenderer::ProcessVisibility(bool meshes_only) const
{
	// determine visibility
	Box frustum = GetFrustum();
	Vector3 viewer = Vector3(0, 0, 0);
	viewer.Transform(GetInvViewMatrix());
	VisibleItems vi = curr_scene->GetVisibleItems(frustum, viewer);
	std::vector<RObject*> robj = vi.objects;

	// create 2 sets of objects to render: skinned and rigid
	vr[curr_vr].skinned_obj.clear();
	vr[curr_vr].rigid_obj.clear();
	for (unsigned int i=0; i<robj.size(); i++)
	{
		if (robj[i]->IsSkinned())
		{
			vr[curr_vr].skinned_obj.push_back(robj[i]);
		}
		else
		{
			vr[curr_vr].rigid_obj.push_back(robj[i]);
		}
	}
	if (meshes_only) return;

	//lights = vi.lights;

	// sort all pvs -specific lights by pvs
	// keep global lights in a separate list
	vr[curr_vr].global_lights.clear();
	std::map<PVS*, std::vector<RLight*> >::iterator it;
	for (it=vr[curr_vr].pvs_specific_lights.begin(); it!=vr[curr_vr].pvs_specific_lights.end(); it++)
	{
		it->second.clear();
	}
	vr[curr_vr].pvs_specific_lights.clear();
	for (unsigned int i=0; i<vi.lights.size(); i++)
	{
		RLight *lt = vi.lights[i];
		if (lt->IsPVSSpecific())
		{
			vr[curr_vr].pvs_specific_lights[lt->GetPVS()].push_back(lt);
		}
		else
		{
			vr[curr_vr].global_lights.push_back(lt);
		}
	}

	vr[curr_vr].active_sets = vi.active_sets;
	ConPrintf("active sets: %d\n", vr[curr_vr].active_sets.size());
}

void SimpleRenderer::PVSPass(PVS *set) const
{

	if (stencil_ref >= 255)
	{
		stencil_ref = 1;
	}
	stencil_ref ++;

	bool vinside = false;
	if (curr_scene->GetPVSWorld()->IsViewerInside(set))
	{
		vinside = true;
	}

	// write to the stencil buffer
	SetStencil(true);
	SetStencilFunc(CMP_ALWAYS);
	SetStencilPass(STENCIL_REPLACE);
	SetStencilFail(STENCIL_KEEP);
	SetStencilRef(stencil_ref);

	SetDepthTest(true);
	
	bool prev_blending = GetBlending();
	SetBlending(false);

	CullFace prev_culling = GetCulling();

	SetCulling(vinside ? CULL_CCW : CULL_CW);

	SetColorMask(false);
	SetDepthMask(false);

	ImSetApplyXForm(false);
	ImBegin(IM_QUADS);
	ImFixedFunction(true);
	SetModelMatrix(Matrix4x4());
	for (unsigned int i=0; i<set->GetPortalCount(); i++)
	{
		Quad q = set->GetPortal(i).quad;
		ImQuad(q);
	}
	ImEnd();
	ImSetApplyXForm(true);

	SetDepthMask(true);
	SetColorMask(true);
	SetCulling(prev_culling);
	SetBlending(prev_blending);

	SetStencil(true);
	SetStencilRef(stencil_ref);
	SetStencilFunc(vinside ? CMP_NOTEQUAL : CMP_EQUAL);
	//SetStencilFunc(vinside ? CMP_EQUAL : CMP_NOTEQUAL);
	SetStencilPass(STENCIL_KEEP);
	SetStencilFail(STENCIL_KEEP);

	//ImOverlay(Vector2(0, 0), Vector2(1, 1), Color(0, 0, stencil_ref * 0.3f), 0, 0);
}

// Settings
void SimpleRenderer::SetParam(const char *pname, bool val)
{
	if (!strcmp(pname, "ENABLE_SHADOWS"))
	{
		cast_shadows = val;
		return;
	}
}









