#include <vbuffer.h>
#include <textures.h>
#include "scene.h"
#include "texman.h"
#include <gc.h>
#include <errlog.h>

// RObject class implementation

RObject::RObject()
{
	owner = 0;
	mesh = 0;
	material = 0;
	skin = false;
	memset(skin_nodes, 0, sizeof *skin_nodes);
	bone_count = 0;

}

RObject::~RObject()
{
}

bool RObject::IsSkinned() const
{
	return skin;
}

ObjectGroup *RObject::GetOwner() const
{
	return owner;
}

void RObject::SetOwner(ObjectGroup *owner)
{
	this->owner = owner;
}


ComplexMesh *RObject::GetMesh() const
{
	return mesh;
}

bool RObject::SetMesh(ComplexMesh *mesh, ObjectGroup *og)
{
	if(!og) og = owner;

	if(!og->HasMesh(mesh)) {
		return false;
	}
	owner = og;

	random_move = owner->GetRandomMove();
	this->mesh = mesh;
	material = owner->GetMaterial(mesh);
	node = owner->GetNode(mesh);

	// TODO: the mesh must provide functions to determine
	// vertex components
	skin = false;
	if(mesh->geom->GetBoneWeightArray() && mesh->geom->GetBoneIndexArray()) {
        skin = true;
		bone_count = owner->GetBoneCount(mesh);
		
		for(unsigned int i=0; i<bone_count; i++) {
			skin_nodes[i] = owner->GetBone(mesh, i);
		}
	}
	return true;
}

OGNode *RObject::GetNode() const
{
	return node;
}

OGNode *RObject::GetBone(unsigned int idx) const
{
	if(idx >= bone_count) {
		return 0;
	}
	return skin_nodes[idx];
}

unsigned int RObject::GetBoneCount() const
{
	return bone_count;
}

FullMaterial *RObject::GetMaterial() const
{
	return material ? material : GetDefaultMaterial();
}

void RObject::SetMaterial(FullMaterial *mat)
{
	material = mat;
}


Box RObject::GetSpacetimeBoundingBox() const
{
	return mesh->bounding_box;
}

unsigned int RObject::GetCullableType() const
{
	return CULLABLE_MESH;
}



// TODO : check for correct FREE memory and objects

Scene::Scene ()
{
	mesh_manager = NULL;
	active_camera = 0;
	clear_scene = true;
	render_proxy = false;
}

Scene::~Scene ()
{
	Destroy();
	printf("deleting scene collisionobj\n");
}

void Scene::Destroy()
{
	// ===============================================================
	// NOTE: currently class Scene inherits from CollisionObject
	// CollisionObject will delete the og's in the destructor, so
	// dont delete them
	// Change this if:
	// 1 - Scene does not inherit any more from CollisionObject
	// 2 - Scene manages non-collisionobject og's
	// ===============================================================
	// for (unsigned int i=0; i<obj_groups.size(); i++)
	// {
	// 	if (obj_groups[i])
	// 	{
	// 		delete obj_groups[i];
	// 		obj_groups[i] = 0;
	// 	}
	// }


	obj_groups.clear();
	static_objects.clear();
	dynamic_objects.clear();
	dynamic_lights.clear();

	og_name_map.clear();

	pvs_world.Destroy();

	if (mesh_manager)
	{
		delete mesh_manager;
		mesh_manager = 0;
	}
}

//TODO: also make this merge the materials of the objectgroup to the Scene::materials vector
ObjectGroup *Scene::AddObjectGroup(const char *filename, const char *obj_name, bool is_static)
{
	//TODO : FIX THIS WITH IDMANAGER
	ObjectGroup *obj_group = new ObjectGroup ();
	if (obj_group->Load (filename, is_static) == false)
	{
		delete obj_group;
		return NULL;
	}
	return AddObjectGroup(obj_group, obj_name);
}

ObjectGroup *Scene::AddObjectGroupFromFP(VFS_FILE *fp, const char *obj_name, bool is_static)
{
	//TODO : FIX THIS WITH IDMANAGER
	ObjectGroup *obj_group = new ObjectGroup ();
	if (obj_group->LoadFromFP(fp, is_static) == false)
	{
		delete obj_group;
		return NULL;
	}
	
	return AddObjectGroup(obj_group, obj_name);
}

ObjectGroup *Scene::AddObjectGroup(ObjectGroup *og, const char *obj_name)
{
	std::string name = obj_name;
	og_name_map[name] = og;
	obj_groups.push_back (og);
	AddCollisionObject(og);

	return og;
}

bool Scene::BuildScene ()
{
	vis.ClearObjects();

	mesh_manager = new MeshManager ();
	for (unsigned int i = 0; i < obj_groups.size(); i++)
	{
		for (unsigned int j = 0; j < obj_groups[i]->GetMeshCount(); j++)
		{
			mesh_manager->AddGeometry (obj_groups[i]->GetMesh(j)->geom, 0, obj_groups[i]->GetMesh(j)->geom->GetVertexFormat(), &obj_groups[i]->GetMesh(j)->rg_info);
		}
		for (unsigned int j = 0; j < obj_groups[i]->GetLightCount(); j++)
		{
			dynamic_lights.push_back (obj_groups[i]->GetLight(j));
		}
	}
	if (mesh_manager->Build() == false)
	{
		return false;
	}

	for (unsigned int i = 0; i < obj_groups.size(); i++)
	{
		//if objectgroup is static ???

		for (unsigned int j = 0; j < obj_groups[i]->GetMeshCount(); j++)
		{
			RObject smesh;
			smesh.SetMesh(obj_groups[i]->GetMesh(j), obj_groups[i]);

			static_objects.push_back (smesh);
		}
	}

	// calculate spacetime bb's
	for (unsigned int i=0; i<obj_groups.size(); i++)
	{
		obj_groups[i]->CalculateAllKeys();
	}
	for (unsigned int i=0; i<static_objects.size(); i++)
	{
		RObject *robj = &static_objects[i];
		ComplexMesh *m = robj->GetMesh();
		std::vector<Vector3> allv;
		Box b = m->model_space_bounding_box;

		OGNode *node = robj->GetNode();
		if (node->all_keys.size())
		{
			for (unsigned int j=0; j<node->all_keys.size(); j++)
			{
				Matrix4x4 cm = node->all_keys[j];
				Box cb = b;
				cb.Transform(cm);
				for (unsigned int k=0; k<8; k++)
				{
					allv.push_back(cb.GetVertex(k));
				}
			}

			m->bounding_box = Box(allv);
		}
	}
	for (unsigned int i=0; i<obj_groups.size(); i++)
	{
		obj_groups[i]->ClearAllKeys();
	}

	// build pvs world
	for (unsigned int i=0; i<obj_groups.size(); i++)
	{
		ObjectGroup *og = obj_groups[i];
		for (unsigned int j=0; j<og->GetConvexAreaCount(); j++)
		{
			pvs_world.AddConvexArea(og->GetConvexArea(j));
		}
	}
	pvs_world.BuildSets();
	for (unsigned int i=0; i<static_objects.size(); i++)
	{
		if (static_objects[i].random_move)
		{
			dynamic_objects.push_back(static_objects[i]);
		}
		else
		{
			pvs_world.AddCullable(&static_objects[i]);
		}
	}
	pvs_world.BuildWorld();

	// Categorize PVS-specific lights
	for (unsigned int i=0; i<dynamic_lights.size(); i++)
	{
		RLight *lt = dynamic_lights[i];
		if (lt->GetType() == RL_DIRECTIONAL) continue;
		if (!lt->IsPVSSpecific()) continue;
		for (unsigned int j=0; j<pvs_world.GetSetCount(); j++)
		{
			PVS *set = pvs_world.GetSet(j);
			if (set->IsPointInside(lt->GetPosition()))
			{
				lt->SetPVS(set);
				if (!set->IsOutdoor())
					break;
			}
		}
		if (!lt->GetPVS())
		{
			lt->SetPVS(pvs_world.GetOutdoorSet());
		}
	}

	//TODO : sort static_meshes by vb / matid here...
	return true;
}

void Scene::SetActiveCamera(Camera *cam)
{
	active_camera = cam;
}

Camera *Scene::GetActiveCamera()
{
	return active_camera;
}

const Camera *Scene::GetActiveCamera() const
{
	return active_camera;
}

void Scene::Render() const
{
	for(size_t i=0; i<static_objects.size(); i++) {
		ComplexMesh *mesh = static_objects[i].GetMesh();
		FullMaterial *mat = static_objects[i].GetMaterial();

		BindVertexBuffer(mesh->rg_info.vb_id);
		BindIndexBuffer(mesh->rg_info.ib_id);
		BindTexture(0, mat->tex[FMTT_DIFFUSE].tex_id);
		BindTexture(1, mat->tex[FMTT_NORMAL].tex_id);
		RenderIndexedPrimitives(PT_TRIANGLE_LIST, mesh->rg_info.start_index,
			mesh->rg_info.primitive_count, mesh->rg_info.start_vertex, mesh->rg_info.vertex_count);
	}
}

unsigned int Scene::GetStaticObjectCount() const
{
	return (unsigned int)static_objects.size();
}

RObject *Scene::GetStaticObjectList()
{
	if(static_objects.size() > 0) {
		return &static_objects[0];
	}
	return 0;
}

const RObject *Scene::GetStaticObjectList() const
{
	if(static_objects.size() > 0) {
		return &static_objects[0];
	}
	return 0;
}

unsigned int Scene::GetDynamicLightCount() const
{
	return (unsigned int)dynamic_lights.size();
}

RLight *Scene::GetDynamicLight (unsigned int lid)
{
	return lid < dynamic_lights.size() ? dynamic_lights[lid] : 0;
}

const RLight *Scene::GetDynamicLight(unsigned int lid) const
{
	return lid < dynamic_lights.size() ? dynamic_lights[lid] : 0;
}

Color Scene::GetBGColor() const
{
	return bg_color;
}

void Scene::SetBGColor(const Color& c)
{
	bg_color = c;
}

void Scene::SetClearScene(bool cs)
{
	clear_scene = cs;
}

bool Scene::GetClearScene() const
{
	return clear_scene;
}

unsigned int Scene::GetObjectGroupCount() const
{
	return (unsigned int) obj_groups.size();
}

ObjectGroup *Scene::GetObjectGroup(int index)
{
	return obj_groups[index];
}

ObjectGroup *Scene::GetObjectGroup(const std::string &name)
{
	return og_name_map[name];
}

void Scene::SetShowProxy(bool enable)
{
	render_proxy = enable;
}

PickResult Scene::Pick(float point_x, float point_y)
{
	/*
	Vector2 cursor = Vector2 (point_x, point_y);
	// make cursor homogenous coords
	Vector3 cur_hom = Vector3(2.0f * cursor.x - 1.0f, 
		2.0f * (1.0f - cursor.y) - 1.0f, 0.0f);
	
	// transform into view space
	Vector3 cur_view = cur_hom.Transformed(GetProjectionMatrix().Inverse());

	// transform into world space
	Vector3 cur_world = cur_view.Transformed(GetViewMatrix().Inverse());
	Vector3 viewer_world = Vector3(0, 0, 0).Transformed(GetViewMatrix().Inverse());

	Vector3 dir = cur_world - viewer_world;
	dir.Normalize();
	Ray r = Ray(cur_world, dir);

	float min_dist = 1000000.0f;
	int closest_og = -1;
	int closest_obj = -1;
	unsigned int triangle_id = 0;
	Triangle triangle;

	for (unsigned int i=0; i<obj_groups.size(); i++)
	{
		for (unsigned int j=0; j<obj_groups[i]->GetMeshCount(); j++)
		{
			// process intersection
			CollisionObject *col = obj_groups[i]->GetMesh(j)->col_obj;
			if (!col) continue;
			Matrix4x4 mat;
			mat.ResetIdentity();
			col->SetTransform(mat);
			if (!col->Collision(r, true)) continue;
			float dist = DotProduct(col->GetCollisionPoint() - r.GetOrigin(),
				r.GetDirection());
			if (dist < min_dist)
			{
				min_dist = dist;
				closest_og = i;
				closest_obj = j;
				triangle_id = col->GetCollisionTriangleIndex(CT_THIS);
				triangle = col->GetCollisionTriangle(CT_THIS);
			}
		}
	}

	PickResult ret;
	ret.r = r;
	if (closest_og == -1 || closest_obj == -1)
	{
		ret.valid = false;
		return ret;
	}
	ret.valid = true;
	ret.objectgroup_id = closest_og;
	ret.mesh_id = closest_obj;
	ret.objectgroup = obj_groups[closest_og];
	ret.mesh = ret.objectgroup->GetMesh(closest_obj);
	ret.triangle_id = triangle_id;
	ret.triangle = triangle;

	return ret;
	*/
	return PickResult();
}

VisibleItems Scene::GetVisibleItems(const Box &frustum, const Vector3 &viewer) const
{
	CullableList cl = pvs_world.Query(frustum, viewer);
	VisibleItems ret;

	// calculate the smartbox of each og
	for (unsigned int i=0; i<obj_groups.size(); i++)
	{
		obj_groups[i]->CalculateSmartBox();
	}

	for (unsigned int i=0; i<cl.size(); i++)
	{
		if (cl[i]->GetCullableType() == CULLABLE_MESH)
		{
			ret.objects.push_back((RObject*) cl[i]);
		}
	}
	for (unsigned int i=0; i<dynamic_objects.size(); i++)
	{
		if (dynamic_objects[i].IsSkinned())
		{
			// smart box test
			if (!Intersect(frustum, dynamic_objects[i].GetOwner()->GetSmartBox())) 
				continue;
		}
		else
		{
			// bounding box test
			// TODO: test if this works
			Box bb = dynamic_objects[i].GetMesh()->bounding_box;
			bb.Transform(dynamic_objects[i].GetOwner()->GetModelMatrix());
			if (!Intersect(frustum, bb))
				continue;
		}
		Cullable *cull = (Cullable*) &dynamic_objects[i];
		ret.objects.push_back((RObject*) cull);
	}

	// Get visible lights
	std::map<RLight*, bool> lmap;
	for (unsigned int i=0; i<dynamic_lights.size(); i++)
	{
		RLight *lt = dynamic_lights[i];

		Sphere s;
		switch (lt->GetType())
		{
		case RL_OMNI:
			s = Sphere(lt->GetPosition(), lt->GetAttenEnd());
			if (Intersect(s, frustum))
			{
				lmap[lt] = true;
			}
			break;
		case RL_SPOT:
			// TODO: add proper cone-box test
			s = GetOutSphereOfCone(lt->GetPosition(), lt->GetConeCenter(), lt->GetConeRange());
			if (Intersect(s, frustum))
			{
				lmap[lt] = true;
			}
			break;
		case RL_DIRECTIONAL:
			lmap[lt] = true;
			break;
		}
	}

	// for all lights that are pvs - specific, see if
	// they do not belong to a currently visible set
	ret.active_sets = pvs_world.GetActiveSets();
	printf("active sets : %d\n", ret.active_sets.size());
	std::map<RLight*, bool>::iterator it;
	for (it=lmap.begin(); it!=lmap.end(); it++)
	{
		if (it->first->GetType() == RL_DIRECTIONAL) continue;
		if (!it->first->IsPVSSpecific()) continue;
		if (!pvs_world.IsSetActive(it->first->GetPVS()))
		{
			it->second = false;
		}
	}

	for (it=lmap.begin(); it!=lmap.end(); it++)
	{
		if (!it->second) continue;
		ret.lights.push_back(it->first);
	}

	return ret;
}

VisibleItems Scene::GetVisibleItems(const Matrix4x4 &view, const Matrix4x4 &proj) const
{
	Vector3 viewer = Vector3(0, 0, 0).Transformed(view.Inverse());
	Box fr = CalculateViewFrustum(view, proj);
	return GetVisibleItems(fr, viewer);
}

void Scene::RenderPVS()
{
	pvs_world.Render();
}

void Scene::AddLight(RLight *lt)
{
	dynamic_lights.push_back(lt);
}