#ifndef SCENE_H
#define SCENE_H

#include <map>
#include <vector>
#include <libvis.h>
#include <libvfs.h>
#include "objectgroup.h"
#include "meshman.h"
#include "builddefs.h"
#include "collision.h"


struct PickResult
{
	bool valid;
	unsigned int objectgroup_id;
	ObjectGroup *objectgroup;
	unsigned int mesh_id;
	ComplexMesh *mesh;
	unsigned int triangle_id;
	Triangle triangle;
	Ray r;
};

// TODO: rename RObject -> Object
class RObject : public Cullable
{
private:
	ObjectGroup *owner;
	ComplexMesh *mesh;	// TODO: this must change to Mesh *
	FullMaterial *material;
	OGNode *node;

	bool skin;
	OGNode *skin_nodes[MAX_BONES];
	unsigned int bone_count;

public:
	bool cast_shadows;
	bool random_move;

	RObject();
	~RObject();

	bool IsSkinned() const;

	ObjectGroup *GetOwner() const;
	void SetOwner(ObjectGroup *owner);

	ComplexMesh *GetMesh() const;
	bool SetMesh(ComplexMesh *mesh, ObjectGroup *og = 0);

	OGNode *GetNode() const;

	OGNode *GetBone(unsigned int idx) const;
	unsigned int GetBoneCount() const;

	FullMaterial *GetMaterial() const;
	void SetMaterial(FullMaterial *mat);

	// cullable interface
	Box GetSpacetimeBoundingBox() const;
	unsigned int GetCullableType() const;
};

struct VisibleItems
{
	std::vector<RObject*> objects;
	std::vector<RLight*> lights;
	std::vector<PVS*> active_sets;
};

// fwd declaration
class Renderer;
class SimpleRenderer;

class Scene : public CollisionObject
{
protected:
	Vis vis;
	std::vector<ObjectGroup *> obj_groups;
	MeshManager *mesh_manager;
	PVSWorld pvs_world;

	Camera *active_camera;

	std::vector<RObject> static_objects;
	std::vector<RObject> dynamic_objects;
	std::vector<RLight*> dynamic_lights;

	bool clear_scene;
	Color bg_color;

	bool render_proxy;

	std::map<std::string, ObjectGroup*> og_name_map;
public:
	Scene();
	~Scene();

	ObjectGroup *AddObjectGroup(const char *filename, const char *obj_name = "foo", bool is_static = false);
	ObjectGroup *AddObjectGroupFromFP(VFS_FILE *fp, const char *obj_name = "foo", bool is_static = false);
	ObjectGroup *AddObjectGroup(ObjectGroup *og, const char *obj_name = "foo");

	bool BuildScene();
	void Render() const;

	bool Load(const char *fname);
	bool Save(const char *fname);

	unsigned int GetStaticObjectCount() const;
	RObject *GetStaticObjectList();
	const RObject *GetStaticObjectList() const;

	unsigned int GetDynamicLightCount() const;
	RLight *GetDynamicLight(unsigned int lid);
	const RLight *GetDynamicLight(unsigned int lid) const;

	void SetActiveCamera(Camera *cam);
	Camera *GetActiveCamera();
	const Camera *GetActiveCamera() const;

	Color GetBGColor() const;
	void SetBGColor(const Color& c);

	void SetClearScene(bool cs);
	bool GetClearScene() const;

	unsigned int GetObjectGroupCount() const;
	ObjectGroup *GetObjectGroup(int index);
	ObjectGroup *GetObjectGroup(const std::string &name);

	void SetShowProxy(bool enable);

	PickResult Pick(float x, float y);

	VisibleItems GetVisibleItems(const Box &frustum, const Vector3 &viewer) const;
	VisibleItems GetVisibleItems(const Matrix4x4 &view, const Matrix4x4 &proj) const;

	void RenderPVS();

	void Destroy();

	const PVSWorld *GetPVSWorld() const {return (const PVSWorld*)&pvs_world;}

	void AddLight(RLight *lt);

	// add all renderers here
	// ----------------------
	friend class Renderer;
	friend class SimpleRenderer;
};


#endif //SCENE_H

