#include <string.h>
#include <errno.h>
#include <errlog.h>
#include "scene.h"
#include "expat/expat.h"

enum XmlTag {
	TAG_SCENE,			//  root
	TAG_OBJGROUP,		//  |- object group
	TAG_OBJGRP_MATREF,	//  |  \- material reference
	TAG_MAT,			//  \- material
	TAG_MAT_ATTR,		//     \- material attribute

	TAG_COUNT
};
#define TAG_UNKNOWN		TAG_COUNT

static struct XmlSymb {
	char *name;
	XmlTag tag;
} symb_table[] = {
	{"scene", TAG_SCENE},
	{"objgroup", TAG_OBJGROUP},
	{"matref", TAG_OBJGRP_MATREF},
	{"mat", TAG_MAT},
	{"attr", TAG_MAT_ATTR}
};

static void Start(void *udata, const XML_Char *name, const XML_Char **atts);
static void End(void *udata, const XML_Char *name);
static XmlTag GetTag(const char *name);
static const char *GetAttrStr(const XML_Char **atts, const char *aname);

static bool IsNumber(const char *str);
static bool IsVector(const char *str);
static Vector4 ParseVec(const char *str);

static const char *ColorStr(const Color &col);

static const char *ind(int lvl);

bool Scene::Load(const char *fname)
{
	VFS_FILE *fp = vfs_open(fname, "r");
	if(!fp) {
		Error("Scene::Load failed to open %s: %s\n", fname, strerror(errno));
		return false;
	}

	XML_Parser xml = XML_ParserCreate("US-ASCII");
	if(!xml) {
		Error("Scene::Load failed while calling XML_ParserCreate\n");
		vfs_close(fp);
		return false;
	}

	XML_SetElementHandler(xml, Start, End);
	//XML_SetUserData(xml, &pstate);
	XML_SetParamEntityParsing(xml, XML_PARAM_ENTITY_PARSING_ALWAYS);

	const size_t bufsz = 1024;
	bool done = false;
	while(!done) {
		char *buf = (char*)XML_GetBuffer(xml, bufsz);
		if(!vfs_gets(buf, bufsz, fp)) {
			done = true;
		}

		if(!XML_ParseBuffer(xml, strlen(buf), done)) {
			Error("Scene::Load XML parsing error: %s\n", XML_ErrorString(XML_GetErrorCode(xml)));
			XML_ParserFree(xml);
			return false;
		}
	}

	XML_ParserFree(xml);
	vfs_close(fp);
	return true;
}

bool Scene::Save(const char *fname)
{
	int lvl = 0;
	VFS_FILE *fp;

	if(!(fp = vfs_open(fname, "w"))) {
		return false;
	}

	vfs_printf(fp, "%s<scene>\n", ind(lvl++));

	/*for(size_t i=0; i<materials.size(); i++) {
		vfs_printf(fp, "%s<mat id=\"%d\">\n", ind(lvl++), i);	// XXX: i?
		vfs_printf(fp, "%s<attr name=\"diffuse\" value=\"%s\"/>\n", ind(lvl), ColorStr(materials[i].diffuse));
		vfs_printf(fp, "%s<attr name=\"specular\" value=\"%s\"/>\n", ind(lvl), ColorStr(materials[i].specular));
		vfs_printf(fp, "%s<attr name=\"shininess\" value=\"%f\"/>\n", ind(lvl), materials[i].power);
		// TODO: more...
		vfs_printf(fp, "%s</mat>\n", ind(--lvl));
	}*/

	for(size_t i=0; i<obj_groups.size(); i++) {
		vfs_printf(fp, "%s<objgroup name=\"%s\"/>\n", ind(lvl), obj_groups[i]->GetFilename());
		//vfs_printf(fp, "%s</objgroup>\n", ind(--lvl));
	}

	vfs_printf(fp, "%s</scene>\n", ind(--lvl));

	vfs_close(fp);
	return true;
}


static void Start(void *udata, const XML_Char *name, const XML_Char **atts)
{
	//ParsingState *ps = (ParsingState*)udata;
	XmlTag tag = GetTag(name);
	const char *str;

	switch(tag) {
	case TAG_SCENE:
		break;

	default:
		Warning("unrecognized tag: %s\n", name);
	}

	//ps->tag_stack.push(tag);
}

static void End(void *udata, const XML_Char *name)
{
	//ParsingState *ps = (ParsingState*)udata;
	XmlTag tag = GetTag(name);

	switch(tag) {
	case TAG_SCENE:
		break;

	default:
		break;
	}

	//ps->tag_stack.pop();
}

static XmlTag GetTag(const char *name) {
	for(int i=0; i<TAG_COUNT; i++) {
		if(strcmp(name, symb_table[i].name) == 0) {
			return symb_table[i].tag;
		}
	}
	return TAG_UNKNOWN;
}

static const char *GetAttrStr(const XML_Char **atts, const char *aname) {
	while(*atts) {
		if(!strcmp(*atts, aname)) {
			return *(atts + 1);
		}
		atts++;
	}
	return 0;
}

static bool IsNumber(const char *str) {
	while(isspace(*str)) str++;
	if(!isdigit(*str) && *str != '.' && *str != '-') {
		return false;
	}
	return true;
}

static bool IsVector(const char *str) {
	if(!IsNumber(str)) {
		return false;
	}
	
	for(int i=0; i<2; i++) {
		if(!(str = strchr(str, ','))) {
			return false;
		}
		str++;

		if(!IsNumber(str)) {
			return false;
		}
	}

	return true;
}

static Vector4 ParseVec(const char *str) {
	double v[4] = {0, 0, 0, 1};

	for(int i=0; i<4; i++) {
		v[i] = atof(str);
		if(!(str = strchr(str, ','))) {
			break;
		}
		str++;
	}

	return Vector4(v[0], v[1], v[2], v[3]);
}

static const char *ColorStr(const Color &col)
{
	static char cstr[128];
	sprintf(cstr, "%.3f, %.3f, %.3f, %.3f", col.r, col.g, col.b, col.a);
	return cstr;
}

// takes an indentation level and returns the prefix tabs needed for that indentation
static const char *ind(int lvl)
{
	static char ind_buf[128];

	int i;
	for(i=0; i<lvl; i++) {
		ind_buf[i] = '\t';
	}
	ind_buf[i] = 0;

	return ind_buf;
}

