// texman.cpp
#include <builddefs.h>

#include "texman.h"
#include <stdlib.h>
#include <map>
#include <caps.h>
#include <string>
#include <image.h>
#include <convimg.h>
#include <math.h>
#include <vector>
#include <errlog.h>

#define VIEW_MIPMAPS

static TextureFormat def_tex_fmt = TFMT_32;
static bool def_mips = true;

static unsigned int BestPow2(unsigned int n)
{
	unsigned int p = 1;
	while (p < n)
	{
		p *= 2;
	}

	unsigned int s = p / 2;
	if (!s) return p;
	unsigned int sd = abs((int)s - (int)n);
	unsigned int ld = abs((int)p - (int)n);

	if (sd < ld) return s;
	return p;
}


//#define LOAD_FROM		"data"
//#define RESIZE_FACTOR	1
unsigned int CreateTextureFromFile(const char *filename, TextureFormat fmt, bool mip)
{
#ifdef LOAD_FROM
	if (filename[0] != '/')
	{
		static char fname[1024];
		char *slash;
		while((slash = strchr(filename, '/')) || (slash = strchr(filename, '\\'))) {
			filename = slash + 1;
		}
		sprintf(fname, "%s/%s", LOAD_FROM, filename);
		filename = fname;
		printf("--- FILENAME: %s\n", filename);
	}
#endif
	unsigned int w, h;
	unsigned int *img = (unsigned int*)ImgLoadImage(filename, &w, &h);
	if (!img) return 0;

#if defined(RESIZE_FACTOR) && (RESIZE_FACTOR > 1)
	// --------------- HACK --------------
	//resample(const unsigned int *psrc, unsigned int *pdst, unsigned int sw, unsigned int sh, unsigned int dw, unsigned int dh);
	int sm_x = w / RESIZE_FACTOR;
	int sm_y = h / RESIZE_FACTOR;
	unsigned int *img_sm = (unsigned int*)malloc(sm_x * sm_y * 4);
	resample(img, img_sm, w, h, sm_x, sm_y);
	free(img);
	img = img_sm;
	w = sm_x;
	h = sm_y;
	// -----------------------------------
#endif

	if ((!IsPow2(w) || !IsPow2(h)) && !GetGcCaps()->npot_textures)
	{
		// resample image to the closest pow2
		unsigned int nw = BestPow2(w);
		unsigned int nh = BestPow2(h);

		unsigned int *new_img = (unsigned int *)malloc(nw * nh * sizeof(unsigned int));
		if (!resample(img, new_img, w, h, nw, nh))
		{
			free(img);
			free(new_img);
			return 0;
		}

		free(img);
		img = new_img;
		w = nw;
		h = nh;
	}

	unsigned int tex_id = CreateTexture(w, h, TEX_2D, fmt, mip);
	if (!tex_id)
	{
		free(img);
		return 0;
	}

	if (!SetTexImage(tex_id, img, w, h, TFMT_32))
	{
		free(img);
		DeleteTexture(tex_id);
		return 0;
	}

	free(img);
	return tex_id;
}

unsigned int CreateCubeTextureFromFile(const char *filename, TextureFormat fmt, bool mip)
{
	unsigned int w, h;
	unsigned int *img, *img_ptr;
	unsigned int tex_id;
	
	if(!(img = (unsigned int *)ImgLoadImage(filename, &w, &h))) {
		return 0;
	}

	if(h < w * 6) {
		Error("CreateCubeTextureFromFile called on something that is not a cubemap\n");
		free(img);
		return 0;
	}
	img_ptr = img;
	h /= 6;

	if(!(tex_id = CreateTexture(w, h, TEX_CUBE, fmt, mip))) {
		free(img);
		return 0;
	}

	for(int i=0; i<6; i++) {
		if(!SetTexImage(tex_id, img_ptr, w, h, TFMT_32, i)) {
			free(img);
			DeleteTexture(tex_id);
			return 0;
		}
		img_ptr += w * h;
	}

	free(img);
	return tex_id;

}

struct TexInfo
{
	unsigned int priority;
	unsigned int ref_count;

	std::string filename;
};

std::map<std::string, unsigned int> name_map;
std::map<unsigned int, TexInfo> info_map;

// get a managed texture (or create it and add it to managed)
unsigned int GetTexture(const char *filename, unsigned int priority)
{
	unsigned int tex_id = name_map[filename];
	if(!tex_id) {
		tex_id = CreateTextureFromFile(filename, def_tex_fmt, def_mips);
		name_map[filename] = tex_id;
		TexInfo nfo;
		nfo.filename = filename;
		nfo.priority = priority;
		nfo.ref_count = 0;
		info_map[tex_id] = nfo;
	} else {
		if(GetTextureType(tex_id) != TEX_2D) {
			Error("GetTexture: %s is already loaded, but it's not a TEX_2D!\n", filename);
			return 0;
		}
	}

	SetTexturePriority(tex_id, priority);
	info_map[tex_id].ref_count++;
	return tex_id;
}

unsigned int GetCubeTexture(const char *filename, unsigned int priority)
{
	unsigned int tex_id = name_map[filename];
	if(!tex_id) {
		tex_id = CreateCubeTextureFromFile(filename, def_tex_fmt, def_mips);
		name_map[filename] = tex_id;
		info_map[tex_id].filename = filename;
	} else {
		if(GetTextureType(tex_id) != TEX_CUBE) {
			Error("GetCubeTexture: %s is already loaded but it's not a cube map!\n", filename);
			return 0;
		}
	}

	SetTexturePriority(tex_id, priority);
	info_map[tex_id].ref_count++;
	return tex_id;
}

void SetTexturePriority(const char *filename, unsigned int priority)
{
	SetTexturePriority(name_map[filename], priority);
}
void SetTexturePriority(unsigned int tex_id, unsigned int priority)
{
	info_map[tex_id].priority = priority;
}

// estimated texture memory (for all managed textures)
unsigned int TexMemoryUsed()
{
	unsigned int bytes = 0;
	std::map<std::string, unsigned int>::iterator it;
	for (it = name_map.begin(); it!= name_map.end(); it++)
	{
		bytes += (unsigned int)GetTextureMemUsage(it->second);
	}
	return bytes;
}

// release all managed textures
void ReleaseTextures()
{
	std::map<std::string, unsigned int>::iterator it;
	for (it=name_map.begin(); it!=name_map.end(); it++)
	{
		if (it->second)
		{
			DeleteTexture(it->second);
		}
		it->second = 0;
	}
	name_map.clear();
	info_map.clear();
}

void ReleaseTexture(const char *filename)
{
	ReleaseTexture(name_map[filename]);
}

void ReleaseTexture(unsigned int tex_id)
{
	info_map[tex_id].ref_count--;
	if (!info_map[tex_id].ref_count)
	{
		// release texture
		DeleteTexture(tex_id);
		std::string filename = info_map[tex_id].filename;
		name_map.erase(filename);
		info_map.erase(tex_id);
	}
}


// texman state
void SetDefTexFormat(TextureFormat fmt)
{
	def_tex_fmt = fmt;
}
void SetDefMipmaps(bool enable)
{
	def_mips = enable;
}

// texture conversion
struct TexConfig
{
	unsigned int w, h;
	bool mip;
	TextureFormat fmt;

	bool operator ==(const TexConfig& b) const
	{
		return w == b.w && h == b.h && mip == b.mip && fmt == b.fmt;
	}
};

// ------------------------------
// 30% error in size calculations
static float size_error = 0.3f;
// ------------------------------ 

static unsigned int Mips(unsigned int w, unsigned int h)
{
	unsigned int m = w;
	if (m > h) m = h;
	return (int) log2((float)m) + 1;
}

static unsigned int ConfBytes(const TexConfig &conf)
{
	unsigned int lvl0 = img_size(conf.w, conf.h, conf.fmt);
	if(conf.mip) {
		unsigned int levels = Mips(conf.w, conf.h);
		lvl0 = (unsigned int)(lvl0 * (1.0 - pow(0.25, (double) (levels + 1))) / 0.75);
	}
	return lvl0;
}

// checks if this configuration fits in the specified bytes (+ size_error)
static bool CheckTexConfig(const TexConfig &conf, unsigned int bytes)
{
	unsigned int err_bytes = (unsigned int) (bytes * (1.0f + size_error) + 0.5f);
	if (ConfBytes(conf) < err_bytes) return true;
	return false;
}

static TexConfig FitTex(const TexConfig &conf, unsigned int bytes)
{
	TexConfig ret = conf;
	if (!ret.w) ret.w = 1;
	if (!ret.h) ret.h = 1;
	if (conf.w == 1 || conf.h == 1) return conf;
	
	ret.fmt = TFMT_32;

	if (!CheckTexConfig(ret, bytes)) ret.fmt = TFMT_16;
	if (!CheckTexConfig(ret, bytes)) ret.fmt = TFMT_COMPRESSED_24;
	if (!CheckTexConfig(ret, bytes))
	{
		ret.w /= 2;
		ret.h /= 2;
		return FitTex(ret, bytes);
	}

	return ret;
}

static size_t FitTexture(unsigned int tex_id, unsigned int bytes)
{
	if (!tex_id) return 0;
 	TexConfig conf;
	conf.w = GetTextureWidth(tex_id);
	conf.h = GetTextureHeight(tex_id);
	conf.fmt = GetTextureFormat(tex_id);
	conf.mip = GetMipLevels(tex_id) == 1 ? false : true;

	TexConfig conf2 = FitTex(conf, bytes);
	if (conf2 == conf)
		return GetTextureMemUsage(tex_id);
	conf = conf2;

	// convert the texture
	unsigned int *img = (unsigned int*) malloc(GetTextureWidth(tex_id) * 
		GetTextureHeight(tex_id) * sizeof(unsigned int));
	GetTexImage(tex_id, img, TFMT_32);

	// resample if nesessary
	if (GetTextureWidth(tex_id) != conf.w || GetTextureHeight(tex_id) != conf.h)
	{
		unsigned int *rsmp = (unsigned int*) malloc(conf.w * conf.h * sizeof(unsigned int));
		resample(img, rsmp, GetTextureWidth(tex_id), GetTextureHeight(tex_id), conf.w, conf.h);
		free(img);
		img = rsmp;
	}

	// recreate the texture
	CreateTexture(conf.w, conf.h, TEX_2D, conf.fmt, conf.mip, tex_id);
	SetTexImage(tex_id, img, conf.w, conf.h, TFMT_32);

	return GetTextureMemUsage(tex_id);
}

unsigned int FitTextures(unsigned int quality, unsigned int bytes)
{
	std::map<unsigned int, std::vector<unsigned int> > pr_map;
	std::map<unsigned int, TexInfo>::iterator it;
	unsigned int max_pr = 0;
	for (it = info_map.begin(); it != info_map.end(); it++)
	{
		pr_map[it->second.priority].push_back(it->first);
		if (it->second.priority > max_pr)
		{
			max_pr = it->second.priority;
		}
	}

	for (unsigned int i=0; i<= max_pr; i++)
	{
		if (i <= quality) continue;

		// calculate bytes used
		unsigned int bused = 0;
		for (unsigned int j=0; j<i; j++)
		{
			for (unsigned int k=0; k<pr_map[j].size(); k++)
			{
				bused += (unsigned int) GetTextureMemUsage(pr_map[j][k]);
			}
		}

		if (bused >= bytes) return TexMemoryUsed();
		unsigned int brem = bytes - bused;

		unsigned int scale_bytes = TexMemoryUsed() - bused;
		for (unsigned int l=0; l<pr_map[i].size(); l++)
		{
			int bytes_to_fit = (int)(((double)GetTextureMemUsage(pr_map[i][l]) * (double)brem) / (double)scale_bytes);
			FitTexture(pr_map[i][l], bytes_to_fit);
		}
	}

	return TexMemoryUsed();
}
