// routines that convert images ftom one pixel format to another

#include <color_bits.h>
#include <math3d.h>
#include "convimg.h"
#include "squish/squish.h"


#ifdef LITTLE_ENDIAN
#define ALPHA_BYTE 3
#else
#define ALPHA_BYTE 0
#endif

extern "C" bool conv_32_to_24(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;

	unsigned char *src = (unsigned char*) psrc;
	unsigned char *dst = (unsigned char*) pdst;

	for (unsigned int i=0; i<4 * size; i++)
	{
		if (i % 4 == ALPHA_BYTE)
		{
			src++;
			continue;
		}
		*dst++ = *src++;
	}

	return true;
}

extern "C" bool conv_24_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;

	unsigned char *src = (unsigned char*) psrc;
	unsigned char *dst = (unsigned char*) pdst;

	for (unsigned int i=0; i<4 * size; i++)
	{
		if (i % 4 == ALPHA_BYTE)
		{
			*dst = 0xff;
			dst++;
			continue;
		}
		*dst++ = *src++;
	}

	return true;
}

extern "C" bool conv_32_to_16(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;
	
	unsigned int *src = (unsigned int*) psrc;
	unsigned short *dst = (unsigned short*) pdst;

	for (unsigned int i=0; i<size; i++)
	{
		int r = UNPACK_RED32(*src) >> 3;
		int g = UNPACK_GREEN32(*src) >> 2;
		int b = UNPACK_BLUE32(*src) >> 3;

		*dst++ = PACK_COLOR16(r, g, b);
		src++;
	}

	return true;
}

extern "C" bool conv_16_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;

	unsigned short *src = (unsigned short*) psrc;
	unsigned int *dst = (unsigned int*) pdst;

	for (unsigned int i=0; i<size; i++)
	{
		int r = UNPACK_RED16(*src) << 3;
		int g = UNPACK_GREEN16(*src) << 2;
		int b = UNPACK_BLUE16(*src) << 3;

		*dst++ = PACK_COLOR32(r, g, b, 0xff);
		src++;
	}

	return true;
}

extern "C" bool conv_32_to_15(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;
	
	unsigned int *src = (unsigned int*) psrc;
	unsigned short *dst = (unsigned short*) pdst;

	for (unsigned int i=0; i<size; i++)
	{
		int r = UNPACK_RED32(*src) >> 3;
		int g = UNPACK_GREEN32(*src) >> 3;
		int b = UNPACK_BLUE32(*src) >> 3;

		*dst++ = PACK_COLOR15(r, g, b);
		src++;
	}

	return true;
}

extern "C" bool conv_15_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;

	unsigned int size = w * h;

	unsigned short *src = (unsigned short*) psrc;
	unsigned int *dst = (unsigned int*) pdst;

	for (unsigned int i=0; i<size; i++)
	{
		int r = UNPACK_RED15(*src) << 3;
		int g = UNPACK_GREEN15(*src) << 3;
		int b = UNPACK_BLUE15(*src) << 3;
		*dst++ = PACK_COLOR32(r, g, b, 0xff);
		src++;
	}

	return true;
}

extern "C" bool conv_32_to_8(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if(!psrc || !pdst) return false;

	int size = w * h;

	unsigned char *dst = (unsigned char*)pdst;
	unsigned int *src = (unsigned int*)psrc;

	for(int i=0; i<size; i++) {
		int r = UNPACK_RED32(*src);
		int g = UNPACK_GREEN32(*src);
		int b = UNPACK_BLUE32(*src);
		*dst++ = (unsigned char)(r * 0.299 + g * 0.587 + b * 0.114);
		src++;
	}
	return true;
}

extern "C" bool conv_8_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	if(!psrc || !pdst) return false;

	int size = w * h;

	unsigned int *dst = (unsigned int*)pdst;
	unsigned char *src = (unsigned char*)psrc;

	for(int i=0; i<size; i++) {
		*dst++ = PACK_COLOR32(*src, *src, *src, 0xff);
		src++;
	}
	return true;
}

static inline unsigned int vec4_to_rgb32(Vector4 v)
{
	int r = (int)((v.x /** 0.5f + 0.5f*/) * 255.0f);
	int g = (int)((v.y /** 0.5f + 0.5f*/) * 255.0f);
	int b = (int)((v.z /** 0.5f + 0.5f*/) * 255.0f);
	int a = (int)((v.w /** 0.5f + 0.5f*/) * 255.0f);
	return PACK_COLOR32(r, g, b, a);
}

static void convert_bgra_to_rgba(unsigned int *img, int xsz, int ysz)
{
	int pixnum = xsz * ysz;

	for(int i=0; i<pixnum; i++) {
		unsigned int pixel = img[i];
		int r = UNPACK_RED32(pixel);
		int g = UNPACK_GREEN32(pixel);
		int b = UNPACK_BLUE32(pixel);
		int a = UNPACK_ALPHA32(pixel);

		/* this is how squish expects the 32bit pixel (TODO: verify on big endian) */
		img[i] = r | (g << 8) | (b << 16) | (a << 24);
	}
}

extern "C" bool conv_32_to_comp24(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	convert_bgra_to_rgba((unsigned int*)psrc, w, h);
	squish::CompressImage((const unsigned char*)psrc, w, h, pdst, squish::kDxt1 | squish::kColourMetricPerceptual | squish::kColourRangeFit);
	return true;
}

extern "C" bool conv_comp24_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	DecompressImage((unsigned char*)pdst, w, h, psrc, squish::kDxt1);
	convert_bgra_to_rgba((unsigned int*)pdst, w, h);
	return true;
}

extern "C" bool conv_32_to_comp32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	convert_bgra_to_rgba((unsigned int*)psrc, w, h);
	squish::CompressImage((const unsigned char*)psrc, w, h, pdst, squish::kDxt5 | squish::kColourMetricPerceptual | squish::kColourRangeFit);
	return true;
}

extern "C" bool conv_comp32_to_32(const void *psrc, void *pdst, unsigned int w, unsigned int h)
{
	DecompressImage((unsigned char*)pdst, w, h, psrc, squish::kDxt5);
	convert_bgra_to_rgba((unsigned int*)pdst, w, h);
	return true;
}


static inline unsigned int get_texel(unsigned int x, unsigned int y, unsigned int w, unsigned int h)
{
	if (x>=w) x = w - 1;
	if (y>=h) y = h - 1;

	return x + y*w;
}

static Vector4 sample(const Vector4 *buffer, unsigned int w, unsigned int h, float u, float v)
{
	//u -= floor(u);
	//v -= floor(v);

	// clamp
	if (u > 1.0f) u = 1.0f;
	if (u < 0.0f) u = 0.0f;
	if (v > 1.0f) v = 1.0f;
	if (v < 0.0f) v = 0.0f;

	u *= w-1;
	v *= h-1;

	unsigned int x = (unsigned int) u;
	unsigned int y = (unsigned int) v;

	u -= x;
	v -= y;

	// get 4 neighbor pixels
	Vector4 n[2][2];
	n[0][0] = buffer[get_texel(x + 0, y + 0, w, h)];
	n[0][1] = buffer[get_texel(x + 1, y + 0, w, h)];
	n[1][0] = buffer[get_texel(x + 0, y + 1, w, h)];
	n[1][1] = buffer[get_texel(x + 1, y + 1, w, h)];

	n[0][0].x = Lerp(n[0][0].x, n[0][1].x, u);
	n[0][0].y = Lerp(n[0][0].y, n[0][1].y, u);
	n[0][0].z = Lerp(n[0][0].z, n[0][1].z, u);
	n[0][0].w = Lerp(n[0][0].w, n[0][1].w, u);

	n[1][0].x = Lerp(n[1][0].x, n[1][1].x, u);
	n[1][0].y = Lerp(n[1][0].y, n[1][1].y, u);
	n[1][0].z = Lerp(n[1][0].z, n[1][1].z, u);
	n[1][0].w = Lerp(n[1][0].w, n[1][1].w, u);

	Vector4 val;
	val.x = Lerp(n[0][0].x, n[1][0].x, v);
	val.y = Lerp(n[0][0].y, n[1][0].y, v);
	val.z = Lerp(n[0][0].z, n[1][0].z, v);
	val.w = Lerp(n[0][0].w, n[1][0].w, v);

	//n[0][0] = Lerp(n[0][0], n[0][1], u);
	//n[1][0] = Lerp(n[1][0], n[1][1], u);

	return val;//Lerp(n[0][0], n[1][0], v);
}

static inline unsigned int avg4(unsigned int a, unsigned int b, unsigned int c, unsigned int d)
{
	unsigned int ret;
	unsigned char *ca = (unsigned char*) &a;
	unsigned char *cb = (unsigned char*) &b;
	unsigned char *cc = (unsigned char*) &c;
	unsigned char *cd = (unsigned char*) &d;
	unsigned char *cr = (unsigned char*) &ret;
	for (unsigned int i=0; i<4; i++)
	{
		unsigned int sum = ca[i] + cb[i] + cc[i] + cd[i];
		cr[i] = sum >> 2;
	}

	return ret;
}

static bool resample_mip(const unsigned int *psrc, unsigned int *pdst, 
						 unsigned int w, unsigned int h)
{
	if (!psrc || !pdst) return false;
	if (!w || !h) return false;
	for (unsigned int j=0; j<h / 2; j++)
	{
		for (unsigned int i=0; i<w / 2; i++)
		{
			*pdst = avg4(
				psrc[2 * i + 0 + (2 * j + 0) * w],
				psrc[2 * i + 0 + (2 * j + 1) * w],
				psrc[2 * i + 1 + (2 * j + 0) * w],
				psrc[2 * i + 1 + (2 * j + 1) * w]);
			pdst++;
		}
	}

	return true;
}

// XXX: Check if we damage alpha channel during resample
extern "C" bool resample(const unsigned int *psrc, unsigned int *pdst, unsigned int sw, unsigned int sh,
			  unsigned int dw, unsigned int dh)
{
	if (sw == 2 * dw && sh == 2 * dh)
	{
		return resample_mip(psrc, pdst, sw, sh);
	}

	if (!psrc || !pdst) return false;
	if (!sw || !sh || !dw || !dh) return false;

	// convert to float
	const unsigned int *src = psrc;
	Vector4 *fp = new Vector4[sw * sh];
	Vector4 *fp_dst = new Vector4[dw * dh];
	Vector4 *fpp = fp;
	for (unsigned int j=0; j<sh; j++)
	{
		for (unsigned int i=0; i<sw; i++)
		{
			*fpp = Vector4(UNPACK_RED32(*src) / 255.0f,
				UNPACK_GREEN32(*src) / 255.0f,
				UNPACK_BLUE32(*src) / 255.0f,
				UNPACK_ALPHA32(*src) / 255.0f);
			fpp++;
			src++;
		}
	}

	// resample
	Vector4 *fpp_dst = fp_dst;
	for (unsigned int j=0; j<dh; j++)
	{
		for (unsigned int i=0; i<dw; i++)
		{
			*fpp_dst = sample(fp, sw, sh, i / (float) (dw-1), j / (float) (dh-1));
			fpp_dst ++;
		}
	}

	// convert back to 32 bit
	for (unsigned int j=0; j<dh; j++)
	{
		for (unsigned int i=0; i<dw; i++)
		{
			pdst[i + j*dw] = vec4_to_rgb32(fp_dst[i + j*dw]);
		}
	}
		
	// cleanup
	delete [] fp;
	delete [] fp_dst;

	return true;
}

typedef bool (*conv_func)(const void *psrc, void *pdst, unsigned int w, unsigned int h);

static conv_func get_func_from32(TextureFormat dst_fmt)
{
	switch (dst_fmt)
	{
	case TFMT_32:
		return 0;

	case TFMT_24:
		return conv_32_to_24;

	case TFMT_16:
		return conv_32_to_16;

	case TFMT_8:
		return conv_32_to_8;

	case TFMT_FLOAT_32:
		return 0;

	case TFMT_COMPRESSED_32:
		return 0;

	case TFMT_COMPRESSED_24:
		return conv_32_to_comp24;

	default:
		break;
	}

	return 0;
}

static conv_func get_func_to32(TextureFormat src_fmt)
{
	switch (src_fmt)
	{
	case TFMT_32:
		return 0;

	case TFMT_24:
		return conv_24_to_32;

	case TFMT_16:
		return conv_16_to_32;

	case TFMT_8:
		return conv_8_to_32;

	case TFMT_FLOAT_32:
		return 0;

	case TFMT_COMPRESSED_32:
		return 0;

	case TFMT_COMPRESSED_24:
		return conv_comp24_to_32;

	default:
		break;
	}

	return 0;
}

extern "C" bool convert(const void *psrc, void *pdst, unsigned int w, unsigned int h, TextureFormat src_fmt, TextureFormat dst_fmt)
{
	if (!psrc || !pdst) return false;

	if (src_fmt == dst_fmt)
	{
		memcpy(pdst, psrc, img_size(w, h, src_fmt));
		return true;
	}

	if (src_fmt == TFMT_32)
	{
		conv_func func = get_func_from32(dst_fmt);
		if (!func) return false;
		return func(psrc, pdst, w, h);
	}
	if (dst_fmt == TFMT_32)
	{
		conv_func func = get_func_to32(src_fmt);
		if (!func) return false;
		return func(psrc, pdst, w, h);
	}

	// try to convert first to 32 bit
	unsigned int *tmp32 = new unsigned int[w * h];
	//conv_func func = get_func_to32(src_fmt);
	if (!convert(psrc, tmp32, w, h, src_fmt, TFMT_32))
	{
		delete [] tmp32;
		return false;
	}

	if (!convert(tmp32, pdst, w, h, TFMT_32, dst_fmt))
	{
		delete [] tmp32;
		return false;
	}

	delete [] tmp32;
	return true;
}

extern "C" unsigned int img_size(unsigned int w, unsigned int h, TextureFormat fmt)
{
	unsigned int blocks;
	switch (fmt)
	{
	case TFMT_32:
		return 4 * w * h;

	case TFMT_24:
		return 3 * w * h;

	case TFMT_16:
		return 2 * w * h;

	case TFMT_8:
		return w * h;

	case TFMT_FLOAT_32:
		return 4 * w * h;

	case TFMT_COMPRESSED_32:
		w += 4 - (w % 4);
		h += 4 - (h % 4);
		blocks = w * h / 16;
		return 16 * blocks;

	case TFMT_COMPRESSED_24:
		w += 4 - (w % 4);
		h += 4 - (h % 4);
		blocks = w * h / 16;
		return 8 * blocks;

	default:
		break;
	}

	return 0;
}
