#include <stdio.h>
#include <stdlib.h>
#include <libvfs.h>
#include <errlog.h>
#include "image.h"

#define FMT_DECL(x) \
	int Check##x(VFS_FILE *fp); \
	void *Load##x(VFS_FILE *fp, unsigned int *xsz, unsigned int *ysz); \
	int Save##x(VFS_FILE *fp, void *pixels, unsigned int xsz, unsigned int ysz)

#define FMT_FUNC(x) \
	{ Check##x, Load##x, Save##x }

#ifdef IMGLIB_USE_PNG
FMT_DECL(Png);
#endif
#ifdef IMGLIB_USE_JPEG
FMT_DECL(Jpeg);
#endif
#ifdef IMGLIB_USE_TGA
FMT_DECL(Tga);
#endif
#ifdef IMGLIB_USE_PPM
FMT_DECL(Ppm);
#endif
#ifdef IMGLIB_USE_DDS
FMT_DECL(Dds);
#endif
#ifdef IMGLIB_USE_RGBE
FMT_DECL(Rgbe);
#endif


struct {
	int (*Check)(VFS_FILE*);
	void *(*Load)(VFS_FILE*, unsigned int*, unsigned int*);
	int (*Save)(VFS_FILE*, void*, unsigned int, unsigned int);
} fmt_func[] = {
#ifdef IMGLIB_USE_PNG
	FMT_FUNC(Png),
#endif
#ifdef IMGLIB_USE_JPEG
	FMT_FUNC(Jpeg),
#endif
#ifdef IMGLIB_USE_TGA
	FMT_FUNC(Tga),
#endif
#ifdef IMGLIB_USE_PPM
	FMT_FUNC(Ppm),
#endif
#ifdef IMGLIB_USE_DDS
	FMT_FUNC(Dds),
#endif
#ifdef IMGLIB_USE_RGBE
	FMT_FUNC(Rgbe),
#endif
	{0, 0, 0}
};



static unsigned int save_flags, load_flags;
static int mip_level;

void *ImgLoadImage(const char *fname, unsigned int *xsz, unsigned int *ysz) {
	int i;
	VFS_FILE *file;
	void *img = 0;

	if(!(file = vfs_open(fname, "rb"))) {
		Error("Image loading error: could not open file %s\n", fname);
		return 0;
	}

	for(i=0; fmt_func[i].Check; i++) {
		if(fmt_func[i].Check(file)) {
			img = fmt_func[i].Load(file, xsz, ysz);
			break;
		}
	}
	
	vfs_close(file);
	return img;
}

void ImgFreeImage(void *img) {
	free(img);
}

int ImgSaveImage(const char *fname, void *pixels, unsigned int xsz, unsigned int ysz, unsigned int fmt) {
	VFS_FILE *fp;
	int res = 0;

	if(!(fp = vfs_open(fname, "wb"))) {
		Error("Image saving error: could not open file %s for writing\n", fname);
		return -1;
	}

	res = fmt_func[fmt].Save(fp, pixels, xsz, ysz);
	vfs_close(fp);
	return res;
}


void ImgSetImageSaveFlags(unsigned int flags)
{
	save_flags = flags;
}

unsigned int ImgGetImageSaveFlags(void)
{
	return save_flags;
}

void ImgSetImageLoadFlags(unsigned int flags)
{
	load_flags = flags;
}

unsigned int ImgGetImageLoadFlags(void)
{
	return load_flags;
}

void ImgSetMipmapLoadLevel(int mip)
{
	mip_level = mip;
}

int ImgGetMipmapLoadlevel(void)
{
	return mip_level;
}
