#include <stdio.h>
#include "image.h"

#ifdef IMGLIB_USE_DDS

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "convimg.h"
#include "color_bits.h"
#include <int_types.h>
#include <libvfs.h>
#include <errlog.h>

#define MKFOURCC(a, b, c, d) \
	((uint32_t)((a) & 0xff) | \
	 ((uint32_t)((b) & 0xff) << 8) | \
	 ((uint32_t)((c) & 0xff) << 16) | \
	 ((uint32_t)((d) & 0xff) << 24))

enum {
	DDS_DXT1		= MKFOURCC('D', 'X', 'T', '1'),
	DDS_DXT2		= MKFOURCC('D', 'X', 'T', '2'),
	DDS_DXT3		= MKFOURCC('D', 'X', 'T', '3'),
	DDS_DXT4		= MKFOURCC('D', 'X', 'T', '4'),
	DDS_DXT5		= MKFOURCC('D', 'X', 'T', '5')
};

enum {
	DDSD_CAPS			= 0x00000001,
	DDSD_HEIGHT			= 0x00000002,
	DDSD_WIDTH			= 0x00000004,
	DDSD_PITCH			= 0x00000008,
	DDSD_PIXELFORMAT	= 0x00001000,
	DDSD_MIPMAPCOUNT	= 0x00020000,
	DDSD_LINEARSIZE		= 0x00080000,
	DDSD_DEPTH			= 0x00800000
};

enum {
	DDSPF_ALPHAPIXELS	= 0x01,
	DDSPF_FOURCC		= 0x02,
	DDSPF_RGB			= 0x40
};

enum {
	DDSCAPS_COMPLEX		= 0x00000008,
	DDSCAPS_TEXTURE		= 0x00001000,
	DDSCAPS_MIPMAP		= 0x00400000
};

enum {
	DDSCAPS2_CUBEMAP	= 0x00000200,
	DDSCAPS2_CUBEMAP_PX	= 0x00000400,
	DDSCAPS2_CUBEMAP_NX	= 0x00000800,
	DDSCAPS2_CUBEMAP_PY	= 0x00001000,
	DDSCAPS2_CUBEMAP_NY	= 0x00002000,
	DDSCAPS2_CUBEMAP_PZ	= 0x00004000,
	DDSCAPS2_CUBEMAP_NZ	= 0x00008000,
	DDSCAPS2_VOLUME		= 0x00200000
};

struct dds_pixel_format {
	uint32_t size;
	uint32_t flags;
	uint32_t fourcc;
	uint32_t bit_count;
	uint32_t rmask, gmask, bmask, amask;
};

struct dds_caps {
	uint32_t caps1, caps2;
	uint32_t unused[2];
};

struct dds_surf_header {
	uint32_t size;				/* size of the structure (?!) */
	uint32_t flags;
	uint32_t height, width;
	uint32_t pitch_or_size;		/* for uncompressed images the pitch, otherwise the full image size */
	uint32_t depth;				/* for volume textures, the depth of the volume */
	uint32_t mipmaps;			/* mipmap count */
	uint32_t unused[11];
	struct dds_pixel_format pixfmt;
	struct dds_caps caps;
	uint32_t unused2;
};

int CheckDds(VFS_FILE *fp)
{
	char magic[4];
	vfs_seek(fp, 0, VFS_SEEK_SET);
	vfs_read(magic, 1, 4, fp);
	return magic[0] == 'D' && magic[1] == 'D' && magic[2] == 'S' && magic[3] == ' ';
}

static int read_dds_surf_header(VFS_FILE *fp, struct dds_surf_header *hdr)
{
	hdr->size = vfs_read_int32_le(fp);
	hdr->flags = vfs_read_int32_le(fp);
	hdr->height = vfs_read_int32_le(fp);
	hdr->width = vfs_read_int32_le(fp);
	hdr->pitch_or_size = vfs_read_int32_le(fp);
	hdr->depth = vfs_read_int32_le(fp);
	hdr->mipmaps = vfs_read_int32_le(fp);
	vfs_read(hdr->unused, 4, 11, fp);
	
	/* read pixelformat structure */
	hdr->pixfmt.size = vfs_read_int32_le(fp);
	hdr->pixfmt.flags = vfs_read_int32_le(fp);
	hdr->pixfmt.fourcc = vfs_read_int32_le(fp);
	hdr->pixfmt.bit_count = vfs_read_int32_le(fp);
	hdr->pixfmt.rmask = vfs_read_int32_le(fp);
	hdr->pixfmt.gmask = vfs_read_int32_le(fp);
	hdr->pixfmt.bmask = vfs_read_int32_le(fp);
	hdr->pixfmt.amask = vfs_read_int32_le(fp);

	/* read caps structure */
	hdr->caps.caps1 = vfs_read_int32_le(fp);
	hdr->caps.caps2 = vfs_read_int32_le(fp);
	vfs_read(hdr->caps.unused, 4, 2, fp);

	hdr->unused2 = vfs_read_int32_le(fp);
	return 0;
}

/*static int CalcMipChainSize(int level_size, int levels)
{
	int i, size = 0;

	for(i=0; i<levels; i++) {
		size += level_size;
		level_size /= 4;
	}
	return size;
}*/

static int ImageSize(int xsz, int ysz, int fourcc, int bpp)
{
	int size;

	if(xsz < 1) xsz = 1;
	if(ysz < 1) ysz = 1;

	if(fourcc == DDS_DXT1 || fourcc == DDS_DXT5) {
		size = xsz * ysz;
		if(fourcc == DDS_DXT1) {
			size /= 2;
		}
		if(size < 8) size = 8;
	} else {
		size = xsz * ysz * (bpp / 8);
	}

	return size;
}

static int CalcMipChainSize(struct dds_surf_header *hdr)
{
	int i, size = 0;
	int xsz = hdr->width;
	int ysz = hdr->height;

	for(i=0; i<hdr->mipmaps; i++) {
		size += ImageSize(xsz, ysz, hdr->pixfmt.fourcc, hdr->pixfmt.bit_count);
		xsz /= 2;
		ysz /= 2;
	}
	return size;
}

/* loads a compressed image starting from the current file pointer */
static void *LoadCompImg(VFS_FILE *fp, struct dds_surf_header *hdr, int *size)
{
	void *img, *raw;
	unsigned int load_flags = ImgGetImageLoadFlags();

	if(!(raw = malloc(hdr->pitch_or_size))) {
		Error("failed to allocate memory\n");
		return 0;
	}
	if(vfs_read(raw, 1, hdr->pitch_or_size, fp) != hdr->pitch_or_size) {
		Error("end of file while reading compressed dds image\n");
		free(raw);
		return 0;
	}

	/* if original pixel format requested, just return the raw data */
	if(load_flags & IMG_LOAD_FMT_ORIG) {
		*size = hdr->pitch_or_size;
		return raw;
	}

	/* otherwise convert to RGBA in the img buffer */
	if(!(img = malloc(hdr->width * hdr->height * 4))) {
		Error("failed to allocate memory\n");
		return 0;
	}

	if(hdr->pixfmt.fourcc == DDS_DXT1) {
		conv_comp24_to_32(raw, img, hdr->width, hdr->height);
	} else if(hdr->pixfmt.fourcc == DDS_DXT5) {
		conv_comp32_to_32(raw, img, hdr->width, hdr->height);
	} else {
		Error("unsupported compressed image format, cannot convert to 32bit rgba\n");
		free(img);
		free(raw);
		return 0;
	}

	free(raw);
	*size = hdr->width * hdr->height * 4;
	return img;
}

/* loads an uncompressed image from the current file pointer */
static void *LoadUncompImg(VFS_FILE *fp, struct dds_surf_header *hdr, int *size)
{
	void *img;
	unsigned char *raw, *scanline;
	int i, pixel_size;
	unsigned int load_flags = ImgGetImageLoadFlags();

	if(hdr->pixfmt.bit_count < 16) {
		Error("unsupported uncompressed image format\n");
		return 0;
	}

	pixel_size = hdr->pixfmt.bit_count / 8;
	if(!(raw = malloc(hdr->width * hdr->height * pixel_size))) {
		Error("failed to allocate memory\n");
		return 0;
	}

	if(!(scanline = malloc(hdr->pitch_or_size))) {
		Error("failed to allocate memory\n");
		free(raw);
		return 0;
	}

	for(i=0; i<hdr->height; i++) {
		if(vfs_read(scanline, 1, hdr->pitch_or_size, fp) < hdr->pitch_or_size) {
			Error("end of file while reading uncompressed dds image\n");
			free(raw);
			free(scanline);
			return 0;
		}
		memcpy(raw + i * hdr->width * pixel_size, scanline, hdr->width * pixel_size);
	}
	free(scanline);

	/* if original pixel format requested or the original format is 32bpp, just return raw */
	if(load_flags & IMG_LOAD_FMT_ORIG || hdr->pixfmt.bit_count == 32) {
		*size = hdr->width * hdr->height * pixel_size;
		return raw;
	}

	if(!(img = malloc(hdr->width * hdr->height * 4))) {
		Error("failed to allocate memory\n");
		free(raw);
		return 0;
	}

	if(hdr->pixfmt.bit_count == 16) {
		conv_16_to_32(raw, img, hdr->width, hdr->height);
	} else if(hdr->pixfmt.bit_count == 24) {
		conv_24_to_32(raw, img, hdr->width, hdr->height);
	} else {
		Error("unsupported uncompressed pixel format, can't convert to 32bpp\n");
		free(raw);
		free(img);
		return 0;
	}

	free(raw);
	*size = hdr->width * hdr->height * 4;
	return img;
}


void *LoadDds(VFS_FILE *fp, unsigned int *xsz, unsigned int *ysz)
{
	struct dds_surf_header hdr;
	unsigned int *img;
	int imgw, imgh;
	int mip_level, tmp;
	void* (*load_func)(VFS_FILE*, struct dds_surf_header*, int*);

	/* find out which mipmap level to load, 0 means original image */
	/* XXX: ignored for now, always load 0 */
	mip_level = 0;/*ImgGetMipmapLoadlevel();*/

	vfs_seek(fp, 0, VFS_SEEK_SET);
	if(!CheckDds(fp)) {
		return 0;
	}

	read_dds_surf_header(fp, &hdr);

	/* one of these must be set, if not, guess ... */
	if(!(hdr.flags & DDSD_PITCH) && !(hdr.flags & DDSD_LINEARSIZE)) {
		/*Warning("malformed dds file, missing both DDSD_PITCH and DDSD_LINEARSIZE, trying to guess...\n");*/
		if(hdr.pixfmt.flags & DDSPF_FOURCC) {
			hdr.flags |= DDSD_LINEARSIZE;
		} else if(hdr.pixfmt.flags & DDSPF_RGB) {
			hdr.flags |= DDSD_PITCH;
		} else {
			/* try to guess ... */
			if(hdr.pixfmt.fourcc == DDS_DXT1 || hdr.pixfmt.fourcc == DDS_DXT5) {
				hdr.pixfmt.flags |= DDSPF_FOURCC;
				hdr.flags |= DDSD_LINEARSIZE;
			} else if(hdr.pixfmt.bit_count == 16 || hdr.pixfmt.bit_count == 24 || hdr.pixfmt.bit_count == 32) {
				hdr.pixfmt.flags |= DDSPF_RGB;
				hdr.flags |= DDSD_PITCH;
			} else {
				Error("unsupported DDS format or exceedingly malformed file\n");
				return 0;
			}
		}
	}

	if(hdr.pitch_or_size == 0) {
		/*Warning("malformed dds file, missing pitch/linearsize, trying to guess...\n");*/
		if(hdr.pixfmt.flags & DDSPF_FOURCC) {
			hdr.pitch_or_size = ImageSize(hdr.width, hdr.height, hdr.pixfmt.fourcc, 0);
		} else {
			hdr.pitch_or_size = hdr.width * (hdr.pixfmt.bit_count / 8);
		}
	}

	load_func = (hdr.flags & DDSD_LINEARSIZE) ? LoadCompImg : LoadUncompImg;

	imgw = hdr.width;
	imgh = hdr.height;

	if(mip_level > 0) {
		imgw /= 2 * mip_level;
		imgh /= 2 * mip_level;
	}

	/* a complex surface might have mipmaps or be a cubemap, or both */
	if(hdr.caps.caps1 & DDSCAPS_COMPLEX) {
		if(hdr.caps.caps2 & DDSCAPS2_CUBEMAP) {
			int i, sz0, full_size, mip_chain_size;
			void *face_img;
			unsigned char *img_fptr;

			/* for all cubemap faces ... */
			for(i=0; i<6; i++) {
				long fpos_start = vfs_tell(fp);

				/* load the face image ... */
				if(!(face_img = load_func(fp, &hdr, &sz0))) {
					free(img);
					return 0;
				}

				if(i == 0) {	/* first time */
					full_size = sz0 * 6;
					mip_chain_size = CalcMipChainSize(&hdr);

					if(!(img = malloc(full_size))) {
						Error("failed to allocate memory\n");
						return 0;
					}
					img_fptr = (unsigned char*)img;
				}

				/* skip the rest of the mipmaps for this face ... */
				if(mip_chain_size) {
					vfs_seek(fp, mip_chain_size - (vfs_tell(fp) - fpos_start), VFS_SEEK_CUR);
				}

				/* copy it to the full cubemap image and advance the target pointer ... */
				memcpy(img_fptr, face_img, sz0);
				img_fptr += sz0;

				free(face_img);
			}

			if(xsz) *xsz = imgw;
			if(ysz) *ysz = imgh * 6;
			return img;
		}
	}

	if(xsz) *xsz = imgw;
	if(ysz) *ysz = imgh;
	return load_func(fp, &hdr, &tmp);
}

int SaveDds(VFS_FILE *fp, void *pixels, unsigned int xsz, unsigned int ysz)
{
	Error("dds saving not implemented yet\n");
	return 1;
}

#endif	/* IMGLIB_USE_DDS */
