#include <stdio.h>
#include "image.h"

#ifdef IMGLIB_USE_PNG

#include <stdlib.h>
#include "libpng/png.h"
#include "color_bits.h"
#include <int_types.h>
#include <libvfs.h>
#include <errlog.h>

#define FILE_SIG_BYTES	8

#ifdef USE_VFS
static void ReadFunc(png_struct *png, unsigned char *data, size_t len);
static void WriteFunc(png_struct *png, unsigned char *data, size_t len);
static void FlushFunc(png_struct *png);
#endif

int CheckPng(VFS_FILE *fp)
{
	unsigned char sig[FILE_SIG_BYTES];

	vfs_read(sig, 1, FILE_SIG_BYTES, fp);

	return png_sig_cmp(sig, 0, FILE_SIG_BYTES) == 0 ? 1 : 0;
}

void *LoadPng(VFS_FILE *fp, unsigned int *xsz, unsigned int *ysz)
{
	png_struct *png_ptr;
	png_info *info_ptr;
	unsigned int i;
	uint32_t **lineptr, *pixels;
	int channel_bits, color_type, ilace_type, compression, filtering;
	
	if(!(png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0))) {
		return 0;
	}
	
	if(!(info_ptr = png_create_info_struct(png_ptr))) {
		png_destroy_read_struct(&png_ptr, 0, 0);
		return 0;
	}
	
	if(setjmp(png_jmpbuf(png_ptr))) {		
		png_destroy_read_struct(&png_ptr, &info_ptr, 0);
		return 0;
	}
	
#ifndef USE_VFS
	png_init_io(png_ptr, vfs_get_real_file(fp));
#else
	png_set_read_fn(png_ptr, fp, ReadFunc);
#endif

	png_set_sig_bytes(png_ptr, FILE_SIG_BYTES);
	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_BGR, 0);
		
	png_get_IHDR(png_ptr, info_ptr, (png_uint_32*)xsz, (png_uint_32*)ysz, &channel_bits, &color_type, &ilace_type, &compression, &filtering);
	pixels = malloc(*xsz * *ysz * sizeof(uint32_t));
	
	lineptr = (uint32_t**)png_get_rows(png_ptr, info_ptr);
	
	for(i=0; i<*ysz; i++) {
		
		switch(color_type) {
			case PNG_COLOR_TYPE_RGB:
			{
				unsigned int j;
				unsigned char *ptr = (unsigned char*)lineptr[i];

				for(j=0; j<*xsz; j++) {
					uint32_t pixel;
					pixel = PACK_COLOR24(*(ptr+2), *(ptr+1), *ptr);
					ptr+=3;
					pixels[i * *xsz + j] = pixel;			
				}
			}
			break;
			
		case PNG_COLOR_TYPE_RGB_ALPHA:
			memcpy(&pixels[i * *xsz], lineptr[i], *xsz * sizeof(uint32_t));
			break;

		case PNG_COLOR_TYPE_GRAY:
		case PNG_COLOR_TYPE_GRAY_ALPHA:
		case PNG_COLOR_TYPE_PALETTE:
			
		default:
			png_destroy_read_struct(&png_ptr, &info_ptr, 0);
			return 0;
		}
				
	}
	
	png_destroy_read_struct(&png_ptr, &info_ptr, 0);
	
	return pixels;
}

int SavePng(VFS_FILE *fp, void *pixels, unsigned int imgx, unsigned int imgy) {
	png_struct *png;
	png_info *info;
	png_text txt;
	int i, j, coltype, xsz, ysz;
	uint32_t *pixptr;
	unsigned char **rows;
	unsigned int save_flags = ImgGetImageSaveFlags();

	xsz = imgx;
	ysz = imgy;

	txt.compression = PNG_TEXT_COMPRESSION_NONE;
	txt.key = "Software";
	txt.text = "mindlapse libimage (fx2)";
	txt.text_length = 0;

	if(!(png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0))) {
		return -1;
	}
	
	if(!(info = png_create_info_struct(png))) {
		png_destroy_write_struct(&png, 0);
		return -1;
	}

	if(setjmp(png_jmpbuf(png))) {
		png_destroy_write_struct(&png, &info);
		return -1;
	}

#ifndef USE_VFS
	png_init_io(png, vfs_get_real_file(fp));
#else
	png_set_write_fn(png, fp, WriteFunc, FlushFunc);
#endif

	coltype = (save_flags & IMG_SAVE_ALPHA) ? PNG_COLOR_TYPE_RGB_ALPHA : PNG_COLOR_TYPE_RGB;
	png_set_IHDR(png, info, xsz, ysz, 8, coltype, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_set_text(png, info, &txt, 1);

	if(!(rows = malloc(ysz * sizeof *rows))) {
		png_destroy_write_struct(&png, &info);
		return -1;
	}

	pixptr = save_flags & IMG_SAVE_INVERT ? (uint32_t*)pixels + (ysz - 1) * xsz : pixels;
	for(i=0; i<ysz; i++) {
		if(coltype == PNG_COLOR_TYPE_RGB) {
			rows[i] = malloc(xsz * 3);
			for(j=0; j<xsz; j++) {
				rows[i][j * 3] = (pixptr[j] & BLUE_MASK32) >> BLUE_SHIFT32;
				rows[i][j * 3 + 1] = (pixptr[j] & GREEN_MASK32) >> GREEN_SHIFT32;
				rows[i][j * 3 + 2] = (pixptr[j] & RED_MASK32) >> RED_SHIFT32;
			}
		} else {
			rows[i] = (unsigned char*)pixptr;
		}
		pixptr += save_flags & IMG_SAVE_INVERT ? -xsz : xsz;
	}

	png_set_rows(png, info, rows);
	png_write_png(png, info, PNG_TRANSFORM_BGR, 0);

	if(coltype == PNG_COLOR_TYPE_RGB) {
		for(i=0; i<ysz; i++) {
			free(rows[i]);
		}
	}
	free(rows);

	png_write_end(png, info);
	png_destroy_write_struct(&png, &info);

	return 0;
}

#ifdef USE_VFS
static void ReadFunc(png_struct *png, unsigned char *data, size_t len)
{
	VFS_FILE *fp = (VFS_FILE*)png_get_io_ptr(png);
	vfs_read(data, 1, len, fp);
}

static void WriteFunc(png_struct *png, unsigned char *data, size_t len)
{
	VFS_FILE *fp = (VFS_FILE*)png_get_io_ptr(png);
	vfs_write(data, 1, len, fp);
}

static void FlushFunc(png_struct *png)
{
	VFS_FILE *fp = (VFS_FILE*)png_get_io_ptr(png);
	vfs_flush(fp);
}
#endif

#endif	/* IMGLIB_USE_PNG */
