#include <stdio.h>
#include "image.h"

#ifdef IMGLIB_USE_TGA

#include <stdlib.h>
#include <string.h>

#include <color_bits.h>
#include <byteorder.h>
#include <libvfs.h>
#include <errlog.h>

enum {
	IMG_NONE,
	IMG_CMAP,
	IMG_RGBA,
	IMG_BW,

	IMG_RLE_CMAP = 9,
	IMG_RLE_RGBA,
	IMG_RLE_BW
};

#define IS_RLE(x)	((x) >= IMG_RLE_CMAP)
#define IS_RGBA(x)	((x) == IMG_RGBA || (x) == IMG_RLE_RGBA)

struct TgaHeader {
	uint8_t idlen;			/* id field length */
	uint8_t cmap_type;		/* color map type (0:no color map, 1:color map present) */
	uint8_t img_type;		/* image type: 
							 *  0: no image data
							 *  1: uncomp. color-mapped		 9: RLE color-mapped
							 *  2: uncomp. true color		10: RLE true color
							 *  3: uncomp. black/white		11: RLE black/white */	
	uint16_t cmap_first;	/* color map first entry index */
	uint16_t cmap_len;		/* color map length */
	uint8_t cmap_entry_sz;	/* color map entry size */
	uint16_t img_x;			/* X-origin of the image */
	uint16_t img_y;			/* Y-origin of the image */
	uint16_t img_width;		/* image width */
	uint16_t img_height;	/* image height */
	uint8_t img_bpp;		/* bits per pixel */
	uint8_t img_desc;		/* descriptor: 
							 * bits 0 - 3: alpha or overlay bits
							 * bits 5 & 4: origin (0 = bottom/left, 1 = top/right)
							 * bits 7 & 6: data interleaving */	
};

struct TgaFooter {
	uint32_t ext_off;		/* extension area offset */
	uint32_t devdir_off;	/* developer directory offset */
	char sig[18];				/* signature with . and \0 */
};

/*static void print_tga_info(struct TgaHeader *hdr);*/

int CheckTga(VFS_FILE *fp) {
	struct TgaFooter foot;
	
	vfs_seek(fp, -18, VFS_SEEK_END);
	vfs_read(foot.sig, 1, 18, fp);

	foot.sig[17] = 0;
	return strcmp(foot.sig, "TRUEVISION-XFILE.") == 0 ? 1 : 0;
}

static uint32_t ReadPixel(VFS_FILE *fp, int rdalpha) {
	int r, g, b, a;
	b = vfs_getc(fp);
	g = vfs_getc(fp);
	r = vfs_getc(fp);
	a = rdalpha ? vfs_getc(fp) : 0xff;
	return PACK_COLOR32(r, g, b, a);
}


void *LoadTga(VFS_FILE *fp, unsigned int *xsz, unsigned int *ysz) {
	struct TgaHeader hdr;
	unsigned long x, y, sz;
	unsigned int i;
	uint32_t *pix, ppixel;
	int rle_mode, rle_pix_left = 0;
	int rdalpha;

	/* read header */
	vfs_seek(fp, 0, VFS_SEEK_SET);
	hdr.idlen = vfs_getc(fp);
	hdr.cmap_type = vfs_getc(fp);
	hdr.img_type = vfs_getc(fp);
	hdr.cmap_first = vfs_read_int16_le(fp);
	hdr.cmap_len = vfs_read_int16_le(fp);
	hdr.cmap_entry_sz = vfs_getc(fp);
	hdr.img_x = vfs_read_int16_le(fp);
	hdr.img_y = vfs_read_int16_le(fp);
	hdr.img_width = vfs_read_int16_le(fp);
	hdr.img_height = vfs_read_int16_le(fp);
	hdr.img_bpp = vfs_getc(fp);
	hdr.img_desc = vfs_getc(fp);

	if(vfs_eof(fp)) {
		return 0;
	}

	/*print_tga_info(&hdr);*/
	
	/* only read true color images */
	if(!IS_RGBA(hdr.img_type)) {
		Error("only true color tga images supported\n");
		return 0;
	}
	vfs_seek(fp, hdr.idlen, VFS_SEEK_CUR); /* skip the image ID */

	/* skip the color map if it exists */
	if(hdr.cmap_type == 1) {
		vfs_seek(fp, hdr.cmap_len * hdr.cmap_entry_sz / 8, VFS_SEEK_CUR);
	}

	x = hdr.img_width;
	y = hdr.img_height;
	sz = x * y;
	if(!(pix = malloc(sz * 4))) {
		return 0;
	}

	rdalpha = hdr.img_desc & 0xf;

	for(i=0; i<y; i++) {
		uint32_t *ptr;
		unsigned int j;

		ptr = pix + ((hdr.img_desc & 0x20) ? i : y-(i+1)) * x;

		for(j=0; j<x; j++) {
			/* if the image is raw, then just read the next pixel */
			if(!IS_RLE(hdr.img_type)) {
				ppixel = ReadPixel(fp, rdalpha);
			} else {
                /* otherwise, for RLE... */

				/* if we have pixels left in the packet ... */
				if(rle_pix_left) {
					/* if it's a raw packet, read the next pixel, otherwise keep the same */
					if(!rle_mode) {
						ppixel = ReadPixel(fp, rdalpha);
					}
					rle_pix_left--;
				} else {
					/* read the RLE packet header */
					unsigned char packet_hdr = vfs_getc(fp);
					rle_mode = (packet_hdr & 128);		/* last bit shows the mode for this packet (1: rle, 0: raw) */
					rle_pix_left = (packet_hdr & ~128);	/* the rest gives the count of pixels minus one (we also read one here, so no +1) */
					ppixel = ReadPixel(fp, rdalpha);	/* and read the first pixel of the packet */
				}
			}

			*ptr++ = ppixel;

			if(vfs_eof(fp)) break;
		}
	}

	*xsz = x;
	*ysz = y;
	return pix;
}

int SaveTga(VFS_FILE *fp, void *pixels, unsigned int xsz, unsigned int ysz) {
	struct TgaHeader hdr;
	struct TgaFooter ftr;
	unsigned long pix_count = xsz * ysz;
	uint32_t *pptr = pixels;
	unsigned long save_flags;
	unsigned int i;

	save_flags = ImgGetImageSaveFlags();

	memset(&hdr, 0, sizeof hdr);
	hdr.img_type = 2;
	hdr.img_width = (uint16_t)xsz;
	hdr.img_height = (uint16_t)ysz;

	if(save_flags & IMG_SAVE_ALPHA) {
		hdr.img_bpp = 32;
		hdr.img_desc = 8 | 0x20;	/* 8 alpha bits, origin top-left */
	} else {
		hdr.img_bpp = 24;
		hdr.img_desc = 0x20;		/* no alpha bits, origin top-left */
	}

	if(save_flags & IMG_SAVE_INVERT) {
		hdr.img_desc ^= 0x20;
	}

	ftr.ext_off = 0;
	ftr.devdir_off = 0;
	strcpy(ftr.sig, "TRUEVISION-XFILE.");

	/* write the header */
	
	vfs_write(&hdr.idlen, 1, 1, fp);
	vfs_write(&hdr.cmap_type, 1, 1, fp);
	vfs_write(&hdr.img_type, 1, 1, fp);
	vfs_write_int16_le(fp, hdr.cmap_first);
	vfs_write_int16_le(fp, hdr.cmap_len);
	vfs_write(&hdr.cmap_entry_sz, 1, 1, fp);
	vfs_write_int16_le(fp, hdr.img_x);
	vfs_write_int16_le(fp, hdr.img_y);
	vfs_write_int16_le(fp, hdr.img_width);
	vfs_write_int16_le(fp, hdr.img_height);
	vfs_write(&hdr.img_bpp, 1, 1, fp);
	vfs_write(&hdr.img_desc, 1, 1, fp);

	/* write the pixels */
	for(i=0; i<pix_count; i++) {
		vfs_putc((*pptr >> BLUE_SHIFT32) & 0xff, fp);
		vfs_putc((*pptr >> GREEN_SHIFT32) & 0xff, fp);
		vfs_putc((*pptr >> RED_SHIFT32) & 0xff, fp);
		
		if(save_flags & IMG_SAVE_ALPHA) {
			vfs_putc((*pptr >> ALPHA_SHIFT32) & 0xff, fp);
		}
		
		pptr++;
	}

	/* write the footer */
	vfs_write_int32_le(fp, ftr.ext_off);
	vfs_write_int32_le(fp, ftr.devdir_off);
	vfs_puts(ftr.sig, fp);
	vfs_putc(0, fp);

	return 0;
}

/*static void print_tga_info(struct TgaHeader *hdr) {
	static const char *img_type_str[] = {
		"no image data",
		"uncompressed color-mapped",
		"uncompressed true color",
		"uncompressed black & white",
		"", "", "", "", "",
		"RLE color-mapped",
		"RLE true color",
		"RLE black & white"
	};
	
	printf("id field length: %d\n", (int)hdr->idlen);
	printf("color map present: %s\n", hdr->cmap_type ? "yes" : "no");
	printf("image type: %s\n", img_type_str[hdr->img_type]);
	printf("color-map start: %d\n", (int)hdr->cmap_first);
	printf("color-map length: %d\n", (int)hdr->cmap_len);
	printf("color-map entry size: %d\n", (int)hdr->cmap_entry_sz);
	printf("image origin: %d, %d\n", (int)hdr->img_x, (int)hdr->img_y);
	printf("image size: %d, %d\n", (int)hdr->img_width, (int)hdr->img_height);
	printf("bpp: %d\n", (int)hdr->img_bpp);
	printf("attribute bits (alpha/overlay): %d\n", (int)(hdr->img_desc & 0xf));
	printf("origin: %s-", (hdr->img_desc & 0x20) ? "top" : "bottom");
	printf("%s\n", (hdr->img_desc & 0x10) ? "right" : "left");
}*/

#endif	/* IMGLIB_USE_TGA */
