/*
 * Post-Processing Effects
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <badcfg.h> 
#include <libvfs.h>
#include <libgc.h>
#include <immed.h>
#include <uniforms.h>
#include <libcommon/timer.h>
#include <libcommon/errlog.h>
#include "libppe.h"

static PFXEffect		**effect = NULL;
static unsigned int		effects = 0;
static PFXShader		**pshader = NULL;
static unsigned int		pshaders = 0;
static float			hpx, hpy;

static float time(float scale)
{
	return sin((float)GetMSec()*Pi/1000.0f*scale);
}

#define ARG(a)	(a.time?time:a.value)

// PFXBuffer
PFXBuffer::PFXBuffer()
{
	ownrt = false;
	rt = 0;
}

PFXBuffer::~PFXBuffer()
{
	if (ownrt) DeleteTexture(rt);
}

PFXBuffer *PFXBuffer::Create(int width, int height)
{
	PFXBuffer	*buff = new PFXBuffer();
	buff->width = width;
	buff->height = height;
	buff->rt = CreateRenderTarget(width, height, TFMT_32);
	return buff;
}

// PFXShader
PFXShader::~PFXShader()
{
}

// PFXPixelShader
PFXPixelShader::PFXPixelShader(char *progname)
{
	char		tmp[1024];
	VFS_FILE	*f;
	name = strdup("");
	for (int i=strlen(progname)-1;i>=0;i--)
		if (progname[i] == '/' || progname[i] == '\\')
		{
			strcpy(tmp, progname + i + 1);
			for (int j=0;tmp[j];j++)
				if (tmp[j] == '.')
				{
					tmp[j] = 0;
					free(name);
					name = strdup(tmp);
					break;
				}
			break;
		}

	prog = 0;
	f = vfs_open(progname, "rb");
	code = NULL;
	if (f)
	{
		long	size = vfs_filesize(f);
		code = (char*)malloc(size + 1);
		vfs_read(code, size, 1, f);
		code[size] = 0;
		vfs_close(f);
	}
}

PFXPixelShader::~PFXPixelShader()
{
	ReleaseGPUProgram(prog);
	free(name);
}

void PFXPixelShader::Apply(PFXBuffer *in, PFXBuffer *out, PFXArg *arg, int args, float time)
{
	if (!prog && !code) return;

	if (!prog)
	{
		prog = CreateGPUProgram(
			// vertex shader
			"uniform float4x4 STATE_MODELVIEWPROJECTION_MATRIX; "
			"uniform float4x4 STATE_TEXTURE0_MATRIX; "

			"struct a2v "
			"{ "
			"	float4 pos	: POSITION; "
			"	float2 tc	: TEXCOORD0; "
			"	float4 col	: COLOR; "
			"}; "

			"a2v main(a2v IN) "
			"{ "
			"	a2v OUT; "

			"	OUT.pos = mul(IN.pos, STATE_MODELVIEWPROJECTION_MATRIX); "
			"	float4 tc4; "
			"	tc4.xy = IN.tc.xy; "
			"	tc4.z = tc4.w = 1; "

			"	OUT.tc = (float2) mul(tc4, STATE_TEXTURE0_MATRIX); "

			"	return OUT; "
			"} ", code);
		free(code);
		if (prog == 0) Error("Failed to compile the '%s' post-processing pixel shader.\n", name);
	}


	ImFixedFunction(false);
	SetRenderTarget(out->rt);
	for (int i=0; i<args; i++)
	{
		unsigned int	uf = FindUniform(arg[i].name);
		float			f;
		if (uf)
		{
			f = ARG(arg[i]);
			SetUniform(uf, &f);
		}
	}
	ImOverlay(Vector2(-hpx, -hpy), Vector2(1 - hpx, 1 - hpy), Color(1, 1, 1, 1), in->rt, prog);
	SetRenderTarget(0); //TODO: use GetRenderTarget once available
}

// PFXReference
PFXReference::PFXReference()
{
	arg = NULL;
	args = 0;
}

PFXReference::~PFXReference()
{
	if (arg)
	{
		for (int i=0; i<args;i++)
			free(arg[i].name);
		free(arg);
	}
}

void PFXReference::AddArg(char *name, char *value)
{
	arg = (PFXArg*)realloc(arg, sizeof(PFXArg)*(args + 1));
	arg[args].name = strdup(name);
	if (value[0] == 't')
	{
		arg[args].value = 0.0f;
		arg[args].time = true;
	}
	else
	{
		arg[args].value = (float)atof(value);
		arg[args].time = false;
	}
	args++;
}

// PFXEffect
PFXEffect::PFXEffect()
{
	name = NULL;
	shader = NULL;
	arg = NULL;
	args = 0;
	ref = NULL;
	refs = 0;
	timescale = 1.0f;
	posteffect = true;
}

PFXEffect::~PFXEffect()
{
	if (arg)
	{
		for (int i=0; i<args;i++)
			free(arg[i].name);
		free(arg);
	}
	if (ref)
	{
		for (int i=0; i<refs;i++)
			delete ref[i];
		free(ref);
	}
	if (name) free(name);
}

void PFXEffect::LoadFromPPFX(void *_node)
{
	BadCFGNode	*node = (BadCFGNode*)_node;
	name = strdup(node->value);
	
	for (BadCFGNode *child=node->child; child; child=child->next)
	{
		if (!strcmp(child->name, "shader"))
		{
			shader = GetPFXShader(child->value);
			if (!shader) printf("Warning: could not find post-processing shader '%s'\n", child->value);
		}
		else if (!strcmp(child->name, "argument"))
		{
			AddArg(child->value,
			       getBadCFGValue(child, "default", "0.0"));
		}
		else if (!strcmp(child->name, "effect"))
		{
			PFXEffect		*rfx = GetPFXEffect(child->value);
			PFXReference	*ref;
			if (!rfx)
			{
				printf("Warning: effect'%s' was not defined!\n",
					child->value);
				continue;
			}
			ref = AddRef(rfx);
			BadCFGNode	*an = getBadCFGNode(child, "timescale");
			if (an)
				ref->timescale = (float)atof(an->value);
			else
				ref->timescale = rfx->timescale;
			
			for (int i=0; i<rfx->args; i++)
			{
				an = getBadCFGNode(child, rfx->arg[i].name);
				if (an)
					ref->AddArg(an->name, an->value);
				else
				{
					if (rfx->arg[i].time)
						ref->AddArg(rfx->arg[i].name, "time");
					else
					{
						char	tmp[100];
						sprintf(tmp, "%f", rfx->arg[i].value);
						ref->AddArg(rfx->arg[i].name, tmp);
					}
				}
			}
		}
		else if (!strcmp(child->name, "timescale"))
		{
			timescale = (float)atof(child->value);
		}
	}
}

void PFXEffect::AddArg(char *name, char *value)
{
	char	realname[1024];
	arg = (PFXArg*)realloc(arg, sizeof(PFXArg)*(args + 1));
	sprintf(realname, "%s_%s", this->name, name);
	arg[args].name = strdup(realname);
	if (value[0] == 't')
	{
		arg[args].value = 0.0f;
		arg[args].time = true;
	}
	else
	{
		arg[args].value = (float)atof(value);
		arg[args].time = false;
	}
	args++;

	unsigned int	uf = FindUniform(realname);
	if (!uf) AddUniform(realname, 1);
}

PFXReference *PFXEffect::AddRef(PFXEffect *fx)
{
	ref = (PFXReference**)realloc(ref, sizeof(PFXReference*)*(refs + 1));
	ref[refs] = new PFXReference();
	ref[refs]->fx = fx;
	return ref[refs++];
}

// PFXSet
PFXSet::PFXSet(PFXEffect *fx)
{
	root = fx;
	ref = NULL;
	refs = 0;
	tmpb = NULL;
	Flatten();
}

PFXSet::~PFXSet()
{
	if (ref)
		free(ref);
	
	if (tmpb) delete tmpb;
}

void PFXSet::AddRef(PFXReference *rfc)
{
	ref = (PFXReference**)realloc(ref, sizeof(PFXReference*)*(refs + 1));
	ref[refs++] = rfc;
}

void PFXSet::AddEffect(PFXEffect *fx)
{
	for (int i=0; i<fx->refs; i++)
	{
		if (!fx->posteffect && fx->ref[i]->fx->shader)
			AddRef(fx->ref[i]);
		if (fx->ref[i]->fx->refs)
			AddEffect(fx->ref[i]->fx);
		if (fx->posteffect && fx->ref[i]->fx->shader)
			AddRef(fx->ref[i]);
	}
}

void PFXSet::Flatten()
{
	AddEffect(root);
}

void PFXSet::Apply(PFXBuffer *in, PFXBuffer *out)
{
	PFXBuffer		*b1, *b2;
	PFXBuffer		*tmp;
	bool			reverse = false;
	float			fv;
	GCInitParams	gcip = GetInitParams();

	fv = 1.0/gcip.dmode.width;
	SetUniform(FindUniform("ppfx_pixel_width"), &fv);
	fv = 1.0/gcip.dmode.height;
	SetUniform(FindUniform("ppfx_pixel_height"), &fv);

	if (!tmpb || tmpb->width != in->width || tmpb->height != in->height)
	{
		if (tmpb) delete tmpb;
		tmpb = PFXBuffer::Create(in->width, in->height);
	}
	
	b1 = in;
	b2 = tmpb;

	BindTexture(1, in->rt);
	
	#define SWAPPFXBUFFERS { \
		if (b1 == in) b1 = out; \
		tmp = b1; b1 = b2; b2 = tmp; reverse = !reverse; \
	}
	if (!root->posteffect && root->shader)
	{
		root->shader->Apply(b1, b2, root->arg, root->args, time(root->timescale));
		if (refs)
			SWAPPFXBUFFERS
	}
	for (int i=0; i<refs; i++)
	{
		ref[i]->fx->shader->Apply(b1, b2, ref[i]->arg, ref[i]->args, time(ref[i]->timescale));
		if (i < refs-1)
			SWAPPFXBUFFERS
	}
	if (root->posteffect && root->shader)
	{
		if (refs)
			SWAPPFXBUFFERS
		root->shader->Apply(b1, b2, root->arg, root->args, time(root->timescale));
	}
	#undef SWAPPFXBUFFERS
	
	if (b2 == tmpb)
	{
		ImFixedFunction(true);
		SetRenderTarget(out->rt);
		ImOverlay(Vector2(-hpx, -hpy), Vector2(1 - hpx, 1 - hpy), Color(1, 1, 1, 1), tmpb->rt, 0);
		ImFixedFunction(false);
	}
}

// Functions
static void LoadPPFX(char *filename)
{
	BadCFGNode	*ppfx;
	VFS_FILE	*f = vfs_open(filename, "rb");
	long		size = vfs_filesize(f);
	char		*code = (char*)malloc(size + 1);
	vfs_read(code, size, 1, f);
	code[size] = 0;
	vfs_close(f);

	ppfx = parseBadCFGCode(code);
	free(code);

	if (!ppfx)
	{
		printf("Warning: failed to load post-processing effect library '%s'!\n", filename);
		return;
	}
	
	for (BadCFGNode *child=ppfx->child; child; child=child->next)
	{
		if (!strcmp(child->name, "effect"))
		{
			if (GetPFXEffect(child->name)) continue;
			
			PFXEffect	*fx = new PFXEffect();
			fx->LoadFromPPFX(child);
			effect = (PFXEffect**)realloc(effect, sizeof(PFXEffect*)*(effects + 1));
			effect[effects++] = fx;
		}
		else if (!strcmp(child->name, "import"))
		{
			char	tmp[1024];
			// replace with engine stuff
			sprintf(tmp, "ppfx/%s.t7ppfx", child->value);
			LoadPPFX(tmp);
		}
	}
}

static void AddShader(PFXShader *ps)
{
	pshader = (PFXShader**)realloc(pshader, sizeof(PFXShader*)*(pshaders + 1));
	pshader[pshaders++] = ps;
}

void InitPostProcessing()
{
	AddUniform("ppfx_pixel_width", 1);
	AddUniform("ppfx_pixel_height", 1);

	// load pixel shaders
	VFS_DIR		*dir = vfs_opendir("ppfx");
	vfs_dirent	*de;
	if (dir)
	{
		printf("Scanning ppfx:\n");
		while ((de = vfs_readdir(dir)))
		{
			char	tmp[1024];
			if (!strstr(de->d_name, ".fs")) continue;
			sprintf(tmp, "ppfx/%s", de->d_name);
			PFXPixelShader	*ps = new PFXPixelShader(tmp);
			AddShader(ps);
		}
		vfs_closedir(dir);
	}

	// loader effect descriptors
	dir = vfs_opendir("ppfx");
	if (!dir) return;
	while ((de = vfs_readdir(dir)))
	{
		char	tmp[1024];
		if (!strstr(de->d_name, ".t7ppfx")) continue;
		sprintf(tmp, "ppfx/%s", de->d_name);
		LoadPPFX(tmp);
	}
	vfs_closedir(dir);

	GCInitParams	ip = GetInitParams();
	hpx = 0.5/(float)ip.dmode.width;
	hpy = 0.5/(float)ip.dmode.height;
}

void ShutdownPostProcessing()
{
	for (unsigned int i=0; i<effects; i++)
		delete effect[i];
	free(effect);
	effect = NULL;
	effects = 0;

	for (unsigned int i=0; i<pshaders; i++)
		delete pshader[i];
	free(pshader);
	pshader = NULL;
	pshaders = 0;
}

PFXEffect *GetPFXEffect(char *name)
{
	// TODO: replace with hashmap
	for (unsigned int i=0; i<effects; i++)
		if (!strcmp(effect[i]->name, name))
			return effect[i];
	return NULL;
}

PFXShader *GetPFXShader(char *name)
{
	// TODO: replace with hashmap
	for (unsigned int i=0; i<pshaders; i++)
		if (!strcmp(pshader[i]->name, name))
			return pshader[i];
	return NULL;
}
