#include <builddefs.h>
#ifdef BUILD_WIN32
#include <stack>
#include <windows.h>
#include "libsys.h"

#ifndef WM_MOUSEWHEEL
#define WM_MOUSEWHEEL	0x020A
#endif
#ifndef WHEEL_DELTA
#define WHEEL_DELTA		120
#endif
#ifndef GET_WHEEL_DELTA_WPARAM 
#define GET_WHEEL_DELTA_WPARAM(wp)	((short)HIWORD(wp)) 
#endif 

#define CLASS_NAME	"t7gfxwin"

static void CalculateWindowSize(int w, int h, bool border, int *winwidth, int *winheight, int *cleft, int *ctop);
static LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static int vkey_to_keysym(unsigned int vkey);

static HWND win = NULL;
static int width, height;
static bool win_external = false;

extern std::stack<Handlers> sys_handlers;

bool SysCreateWindow(unsigned int w, unsigned int h, bool border, void *data)
{
	if (data == NULL)
	{
		win_external = false;
		width = w;
		height = h;

		WNDCLASS wclass;
		memset(&wclass, 0, sizeof(WNDCLASS));
		wclass.style = CS_VREDRAW | CS_HREDRAW | CS_OWNDC;
		wclass.lpfnWndProc = WindowProc;
		wclass.hInstance = GetModuleHandle(NULL);
		wclass.lpszClassName = CLASS_NAME;
	
		if (RegisterClass (&wclass) == 0)
		{
			return false;
		}
		int disp_left, disp_top;
		int win_width, win_height;

		CalculateWindowSize (w, h,  border, &win_width, &win_height, &disp_left, &disp_top);

		DWORD wstyle;
		if (border == false)
		{
			wstyle = WS_POPUP | WS_VISIBLE;
		}
		else
		{
		    wstyle = WS_BORDER | WS_CAPTION | WS_MINIMIZEBOX | WS_POPUP | WS_SYSMENU | WS_VISIBLE;
		}
	
		win = CreateWindow (CLASS_NAME,
							"Track7 Games",
							wstyle,
							disp_left, disp_top,
							win_width, win_height,
							0,
							NULL,
							GetModuleHandle (NULL),
							0);
		if (win == NULL)
		{
		    SysCloseWindow();
			return false;
		}
		SysShowWindow ();
		SysSetForegroundWindow();
		SetActiveWindow(win);
	}
	else
	{
		win_external = true;
		win = (HWND) data;

		RECT rect;
		memset (&rect, 0, sizeof (rect));

		GetClientRect (win, &rect);

		width = rect.right;
		height = rect.bottom;
	}

	/* set the initial handlers to zero */
	Handlers handlers;
	memset(&handlers, 0, sizeof handlers);
	sys_handlers.push(handlers);

	return true;
}

void SysCloseWindow()
{
	if(win_external) {
		if(win) {
	        DestroyWindow(win);
		}
		UnregisterClass(CLASS_NAME, GetModuleHandle(NULL));
	}
	win = 0;
}

void SysSetWinTitle(const char *title)
{
	if(win) {
		 SetWindowText(win, title);
	}
}

void SysShowWindow()
{
	ShowWindow(win, SW_SHOW);
}

void SysSetForegroundWindow()
{
	SetForegroundWindow(win);
}

void *SysGetWindow()
{
	return (void*)win;
}


int SysGetWindowWidth()
{
	return width;
}

int SysGetWindowHeight()
{
	return height;
}

void SysShowCursor(bool show)
{
	ShowCursor(show);
}

bool SysProcessEvents(bool block)
{
	XPadEvent xpev;
	UpdateXPad();
	while(GetXPadEvent(xpev)) {
		if(sys_handlers.top().xpad) {
			sys_handlers.top().xpad(xpev);
		}
	}

	MSG msg;

	/* if block is true, wait for at least one event, then go into a PeekMessage loop
	 * to handle all available events before returning.
	 */
	if(block) {
		if(!GetMessage(&msg, NULL, 0, 0)) {
			return false;
		}
		
		do {
			if(msg.message == WM_QUIT) {
				return false;
			}
            TranslateMessage(&msg);
			DispatchMessage(&msg);
		} while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE));
	} else {
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if(msg.message == WM_QUIT) {
				return false;
			}
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return true;
}

static void CalculateWindowSize(int w, int h, bool border, int *winwidth, int *winheight, int *cleft, int *ctop)
{
	*cleft = 0;
	*ctop = 0;

	*winwidth = w;
	*winheight = h;

	DWORD  wstyle;
	if (border == false)
	{
		wstyle = WS_POPUP | WS_VISIBLE;
	}
	else
	{
		wstyle = WS_BORDER | WS_CAPTION | WS_MINIMIZEBOX | WS_POPUP | WS_SYSMENU | WS_VISIBLE;
	}

	RECT  clientArea;
	clientArea.left = 0;
	clientArea.top = 0;
	clientArea.right = w;
	clientArea.bottom = h;

	if (AdjustWindowRect(&clientArea, wstyle, FALSE) == 0)
	{
		return;
	}

	*winwidth  = clientArea.right - clientArea.left;
	*winheight = clientArea.bottom - clientArea.top;

	DEVMODE devMode;
	memset (&devMode, 0, sizeof (DEVMODE));
	devMode.dmSize = sizeof (DEVMODE);

	if (EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &devMode) == 0)
	{
		return;
	}

	*cleft = (devMode.dmPelsWidth - *winwidth) / 2;
	if ((*cleft < 0) || (*cleft > (int)(devMode.dmPelsWidth / 2)))
	{
		*cleft = 0;
	}

	*ctop = (devMode.dmPelsHeight - *winheight) / 2;
	if ((*ctop < 0) || (*ctop > (int)(devMode.dmPelsHeight / 2)))
	{
		*ctop = 0;
	}
}

static WIN32PROC win32_hook = 0;

void SysSetWin32Hook(WIN32PROC hook)
{
	win32_hook = hook;
}

static LRESULT CALLBACK WindowProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	if (win32_hook)
	{
		LRESULT	r = (LRESULT) win32_hook((int)hwnd, (UINT) msg, (WPARAM)wparam, (LPARAM)lparam);
		if (r) return r;
	}

	switch (msg)
	{
	case WM_PAINT:
		if(sys_handlers.top().display) {
			sys_handlers.top().display();
		}
		ValidateRect(hwnd, 0);
		break;

    case WM_KEYDOWN:
    case WM_KEYUP:
		if(sys_handlers.top().keyb) {
			sys_handlers.top().keyb(MapVirtualKey((unsigned int)wparam, 2), msg == WM_KEYDOWN);
        }
		break;

	case WM_LBUTTONDOWN:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_LEFT, true);
		}
		break;
		
	case WM_RBUTTONDOWN:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_RIGHT, true);
		}
		break;

	case WM_MBUTTONDOWN:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_MIDDLE, true);
		}
		break;
	
	case WM_LBUTTONUP:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_LEFT, false);
		}
		break;

	case WM_MOUSEWHEEL:
		if(sys_handlers.top().button) {
			int delta = GET_WHEEL_DELTA_WPARAM(wparam);
			if (delta > 0) { // up
				for (int i=0; i<delta; i+=WHEEL_DELTA)
				{
					sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_WHEELUP, true);
					sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_WHEELUP, false);
				}
			} else {
				for (int i=0; i>delta; i-=WHEEL_DELTA)
				{
					sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_WHEELDOWN, true);
					sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_WHEELDOWN, false);
				}
			}
		}
		break;

	case WM_RBUTTONUP:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_RIGHT, false);
		}
		break;

	case WM_MBUTTONUP:
		if(sys_handlers.top().button) {
			sys_handlers.top().button(LOWORD(lparam), HIWORD(lparam), BN_MIDDLE, false);
		}
		break;

	case WM_MOUSEMOVE:
		if(sys_handlers.top().motion) {
			sys_handlers.top().motion(LOWORD(lparam), HIWORD(lparam));
        }
		break;

	case WM_CLOSE:
		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hwnd, msg, wparam, lparam);
    }

    return 0;
}

#else
int libsys_w32_cpp_silence_stupid_compiler_warning_about_empty_files = 0;
#endif
