#include <builddefs.h>
#include "xpad.h"

#ifdef USE_XBOX_PAD
#include <vector>
#include <windows.h>
#include <XInput.h>

static unsigned int packet[4] = {0, 0, 0, 0};
static XPadState pads[4], old_pads[4];
static std::vector<XPadEvent> msg;

static float Analog(int value, unsigned int deadzone, int max_val)
{
	int sn = 1;
	if (value < 0) 
	{
		value = -value;
		sn = -1;
	}
	value -= deadzone;
	if (value < 0) value = 0;

	// normalize
	float v = (float)value / (float) (max_val - deadzone);
	if (v > 1.0f) v = 1.0f;

	return v * sn;
}

static XPadState XInputToState(const XINPUT_STATE &s)
{
	XPadState ret;
	
	// pad
	ret.keys[XPAD_UP] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_UP) ret.keys[XPAD_UP] = true;
	ret.keys[XPAD_DOWN] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_DOWN) ret.keys[XPAD_DOWN] = true;
	ret.keys[XPAD_LEFT] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_LEFT) ret.keys[XPAD_LEFT] = true;
	ret.keys[XPAD_RIGHT] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_RIGHT) ret.keys[XPAD_RIGHT] = true;

	// func
	ret.keys[XPAD_BACK] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_BACK) ret.keys[XPAD_BACK] = true;
	ret.keys[XPAD_START] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_START) ret.keys[XPAD_START] = true;

	// buttons
	ret.keys[XPAD_A] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_A) ret.keys[XPAD_A] = true;
	ret.keys[XPAD_B] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_B) ret.keys[XPAD_B] = true;
	ret.keys[XPAD_X] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_X) ret.keys[XPAD_X] = true;
	ret.keys[XPAD_Y] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_Y) ret.keys[XPAD_Y] = true;
	ret.keys[XPAD_L_THUMB] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_LEFT_THUMB) ret.keys[XPAD_L_THUMB] = true;
	ret.keys[XPAD_R_THUMB] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_THUMB) ret.keys[XPAD_R_THUMB] = true;
	ret.keys[XPAD_L_SHOULDER] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER) ret.keys[XPAD_L_SHOULDER] = true;
	ret.keys[XPAD_R_SHOULDER] = false;
	if (s.Gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER) ret.keys[XPAD_R_SHOULDER] = true;

	// analogs
	ret.analogs[XPAD_L_THUMB_X] = Analog(s.Gamepad.sThumbLX, XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE, 32767);
	ret.analogs[XPAD_L_THUMB_Y] = Analog(s.Gamepad.sThumbLY, XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE, 32767);
	ret.analogs[XPAD_R_THUMB_X] = Analog(s.Gamepad.sThumbRX, XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE, 32767);
	ret.analogs[XPAD_R_THUMB_Y] = Analog(s.Gamepad.sThumbRY, XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE, 32767);
	ret.analogs[XPAD_L_TRIGGER] = Analog(s.Gamepad.bLeftTrigger, XINPUT_GAMEPAD_TRIGGER_THRESHOLD, 255);
	ret.analogs[XPAD_R_TRIGGER] = Analog(s.Gamepad.bRightTrigger, XINPUT_GAMEPAD_TRIGGER_THRESHOLD, 255);

	return ret;
}

static void GenerateMessages(unsigned int which)
{
	if (!pads[which].connected) return;
	if (!old_pads[which].connected)
	{
		XPadEvent new_msg;
		new_msg.which = which;
		new_msg.type = XPAD_CONNECTED;
		msg.push_back(new_msg);
	}
	for (unsigned int i=0; i<14; i++)
	{
		if (pads[which].keys[i] == old_pads[which].keys[i]) continue;
		XPadEvent new_msg;
		new_msg.key = (XPadKeys)i;
		new_msg.which = which;
		if (pads[which].keys[i]) new_msg.type = XPAD_PRESSED;
		else new_msg.type = XPAD_RELEASED;
		msg.push_back(new_msg);
	}
	for (unsigned int i=0; i<6; i++)
	{
		if (pads[which].analogs[i] == old_pads[which].analogs[i]) continue;
		XPadEvent new_msg;
		new_msg.analog = (XPadAnalogs)i;
		new_msg.which = which;
		new_msg.type = XPAD_MOVED;
		new_msg.move = pads[which].analogs[i] - old_pads[which].analogs[i];
		msg.push_back(new_msg);
	}
}

void UpdateXPad()
{
	for (unsigned int i=0; i<4; i++)
	{
		XINPUT_STATE state;
		memset(&state, 0, sizeof(XINPUT_STATE));

		unsigned int res = XInputGetState(i, &state);		
		if(res == ERROR_SUCCESS)
		{
			old_pads[i] = pads[i];
			pads[i] = XInputToState(state);
			pads[i].connected = true;
			if (packet[i] != state.dwPacketNumber)
			{
				GenerateMessages(i);
			}
			packet[i] = state.dwPacketNumber;
		}
		else
		{
			if (pads[i].connected)
			{
				XPadEvent new_msg;
				new_msg.which = i;
				new_msg.type = XPAD_DISCONNECTED;
				msg.push_back(new_msg);
			}
			pads[i].connected = false;
		}

	}
}

bool GetXPadEvent(XPadEvent &message)
{
	if (!msg.size())
	{
		return false;
	}
	message = msg[0];
	msg.erase(msg.begin(), msg.begin() + 1);
	return true;
}

XPadState GetXPadState(unsigned int which)
{
	if (which >= 4)
	{
		XPadState s;
		s.connected = false;
		return s;
	}
	return pads[which];
}

void SetXPadSpeed(unsigned int which, float left, float right)
{
	if (which >= 4) return;
	int lval = (int)(left * 65535);
	if (lval > 65535) lval = 65535;
	if (lval < 0) lval = 0;
	int rval = (int)(right * 65535);
	if (rval > 65535) rval = 65535;
	if (rval < 0) rval = 0;

	XINPUT_VIBRATION vibration;
	memset(&vibration, 0, sizeof(XINPUT_VIBRATION));
	vibration.wLeftMotorSpeed = lval;
	vibration.wRightMotorSpeed = rval;
	XInputSetState(which, &vibration);
}
#else	// ------ no xpad support -------

#include <string.h>

void UpdateXPad() {}

bool GetXPadEvent(XPadEvent &message)
{
	return false;
}

XPadState GetXPadState(unsigned int which)
{
	XPadState st;
	memset(&st, 0, sizeof st);
	return st;
}

void SetXPadSpeed(unsigned int which, float left, float right) {}

#endif
