#include <libgc.h>
#include <libsys.h>
#include <math3d.h>
#include <errlog.h>
#include "utk.h"

/* TODO:
 * - Optimize drawing (qurashee idea).
 *   Make everything drawable by quads, so that we only need to do one
 *   ImBegin/ImEnd for the whole GUI (?). In order for this to work, make a
 *   draw queue and set an increasing Z-depth for each quad. Then batch up
 *   images/text (which need state switching) for the end and draw them with
 *   zbuffering.
 */

#define TEXT_PT_SIZE 18

static int fb_w, fb_h;

static utk::Widget *root;
static unsigned int font;
static float max_descent;

static SysKeyboardHandler orig_keyb;
static SysMotionHandler orig_motion;
static SysButtonHandler orig_button;

// state
static struct {
    float color_r, color_g, color_b, color_a;
	int clip_x1, clip_y1, clip_x2, clip_y2;
	int tex_w, tex_h;
	unsigned int tex_id;
	unsigned int *bitmap;
} ugfx;

// saved GC state to restore after drawing
static struct {
    bool fixed_function, depth_test, lighting, scissor, blending;
	CullFace cull_mode;
} gc_state;


static void save_gc_state();
static void restore_gc_state();

static void init_bitmap_tex();
static void utk_state_begin();
static void utk_state_end();

static void utk_color(int r, int g, int b, int a);
static void utk_clip(int x1, int y1, int x2, int y2);
static void utk_image(int x, int y, void *pix, int xsz, int ysz);
static void utk_rect(int x1, int y1, int x2, int y2);
static void utk_line(int x1, int y1, int x2, int y2, int width);
static int utk_text_spacing();
static void utk_text(int x, int y, const char *txt, int sz);
static int utk_text_width(const char *txt, int sz);

static void utk_keyb(int key, bool pressed);
static void utk_motion(int x, int y);
static void utk_button(int x, int y, int bn, bool pressed);

bool InitUTK()
{
	if(!(font = CreateFont(DEV_DATA_PATH "/fonts/georgia.ttf", TEXT_PT_SIZE))) {
		Error("InitUTK: failed to load font\n");
		return false;
	}
	BindFont(font);
	max_descent = GetMaxDescent();

	fb_w = GetInitParams().dmode.width;
	fb_h = GetInitParams().dmode.height;

	ugfx.clip_x1 = ugfx.clip_y1 = 0;
	ugfx.clip_x2 = fb_w;
	ugfx.clip_y2 = fb_h;

	ugfx.color_r = ugfx.color_g = ugfx.color_b = ugfx.color_a = 1.0;

	utk::gfx::color = utk_color;
	utk::gfx::clip = utk_clip;
	utk::gfx::image = utk_image;
	utk::gfx::rect = utk_rect;
	utk::gfx::line = utk_line;
	utk::gfx::text = utk_text;
	utk::gfx::text_spacing = utk_text_spacing;
	utk::gfx::text_width = utk_text_width;

	if(!(root = utk::init(fb_w, fb_h))) {
		Error("InitUTK: failed to initialize utk\n");
		return false;
	}
	return true;
}

void DestroyUTK()
{
	utk::close();
}

void DrawUTK()
{
	utk_state_begin();
	utk::draw();
	utk_state_end();
}

void SetUTKHandlers(bool pass_through)
{
	orig_keyb = pass_through ? SysGetKeyboardHandler() : 0;
	orig_motion = pass_through ? SysGetMotionHandler() : 0;
	orig_button = pass_through ? SysGetButtonHandler() : 0;

	SysSetKeyboardHandler(utk_keyb);
	SysSetMotionHandler(utk_motion);
	SysSetButtonHandler(utk_button);
}

static void save_gc_state()
{
	gc_state.depth_test = GetDepthTest();
	gc_state.cull_mode = GetCulling();
	gc_state.lighting = GetLighting();
	gc_state.fixed_function = ImFixedFunctionUsed();
	gc_state.blending = GetBlending();
	gc_state.scissor = GetScissorTest();
}

static void restore_gc_state()
{
	SetScissorTest(gc_state.scissor);
	SetBlending(gc_state.blending);
	ImFixedFunction(gc_state.fixed_function);
	SetDepthTest(gc_state.depth_test);
	SetCulling(gc_state.cull_mode);
	SetLighting(gc_state.lighting);
}

static void init_bitmap_tex()
{
	if(!ugfx.tex_w || !ugfx.tex_h) return;
	if(ugfx.tex_id) {
		DeleteTexture(ugfx.tex_id);
	}
	if(ugfx.bitmap) {
		delete [] ugfx.bitmap;
	}

	ugfx.tex_w = NextPow2(ugfx.tex_w);
	ugfx.tex_h = NextPow2(ugfx.tex_h);
	ugfx.tex_id = CreateTexture(ugfx.tex_w, ugfx.tex_h, TEX_2D, TFMT_32, false, 0);
	ugfx.bitmap = new unsigned int[ugfx.tex_w * ugfx.tex_h];
}

static void utk_state_begin()
{
	save_gc_state();

	SetDepthTest(false);
	SetCulling(CULL_NONE);
	SetLighting(false);
	ImFixedFunction(true);
	SetBlending(true);
	SetBlendFunc(BLEND_SRC_ALPHA, BLEND_ONE_MINUS_SRC_ALPHA);

	// setup orthographic projection
	ImMatrixMode(IM_PROJECTION);
	ImPushMatrix();
	ImLoadMatrix(OrthoProj(0, (float)fb_w, 0, (float)fb_h, 0, 10));
	ImMatrixMode(IM_VIEW);
	ImPushMatrix();
	ImLoadIdentity();
	ImMatrixMode(IM_MODEL);
	ImPushMatrix();
	ImLoadIdentity();

	SetScissorTest(true);
}

static void utk_state_end()
{
	ImMatrixMode(IM_PROJECTION);
	ImPopMatrix();
	ImMatrixMode(IM_VIEW);
	ImPopMatrix();
	ImMatrixMode(IM_MODEL);
	ImPopMatrix();

	restore_gc_state();
}

static void utk_color(int r, int g, int b, int a)
{
	ugfx.color_r = r / 255.0f;
	ugfx.color_g = g / 255.0f;
	ugfx.color_b = b / 255.0f;
	ugfx.color_a = a / 255.0f;
}

static void utk_clip(int x1, int y1, int x2, int y2)
{
	ugfx.clip_x1 = x1;
	ugfx.clip_y1 = y1;
	ugfx.clip_x2 = x2;
	ugfx.clip_y2 = y2;
	SetScissorRect(ugfx.clip_x1, ugfx.clip_y1, ugfx.clip_x2, ugfx.clip_y2);
}

static void utk_image(int x, int y, void *pix, int xsz, int ysz)
{
	if(!xsz || !ysz) return;

	bool create_new_tex = false;
	if(xsz >= ugfx.tex_w) {
		ugfx.tex_w = xsz;
		create_new_tex = true;
	}
	if(ysz >= ugfx.tex_h) {
		ugfx.tex_h = ysz;
		create_new_tex = true;
	}
	if(create_new_tex) init_bitmap_tex();
	
	// update texture image
	unsigned int *dest = ugfx.bitmap;
	unsigned int *src = (unsigned int*)pix;

	for(int i=0; i<xsz*ysz; i++) {
		*dest++ = *src++;
	}
	SetTexImage(ugfx.tex_id, ugfx.bitmap, ugfx.tex_w, ugfx.tex_h, TFMT_32);

	BindTexture(0, ugfx.tex_id);

	// draw quad
	ImBegin(IM_QUADS);
	{
		ImColor(1, 1, 1, 1);

		ImTexCoord(0, 0);
		ImVertex((float)x, (float)y, 1);
		
		ImTexCoord((float)xsz / (float)ugfx.tex_w, 0);
		ImVertex((float)(x + xsz), (float)y, 1);
		
		ImTexCoord((float)xsz / (float)ugfx.tex_w, (float)ysz / (float)ugfx.tex_h);
		ImVertex((float)(x + xsz), (float)(y + ysz), 1);
		
		ImTexCoord(0, (float)ysz / (float)ugfx.tex_h);
		ImVertex((float)x, (float)(y + ysz), 1);
	}
	ImEnd();

	BindTexture(0, 0);
}

static void utk_rect(int x1, int y1, int x2, int y2)
{
	ImBegin(IM_QUADS);
	ImColor(Color(ugfx.color_r, ugfx.color_g, ugfx.color_b, ugfx.color_a));
	ImVertex((float)x1, (float)y1, 1);
	ImVertex((float)x2, (float)y1, 1);
	ImVertex((float)x2, (float)y2, 1);
	ImVertex((float)x1, (float)y2, 1);
	ImEnd();
}

static void utk_line(int x1, int y1, int x2, int y2, int width)
{
	ImBegin(IM_LINES);
	ImColor(ugfx.color_r, ugfx.color_g, ugfx.color_b, ugfx.color_a);
	ImVertex((float)x1, (float)y1, 1.0f);
	ImVertex((float)x2, (float)y2, 1.0f);
	ImEnd();
}

static int utk_text_spacing()
{
	BindFont(font);
	return (int)(GetLineAdvance() * (float)fb_h);
}

static void utk_text(int x, int y, const char *txt, int sz)
{
	float fx = (float)x / (float)fb_w;
	float fy = (float)y / (float)fb_h;

	BindFont(font);
	SetTextColor(Color(ugfx.color_r, ugfx.color_g, ugfx.color_b, ugfx.color_a));
	SetTextPos(Vector2(fx, fy - max_descent));
	//SetTextPos(Vector2(0.25, 0.25));
	SetTextSize((float)sz / (float)TEXT_PT_SIZE);
	PrintString(txt);
}

static int utk_text_width(const char *txt, int sz)
{
	BindFont(font);

	float prev_sz = GetTextSize();
	SetTextSize((float)sz / (float)TEXT_PT_SIZE);

	int width = (int)(GetTextWidth(txt) * (float)fb_w);

	SetTextSize(prev_sz);
	return width;
}

// --- std utk event handlers ---

static void utk_keyb(int key, bool pressed)
{
	utk::KeyboardEvent e;
	e.key = key;
	e.pressed = pressed;
	event(&e);

	if(orig_keyb) {
		orig_keyb(key, pressed);
	}
}

static void utk_motion(int x, int y)
{
	utk::MMotionEvent e;
	e.x = x;
	e.y = y;
	event(&e);

	if(orig_motion) {
		orig_motion(x, y);
	}
}

static void utk_button(int x, int y, int bn, bool pressed)
{
	utk::MButtonEvent e;
	e.button = bn;
	e.pressed = pressed;
	e.x = x;
	e.y = y;
	event(&e);

	if(orig_button) {
		orig_button(x, y, bn, pressed);
	}
}

utk::Widget *GetUTKRoot()
{
	return root;
}
