#include <iostream>
#include <vector>
#include <stack>
#include <string>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <assert.h>
#include <errlog.h>
#include "libvfs.h"
#include "tar.h"

#if !defined(GLIBC) && !defined(__GLIBC__)
#include <malloc.h>

#ifdef _MSC_VER
#define alloca _alloca
#endif	// _MSC_VER

#else
#include <alloca.h>
#endif	// glibc

#if defined(BUILD_UNIX)
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#elif defined(BUILD_WIN32)
#include "w32_dirent.h"
#endif	// BUILD_WIN32

using namespace std;

enum node_type {
	NODE_FILE,
	NODE_DIR
};

enum file_type {
	VFILE_NONE,
	VFILE_REAL,
	VFILE_TAR
};

struct vfs_node {
	string name;
	node_type type;
	vector<vfs_node*> children;
	stack<vector<vfs_node*> > old_children;

	/* this is conceptually a union, but the members are full classes with
	 * constructors etc, so it can't be a union...
	 */
	struct {
		tar_file tar;		// if this node represents a file inside a tar archive
		string real_name;	// if this node represents a real file/dir on the filesystem
	} file;
	file_type ftype;
};

struct VFS_FILE {
	vfs_node *node;
	FILE *fp;
};

#define VSIZE(fp)	((fp)->node->file.tar.size)

struct VFS_DIR {
	vfs_node *node;
	int iter;

	struct vfs_dirent dent;
};

/* use C linkage */
extern "C" {

static bool init();
static void clean_up();
static void clean_up_rec(vfs_node *node);
static int build_dir_tree(vfs_node *mnode, const char *path);
static int mount_tar(vfs_node *mnode, const char *tarfile);
static int mount_dir(vfs_node *mnode, const char *dirname);
static vfs_node *openpath_rec(const char *path, vfs_node *node);


// pointers to the root and current directory nodes.
static vfs_node *root, *cwd;

static int stricmp(const char *s1, const char *s2)
{
	char a, b;

	while(*s1 && *s2) {
		a = tolower(*s1++);
		b = tolower(*s2++);

		if(a != b) {
			return a - b;
		}
	}

	if(*s1 == *s2) {
		return 0;
	}
	return *s1 ? 1 : -1;
}

static void slashify(char *a)
{
	while (*a)
	{
		if (*a == '\\') *a = '/';
		a++;
	}
}

static bool init()
{
	atexit(clean_up);
	root = new vfs_node;
	root->name = "/";
	root->type = NODE_DIR;
	cwd = root;
	return true;
}

static void clean_up()
{
	clean_up_rec(root);
}

static void clean_up_rec(vfs_node *node)
{
	if(!node) return;
	for(size_t i=0; i<node->children.size(); i++) {
		clean_up_rec(node->children[i]);
	}
	delete node;
}

static int build_dir_tree(vfs_node *mnode, const char *path)
{
	if(!path || !*path) {
		return -1;
	}

	char *pbuf = (char*)alloca(strlen(path) + 1);
	strcpy(pbuf, path);

	char *next_slash = strchr(pbuf, '/');
	if(next_slash) {
		*next_slash = 0;
	}

	/* see if the directory exists, if not create it */
	for(size_t i=0; i<mnode->children.size(); i++) {
		if(stricmp(pbuf, mnode->children[i]->name.c_str()) == 0) {
			if(mnode->children[i]->type != NODE_DIR) {
				return -1;
			}
			if(next_slash) {
				return build_dir_tree(mnode->children[i], next_slash + 1);
			}
			return 0;
		}
	}

	// does not exist, create it...
	vfs_node *tmp = new vfs_node;
	tmp->name = string(pbuf);
	tmp->type = NODE_DIR;
	tmp->ftype = VFILE_NONE;
	mnode->children.push_back(tmp);
	
	if(next_slash) {
		return build_dir_tree(tmp, next_slash + 1);
	}
	return 0;
}

static int mount_tar(vfs_node *mnode, const char *tarfile)
{
	if(!root) init();

	if(!tarfile) {
		errno = EFAULT;
		return -1;
	}
	if(!*tarfile) {
		errno = ENOENT;
		return -1;
	}

	vector<tar_file> tvec = get_tar_files(tarfile);
	if(tvec.empty()) {
		errno = ENOENT;
		return -1;
	}

	mnode->old_children.push(mnode->children);
	mnode->children.clear();

	for(size_t i=0; i<tvec.size(); i++) {
		if(tvec[i].type == TFT_DIR) {
			build_dir_tree(mnode, tvec[i].filename.c_str());
		}
	}

	for(size_t i=0; i<tvec.size(); i++) {
		if(tvec[i].type == TFT_FILE) {
			vfs_node *parent = mnode;

			char *dirname = (char*)alloca(tvec[i].filename.length() + 1);
			strcpy(dirname, tvec[i].filename.c_str());

			char *last_slash = strrchr(dirname, '/');
			if(last_slash) {
				*last_slash = 0;
				
				if(!(parent = openpath_rec(dirname, mnode))) {
					errno = ENOENT;
					return -1;
				}
			}

			vfs_node *fnode = new vfs_node;
			fnode->name = last_slash ? string(last_slash + 1) : tvec[i].filename;
			fnode->type = NODE_FILE;
			fnode->ftype = VFILE_TAR;
			fnode->file.tar = tvec[i];
			parent->children.push_back(fnode);
		}
	}

	return 0;
}

static int mount_dir(vfs_node *mnode, const char *dirname)
{
	if(!dirname) {
		errno = EFAULT;
		return -1;
	}
	if(!*dirname) {
		errno = ENOENT;
		return -1;
	}

	DIR *dir = opendir(dirname);
	if(!dir) {
		return -1;
	}

	mnode->old_children.push(mnode->children);
	mnode->children.clear();

	dirent *entry;
	while((entry = readdir(dir))) {
		if(strcmp(entry->d_name, ".") == 0 || strcmp(entry->d_name, "..") == 0) {
			continue;
		}
		vfs_node *node = new vfs_node;

		node->name = string(entry->d_name);
		node->ftype = VFILE_REAL;
		
		if(mnode->ftype == VFILE_REAL) {
			node->file.real_name = mnode->file.real_name + string("/") + string(entry->d_name);
		} else {
			/* if the mount node is not a real-file node, then it's the original mount point (first recursive call)
			 * so we'll have to include the real-dir path as a prefix to the child's real-name.
			 */
			int len = (int)strlen(dirname);
			char *dirname_mod = new char[len + 1];
			strcpy(dirname_mod, dirname);
			if(dirname_mod[len - 1] == '/') {
				dirname_mod[len - 1] = 0;
			}
			node->file.real_name = string(dirname_mod) + string("/") + string(entry->d_name);
			delete [] dirname_mod;
		}

		DIR *test = opendir(node->file.real_name.c_str());	// (it was: entry->d_name)
		if(test) {	/* it's a dir ... */
			node->type = NODE_DIR;
			closedir(test);
			//mount_dir(node, entry->d_name);
			mount_dir(node, node->file.real_name.c_str());
		} else {
			node->type = NODE_FILE;
		}

		mnode->children.push_back(node);
	}
	closedir(dir);

	return 0;
}

int vfs_mount(const char *fs, const char *mntpt)
{
	VFS_DIR *mdir;
	if(!(mdir = vfs_opendir(mntpt))) {
		return -1;
	}

	/* check if fs is a tar file, and mount it */
	if(mount_tar(mdir->node, fs) == 0) {
		vfs_closedir(mdir);
		return 0;
	}

	/* check if fs is a system directory and create the whole virtual tree from that point on */
	if(mount_dir(mdir->node, fs) == 0) {
		vfs_closedir(mdir);
		return 0;
	}

	vfs_closedir(mdir);
    return -1;
}

int vfs_umount(const char *mntpt)
{
	VFS_DIR *mdir;
	if(!(mdir = vfs_opendir(mntpt))) {
		return -1;
	}

	if(mdir->node->old_children.empty()) {
		errno = EINVAL;
		vfs_closedir(mdir);
		return -1;
	}

	mdir->node->children = mdir->node->old_children.top();
	mdir->node->old_children.pop();

	vfs_closedir(mdir);
	return 0;
}

static void print_tree_rec(FILE *fp, vfs_node *node, int lvl)
{
	for(int i=0; i<lvl; i++) {
		fputc('\t', fp);
	}
	fprintf(fp, "%s [%s]\n", node->name.c_str(), node->type == NODE_DIR ? "dir" : (node->ftype == VFILE_REAL ? "file (real)" : "file (tar)"));

	for(size_t i=0; i<node->children.size(); i++) {
		print_tree_rec(fp, node->children[i], lvl + 1);
	}
}

void vfs_print_tree(FILE *outfile)
{
	if(!root) init();
	print_tree_rec(outfile, root, 0);
}

/* openpath_rec recursively follows the virtual directory tree following the path given in the first argument.
 * NOTE: if the path starts with /, it starts looking from the root instead of the node passed.
 */
static vfs_node *openpath_rec(const char *path, vfs_node *node)
{
	char	*cpath = (char*)alloca(strlen(path) + 1);
	strcpy(cpath, path);
	if(!root) init();
	slashify(cpath);
	path = cpath;

	if(path[0] == '/') {	// if it starts with a /, start from the root
		node = root;
		path++;
	}

	if(!*path) {		// we've reached the end of the path string
		return node;
	}

	// create a scratch path buffer on the stack
	char *pathbuf = (char*)alloca(strlen(path) + 1);
	strcpy(pathbuf, path);

	// find the next slash and put a null terminator there
	char *next_slash = strchr(pathbuf, '/');
	if(next_slash) {
		*next_slash = 0;
	}

	if(!*pathbuf) {	// empty path component
		return next_slash ? openpath_rec(next_slash + 1, node) : node;
	}

	for(size_t i=0; i<node->children.size(); i++) {
		if(!stricmp(node->children[i]->name.c_str(), pathbuf)) {
			return next_slash ? openpath_rec(next_slash + 1, node->children[i]) : node->children[i];
		}
	}
	return 0;
}


VFS_DIR *vfs_opendir(const char *name)
{
	if(!root) init();

	vfs_node *node = openpath_rec(name, cwd);
	if(!node) {		// not found
		errno = ENOENT;
		return 0;
	}

	if(node->type != NODE_DIR) {	// not a directory
		errno = ENOTDIR;
		return 0;
	}

	VFS_DIR *vdir = new VFS_DIR;
	vdir->node = node;
	vdir->iter = 0;
	return vdir;
}

int vfs_closedir(VFS_DIR *dir)
{
	delete dir;
	return 0;
}

struct vfs_dirent *vfs_readdir(VFS_DIR *dir)
{
	if(dir->iter < (int)dir->node->children.size()) {
		strcpy(dir->dent.d_name, dir->node->children[dir->iter++]->name.c_str());
		return &dir->dent;
	}
	return 0;
}

void vfs_rewinddir(VFS_DIR *dir)
{
	dir->iter = 0;
}


int vfs_chdir(const char *name)
{
	VFS_DIR *dir = vfs_opendir(name);
	if(!dir) return -1;

	cwd = dir->node;
	vfs_closedir(dir);
	return 0;
}

int vfs_mkdir(const char *pathname, unsigned int mode)
{
	if(!root) init();

	if(!pathname) {
		errno = EFAULT;
		return -1;
	}

	if(openpath_rec(pathname, cwd)) {
		errno = EEXIST;
		return -1;
	}
	
	char *pbuf = (char*)alloca(strlen(pathname) + 1);
	if(!pbuf) {
		errno = ENOMEM;
		return -1;
	}
	strcpy(pbuf, pathname);

	char *last_slash = strrchr(pbuf, '/');
	if(last_slash) {
		*last_slash = 0;
	}

	vfs_node *parent = cwd;
	if(last_slash) {
		if(!(parent = openpath_rec(pbuf, cwd))) {
			errno = ENOENT;
			return -1;
		}
	}

	vfs_node *node = new vfs_node;
	node->name = string(last_slash ? last_slash + 1 : pathname);
	node->type = NODE_DIR;
	node->ftype = VFILE_NONE;

	parent->children.push_back(node);
	return 0;
}

struct VFS_FILE *vfs_open(const char *fname, const char *mode)
{
	if(!root) init();

	// first try to open it as a regular file and create a dummy vnode
	FILE *fp;
	if((fp = fopen(fname, mode))) {
		vfs_node *node = new vfs_node;
		node->name = string("DUMMY_NODE");
		node->type = NODE_FILE;
		node->file.real_name = string(fname);
		node->ftype = VFILE_REAL;

		VFS_FILE *vfile = new VFS_FILE;
		vfile->node = node;
		vfile->fp = fp;
		return vfile;
	}

#if defined(WIN32) || defined(__WIN32__)
	/* if this is an absolute windows path, it was certainly meant as a regular
	 * file, fopen failed, so it doesn't exist.
	 */
	if(isalpha(fname[0]) && fname[1] == ':') {
		errno = ENOENT;
		return 0;
	}
#endif
	
	vfs_node *node = openpath_rec(fname, cwd);
	if(!node) {		// not found
		// TODO
		errno = ENOENT;
		return 0;
	}

	if(node->type != NODE_FILE) {	// not a file
		errno = EISDIR;
		return 0;
	}

	VFS_FILE *vfile = new VFS_FILE;
	memset(vfile, 0, sizeof *vfile);
	vfile->node = node;

	if(vfile->node->ftype == VFILE_REAL) {
		if(!(vfile->fp = fopen(vfile->node->file.real_name.c_str(), mode))) {
			delete vfile;
			return 0;
		}
		return vfile;
	} else if(vfile->node->ftype == VFILE_TAR) {
		if(strchr(mode, 'w')) {	// cannot allow writing in packed files
			errno = EPERM;
			delete vfile;
			return 0;
		}

		if(!(vfile->fp = fopen(vfile->node->file.tar.tar_fname.c_str(), mode))) {
			delete vfile;
			return 0;
		}

		fseek(vfile->fp, vfile->node->file.tar.offset, SEEK_SET);	// seek to the begining of the virtual file
		return vfile;
	}

	delete vfile;
	return 0;
}

int vfs_close(VFS_FILE *fp)
{
	if(fp->fp) {
		fclose(fp->fp);
	}

	if(fp->node->name == string("DUMMY_NODE")) {
		delete fp->node;
	}	
	delete fp;	
	return 0;
}

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))

size_t vfs_read(void *buf, size_t sz, size_t nmemb, VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fread(buf, sz, nmemb, fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		tar_file *vfile = &fp->node->file.tar;

		size_t rdsz = sz * nmemb;
		size_t bytes_left = MAX((long)vfile->size - vfs_tell(fp), 0);
        if(!(rdsz = MIN(rdsz, bytes_left))) {
			return 0;
		}
		size_t virt_nmemb = rdsz / sz;	// find the number of elements that can be read untill the virtual eof
		virt_nmemb = fread(buf, sz, virt_nmemb, fp->fp);
		return virt_nmemb;
	}
	return 0;
}

size_t vfs_write(void *buf, size_t sz, size_t nmemb, VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fwrite(buf, sz, nmemb, fp->fp);
	}
	return 0;	// can only write in real files
}

int vfs_seek(VFS_FILE *fp, long offs, int whence)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fseek(fp->fp, offs, whence);
	}
	if(fp->node->ftype == VFILE_TAR) {
		tar_file *vfile = &fp->node->file.tar;
		switch(whence) {
		case VFS_SEEK_SET:
			return fseek(fp->fp, offs + vfile->offset, SEEK_SET);

		case VFS_SEEK_CUR:
			return fseek(fp->fp, offs, SEEK_CUR);

		case VFS_SEEK_END:
			return fseek(fp->fp, vfile->offset + vfile->size + offs, SEEK_SET);

		default:
			return -1;
		}
	}

	return -1;
}

long vfs_tell(VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return ftell(fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		return ftell(fp->fp) - fp->node->file.tar.offset;
	}
	return -1;
}

int vfs_printf(VFS_FILE *fp, const char *fmt, ...)
{
	if(fp->node->ftype == VFILE_REAL) {
        va_list arg_list;
		va_start(arg_list, fmt);
		int res = vfprintf(fp->fp, fmt, arg_list);
		va_end(arg_list);
		return res;
	}
	return -1;
}

int vfs_putc(int c, VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fputc(c, fp->fp);
	}
	return -1;
}

int vfs_puts(const char *s, VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fputs(s, fp->fp);
	}
	return -1;
}

int vfs_getc(VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fgetc(fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		long dif = (long)VSIZE(fp) - vfs_tell(fp);
		if(dif <= 0) {
			return -1;
		}
		return fgetc(fp->fp);
	}
	return -1;
}

int vfs_ungetc(int c, VFS_FILE *fp)
{
	return ungetc(c, fp->fp);	// do I have to handle this specially for virtual files? I think not...
}

char *vfs_gets(char *str, int size, VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fgets(str, size, fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		int c;
		char *buf = str;
		while(--size > 0 && (c = vfs_getc(fp)) >= 0) {
			*buf++ = c;
			if(c == '\n') break;
		}
		*buf = 0;
		return str;
	}

	return 0;	// unreachable unless ftype is invalid
}

int vfs_eof(VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return feof(fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		return vfs_tell(fp) >= (long)VSIZE(fp);
	}
	return -1;
}


int vfs_flush(VFS_FILE *fp)
{
	if(fp->node->ftype == VFILE_REAL) {
		return fflush(fp->fp);
	}
	if(fp->node->ftype == VFILE_TAR) {
		return 0;
	}
	return -1;
}

int vfs_stat(const char *fname, struct vfs_stat *buf)
{
	VFS_FILE *fp;

	if(!fname || !buf) {
		errno = EFAULT;
		return -1;
	}

	if(!(fp = vfs_open(fname, "r"))) {
		return -1;	/* errno is set by vfs_open */
	}

	memset(buf, 0, sizeof *buf);

	assert(fp->node->ftype == VFILE_REAL || fp->node->ftype == VFILE_TAR);
	if(fp->node->ftype == VFILE_REAL) {
#if defined(unix) || defined(__unix__)
		struct stat st;
		if(fstat(fileno(fp->fp), &st) == -1) {
			return -1;
		}
		buf->st_dev = st.st_dev;
		buf->st_ino = st.st_ino;
		buf->st_mode = st.st_mode;
		buf->st_nlink = st.st_nlink;
		buf->st_uid = st.st_uid;
		buf->st_gid = st.st_gid;
		buf->st_rdev = st.st_rdev;
		buf->st_size = st.st_size;
		buf->st_blksize = st.st_blksize;
		buf->st_blocks = st.st_blocks;
		/*buf->st_atime = st.st_atime;
		buf->st_mtime = st.st_mtime;
		buf->st_ctime = st.st_ctime;*/
#elif defined(WIN32) || defined(__WIN32__)
		buf->st_size = filelength(fileno(fp->fp));
		// TODO: add any other applicable fields at some point too
#else
#error "unsupported platform or OS detection failed"
#endif
	} else {	/* virtual file in a tar archive */
		buf->st_size = fp->node->file.tar.size;
		// TODO: add any other applicable fields at some point too
	}

	vfs_close(fp);
	return 0;
}

long vfs_filesize(VFS_FILE *fp)
{
	if (fp->fp)
	{
		size_t	curr = ftell(fp->fp);
		long	size;
		fseek(fp->fp, 0, SEEK_END);
		size = (long)ftell(fp->fp);
		fseek(fp->fp, curr, SEEK_SET);
		return size;
	}
	return VSIZE(fp);
}

FILE *vfs_get_real_file(VFS_FILE *fp)
{
	if(fp->node->type == NODE_FILE && fp->node->ftype == VFILE_REAL) {
		return fp->fp;
	}
	return 0;
}


// ----- reading and writing for various byte orders -----

int8_t vfs_read_int8(VFS_FILE *fp)
{
	int8_t v;
	vfs_read(&v, 1, 1, fp);
	return v;
}

int16_t vfs_read_int16(VFS_FILE *fp)
{
	int16_t v;
	vfs_read(&v, 2, 1, fp);
	return v;
}

int16_t vfs_read_int16_inv(VFS_FILE *fp)
{
	int16_t v;
	vfs_read(&v, 2, 1, fp);
	return v >> 8 | v << 8;
}

int32_t vfs_read_int32(VFS_FILE *fp)
{
	int32_t v;
	vfs_read(&v, 4, 1, fp);
	return v;
}

int32_t vfs_read_int32_inv(VFS_FILE *fp)
{
	int32_t v;
	vfs_read(&v, 4, 1, fp);
	return v >> 24 | (v & 0x00ff0000) >> 8 | (v & 0x0000ff00) << 8 | v << 24;
}

float vfs_read_float(VFS_FILE *fp)
{
	int32_t tmp = vfs_read_int32(fp);
	return *((float*)&tmp);
}

float vfs_read_float_inv(VFS_FILE *fp)
{
	int32_t tmp = vfs_read_int32_inv(fp);
	return *((float*)&tmp);
}

void vfs_write_int8(VFS_FILE *fp, int8_t v)
{
	vfs_write(&v, 1, 1, fp);
}

void vfs_write_int16(VFS_FILE *fp, int16_t v)
{
	vfs_write(&v, 2, 1, fp);
}

void vfs_write_int16_inv(VFS_FILE *fp, int16_t v)
{
	int16_t tmp = v >> 8 | v << 8;
	vfs_write(&tmp, 2, 1, fp);
}

void vfs_write_int32(VFS_FILE *fp, int32_t v)
{
	vfs_write(&v, 4, 1, fp);
}

void vfs_write_int32_inv(VFS_FILE *fp, int32_t v)
{
	int32_t tmp = v >> 24 | (v & 0x00ff0000) >> 8 | (v & 0x0000ff00) << 8 | v << 24;
	vfs_write(&tmp, 4, 1, fp);
}

void vfs_write_float(VFS_FILE *fp, float v)
{
	vfs_write_int32(fp, *((int32_t*)&v));
}

void vfs_write_float_inv(VFS_FILE *fp, float v)
{
	vfs_write_int32_inv(fp, *((int32_t*)&v));
}


/* extern "C" ends */
}
