#ifndef LIB_VFS_H
#define LIB_VFS_H

#include <stdio.h>
#include <time.h>
#include <builddefs.h>
#include <byteorder.h>
#include <int_types.h>

#define VFS_SEEK_SET  SEEK_SET
#define VFS_SEEK_CUR  SEEK_CUR
#define VFS_SEEK_END  SEEK_END 

typedef struct VFS_FILE VFS_FILE;
typedef struct VFS_DIR VFS_DIR;

struct vfs_dirent {
	char d_name[256];
};

struct vfs_stat {
	int st_dev;					/* device */
	unsigned long st_ino;		/* inode */
	unsigned int st_mode;		/* protection */
	unsigned short st_nlink;	/* number of hard links */
	long st_uid;				/* user ID of owner */
	long st_gid;				/* group ID of owner */
	int st_rdev;				/* device type (if inode device) */
	long st_size;				/* total size, in bytes */
	long st_blksize;			/* blocksize for filesystem I/O */
	long st_blocks;				/* number of blocks allocated */
	time_t st_atime;			/* time of last access */
	time_t st_mtime;			/* time of last modification */
	time_t st_ctime;			/* time of last status change */
};

#ifdef __cplusplus
extern "C" {
#endif

/* mount a real dir or a tar archive to a mount point of the VFS */
int vfs_mount(const char *fs, const char *mntpt);
int vfs_umount(const char *mntpt);

/* print the virtual tree to the specified stream */
void vfs_print_tree(FILE *outfile);

/* UNIX-like directory handling & traversal */
VFS_DIR *vfs_opendir(const char *name);
int vfs_closedir(VFS_DIR *dir);
struct vfs_dirent *vfs_readdir(VFS_DIR *dir);
void vfs_rewinddir(VFS_DIR *dir);
int vfs_chdir(const char *name);
int vfs_mkdir(const char *pathname, unsigned int mode);

/* virtual counterparts of the libc I/O functions */
VFS_FILE *vfs_open(const char *fname, const char *mode);
int vfs_close(VFS_FILE *fp);

size_t vfs_read(void *buf, size_t sz, size_t nmemb, VFS_FILE *fp);
size_t vfs_write(void *buf, size_t sz, size_t nmemb, VFS_FILE *fp);

int vfs_seek(VFS_FILE *fp, long offs, int whence);
long vfs_tell(VFS_FILE *fp);

int vfs_printf(VFS_FILE *fp, const char *fmt, ...);

int vfs_putc(int c, VFS_FILE *fp);
int vfs_puts(const char *s, VFS_FILE *fp);
int vfs_getc(VFS_FILE *fp);
char *vfs_gets(char *str, int size, VFS_FILE *fp);
int vfs_ungetc(int c, VFS_FILE *fp);

int vfs_eof(VFS_FILE *fp);

int vfs_flush(VFS_FILE *fp);

int vfs_stat(const char *fname, struct vfs_stat *buf);
long vfs_filesize(VFS_FILE *fp);

/* for sanity checks, don't rely on this */
FILE *vfs_get_real_file(VFS_FILE *fp);

/* use these to be able to load/save in a specific endianess,
 * either named (le/be), system endianess, or inverse of the system endianess.
 */


#ifdef LITTLE_ENDIAN
/* little endian */
#define vfs_read_int8_le(f)			vfs_read_int8(f)
#define vfs_read_int8_be(f)			vfs_read_int8(f)
#define vfs_read_int16_le(f)		vfs_read_int16(f)
#define vfs_read_int16_be(f)		vfs_read_int16_inv(f)
#define vfs_read_int32_le(f)		vfs_read_int32(f)
#define vfs_read_int32_be(f)		vfs_read_int32_inv(f)
#define vfs_read_float_le(f)		vfs_read_float(f)
#define vfs_read_float_be(f)		vfs_read_float_inv(f)
	
#define vfs_write_int8_le(f, v)		vfs_write_int8(f, v)
#define vfs_write_int8_be(f, v)		vfs_write_int8(f, v)
#define vfs_write_int16_le(f, v)	vfs_write_int16(f, v)
#define vfs_write_int16_be(f, v)	vfs_write_int16_inv(f, v)
#define vfs_write_int32_le(f, v)	vfs_write_int32(f, v)
#define vfs_write_int32_be(f, v)	vfs_write_int32_inv(f, v)
#define vfs_write_float_le(f, v)	vfs_write_float(f, v)
#define vfs_write_float_be(f, v)	vfs_write_float_inv(f, v)

#else

/* big endian */
#define vfs_read_int8_be(f)			vfs_read_int8(f)
#define vfs_read_int8_le(f)			vfs_read_int8(f)
#define vfs_read_int16_be(f)		vfs_read_int16(f)
#define vfs_read_int16_le(f)		vfs_read_int16_inv(f)
#define vfs_read_int32_be(f)		vfs_read_int32(f)
#define vfs_read_int32_le(f)		vfs_read_int32_inv(f)
#define vfs_read_float_be(f)		vfs_read_float(f)
#define vfs_read_float_le(f)		vfs_read_float_inv(f)
	
#define vfs_write_int8_be(f, v)		vfs_write_int8(f, v)
#define vfs_write_int8_le(f, v)		vfs_write_int8(f, v)
#define vfs_write_int16_be(f, v)	vfs_write_int16(f, v)
#define vfs_write_int16_le(f, v)	vfs_write_int16_inv(f, v)
#define vfs_write_int32_be(f, v)	vfs_write_int32(f, v)
#define vfs_write_int32_le(f, v)	vfs_write_int32_inv(f, v)
#define vfs_write_float_be(f, v)	vfs_write_float(f, v)
#define vfs_write_float_le(f, v)	vfs_write_float_inv(f, v)

#endif	/* LITTLE_ENDIAN */


int8_t vfs_read_int8(VFS_FILE *fp);
int16_t vfs_read_int16(VFS_FILE *fp);
int16_t vfs_read_int16_inv(VFS_FILE *fp);
int32_t vfs_read_int32(VFS_FILE *fp);
int32_t vfs_read_int32_inv(VFS_FILE *fp);
float vfs_read_float(VFS_FILE *fp);
float vfs_read_float_inv(VFS_FILE *fp);

void vfs_write_int8(VFS_FILE *fp, int8_t v);
void vfs_write_int16(VFS_FILE *fp, int16_t v);
void vfs_write_int16_inv(VFS_FILE *fp, int16_t v);
void vfs_write_int32(VFS_FILE *fp, int32_t v);
void vfs_write_int32_inv(VFS_FILE *fp, int32_t v);
void vfs_write_float(VFS_FILE *fp, float v);
void vfs_write_float_inv(VFS_FILE *fp, float v);

#ifdef __cplusplus
}
#endif

#endif	/* LIB_VFS_H */
