// octree.cpp

#include "octree.h"
#include <progress.h>

static unsigned int cull_count = 0;
unsigned int GetCullCount()
{
	return cull_count;
}

void ResetCullCount()
{
	cull_count = 0;
}


// AlignedBox class implementation
// ===============================

void AlignedBox::Update()
{
	box = Box(from, to);
}

AlignedBox::AlignedBox()
{
	from = to = Vector3(0, 0, 0);
}

AlignedBox::AlignedBox(Vector3 from, Vector3 to)
{
	this->from = from;
	this->to = to;
	Update();
}

Vector3 AlignedBox::GetFrom() const
{
	return from;
}

Vector3 AlignedBox::GetTo() const
{
	return to;
}


AlignedBox AlignedBox::GetSubBox(OctreeBox sb) const
{
	Vector3 middle = from + 0.5f * (to - from);
	switch (sb)
	{
		case OCTREE_FRONT_TOP_LEFT:
			return AlignedBox(
				Vector3(from.x,		middle.y,	from.z),
				Vector3(middle.x,	to.y,		middle.z));
		case OCTREE_FRONT_TOP_RIGHT:
			return AlignedBox(
				Vector3(middle.x,	middle.y,	from.z),
				Vector3(to.x,		to.y,		middle.z));
		case OCTREE_FRONT_BOTTOM_LEFT:
			return AlignedBox(
				Vector3(from.x,		from.y,		from.z),
				Vector3(middle.x,	middle.y,	middle.z));
		case OCTREE_FRONT_BOTTOM_RIGHT:
			return AlignedBox(
				Vector3(middle.x,	from.y,		from.z),
				Vector3(to.x,		middle.y,	middle.z));

		case OCTREE_BACK_TOP_LEFT:
			return AlignedBox(
				Vector3(from.x,		middle.y,	middle.z),
				Vector3(middle.x,	to.y,		to.z));
		case OCTREE_BACK_TOP_RIGHT:
			return AlignedBox(
				Vector3(middle.x,	middle.y,	middle.z),
				Vector3(to.x,		to.y,		to.z));
		case OCTREE_BACK_BOTTOM_LEFT:
			return AlignedBox(
				Vector3(from.x,		from.y,		middle.z),
				Vector3(middle.x,	middle.y,	to.z));
		case OCTREE_BACK_BOTTOM_RIGHT:
			return AlignedBox(
				Vector3(middle.x,	from.y,		middle.z),
				Vector3(to.x,		middle.y,	to.z));

		default:
			return AlignedBox();
	}
	return AlignedBox();
}


// OctreeNode class implementation
// ===============================

void OctreeNode::Init()
{
	aabb = AlignedBox();
	depth = max_depth = max_cullables = 16;
	parent = 0;
	for (unsigned int i=0; i<8; i++)
	{
		children[i] = 0;
	}
}

OctreeNode::OctreeNode()
{
	Init();
}

OctreeNode::OctreeNode(const AlignedBox &aabb, unsigned int max_depth,
		unsigned int max_cullables, unsigned int depth)
{
	OctreeNode();
	this->aabb = aabb;
	this->max_depth = max_depth;
	this->max_cullables = max_cullables;
	this->depth = depth;
}

OctreeNode::~OctreeNode()
{
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
			delete children[i];
	}
}

void OctreeNode::AddCullable(Cullable *c)
{
	if (!c) return;
	// avoid duplicates
	RemoveCullable(c);
	cullables.push_back(c);
}

void OctreeNode::RemoveCullable(Cullable *c)
{
	for (unsigned int i=0; i<cullables.size(); i++)
	{
		if (cullables[i] == c)
		{
			cullables.erase(cullables.begin() + i, 
				cullables.begin() + i + 1);
			return;
		}
	}
}


void OctreeNode::FitContained()
{
	Vector3 minv, maxv;
	if (!cullables.size())
	{
		aabb = AlignedBox();
		return;
	}

	minv = maxv = cullables[0]->GetSpacetimeBoundingBox().GetVertex(0);

	for (unsigned int i=0; i<cullables.size(); i++)
	{
		Box b = cullables[i]->GetSpacetimeBoundingBox();
		for (unsigned int j=0; j<8; j++)
		{
			Vector3 v = b.GetVertex(j);

			if (v.x < minv.x) minv.x = v.x;
			if (v.y < minv.y) minv.y = v.y;
			if (v.z < minv.z) minv.z = v.z;

			if (v.x > maxv.x) maxv.x = v.x;
			if (v.y > maxv.y) maxv.y = v.y;
			if (v.z > maxv.z) maxv.z = v.z;
		}
	}

	aabb = AlignedBox(minv, maxv);
}

void OctreeNode::Subdivide()
{
	// TODO: fix this
	// for some weird reason, the constructor won't be called,
	// so we have to zero all children here
	for (unsigned int i=0; i<8; i++)
	{
		children[i] = 0;
	}

	if (cullables.size() <= max_cullables) return;
	if (depth >= max_depth) return;

	std::vector<Cullable*> culls_to_remove;
	std::vector<int> destinations;

	// prepare child boxes
	AlignedBox ch_box[8];
	bool child_exists[8];
	for (unsigned int i=0; i<8; i++)
	{
		ch_box[i] = aabb.GetSubBox((OctreeBox)i);
		child_exists[i] = false;
	}

	for (unsigned int i=0; i<cullables.size(); i++)
	{
		int dst = -1;
		Box bb = cullables[i]->GetSpacetimeBoundingBox();
		for (unsigned int j=0; j<8; j++)
		{
			if (Intersect(ch_box[j].GetBox(), bb))
			{
				if (dst != -1)
				{
					// this cullable intersects more than one children nodes
					// so it remains here
					dst = -1;
					break;
				}

				dst = j;
			}
		} // end for (all children nodes)

		if (dst != -1)
		{
			culls_to_remove.push_back(cullables[i]);
			destinations.push_back(dst);
			child_exists[dst] = true;
		}

	} // end for (all cullables)

	// initialize existing children
	for (unsigned int i=0; i<8; i++)
	{
		if (child_exists[i])
		{
			OctreeNode *new_node = new OctreeNode(ch_box[i], 
				max_depth, max_cullables, depth+1);
			new_node->parent = this;
			children[i] = new_node;
		}
		else
		{
			children[i] = 0;
		}
	}

	// move cullables from parent to children
	for (unsigned int i=0; i<culls_to_remove.size(); i++)
	{
		children[destinations[i]]->AddCullable(culls_to_remove[i]);
		RemoveCullable(culls_to_remove[i]);
	}

	// recursion
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
		{
			children[i]->Subdivide();
		}
	}

	UpdateProgressBar((float) depth / (float) max_depth);
}

// Queries
CullableList OctreeNode::Query(const Sphere &s)
{
	CullableList ret;
	if (!Intersect(s, aabb.GetBox())) return ret;

	// intersection. return all cullables
	for (unsigned int i=0; i<cullables.size(); i++)
	{
		if (Intersect(s, cullables[i]->GetSpacetimeBoundingBox()))
			ret.push_back(cullables[i]);
	}

	// recursion
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
		{
			ret = ConcatCullLists(ret, children[i]->Query(s));
		}
	}

	return ret;
}

inline bool BoxBox(const Box &b1, const Box &b2)
{
	cull_count ++;
	return Intersect(b1, b2);
}

CullableList OctreeNode::Query(const Box &b)
{
	CullableList ret;
	if (!BoxBox(b, aabb.GetBox())) return ret;

	// intersection. return all cullables
	for (unsigned int i=0; i<cullables.size(); i++)
	{
		if (BoxBox(b, cullables[i]->GetSpacetimeBoundingBox()))
			ret.push_back(cullables[i]);
	}

	// recursion
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
		{
			ret = ConcatCullLists(ret, children[i]->Query(b));
		}
	}

	return ret;
}

CullableList OctreeNode::Query(const Ray &r)
{
	CullableList ret;
	if (!Intersect(r, aabb.GetBox())) return ret;

	// intersection. return all cullables
	for (unsigned int i=0; i<cullables.size(); i++)
	{
		if (Intersect(r, cullables[i]->GetSpacetimeBoundingBox()))
			ret.push_back(cullables[i]);
	}

	// recursion
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
		{
			ret = ConcatCullLists(ret, children[i]->Query(r));
		}
	}

	return ret;
}

CullableList OctreeNode::Query(const Vector3 &p)
{
	CullableList ret;
	if (!Intersect(p, aabb.GetBox())) return ret;

	// intersection. return all cullables
	for (unsigned int i=0; i<cullables.size(); i++)
	{
		if (Intersect(p, cullables[i]->GetSpacetimeBoundingBox()))
			ret.push_back(cullables[i]);
	}

	// recursion
	for (unsigned int i=0; i<8; i++)
	{
		if (children[i])
		{
			ret = ConcatCullLists(ret, children[i]->Query(p));
		}
	}

	return ret;
}
