// octree.h

/*
Manual:
=======
1 - derive the class to be culled from Cullable
2 - Provide the virtuals of Cullable, update enum CullableType, if needed
3 - Init octree:
	a - OctreeNode oct(AlignedBox(), max_depth, max_cullables);
		(max_cullables is the maximum number of cullables per cell)
		(max depth is the maximum depth of the octree)
	b - for all your cullables: oct.AddCullable(Cullable *cb);
	c - oct.FitContained(); to fit all the contained cullables
	d - oct.Subdivide(); to recursively subdivide the cell until "max cullables" 
		is achieved, or subdivision reached "max_depth".

4 - Perform queries:
	oct.Query(Sphere or Box or Ray or Point);

	a std::vector<Cullable*> (typedef CullableList) is returned
*/

#ifndef _OCTREE_H_
#define _OCTREE_H_

#include "cullable.h"
#include <math3d.h>
#include <geom.h>
#include <vector>

enum OctreeBox
{
	OCTREE_FRONT_TOP_LEFT = 0,
	OCTREE_FRONT_TOP_RIGHT,
	OCTREE_FRONT_BOTTOM_LEFT,
	OCTREE_FRONT_BOTTOM_RIGHT,
	OCTREE_BACK_TOP_LEFT,
	OCTREE_BACK_TOP_RIGHT,
	OCTREE_BACK_BOTTOM_LEFT,
	OCTREE_BACK_BOTTOM_RIGHT
};

class AlignedBox
{
protected:
	Vector3 from, to;
	void Update();
	Box box;
public:
	AlignedBox();
	AlignedBox(Vector3 from, Vector3 to);

	Vector3 GetFrom() const;
	Vector3 GetTo() const;

	Box GetBox() const {return box;};
	AlignedBox GetSubBox(OctreeBox sb) const;
};

class OctreeNode
{
protected:
	AlignedBox aabb;
	unsigned int depth, max_depth, max_cullables;
	OctreeNode *parent;
	OctreeNode *children[8];

	CullableList cullables;
public:
	OctreeNode();
	OctreeNode(const AlignedBox &aabb, unsigned int max_depth=8,
		unsigned int max_cullables=4, unsigned int depth=0);
	~OctreeNode();

	unsigned int GetCullableCount() const {return (unsigned int) cullables.size();}
	Cullable * GetCullable(unsigned int which) const {return cullables[which];}
	
	void AddCullable(Cullable *c);
	void RemoveCullable(Cullable *c);

	void FitContained();
	void Subdivide();

	void Init();

	AlignedBox GetBox() const {return aabb;}
	OctreeNode *GetChild(unsigned int which) {return children[which];}
	unsigned int GetDepth() const {return depth;}

	// Queries
	CullableList Query(const Sphere &s);
	CullableList Query(const Box &b);
	CullableList Query(const Ray &r);
	CullableList Query(const Vector3 &p);
};

unsigned int GetCullCount();
void ResetCullCount();

#endif // ndef _OCTREE_H_
