// pvs.cpp

#include "pvs.h"
#include <map>
#include <progress.h>
#include <gc.h>
#include <immed.h>

PVS::PVS()
{
	box_ok = false;
	convex_area = 0;
	OctreeNode::Init();
}

PVS::~PVS()
{

}

// cullable interface
Box PVS::GetSpacetimeBoundingBox() const
{
	if (!box_ok)
	{
		if (convex_area) bb = convex_area->GetBoundingBox();
	}
	return bb;
}

unsigned int PVS::GetCullableType() const
{
	return CULLABLE_PVS;
}

// Queries
CullableList PVS::Query(const Box &frustum, const Vector3 &viewer)
{
	CullableList ret;

	if (frame == parent_world->GetFrame())
	{
		// we have already checked this room
		return ret;
	}

	frame = parent_world->GetFrame();

	CullableList tmp = OctreeNode::Query(frustum); // Octree::Query();

	for (unsigned int i=0; i<tmp.size(); i++)
	{
		if (tmp[i]->frame != parent_world->GetFrame())
		{
			// add the cullable to the visible list
			ret.push_back(tmp[i]);
			tmp[i]->frame = parent_world->GetFrame();
		}
	}

	// this set is considered active
	parent_world->active_sets[this] = true;

	// recursion
	for (unsigned int i=0; i<portals.size(); i++)
	{
		Box reduced_frustum;
		if (IsPortalVisible(portals[i].quad, frustum, viewer, reduced_frustum))
		{
			ret = ConcatCullLists(ret, portals[i].dst->Query(reduced_frustum, viewer));
		}
	}

	return ret;
}

bool PVS::IsOutdoor() const
{
	return (parent_world->outdoor_set == this);
}

// PVSWorld class implementation
void PVSWorld::RemovePartial(Cullable *cull)
{
	CullableList cl;
	for (unsigned int i=0; i<partial_cullables.size(); i++)
	{
		if (partial_cullables[i] != cull)
		{
			cl.push_back(partial_cullables[i]);
		}
	}
	partial_cullables = cl;
}

PVSWorld::PVSWorld()
{
	frame = 0;
	outdoor_set = new (std::nothrow) PVS();
	outdoor_set->parent_world = this;
}

PVSWorld::~PVSWorld()
{
	Destroy();
}

void PVSWorld::Destroy()
{
	if (pvs_octree)
	{
		delete pvs_octree;
		pvs_octree = 0;
	}

	for (unsigned int i=0; i<sets.size(); i++)
	{
		if (sets[i])
		{
			delete sets[i];
			sets[i] = 0;
		}
	}


	if (outdoor_set)
	{
		delete outdoor_set;
		outdoor_set = 0;
	}

	sets.clear();
	partial_cullables.clear();
}

void PVSWorld::BuildSets()
{
	Build(); // PortalBuilder::Build();
	
	// init all sets and remember the convex area they wrap
	std::map<ConvexArea*, PVS*> pvs_area_map;
	for (unsigned int i=0; i<area_list.size(); i++)
	{
		if (area_list[i]->IsLink()) continue;
	
		PVS *curr_pvs = new PVS();
		curr_pvs->parent_world = this;
		sets.push_back(curr_pvs);

		curr_pvs->convex_area = area_list[i];
		pvs_area_map[area_list[i]] = curr_pvs;
	}

	// wrap all convex areas around PVS sets
	for (unsigned int i=0; i<sets.size(); i++)
	{
		PVS *curr_pvs = sets[i];

		// add portals to the pvs 
		for (unsigned int j=0; j<connection_list.size(); j++)
		{
			PortalConnection curr_conn = connection_list[j];
			if (curr_conn.src_convex != curr_pvs->convex_area &&
				curr_conn.dst_convex != curr_pvs->convex_area)
				continue;

			Portal added_portal, other_portal;
			added_portal.quad = *curr_conn.dst_quad;
			added_portal.dst = pvs_area_map[curr_conn.dst_convex];
			other_portal.quad = *curr_conn.src_quad;
			other_portal.dst = pvs_area_map[curr_conn.src_convex];
			if (curr_conn.dst_convex == curr_pvs->convex_area)
			{
				Portal tmp;
				tmp = added_portal;
				added_portal = other_portal;
				other_portal = tmp;
			}

			if (!added_portal.dst)
			{
				added_portal.dst = outdoor_set;
				outdoor_set->portals.push_back(other_portal);
			}

			curr_pvs->portals.push_back(added_portal);
		}
	}

	// create an octree of the PVS sets
	pvs_octree = new OctreeNode();
	pvs_octree->Init();
	for (unsigned int i=0; i<sets.size(); i++)
	{
		printf("adding set to octree: %s\n", sets[i]->convex_area->GetName());
		for (unsigned int k=0; k<8; k++)
		{
			Vector3 v = sets[i]->GetSpacetimeBoundingBox().GetVertex(k);
			printf("\t%f, %f, %f\n", v.x, v.y, v.z);
		}
		pvs_octree->AddCullable((Cullable*) sets[i]);
	}
	pvs_octree->FitContained();
	pvs_octree->Subdivide();
}

void PVSWorld::AddCullable(Cullable *cullable)
{
	bool added_to_room = false;

	// create a ConvexArea from the cullable's bounding box
	ConvexArea ca;
	for (unsigned int i=0; i<6; i++)
	{
		Quad q = cullable->GetSpacetimeBoundingBox().GetQuad(i);
		Triangle t1 = q.GetTriangle(0);
		Triangle t2 = q.GetTriangle(1);
		ca.AddTriangle(t1.GetVertex(0), t1.GetVertex(1), t1.GetVertex(2));
		ca.AddTriangle(t2.GetVertex(0), t2.GetVertex(1), t2.GetVertex(2));
	}

	for (unsigned int i=0; i<sets.size(); i++)
	{
		if (ca.ConvexIntersects(sets[i]->convex_area))
		{
			added_to_room = true;
			sets[i]->AddCullable(cullable);
			cullable->SetOwner(sets[i]);
		}


		if (!ca.IsEnclosed(sets[i]->convex_area))
		{
			// this is a potentially partial outdoor
			partial_cullables.push_back(cullable);
		}
		else
		{
			// this is entirely inside a room
			added_to_room = true;
			sets[i]->AddCullable(cullable);
			cullable->SetOwner(sets[i]);
			RemovePartial(cullable);
		}
	}

	if (!added_to_room)
	{
		// add this cullable to the environment
		outdoor_set->AddCullable(cullable);
		cullable->SetOwner(outdoor_set);
	}
}

void PVSWorld::BuildWorld()
{
	// perform per-triangle testing to see if any part of 
	// any cullable is outdoors
	UpdateProgressBar(0, "Extensive tests for partial cullables\n");
	for (unsigned int i=0; i<partial_cullables.size(); i++)
	{
		Cullable *curr = partial_cullables[i];
		for (unsigned int j=0; j<curr->GetCullableTriangleCount(); j++)
		{
			Triangle curr_triangle = curr->GetCullableTriangle(j);
			bool success = false;
			for (unsigned int k=0; k<area_list.size(); k++)
			{
				ConvexArea *curr_area = area_list[k];
				if (curr_area->IsLink()) continue;
				if (curr_area->Encloses(&curr_triangle))
				{
					success = true;
				}
			} 

			if (!success)
			{
				// at least one triangle can see the outside world
				// add it to the outdoor set
				outdoor_set->AddCullable(curr);
				curr->SetOwner(outdoor_set);
				break;
			}
		} // end triangle loop
		UpdateProgressBar((float) i / (float) partial_cullables.size());
	} // end cullable loop
	CompleteProgressTask();

	partial_cullables.erase(partial_cullables.begin(), partial_cullables.end());

	// build octree foreach set (and outdoor)
	UpdateProgressBar(0, "Building PVS set octrees\n");
	for (unsigned int i=0; i<sets.size(); i++)
	{
		sets[i]->FitContained();
		sets[i]->Subdivide();
		UpdateProgressBar((float) i / (float) sets.size());
	}
	CompleteProgressTask();
	outdoor_set->FitContained();
	outdoor_set->Subdivide();
}

unsigned int PVSWorld::GetSetCount() const
{
	return (unsigned int) sets.size();
}

PVS *PVSWorld::GetSet(unsigned int which)
{
	if (which >= sets.size()) return outdoor_set;
	return sets[which];
}

PVS *PVSWorld::GetOutdoorSet()
{
	return outdoor_set;
}

void PVSWorld::Render()
{
	//PortalBuilder::Render();
	Matrix4x4 model = GetModelMatrix();
	SetModelMatrix(Matrix4x4());

	for  (unsigned int i=0; i<sets.size(); i++)
	{
		sets[i]->convex_area->Render();
		ImSetApplyXForm(false);
		ImBegin(IM_QUADS);
		ImColor(0, 0.1, 0);
		for (unsigned int j=0; j<sets[i]->portals.size(); j++)
		{
			ImQuad(sets[i]->portals[j].quad);
		}
		ImEnd();
		ImSetApplyXForm(true);
	}
	Vector3 viewer = Vector3(0, 0, 0).Transformed(GetViewMatrix().Inverse());
	Box fr = CalculateViewFrustum(GetViewMatrix(), GetProjectionMatrix());
	CullableList cl = Query(fr, viewer);
	ImSetApplyXForm(false);
	ImBegin(IM_LINES);
	ImColor(0, 0, 0.2);
	for (unsigned int i=0; i<cl.size(); i++)
	{
		Box b = cl[i]->GetSpacetimeBoundingBox();
		ImWireBox(b);
	}
	ImEnd();
	ImSetApplyXForm(true);

	// reset model matrix
	SetModelMatrix(model);
}

// Queries
CullableList PVSWorld::Query(const Box &frustum, const Vector3 &viewer) const
{
	frame ++;

	active_sets.clear();
	viewer_inside.clear();

	CullableList vs = pvs_octree->Query(viewer);
	CullableList visible_sets;
	for (unsigned int i=0; i<vs.size(); i++)
	{
		if (((PVS*)vs[i])->convex_area->IsPointInside(viewer))
		{
			visible_sets.push_back(vs[i]);
			viewer_inside[(PVS*)vs[i]] = true;
		}
	}

	if (!visible_sets.size())
	{
		if (!outdoor_set)
		{
			CullableList ret;
			return ret;
		}
		viewer_inside[outdoor_set] = true;
		return outdoor_set->Query(frustum, viewer);
	}

	CullableList ret;

	for (unsigned int i=0; i<visible_sets.size(); i++)
	{
		PVS *curr_pvs = (PVS*)visible_sets[i];
		if (!curr_pvs->convex_area->IsPointInside(viewer))
			continue;

		ret = ConcatCullLists(ret, curr_pvs->Query(frustum, viewer));
	}

	return ret;
}

std::vector<PVS*> PVSWorld::GetActiveSets() const
{
	std::map<PVS*, bool>::iterator it;
	std::vector<PVS*> ret;
	for (it = active_sets.begin(); it!=active_sets.end(); it++)
	{
		if (it->second)
		{
			ret.push_back(it->first);
		}
	}

	return ret;
}

bool PVSWorld::IsSetActive(PVS* set) const
{
	return active_sets[set];
}
