// pvs.h

/* Manual
 * ======
 * 1 - PVSWorld world;
 * 2 - foreach ConvexArea in your world:
 *		if (the convex area is a link) ca->SetLink(true);
 *		world.AddConvexArea(ca);
 * 3 - world.BuildSets(); // Generates portals. May take a while
 * 4 - foreach cullable:
 *		world.AddCullable(cullable);
 * 5 - world.BuildWorld(); // Builds octrees and decides which 
 *						   // objects should go to the outdoor set
 *                         // (may also take a while)
 * 6 - world.Query(frustum, viewer); // returns a CullableList. Render it!
*/
#ifndef _PVS_H_
#define _PVS_H_

#include "portals.h"
#include "octree.h"
#include <vector>
#include <map>
#include <math3d.h>

class PVS;

struct Portal
{
	Quad quad;
	PVS *dst;
};

class PVSWorld;

class PVS : public OctreeNode, public Cullable
{
protected:
	ConvexArea *convex_area;
	std::vector<Portal> portals;
	PVSWorld *parent_world;
	mutable Box bb;
	bool box_ok;
public:
	PVS();
	virtual ~PVS();

	// cullable interface
	Box GetSpacetimeBoundingBox() const;
	unsigned int GetCullableType() const;

	// Queries
	CullableList Query(const Box &frustum, const Vector3 &viewer);

	bool IsPointInside(const Vector3 &p) {if (!convex_area) return true;return convex_area->IsPointInside(p);}

	bool IsOutdoor() const;

	unsigned int GetPortalCount() {return (unsigned int)portals.size();}
	Portal GetPortal(unsigned int which) {return portals[which];}

	friend class PVSWorld;
};

class PVSWorld : public PortalBuilder
{
protected:
	OctreeNode *pvs_octree;
	std::vector<PVS*> sets;
	PVS *outdoor_set;

	mutable std::map<PVS*, bool> viewer_inside;
	mutable unsigned int frame;

	CullableList partial_cullables;

	void RemovePartial(Cullable *cull);
	mutable std::map<PVS*, bool> active_sets;
public:
	PVSWorld();
	~PVSWorld();

	void BuildSets();

	void Destroy();

	// This will be called AFTER BuildSets()
	// and BEFORE BuildWorld()
	void AddCullable(Cullable *cullable);

	void BuildWorld();

	unsigned int GetSetCount() const;
	PVS *GetSet(unsigned int which);
	PVS *GetOutdoorSet();

	unsigned int GetFrame() {return frame;}

	void Render();

	// Queries
	CullableList Query(const Box &frustum, const Vector3 &viewer) const;

	// This is valid only after Query()
	std::vector<PVS*> GetActiveSets() const;
	bool IsSetActive(PVS* set) const;
	bool IsViewerInside(PVS *set) const {return viewer_inside[set];}

	friend class PVS;
};

#endif // ndef _PVS_H_

