// approx.cpp

#include "approx.h"
#include "stat.h"
#include "math3d.h"

Ray ApproximateRay(std::vector<Vector3> points)
{
    Ray ret(Vector3(0, 0, 0), Vector3(0, 0, 0));
	int size = (int) points.size();
	float inv_size = 1.0f / size;

    ret.SetOrigin(Average(points));

    float sum_xx = 0.0f, sum_xy = 0.0f, sum_xz = 0.0f;
    float sum_yy = 0.0f, sum_yz = 0.0f, sum_zz = 0.0f;
    for (int i = 0; i < size; i++) 
    {
        Vector3 dif = points[i] - ret.GetOrigin();
        sum_xx += dif.x * dif.x;
        sum_xy += dif.x * dif.y;
        sum_xz += dif.x * dif.z;
        sum_yy += dif.y * dif.y;
        sum_yz += dif.y * dif.z;
        sum_zz += dif.z * dif.z;
    }

    sum_xx *= inv_size;
    sum_xy *= inv_size;
    sum_xz *= inv_size;
    sum_yy *= inv_size;
    sum_yz *= inv_size;
    sum_zz *= inv_size;

	float es[3][3];
	float diag[3];
	float subd[3];

    es[0][0] = sum_yy + sum_zz;
    es[0][1] = -sum_xy;
    es[0][2] = -sum_xz;
    es[1][0] = es[0][1];
    es[1][1] = sum_xx + sum_zz;
    es[1][2] = -sum_yz;
    es[2][0] = es[0][2];
    es[2][1] = es[1][2];
    es[2][2] = sum_xx + sum_yy;

	float m00 = es[0][0];
	float m01 = es[0][1];
	float m02 = es[0][2];
	float m11 = es[1][1];
	float m12 = es[1][2];
	float m22 = es[2][2];

	diag[0] = m00;
	subd[2] = (float)0.0;
	if (fabs(m02) > XSmallNumber)
	{
		float length = sqrt(SQ(m01) + SQ(m02));
		float inv_length = 1.0f / length;
		m01 *= inv_length;
		m02 *= inv_length;
		float q = 2.0f * m01 * m12 + m02 * (m22 - m11);
		diag[1] = m11 + m02 * q;
		diag[2] = m22 - m02 * q;
		subd[0] = length;
		subd[1] = m12 - m01 * q;
		es[0][0] = 1.0f;
		es[0][1] = 0.0f;
		es[0][2] = 0.0f;
		es[1][0] = 0.0f;
		es[1][1] = m01;
		es[1][2] = m02;
		es[2][0] = 0.0f;
		es[2][1] = m02;
		es[2][2] = -m01;
	}
	else
	{
		diag[1] = m11;
		diag[2] = m22;
		subd[0] = m01;
		subd[1] = m12;
		es[0][0] = 1.0f;
		es[0][1] = 0.0f;
		es[0][2] = 0.0f;
		es[1][0] = 0.0f;
		es[1][1] = 1.0f;
		es[1][2] = 0.0f;
		es[2][0] = 0.0f;
		es[2][1] = 0.0f;
		es[2][2] = 1.0f;
	}
	
	const int max_iter = 32;

	for (int i0 = 0; i0 < 3; i0++)
	{
		int i1;
		for (i1 = 0; i1 < max_iter; i1++)
		{
			int i2;
			for (i2 = i0; i2 <= 1; i2++)
			{
				float tmp = fabs(diag[i2]) + fabs(diag[i2 + 1]);
				if ( fabs(subd[i2]) + tmp == tmp )
					break;
			}
			if (i2 == i0)
			{
				break;
			}

			float g = (diag[i0 + 1] - diag[i0]) / 2.0f * subd[i0];
			float r = sqrt(SQ(g) + 1.0f);
			if (g < 0.0f)
			{
				g = diag[i2] - diag[i0] + subd[i0] / (g - r);
			}
			else
			{
				g = diag[i2] - diag[i0] + subd[i0] / (g + r);
			}
			float sine = 1.0f, cosn = 1.0f, p = 0.0f;
			for (int i3 = i2-1; i3 >= i0; i3--)
			{
				float f = sine * subd[i3];
				float b = cosn * subd[i3];
				if (fabs(f) >= fabs(g))
				{
					cosn = g / f;
					r = sqrt(SQ(cosn) + 1.0f);
					subd[i3 + 1] = f * r;
					sine = 1.0f / r;
					cosn *= sine;
				}
				else
				{
					sine = f / g;
					r = sqrt(SQ(sine) + 1.0f);
					subd[i3 + 1] = g * r;
					cosn = 1.0f/r;
					sine *= cosn;
				}
				g = diag[i3 + 1] - p;
				r = (diag[i3] - g) * sine + 2.0f * b * cosn;
				p = sine * r;
				diag[i3 + 1] = g + p;
				g = cosn * r - b;

				for (int i4 = 0; i4 < 3; i4++)
				{
					f = es[i4][i3 + 1];
					es[i4][i3 + 1] = sine * es[i4][i3] + cosn * f;
					es[i4][i3] = cosn * es[i4][i3] - sine * f;
				}
			}
			diag[i0] -= p;
			subd[i0] = g;
			subd[i2] = 0.0f;
		}
	}

	for (int i0 = 0, i1; i0 <= 1; i0++)
	{
		i1 = i0;
		float mx = diag[i1];
		int i2;
		for (i2 = i0 + 1; i2 < 3; i2++)
		{
			if (diag[i2] > mx)
			{
				i1 = i2;
				mx = diag[i1];
			}
		}

		if (i1 != i0)
		{
			diag[i1] = diag[i0];
			diag[i0] = mx;

			for (i2 = 0; i2 < 3; i2++)
			{
				float tmp = es[i2][i0];
				es[i2][i0] = es[i2][i1];
				es[i2][i1] = tmp;
			}
		}
	}

    Vector3 dir(es[0][2], es[1][2], es[2][2]);
	ret.SetDirection(dir);

    return ret;
}
