#include "camera.h"

// Camera class implementation
Camera::Camera()
{
	this->aspect = 1.333333;
	this->fov = 60.0f;
	this->near_clip = 2.0f;
	this->far_clip = 8000.0f;
}
Camera::~Camera()
{
	
}

float Camera::GetAspect() const
{
	return aspect;
}
void Camera::SetAspect(float aspect)
{
	this->aspect = aspect;
}

float Camera::GetFOV() const
{
	return fov;
}
void Camera::SetFOV(float fov)
{
	this->fov = fov;
}

float Camera::GetNearClip() const
{
	return near_clip;
}
void Camera::SetNearClip(float near_clip)
{
	this->near_clip = near_clip;
}

float Camera::GetFarClip() const
{
	return far_clip;
}
void Camera::SetFarClip(float far_clip)
{
	this->far_clip = far_clip;
}

Matrix4x4 Camera::GetProjectionMatrix() const
{
	return PerspectiveProj(fov, aspect, near_clip, far_clip);
}

Box Camera::GetFrustum () const
{
    return CalculateViewFrustum(GetViewMatrix(), GetProjectionMatrix());
}

Vector3 Camera::GetViewpoint() const
{
	Vector3 pos = Vector3(0, 0, 0);

	pos.Transform(GetViewMatrix().Inverse());
	return pos;
}

Matrix4x4 Camera::GetViewMatrixCube(unsigned int face) const
{
	PRSCamera prs_cam;
	prs_cam.SetPosition(GetViewpoint());

	unsigned int f = face % 6;

	switch (f)
	{
	case 0:
		// +x
		prs_cam.SetRotation(Quaternion(Vector3(0, 1, 0), -Pi / 2.0f));
		break;
	case 1:	
		// -x
		prs_cam.SetRotation(Quaternion(Vector3(0, 1, 0), Pi / 2.0f));
		break;

	case 2:
		// +y
		prs_cam.SetRotation(Quaternion(Vector3(1, 0, 0), Pi / 2.0f));
		break;
	case 3:
		// -y
		prs_cam.SetRotation(Quaternion(Vector3(1, 0, 0), -Pi / 2.0f));
		break;

	case 4:
		// +z
		prs_cam.SetRotation(Quaternion(Vector3(0, 1, 0), 0));
		break;
	case 5:
		// -z
		prs_cam.SetRotation(Quaternion(Vector3(0, 1, 0), Pi));
		break;
	}

	return prs_cam.GetViewMatrix();
}

Matrix4x4 Camera::GetProjectionMatrixCube() const
{
	return PerspectiveProj(90.0f, 1.0f, near_clip, far_clip);
}

// MatrixCamera class implementation
// animated cameras mounted in a hierarchy need to store their transform
// as a matrix

MatrixCamera::MatrixCamera()
{
	view_matrix = 0;
}
MatrixCamera::~MatrixCamera()
{

}

void MatrixCamera::SetViewMatrix(Matrix4x4 *view)
{
	view_matrix = view;
}

Matrix4x4 MatrixCamera::GetViewMatrix() const
{
	Matrix4x4 ret;
	ret.ResetIdentity();
	ret.Rotate(Vector3(1, 0, 0), -Pi / 2.0f);
	ret *= (*view_matrix).Inverse();
	return ret;
}

// PRSCamera class implementation
// free cameras (not mounted in any hierarchy) specified by the user
PRSCamera::PRSCamera()
{

}
PRSCamera::~PRSCamera()
{

}

Vector3 PRSCamera::GetViewpoint() const
{
	return GetPosition();
}

Matrix4x4 PRSCamera::GetViewMatrix() const
{
	return GetTransform().Inverse();
}

// AxisCamera class implementation
AxisCamera::AxisCamera ()
{
	position = Vector3 (0.0f, 0.0f, 0.0f);
	up = Vector3 (0.0f, 1.0f, 0.0f);
	look = Vector3 (0.0f, 0.0f, 1.0f);
	right = Vector3 (1.0f, 0.0f, 0.0f);

	SetAspect (1024.0f / 768.0f);
	SetFOV (CAMERA_STD_FOV);
	SetFarClip (CAMERA_FAR_PLANE);
}

AxisCamera::~AxisCamera ()
{
}


void AxisCamera::MoveLook (float units)
{
	position += look * units;
}

void AxisCamera::MoveUp (float units)
{
	position += up * units;
}

void AxisCamera::MoveRight (float units)
{
	position += right * units;
}

void AxisCamera::RotateAroundUp (float angles)
{
	Matrix4x4 m;
	m.SetRotation (up, DEG_TO_RAD (angles));
	look.Transform (m);
	right.Transform (m);
}

void AxisCamera::RotateAroundRight (float angles)
{
	Matrix4x4 m;
	m.SetRotation (right, DEG_TO_RAD (angles));
	look.Transform (m);
	up.Transform (m);
}

void AxisCamera::RotateAroungLook (float angles)
{
	Matrix4x4 m;
	m.SetRotation (look, DEG_TO_RAD (angles));
	up.Transform (m);
	right.Transform (m);
}

void AxisCamera::SetPosition (const Vector3 &pos)
{
	position = pos;
}

void AxisCamera::SetUp (const Vector3 &up)
{
	this->up = up;
}

void AxisCamera::SetLook (const Vector3 &look)
{
	this->look = look;
}

Vector3 AxisCamera::GetPosition () const
{
	return position;
}

Vector3 AxisCamera::GetUp () const
{
	return up;
}

Vector3 AxisCamera::GetLook () const
{
	return look;
}

Vector3 AxisCamera::GetRight () const
{
	return right;
}

Vector3 AxisCamera::GetViewpoint() const
{
	return GetPosition();
}

Matrix4x4 AxisCamera::GetViewMatrix () const
{
    Matrix4x4 m;

    Vector3 At;
    At = position + look;

    Vector3 zaxis = (At - position);
    zaxis.Normalize();

    Vector3 xaxis = CrossProduct (up, zaxis);
    xaxis.Normalize ();

    Vector3 yaxis = CrossProduct (zaxis, xaxis);

    m[0][0] = xaxis.x;           m[1][0] = yaxis.x;           m[2][0] = zaxis.x;          m[3][0] = 0;
    m[0][1] = xaxis.y;           m[1][1] = yaxis.y;           m[2][1] = zaxis.y;          m[3][1] = 0;
    m[0][2] = xaxis.z;           m[1][2] = yaxis.z;           m[2][2] = zaxis.z;          m[3][2] = 0;
    m[0][3] = -DotProduct(xaxis, position);  m[1][3] = -DotProduct(yaxis, position);  m[2][3] = -DotProduct(zaxis, position);  m[3][3] = 1;

    return m;
}

TargetCamera::TargetCamera(const Vector3 &pos, const Vector3 &target)
{
	this->pos = pos;
	this->target = target;
	up = Vector3(0, 1, 0);
}

TargetCamera::~TargetCamera()
{
}

void TargetCamera::SetPosition(const Vector3 &pos)
{
	this->pos = pos;
}

Vector3 TargetCamera::GetPosition() const
{
	return pos;
}

void TargetCamera::SetTarget(const Vector3 &target)
{
	this->target = target;
}

Vector3 TargetCamera::GetTarget() const
{
	return target;
}

void TargetCamera::SetUp(const Vector3 &up)
{
	this->up = up;
}

Vector3 TargetCamera::GetUp() const
{
	return up;
}

Vector3 TargetCamera::GetViewpoint() const
{
	return GetPosition();
}

Matrix4x4 TargetCamera::GetViewMatrix() const
{
	float tx, ty, tz;
	Vector3 i, j, k;

	k = (target - pos).Normalized();
	j = up.Normalized();
	i = CrossProduct(j, k);
	j = CrossProduct(k, i);

	tx = -DotProduct(i, pos);
	ty = -DotProduct(j, pos);
	tz = -DotProduct(k, pos);

	Matrix4x4 m;
	m[0][0] = i.x; m[0][1] = i.y; m[0][2] = i.z; m[0][3] = tx;
	m[1][0] = j.x; m[1][1] = j.y; m[1][2] = j.z; m[1][3] = ty;
	m[2][0] = k.x; m[2][1] = k.y; m[2][2] = k.z; m[2][3] = tz;
	m[3][0] = 0.0; m[3][1] = 0.0; m[3][2] = 0.0; m[3][3] = 1.0;
	return m;
}

DoomCamera::DoomCamera()
{
	trans_sensitivity = rot_sensitivity = 1;
	rotate = false;
	rx = ry = 0;
}

void DoomCamera::RelativeMove(const Vector3 &move)
{
	SetPosition(GetPosition() + 
		move.Transformed(GetRotation().GetRotationMatrix()));
}

void DoomCamera::KeyHandler(int key, bool down)
{
	if(down) {
		switch(key) {
		case 'w':
		case 'W':
            RelativeMove(Vector3(0, 0, trans_sensitivity));
			break;

		case 's':
		case 'S':
			RelativeMove(Vector3(0, 0, -trans_sensitivity));
			break;

		case 'a':
		case 'A':
			RelativeMove(Vector3(-trans_sensitivity, 0, 0));
			break;

		case 'd':
		case 'D':
            RelativeMove(Vector3(trans_sensitivity, 0, 0));
			break;

		case 'q':
		case 'Q':
            RelativeMove(Vector3(0, trans_sensitivity, 0));
			break;

		case 'e':
		case 'E':
			RelativeMove(Vector3(0, -trans_sensitivity, 0));
			break;

		default:
			break;
		}
	}
}

void DoomCamera::Update()
{
	SetRotation(Quaternion(Vector3(1, 0, 0), -ry) *
				Quaternion(Vector3(0, 1, 0), -rx));
}

void DoomCamera::MotionHandler(int x, int y)
{
	static int oldx = x, oldy = y;

	if (rotate)
	{
		float xrad = (x - oldx) * rot_sensitivity / 100.0f;
		float yrad = (y - oldy) * rot_sensitivity / 100.0f;
	
		rx += xrad;
		ry += yrad;

		Update();	
	}

	oldx = x;
	oldy = y;
}

void DoomCamera::ButtonHandler(int x, int y, int keys, bool pressed)
{
	if (keys == 2)
	{
		if (pressed) rotate = true;
		else rotate = false;
	}
}
