
#ifndef CAMERA_H
#define CAMERA_H

#include "math3d.h"
#include "geom.h"

#define CAMERA_STD_FOV    70.0f
#define CAMERA_FAR_PLANE  4000.0f

class Camera
{
protected:
	float aspect, fov, near_clip, far_clip;
public:
	Camera();
	virtual ~Camera();

	float GetAspect() const;
	void SetAspect(float aspect);

	float GetFOV() const;
	void SetFOV(float fov);

	float GetNearClip() const;
	void SetNearClip(float near_clip);

	float GetFarClip() const;
	void SetFarClip(float far_clip);

	Matrix4x4 GetProjectionMatrix() const;
	Box GetFrustum () const;

	virtual Vector3 GetViewpoint() const;

	virtual Matrix4x4 GetViewMatrix() const = 0;

	Matrix4x4 GetViewMatrixCube(unsigned int face) const;
	Matrix4x4 GetProjectionMatrixCube() const;
};

class MatrixCamera : public Camera
{
protected:
	Matrix4x4 *view_matrix;
public:
	MatrixCamera();
	virtual ~MatrixCamera();

	void SetViewMatrix(Matrix4x4 *view);

	virtual Matrix4x4 GetViewMatrix() const;
};

class PRSCamera : public Camera, public PRS
{
protected:
public:
	PRSCamera();
	virtual ~PRSCamera();

	virtual Vector3 GetViewpoint() const;
	virtual Matrix4x4 GetViewMatrix() const;
};

class AxisCamera : public Camera
{
protected:
	Vector3 position;
	Vector3 up, look, right;
public:
	AxisCamera ();
	virtual ~AxisCamera ();

	void MoveLook (float units);
	void MoveUp (float units);
	void MoveRight (float units);

	void RotateAroundUp (float angles);
	void RotateAroundRight (float angles);
	void RotateAroungLook (float angles);

	void SetPosition (const Vector3 &pos);
	void SetUp (const Vector3 &up);
	void SetLook (const Vector3 &look);
	Vector3 GetPosition () const;
	Vector3 GetUp () const;
	Vector3 GetLook () const;
	Vector3 GetRight () const;

	virtual Vector3 GetViewpoint() const;
	virtual Matrix4x4 GetViewMatrix() const;
};

class TargetCamera : public Camera {
protected:
	Vector3 pos, target, up;

public:
	TargetCamera(const Vector3 &pos = Vector3(0, 0, -1), const Vector3 &target = Vector3(0, 0, 0));
	virtual ~TargetCamera();

	virtual void SetPosition(const Vector3 &pos);
	virtual void SetTarget(const Vector3 &target);
	virtual void SetUp(const Vector3 &up);
	virtual Vector3 GetPosition() const;
	virtual Vector3 GetTarget() const;
	virtual Vector3 GetUp() const;

	virtual Vector3 GetViewpoint() const;
	virtual Matrix4x4 GetViewMatrix() const;
};

class DoomCamera : public PRSCamera
{
protected:
	bool rotate;
	float rx, ry;
public:
	float trans_sensitivity, rot_sensitivity;

	DoomCamera();

	float GetRX() const {return rx;}
	float GetRY() const {return ry;}

	void SetRX(float r) {rx = r;}
	void SetRY(float r) {ry = r;}

	void RelativeMove(const Vector3 &move);

	void Update();

	void KeyHandler(int key, bool down);
	void MotionHandler(int x, int y);
	void ButtonHandler(int x, int y, int keys, bool pressed);
};


#endif //CAMERA_H

