// frustum.cpp

#include <cfloat>
#include <string.h>
#include <math.h>
#include "geom.h"

#define UNDER_PLANE(v,p) (DotProduct(v-p.GetPos(),p.GetNormal())<0)

// Plane class implementation
// --------------------------
Plane::Plane()
{
	pos = Vector3(0, 0, 0);
	normal = Vector3(0, 1, 0);
}
	
// plane equation is ax + by + cz + d = 0
// *not* ax + by + cz = d !!!
Plane::Plane(float a, float b, float c, float d)
{
	Vector3 normal = Vector3(a, b, c);
    float len = normal.Length();
    normal /= len;

	Vector3 pos = d * len * normal;
	this->pos = pos;
	this->normal = normal;
}

Plane::Plane(const Vector4 &v)
{
	*this = Plane(v.x, v.y, v.z, v.w);
}

Plane::Plane(const Vector3& normal, const Vector3& pos)
{
	this->normal = normal;
	this->pos = pos;
}

Vector3 Plane::GetPos() const
{
	return pos;
}
	
void Plane::SetPos(const Vector3& pos)
{
	this->pos = pos;
}

/*
this was inlined
Vector3 Plane::GetNormal() const
{
	return normal;
}
*/

void Plane::SetNormal(const Vector3& normal)
{
	this->normal = normal;
}

unsigned int Plane::ClassifyPoint (const Vector3 &point) const
{
	float result;
    Vector3 vec1 = pos;
    Vector3 direction = vec1-point;

    result = DotProduct (normal, direction);

    if (result< -0.01)
    {
        return CP_FRONT;
    }
    if (result> 0.01)
    {
        return CP_BACK;
    }
    return CP_ONPLANE;

}


// Line class implementation
// -------------------------
Line::Line()
{
	verts[0] = verts[1] = Vector3(0, 0, 0);
}

Line::Line(const Vector3& a, const Vector3& b)
{
	verts[0] = a;
	verts[1] = b;
}

/* 
This was made inline
Vector3 Line::GetVertex(unsigned int vert) const
{
	if (vert < 2)
	{
		return verts[vert];
	}
	return Vector3();
}*/

void Line::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 2)
	{
		verts[vert] = vertex;
	}
}

Ray Line::GetRay() const
{
	return Ray(verts[0], (verts[1] - verts[0]).Normalized());
}

float Line::GetLength() const
{
	return (verts[1] - verts[0]).Length();		
}

bool Line::Inside(const Vector3 &v) const
{
	Ray r = GetRay();
	float d = r.GetDist(v);
	if (d < 0) return false;
	if (d > GetLength()) return false;
	return true;
}

Vector3 Line::GetMidpoint() const
{
	return 0.5f * (verts[0] + verts[1]);
}

Line Line::GetReverse() const
{
	return Line(verts[1], verts[0]);
}

// Ray class implementation
// ------------------------
Ray::Ray()
{
	origin = Vector3(0, 0, 0);
	direction = Vector3(0, 0, 1);
}

Ray::Ray(const Vector3& origin, const Vector3& direction)
{
	this->origin = origin;
	this->direction = direction.Normalized();
}

Vector3 Ray::GetOrigin() const
{
	return origin;
}

Vector3 Ray::GetDirection() const
{
	return direction;
}

void Ray::SetOrigin(const Vector3& origin)
{
	this->origin = origin;
}

void Ray::SetDirection(const Vector3& direction)
{
	this->direction = direction.Normalized();
}

Vector3 Ray::GetPos(float dist) const
{
	return origin + (dist * direction);
}

float Ray::GetDist(const Vector3 &pos) const
{
	Vector3 op = pos - origin;
	return DotProduct(op, direction);
}

Vector3 Ray::Project(const Vector3 &v) const
{
	return origin + GetDist(v) * direction;
}

// Triangle class implementation
// -----------------------------
Triangle::Triangle()
{
	verts[0] = verts[1] = verts[2] = Vector3(0, 0, 0);
}

Triangle::Triangle(const Vector3& a, const Vector3& b, const Vector3& c)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
}

Vector3 Triangle::GetNormal() const
{
	return (CrossProduct((verts[1] - verts[0]), (verts[2] - verts[0]))).Normalized();
}

Vector3 Triangle::GetVertex(unsigned int vert) const
{
	if (vert < 3)
	{
		return verts[vert];
	}
	return Vector3();
}

void Triangle::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 3)
	{
		verts[vert] = vertex;
	}
}


Plane Triangle::GetPlane() const
{
	return Plane(GetNormal(), verts[0]);
}

Line Triangle::GetLine(unsigned int lin) const
{
	if (lin < 3)
	{
		return Line(verts[lin], verts[(lin + 1) % 3]);
	}
	return Line();
}

Line Triangle::GetSmallLine() const
{
	Line A, B, C;
	A = GetLine(0);
	B = GetLine(1);
	C = GetLine(2);

	float al = A.GetLength();
	float bl = B.GetLength();
	float cl = C.GetLength();

	if (al < bl && al < cl ) return A;
	if (bl < cl) return B;
	return C;
}

Line Triangle::GetMediumLine() const
{
	Line A, B, C;
	A = GetLine(0);
	B = GetLine(1);
	C = GetLine(2);

	float al = A.GetLength();
	float bl = B.GetLength();
	float cl = C.GetLength();

	if (al < bl && al > cl ) return A;
	if (al > bl && al < cl ) return A;
	
	if (bl < al && bl > cl ) return B;
	if (bl > al && bl < cl ) return B;
	
	return C;
}

Line Triangle::GetLargeLine() const
{
	Line A, B, C;
	A = GetLine(0);
	B = GetLine(1);
	C = GetLine(2);

	float al = A.GetLength();
	float bl = B.GetLength();
	float cl = C.GetLength();

	if (al > bl && al > cl ) return A;
	if (bl > cl) return B;
	return C;
}

Vector3 Triangle::GetCenter() const
{
	return (verts[0] + verts[1] + verts[2]) / 3.0f;
}

void Triangle::Transform(const Matrix4x4 &xform)
{
	for (unsigned int i=0; i<3; i++)
	{
		verts[i].Transform(xform);
	}
}

Plane Triangle::GetEdgePlane(unsigned int i) const
{
	if (i < 3)
	{
		Line lin = GetLine(i);
		Vector3 c = lin.GetVertex(0) + GetNormal();
		Triangle tri = Triangle(lin.GetVertex(0), c, lin.GetVertex(1));
		return tri.GetPlane();
	}
	return Plane();
}

// Quad class implementation
// -------------------------
Quad::Quad()
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Quad::Quad(const Vector3& a, const Vector3& b, 
		   const Vector3& c, const Vector3& d)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
	verts[3] = d;
}

Vector3 Quad::GetVertex(unsigned int vert) const
{
	if (vert < 4)
	{
		return verts[vert];
	}
	return Vector3();
}

void Quad::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 4)
	{
		verts[vert] = vertex;
	}
}

/*
This was made inline
Line Quad::GetLine(unsigned int lin) const
{
	if (lin < 4)
	{
		return Line(verts[lin], verts[(lin + 1) % 4]);
	}
	return Line();
}
*/
Triangle Quad::GetTriangle(unsigned int tri) const
{
	switch (tri)
	{
		case 0:
			return Triangle(verts[0], verts[1], verts[2]);
		case 1:
			return Triangle(verts[0], verts[2], verts[3]);
	}
	return Triangle();
}

Vector3 Quad::GetNormal() const
{
	return GetTriangle(0).GetNormal();
}

Plane Quad::GetPlane() const
{
	return GetTriangle(0).GetPlane();	
}

Plane Quad::GetEdgePlane(unsigned int i) const
{
	if (i < 4)
	{
		Line lin = GetLine(i);
		Vector3 c = lin.GetVertex(0) + GetNormal();
		Triangle tri = Triangle(lin.GetVertex(0), c, lin.GetVertex(1));
		return tri.GetPlane();
	}
	return Plane();
}

bool Quad::FrustumCheck(const Box &frustum) const
{
	Plane box_planes[6];
	for(int i=0; i<6; i++) {
		box_planes[i] = frustum.GetPlane(i);
	}

	// first check the quad vertices...
	Vector3 qverts[4];
	for(int i=0; i<4; i++) {
		qverts[i] = GetVertex(i);

		bool vertex_inside = true;
		for(int j=2; j<6; j++) {
			unsigned int cp = box_planes[j].ClassifyPoint(qverts[i]);
			if(cp != CP_FRONT && cp != CP_ONPLANE) {
				vertex_inside = false;
				break;
			}
		}

		// if there is at least one vertex inside, return true
		if(vertex_inside) {
			return true;
		}
	}

	// otherwise check the intersection of each frustum quad with the quad
	for(int i=0; i<6; i++) {
		if(Intersect(frustum.GetQuad(i), *this)) {
			return true;
		}
	}

	return false;
}

void Quad::Flip()
{
	Vector3 old[4];
	memcpy(old, verts, 4 * sizeof(Vector3));
	for (unsigned int i=0; i<4; i++)
	{
		verts[i] = old[3 - i];
	}
}

// BoundingVolume class implementation
// -----------------------------------
BoundingVolume::BoundingVolume()
{
	transform = Matrix4x4::identity;
}

BoundingVolume::~BoundingVolume()
{
	for (size_t i=0; i<children.size(); i++)
	{
		delete children[i];
	}
}

// this will be overriden
bool BoundingVolume::FrustumCheck(const Box &frustum, const Matrix4x4 &transform) const {
	return true;
}

bool BoundingVolume::FrustumCheck(const Box& frustum) const
{
	if (!FrustumCheck(frustum, transform))
	{
		return false;
	}

	if (children.size() == 0)
	{
		return true;
	}

	for (unsigned int i=0; i<children.size(); i++)
	{
		if (children[i]->FrustumCheck(frustum))
		{
			return true;
		}
	}

	return false;
}

void BoundingVolume::SetTransform(const Matrix4x4& mat)
{
	this->transform = mat;
	for (unsigned int i=0; i<children.size(); i++)
	{
		children[i]->SetTransform(mat);
	}
}

void BoundingVolume::AddVolume(BoundingVolume* vol)
{
	children.push_back(vol);
}

// Tetrahedron class implementation
// --------------------------------
Tetrahedron::Tetrahedron()
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Tetrahedron::Tetrahedron(const Vector3& a, const Vector3& b, 
						 const Vector3& c, const Vector3& d)
{
	verts[0] = a;
	verts[1] = b;
	verts[2] = c;
	verts[3] = d;
}

Vector3 Tetrahedron::GetVertex(unsigned int vert) const
{
	if (vert < 4)
	{
		return verts[vert];
	}
	return Vector3();
}

void Tetrahedron::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 4)
	{
		verts[vert] = vertex;
	}
}

Triangle Tetrahedron::GetTriangle(unsigned int tri) const
{
	switch (tri)
	{
		case 0:
			return Triangle(verts[0], verts[1], verts[2]);
		case 1:
			return Triangle(verts[0], verts[4], verts[1]);
		case 2:
			return Triangle(verts[1], verts[4], verts[2]);
		case 3:
			return Triangle(verts[2], verts[4], verts[0]);
	}
	return Triangle();
}

bool Tetrahedron::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Tetrahedron transformed = *this;
	transformed.Transform(transform);
	
	unsigned char outside[4];
	memset(outside, 0, 4 * sizeof(unsigned char));
	
	// STEP 1: collect info about which vertex is in/outside which frustum plane
	// if any vertex is inside all frustum planes, then we return success
	for (unsigned int t_vert=0; t_vert<4; t_vert++)
	{
		for (unsigned int f_plane=0; f_plane<6; f_plane++)
		{
			if (UNDER_PLANE(transformed.GetVertex(t_vert), frustum.GetPlane(f_plane)))
			{
				outside[t_vert] |= (1 << f_plane);
			}
		}
		if (!outside[t_vert])
		{
			// this vertex is inside the frustum
			return true;
		}
	}

	// STEP 2: if all vertices are outside *any* plane then return failure
	unsigned char all_verts = outside[0];

	for (unsigned int i=1; i<4; i++)
	{
		all_verts &= outside[i];
	}

	if (all_verts)
	{
		// all verts are outside of at least one plane
		return false;
	}

	// STEP 3: if any quad of the frustum intersects any 
	// triangle of the tetrahedron, return success
	for (unsigned int f_quad=0; f_quad<6; f_quad++)
	{
		for (unsigned int t_tri=0; t_tri<4; t_tri++)
		{
			if (Intersect(transformed.GetTriangle(t_tri), frustum.GetQuad(f_quad)))
			{
				return true;
			}
		}
	}

	// STEP 4: there are 2 cases left: 
	// 1 - the tetrahedron is completely outside the frustum
	// 2 - the tetrahedron contains the frustum
	// if any point of the frustum is outside any plane 
	// of the Tetrahedron then return failure
	for (unsigned int f_vert=0; f_vert<8; f_vert++)
	{
		for (unsigned int t_plane=0; t_plane<4; t_plane++)
		{
			if (UNDER_PLANE(frustum.GetVertex(f_vert), transformed.GetTriangle(t_plane).GetPlane()))
			{
				return false;
			}
		}
	}

	// yup! the bounding tetrahedron *contains* the frustum :)
	return true;
}

void Tetrahedron::Transform(const Matrix4x4& mat)
{
	for (unsigned int i=0; i<4; i++)
	{
		verts[i].Transform(mat);
	}
}

// Box class implementation
// ------------------------
void Box::CalculatePlanes()
{
	for (unsigned int i=0; i<6; i++)
	{
		Quad q = GetQuad(i);
		planes[i] = Plane(q.GetNormal(), q.GetVertex(0));
	}
}

void Box::CalculateVertices()
{
	verts[0] = PointOfIntersection(planes[FP_FRONT], planes[FP_TOP], planes[FP_LEFT]);
	verts[1] = PointOfIntersection(planes[FP_FRONT], planes[FP_BOTTOM], planes[FP_LEFT]);
	verts[2] = PointOfIntersection(planes[FP_FRONT], planes[FP_BOTTOM], planes[FP_RIGHT]);
	verts[3] = PointOfIntersection(planes[FP_FRONT], planes[FP_TOP], planes[FP_RIGHT]);

	verts[4] = PointOfIntersection(planes[FP_BACK], planes[FP_TOP], planes[FP_LEFT]);
	verts[5] = PointOfIntersection(planes[FP_BACK], planes[FP_BOTTOM], planes[FP_LEFT]);
	verts[6] = PointOfIntersection(planes[FP_BACK], planes[FP_BOTTOM], planes[FP_RIGHT]);
	verts[7] = PointOfIntersection(planes[FP_BACK], planes[FP_TOP], planes[FP_RIGHT]);
}

Box::Box()
{
	for (unsigned int i=0; i<6; i++)
	{
		planes[i] = Plane();
	}

	for (unsigned int i=0; i<8; i++)
	{
		verts[i] = Vector3(0, 0, 0);
	}
}

Box::Box(const Plane& front, const Plane& back,
	const Plane& left, const Plane& right,
	const Plane& top, const Plane& bottom)
{
	planes[0] = front;
	planes[1] = back;
	planes[2] = left;
	planes[3] = right;
	planes[4] = top;
	planes[5] = bottom;

	CalculateVertices();
}

Box::Box(const Quad& front, const Quad& back)
{
	verts[0] = front.GetVertex(0);
	verts[1] = front.GetVertex(1);
	verts[2] = front.GetVertex(2);
	verts[3] = front.GetVertex(3);

	verts[4] = back.GetVertex(0);
	verts[5] = back.GetVertex(1);
	verts[6] = back.GetVertex(2);
	verts[7] = back.GetVertex(3);

	CalculatePlanes();
}

// Axis aligned box
Box::Box(const Vector3 &from, const Vector3 &to)
{
	Vector3 size = to - from;
	Quad front(
		from + Vector3(0, size.y, 0),
		from + Vector3(0, 0, 0),
		from + Vector3(size.x, 0, 0),
		from + Vector3(size.x, size.y, 0));
	Quad back(
		front.GetVertex(0) + Vector3(0, 0, size.z),
		front.GetVertex(1) + Vector3(0, 0, size.z),
		front.GetVertex(2) + Vector3(0, 0, size.z),
		front.GetVertex(3) + Vector3(0, 0, size.z));

	*this = Box(front, back);
}


// creates a bounding box for a set of vertices
Box::Box(const std::vector<Vector3> &verts)
{
	// TODO: optimize this with oriented bounding boxes
	if (!verts.size())
	{
		*this = Box(Vector3(0, 0, 0), Vector3(0, 0, 0));
		return;
	}
	
	Vector3 from, to;
	from = to = verts[0];
	for (unsigned int i=1; i<verts.size(); i++)
	{
		if (verts[i].x < from.x) from.x = verts[i].x;
		if (verts[i].y < from.y) from.y = verts[i].y;
		if (verts[i].z < from.z) from.z = verts[i].z;

		if (verts[i].x > to.x) to.x = verts[i].x;
		if (verts[i].y > to.y) to.y = verts[i].y;
		if (verts[i].z > to.z) to.z = verts[i].z;
	}
	
	*this = Box(from, to);
}

Quad Box::GetQuad(unsigned int i) const
{
	switch (i)
	{
		case FP_FRONT:
			return Quad(verts[3], verts[0], verts[1], verts[2]);
		case FP_BACK:
			return Quad(verts[4], verts[7], verts[6], verts[5]);
		case FP_LEFT:
			return Quad(verts[0], verts[4], verts[5], verts[1]);
		case FP_RIGHT:
			return Quad(verts[7], verts[3], verts[2], verts[6]);
		case FP_TOP:
			return Quad(verts[0], verts[3], verts[7], verts[4]);
		case FP_BOTTOM:
			return Quad(verts[5], verts[6], verts[2], verts[1]);
	}
	return Quad();
}

/*
This became inline
Plane Box::GetPlane(unsigned int pln) const
{
	if (pln < 6)
	{
		return planes[pln];
	}
	return Plane();
}
*/

void Box::SetPlane(unsigned int pln, const Plane& plane)
{
	if (pln < 6)
	{
		planes[pln] = plane;
		CalculateVertices();
	}
}

/*
This became inline
Vector3 Box::GetVertex(unsigned int vert) const
{
	if (vert < 8)
	{
		return verts[vert];
	}
	return Vector3();
}
*/

void Box::SetVertex(unsigned int vert, const Vector3& vertex)
{
	if (vert < 8)
	{
		verts[vert] = vertex;
		CalculatePlanes();
	}
}

bool Box::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Box transformed = *this;
	transformed.Transform(transform);
	
	return Intersect(frustum, transformed);
}

void Box::Transform(const Matrix4x4& mat)
{
	for (unsigned int i=0; i<8; i++)
	{
		verts[i] = verts[i].Transformed(mat);
	}

	CalculatePlanes();
}

unsigned int Box::ClassifyWithPlane (const Plane &plane) const
{
	unsigned int front=0, back=0;

	for (int i = 0; i < 8; i ++)
	{
		unsigned int clid = plane.ClassifyPoint (verts[i]);
		switch (clid)
		{
		case CP_FRONT: front++; break;
		case CP_BACK: back++; break;
		case CP_ONPLANE: front++; back++; break;
		}
	}

	if ((front == 8) && (back == 8))
	{
		return CP_ONPLANE;
	}
	if (front == 8)
	{
		return CP_FRONT;
	}
	if (back == 8)
	{
		return CP_BACK;
	}
	return CP_SPLIT;
}

Vector3 Box::GetViewer() const {
	Ray r = RayOfIntersection(GetPlane(FP_TOP), GetPlane(FP_BOTTOM));
	return PointOfIntersection(r, GetPlane(FP_RIGHT));
}

void Box::ConvertToPyramid(const Vector3 &v)
{
	verts[0] = verts[1] = verts[2] = verts[3] = v;
	CalculatePlanes();
}

Vector3 Box::GetCenter() const
{
	Vector3 sum(0, 0, 0);
	for (unsigned int i=0; i<8; i++)
	{
		sum += verts[i];
	}
	return sum / 8;
}

// Sphere class implementation
// ---------------------------
Sphere::Sphere()
{
	radius = 0.0f;
	center = Vector3(0, 0, 0);
}

Sphere::Sphere(const Vector3 center, float radius)
{
	this->center = center;
	this->radius = radius;
}

float Sphere::GetRadius() const
{
	return radius;
}

Vector3 Sphere::GetCenter() const
{
	return center;
}

void Sphere::SetRadius(float radius)
{
	this->radius = radius;
}

void Sphere::SetCenter(const Vector3& center)
{
	this->center = center;
}

bool Sphere::FrustumCheck(const Box& frustum, const Matrix4x4& transform) const
{
	Sphere transformed = *this;
	transformed.Transform(transform);
	
	for (unsigned int i=0; i<6; i++)
	{
		if (UnderPlane(transformed, frustum.GetPlane(i)))
		{
			return false;
		}
	}
	return true;
}

void Sphere::Transform(const Matrix4x4& mat)
{
	// transform radius
	Vector3 v = center + Vector3(radius, 0, 0);
	v.Transform(mat);

	// transform center
	center.Transform(mat);

	radius = (v - center).Length();
}

// utility functions
// -----------------
bool UnderPlane(const Vector3& point, const Plane& plane)
{
	/*
	if (DotProduct(point - plane.GetPos(), plane.GetNormal()) < 0.0f)
	{
		return true;
	}
	return false;
	*/
	return UNDER_PLANE(point, plane);
}

bool UnderPlane(const Triangle& triangle, const Plane& plane)
{
	for (unsigned int i=0; i<3; i++)
	{
		if (!UNDER_PLANE(triangle.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Quad& quad, const Plane& plane)
{
	for (unsigned int i=0; i<4; i++)
	{
		if (!UNDER_PLANE(quad.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Tetrahedron& tetrahedron, const Plane& plane)
{
	for (unsigned int i=0; i<4; i++)
	{
		if (!UNDER_PLANE(tetrahedron.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Box& box, const Plane& plane)
{
	for (unsigned int i=0; i<8; i++)
	{
		if (!UNDER_PLANE(box.GetVertex(i), plane))
		{
			return false;
		}
	}
	return true;
}

bool UnderPlane(const Sphere& sphere, const Plane& plane)
{
	float dist = DistanceFromPlane(sphere.GetCenter(), plane);

	if (dist + sphere.GetRadius() < 0)
	{
		return true;
	}

	return false;
}

bool OverPlane(const Sphere &sphere, const Plane &plane)
{
	float dist = DistanceFromPlane(sphere.GetCenter(), plane);

	if (dist - sphere.GetRadius() > 0)
	{
		return true;
	}

	return false;
}

bool Intersect(const Ray &ray, const Plane &plane)
{
	float dp = DotProduct(ray.GetDirection(), plane.GetNormal());
	if (fabs(dp) < 0.00000001f) return false;
	return true;
}

bool Intersect(const Ray &ray, const Triangle &triangle)
{
	Plane pl = triangle.GetPlane();
	if (!Intersect(ray, pl)) return false;
	
	// ok, it intersects the plane, but is the 
	// intersection inside the triangle?
	Vector3 pt = PointOfIntersection(ray, pl);
	for (int i=0; i<3; i++)
	{
		if (UNDER_PLANE(pt, triangle.GetEdgePlane(i))) return false;
	}
	return true;
}

bool Intersect(const Ray &ray, const Quad &quad)
{
	if (Intersect(ray, quad.GetTriangle(0))) return true;
	return Intersect(ray, quad.GetTriangle(1));
}

bool Intersect(const Line& line, const Triangle& triangle)
{
	Plane pl = triangle.GetPlane();
	if (!(UNDER_PLANE(line.GetVertex(0), pl) ^ UNDER_PLANE(line.GetVertex(1), pl)))
	{
		// both of the line's vertices are in the same side of the
		// plane - no intersection
		return false;
	}

	// get instersection with plane
	Vector3 isc = PointOfIntersection(line.GetRay(), pl);

	// is the intersection inside the triangle?
	for (unsigned int i=0; i<3; i++)
	{
		if (UNDER_PLANE(isc, triangle.GetEdgePlane(i)))
		{
			return false;
		}
	}

	return true;
}

bool Intersect(const Triangle& a, const Triangle& b)
{
	for (unsigned int i=0; i<3; i++)
	{
		if (Intersect(a.GetLine(i), b))
		{
			return true;
		}
	}

	for (unsigned int i=0; i<3; i++)
	{
		if (Intersect(b.GetLine(i), a))
		{
			return true;
		}
	}

	return false;
}

bool Intersect(const Triangle& triangle, const Tetrahedron& tetrahedron)
{
	for (int i=0; i<4; i++)
	{
		Triangle tetr_tri = tetrahedron.GetTriangle(i);
		if (UnderPlane(triangle, tetr_tri.GetPlane()))
		{
			return false;
		}

		if (Intersect(triangle, tetr_tri))
		{
			return true;
		}
	}

	// triangle is contained inside the tetrahedron
	return true;
}

bool Intersect(const Line& line, const Quad& quad)
{
	Plane pl = quad.GetPlane();
	if (!(UNDER_PLANE(line.GetVertex(0), pl) ^ UNDER_PLANE(line.GetVertex(1), pl)))
	{
		// both of the line's vertices are in the same side of the
		// plane - no intersection
		return false;
	}

	// get instersection with plane
	Vector3 isc = PointOfIntersection(line.GetRay(), pl);

	// is the intersection inside the quad?
	for (unsigned int i=0; i<4; i++)
	{
		if (UNDER_PLANE(isc, quad.GetEdgePlane(i)))
		{
			return false;
		}
	}

	return true;
}

bool Intersect(const Triangle& triangle, const Quad& quad)
{
	for (int i=0; i<2; i++)
	{
		if (Intersect(triangle, quad.GetTriangle(i)))
		{
			return true;
		}
	}

	return false;
}

static inline unsigned int FourVertsPlane(Plane pl, Vector3 verts[])
{
	unsigned int under = 0;
	for (unsigned int i=0; i<4; i++)
	{
		if (UNDER_PLANE(verts[i], pl))
		{
			under |= 1 << i;
		}
	}

	return under;
}

static inline unsigned int FourVertsSameSide(Plane pl, Vector3 verts[])
{
	unsigned int under = FourVertsPlane(pl, verts);
	if (!under || under == 15) return true;
	return false;
}

static inline unsigned int FourVertsUnder(Plane pl, Vector3 verts[])
{
	unsigned int under = FourVertsPlane(pl, verts);
	if (under == 15) return true;
	return false;
}

bool Intersect(const Quad& a, const Quad& b)
{
	// collect verts and planes
	static Plane apl, bpl;
	static Vector3 av[4], bv[4];
	apl = a.GetPlane();
	bpl = b.GetPlane();
	for (unsigned int i=0; i<4; i++)
	{
		av[i] = a.GetVertex(i);
		bv[i] = b.GetVertex(i);
	}

	if (FourVertsSameSide(apl, bv)) return false;
	if (FourVertsSameSide(bpl, av)) return false;

	for (unsigned int i=0; i<4; i++)
	{
		if (FourVertsUnder(a.GetEdgePlane(i), bv)) return false;
		if (FourVertsUnder(b.GetEdgePlane(i), av)) return false;
	}

	for (unsigned int i=0; i<4; i++)
	{
		if (Intersect(a.GetLine(i), b)) return true;
		if (Intersect(b.GetLine(i), a)) return true;
	}

	return false;
	
	/*
	//Unoptimized version
	for (int i=0; i<2; i++)
	{
		if (Intersect(a.GetTriangle(i), b))
		{
			return true;
		}
	}
	return false;
	*/
}

bool Intersect(const Line& l, const Plane& p)
{
	bool a = UNDER_PLANE(l.GetVertex(0), p);
	bool b = UNDER_PLANE(l.GetVertex(1), p);

	if (a ^ b)
	{
		return true;
	}

	return false;
}

bool Intersect(const Sphere& s, const Plane& p)
{
	float d = DistanceFromPlane(s.GetCenter(), p);
	if (d <= s.GetRadius()) return true;
	return false;
}


bool Intersect(const Sphere &s, const Triangle &t)
{
	if (!Intersect(s, t.GetPlane())) return false;
	if (UnderPlane(s, t.GetEdgePlane(0))) return false;
	if (UnderPlane(s, t.GetEdgePlane(1))) return false;
	if (UnderPlane(s, t.GetEdgePlane(2))) return false;

	if (Intersect(t.GetLine(0), s)) return true;
	if (Intersect(t.GetLine(1), s)) return true;
	if (Intersect(t.GetLine(2), s)) return true;

	// is the sphere totally inside the triangle?
	if (OverPlane(s, t.GetEdgePlane(0)) &&
		OverPlane(s, t.GetEdgePlane(1)) &&
		OverPlane(s, t.GetEdgePlane(2))) return true;

	return false;
}

bool Intersect(const Ray &ray, const Box &box)
{
	for (int i=0; i<6; i++)
	{
		if (Intersect(ray, box.GetQuad(i))) return true;
	}
	return false;
}

bool Intersect(const Ray &ray, const Sphere &sphere)
{
	Vector3 closest = ray.Project(sphere.GetCenter());
	Vector3 cc = sphere.GetCenter() - closest;
	if (cc.LengthSq() > sphere.GetRadius() * sphere.GetRadius()) return false;
	return true;
}

bool Intersect(const Line &line, const Sphere &sphere)
{
	if (!Intersect(line.GetRay(), sphere)) return false;
	Vector3 p = PointOfIntersection(line.GetRay(), sphere);
	if (line.GetRay().GetDist(p) > line.GetLength()) return false;
	return true;
}

bool Intersect(const Box &box1, const Box &box2)
{
	unsigned char outside[8];
	memset(outside, 0, 8 * sizeof(unsigned char));
	
	// STEP 1: collect info about which vertex is in/outside which frustum plane
	// if any vertex is inside all frustum planes, then we return success
	for (unsigned int b_vert=0; b_vert<8; b_vert++)
	{
		for (unsigned int f_plane=0; f_plane<6; f_plane++)
		{
			if (UNDER_PLANE(box2.GetVertex(b_vert), box1.GetPlane(f_plane)))
			{
				outside[b_vert] |= (1 << f_plane);
			}
		}
		if (!outside[b_vert])
		{
			// this vertex is inside the frustum
			return true;
		}
	}

	// STEP 2: if all vertices are outside *any* plane then return failure
	unsigned char all_verts = outside[0];

	for (unsigned int i=1; i<8; i++)
	{
		all_verts &= outside[i];
	}

	if (all_verts)
	{
		// all verts are outside of at least one plane
		return false;
	}

	// STEP 3: if any quad of the frustum intersects any quad of the box, return success
	for (unsigned int f_quad=0; f_quad<6; f_quad++)
	{
		for (unsigned int b_quad=0; b_quad<6; b_quad++)
		{
			if (Intersect(box1.GetQuad(f_quad), box2.GetQuad(b_quad)))
			{
				return true;
			}
		}
	}

	// STEP 4: there are 2 cases left: 
	// 1 - the box is completely outside the frustum
	// 2 - the box contains the frustum
	// if any point of the frustum is outside any plane 
	// of the box then return failure
	for (unsigned int f_vert=0; f_vert<8; f_vert++)
	{
		for (unsigned int b_plane=0; b_plane<6; b_plane++)
		{
			if (UNDER_PLANE(box1.GetVertex(f_vert), box2.GetPlane(b_plane)))
			{
				return false;
			}
		}
	}

	// yup! the bounding box *contains* the frustum :)
	return true;
}

bool Intersect(const Sphere &sphere, const Box &box)
{
	for (unsigned int i=0; i<6; i++)
	{
		if (UnderPlane(sphere, box.GetPlane(i)))
			return false;
	}

	return true;
}

bool Intersect(const Vector3 &point, const Box &box)
{
	for (unsigned int i=0; i<6; i++)
	{
		if (UNDER_PLANE(point, box.GetPlane(i)))
			return false;
	}

	return true;
}

bool Intersect(const Quad &quad, const Box &box)
{
	for (unsigned int i=0; i<4; i++)
	{
		if (Intersect(quad.GetVertex(i), box))
			return true;
	}

	for (unsigned int i=0; i<6; i++)
	{
		if (Intersect(quad, box.GetQuad(i)))
			return true;
	}

	return false;
}

Ray RayOfIntersection(const Plane& a, const Plane& b)
{
	Vector3 t = CrossProduct(a.GetNormal(), b.GetNormal());
	if(t.Length() < XSmallNumber) {
		return Ray();
	}

	t = CrossProduct(t, a.GetNormal());
	Ray r(a.GetPos(), t);

	Vector3 point = PointOfIntersection(r, b);
	Vector3 dir = CrossProduct(a.GetNormal(), b.GetNormal());
	return Ray(point, dir);
}

Vector3 PointOfIntersection(const Ray& ray, const Plane& plane)
{
	float dot2 = DotProduct(ray.GetDirection(), plane.GetNormal());
	if (fabs(dot2) < 0.0000000001f)
	{
		// oops! something went wrong
		return Vector3();
	}

	Vector3 X = ray.GetOrigin() - plane.GetPos();
	float dot1 = DotProduct(X, plane.GetNormal());
	float t = -(dot1 / dot2);

	return ray.GetPos(t);
}

Vector3 PointOfIntersection(const Ray &ray, const Triangle &triangle)
{
	if (Intersect(ray, triangle))
	{
		return PointOfIntersection(ray, triangle.GetPlane());
	}
	return Vector3();
}

Vector3 PointOfIntersection(const Plane& a, const Plane& b, const Plane& c)
{
	// get the intersection of a and b as a ray
	Ray ab = RayOfIntersection(a, b);

	return PointOfIntersection(ab, c);
}

Vector3 PointOfIntersection(const Ray &ray, const Sphere &sphere)
{
	Vector3 closest = ray.Project(sphere.GetCenter());
	float a = (closest - sphere.GetCenter()).Length();
	float d = ray.GetDist(closest);
	float sq = sphere.GetRadius()*sphere.GetRadius() - a*a;
	if (sq < 0) return Vector3();
	float x = sqrt(sq);
	float dist = d - x;
	if (dist < 0) dist = d + x;
	return ray.GetPos(dist);
}

float DistanceFromPlane(const Vector3& point, const Plane& plane)
{
	return DotProduct(point - plane.GetPos(), plane.GetNormal());
}

Vector3 ProjectToPlane(const Vector3& point, const Plane& plane)
{
	return point - (DistanceFromPlane(point, plane) * plane.GetNormal());
}

Vector3 Project(const Vector3 &point, const Ray &ray)
{
	Vector3 d = ray.GetOrigin() - point;
	float dot = DotProduct(d, ray.GetDirection());
	d -= dot * ray.GetDirection();
	Vector3 ret = point + d;
	return ret;
}
float Distance(const Vector3 &point, const Ray &ray)
{
	return (Project(point, ray) - point).Length();
}

Plane FrustumPlane (const Matrix4x4 &mat, int plane)
{
    int i = plane / 2;
    int neg = plane % 2;

    float a,b,c,d;

    a = mat [3][0] + (neg ? -mat[i][0] : mat [i][0]);
    b = mat [3][1] + (neg ? -mat[i][1] : mat [i][1]);
    c = mat [3][2] + (neg ? -mat[i][2] : mat [i][2]);
    d = mat [3][3] + (neg ? -mat[i][3] : mat [i][3]);

    float len = Vector4 (a, b, c, d).Length();
    a /= len;
    b /= len;
    c /= len;
    d /= len;

    Plane p = Plane (a,b,c,d);
    return p;
}

// Viewing frustum 2
Box CalculateViewFrustum(const Matrix4x4 &view, const Matrix4x4 &proj)
{
	// projection matrix acts on 4d vectors
	Vector4 box_verts[8] = 
	{
		Vector4(-1,  1,  0, 1),
		Vector4(-1, -1,  0, 1),
		Vector4( 1, -1,  0, 1),
		Vector4( 1,  1,  0, 1),

		Vector4(-1,  1,  1, 1),
		Vector4(-1, -1,  1, 1),
		Vector4( 1, -1,  1, 1),
		Vector4( 1,  1,  1, 1)
	};

	// compute view space verts
	for (unsigned int i=0; i<8; i++)
	{
		box_verts[i].Transform(proj.Inverse());
		// w is not guaranteed to be 1
		box_verts[i] /= box_verts[i].w;
	}
	Box fr = Box(
		Quad(box_verts[0],
		box_verts[1],
		box_verts[2],
		box_verts[3]),
		Quad(box_verts[4],
		box_verts[5],
		box_verts[6],
		box_verts[7])
		);

	// transform the box from view to world space
	fr.Transform(view.Inverse());

	return fr;
/*
    Matrix4x4 comboMatrix;
    comboMatrix = proj * view;

    Plane p_planes_0 = FrustumPlane (comboMatrix, 0);
    Plane p_planes_1 = FrustumPlane (comboMatrix, 1);
    Plane p_planes_2 = FrustumPlane (comboMatrix, 2);
    Plane p_planes_3 = FrustumPlane (comboMatrix, 3);
    Plane p_planes_4 = FrustumPlane (comboMatrix, 4);
    Plane p_planes_5 = FrustumPlane (comboMatrix, 5);

    Box b = Box (p_planes_4, p_planes_5,	// front, back
		p_planes_0, p_planes_1,				// left, right
		p_planes_3, p_planes_2);			// top, bottom

	return b;
*/
}

// TODO: remove this
std::vector<Box> reduced_boxes;
std::vector<Quad> reduced_quads;
std::vector<Quad> cancelled_portals;

/* Frustum reduction works by finding for each plane, the closest portal plane,
 * and substituting the plane with another one that passes through that vertex.
 * If any of the vertices are behind the plane (outside) we use the original
 * plane instead.
 */
Box ReduceFrustum(const Quad &portal, const Box &frustum)
{
	Plane new_planes[4];
	Plane far_plane = frustum.GetPlane(FP_BACK);
	Plane near_plane = frustum.GetPlane(FP_FRONT);
	Vector3 view_point = frustum.GetViewer();

	Vector3 up[] = {
		frustum.GetVertex(4) - frustum.GetVertex(5),
		frustum.GetVertex(5) - frustum.GetVertex(4),
		frustum.GetVertex(7) - frustum.GetVertex(4),
		frustum.GetVertex(4) - frustum.GetVertex(7)
	};

	// for each plane of the frustum
	for(int i=0; i<4; i++) {
		Plane fplane = frustum.GetPlane(i + 2);	// don't get front/back

		float min_dist = FLT_MAX;
		Vector3 closest_vert;
		
		// for each vertex of the portal quad
		for(int j=0; j<4; j++) {
			Vector3 v = portal.GetVertex(j);

			/* if any one of the vertices lies behind the plane, we want to use
			 * the original plane.
			 */ 
			if(UNDER_PLANE(v, fplane)) {
				min_dist = -1.0;	
				break;
			} else {
				
				/* Project the vertex to the near clipping plane, and calculate
				 * it's projected distance from the "current" plane. The vertex
				 * with the smallest distance wins.
				 * 
				 * The projection is essential, to emulate a "screen-space"
				 * comparison of the portal's vertices. In world space, the
				 * closest vertex to one of the frustum planes, can be different
				 * from the closest in screen space, due to perspective
				 * distortion.
				 *
				 * In order to project the vertex, we find the distance from the
				 * origin of the near plane, along the ray from the origin to the
				 * vertex of interest. Then using that distance we scale the
				 * direction vector.
				 */
				
				Vector3 dir = (v - view_point).Normalized();
				Vector3 near_norm = near_plane.GetNormal();
				
				float pdist = DistanceFromPlane(Vector3(0, 0, 0), near_plane);
				float ndoto = DotProduct(near_norm, view_point);
				float t = -(ndoto + pdist) / DotProduct(near_norm, dir);
				
				Vector3 pv = dir * t;
				float dist = DistanceFromPlane(pv, fplane);
				
				if(dist < min_dist) {
					min_dist = dist;
					closest_vert = v;
				}
			}
		}

		if(min_dist < 0.0) {
			new_planes[i] = fplane;
		} else {
			// construct the plane from the closest vertex found
			Vector3 v1 = closest_vert - view_point;
			Vector3 v2 = (closest_vert + up[i]) - view_point;

			new_planes[i].SetPos(view_point);
			new_planes[i].SetNormal(CrossProduct(v2, v1));
		}
	}

	float min_dist = FLT_MAX;
	Vector3 closest_vert;
	for(int i=0; i<4; i++) {
		Vector3 v = portal.GetVertex(i);
		float dist = DistanceFromPlane(v, near_plane);
		if(dist < 0.0) {
			min_dist = -1.0;
			break;
		}
		if(dist < min_dist) {
			min_dist = dist;
			closest_vert = v;
		}
	}

	Box ret = Box(min_dist < 0.0 ? near_plane : Plane(near_plane.GetNormal(), closest_vert), frustum.GetPlane(FP_BACK),
			new_planes[0], new_planes[1], new_planes[2], new_planes[3]);

	// TODO: remove this
	reduced_boxes.push_back(ret);
	reduced_quads.push_back(portal);

	return ret;
}

bool IsPortalVisible(const Quad &portal, const Box &frustum, 
					 const Vector3 &viewer, Box &reduced_fr)
{
	Vector3 view(0, 0, 0);
	for (int i=0; i<4; i++)
	{
		view += frustum.GetVertex(i);
	}
	view *= 0.25f;
	Vector3 b = view;
	view -= viewer;

	/*
	if (DotProduct(view, portal.GetNormal()) < 0)
		return false;*/

	if (!Intersect(portal, frustum))
	{
		// see if the portal is between viewer and front plane
		// thus not visible by the frustum but allowing us to see
		// things after it
		Line l(viewer, b);
		if (Intersect(l, portal))
		{
			reduced_fr = frustum;
			return true;
		}
		else
		{
			cancelled_portals.push_back(portal);
			return false;
		}
	}
	
	reduced_fr = ReduceFrustum(portal, frustum);
	//reduced_fr = frustum;

	return true;
}

bool CompareRay(const Ray &r1, const Ray &r2)
{
	// compare by distance
	if (Distance(r1.GetOrigin(), r2) > 0.00001f) return false;
	
	// compare by direction
	if (fabs(DotProduct(r1.GetDirection(), r2.GetDirection())) < 0.99999f)
		return false;

	return true;
}

Sphere Concatenate(const Sphere &a, const Sphere &b)
{
	Sphere A = a, B = b;
	if (a.GetRadius() < b.GetRadius())
	{
		A = b;
		B = a;
	}

	Vector3 AB = B.GetCenter() - A.GetCenter();
	if (AB.Length() < B.GetRadius()) return A;
	float csz = (B.GetRadius() - A.GetRadius() + AB.Length()) / 2.0f;
	Vector3 C = A.GetCenter() + csz * AB.Normalized();
	float crad = (AB.Length() + A.GetRadius() + B.GetRadius()) / 2.0f;
	return Sphere(C, crad);
}


// TODO!!!
// optimize this
Sphere GetOutSphere(const Triangle &tri)
{
	// Get the midpoint of edge 1
	Vector3 midp_1 = tri.GetLine(1).GetMidpoint();
	
	// get the perp ray from midpoint
	Ray perp_ray_1 = Ray(midp_1, tri.GetEdgePlane(1).GetNormal());

	// create a perpendicular plane to the triangle that passes
	// through the above ray
	Plane perp_pl = Triangle(midp_1, midp_1 + tri.GetNormal(), 
		midp_1 + perp_ray_1.GetDirection()).GetPlane();

	// Get the midpoint of edge 2
	Vector3 midp_2 = tri.GetLine(2).GetMidpoint();
	
	// get the perp ray from midpoint
	Ray perp_ray_2 = Ray(midp_2, tri.GetEdgePlane(2).GetNormal());

	// the center of the sphere is the intersection of the perp_pl and perp_ray_2
	Vector3 center = PointOfIntersection(perp_ray_2, perp_pl);

	float radius = (center - tri.GetVertex(0)).Length();

	return Sphere(center, radius);
}

// TODO!!!!:
// Add a cone class and change this function properly
Sphere GetOutSphereOfCone(const Vector3 &vert, const Vector3 &center, float radius)
{
	// Get the ray of the cone
	Ray cr = Line(vert, center).GetRay();

	float x, y, z;
	Vector3 norm = cr.GetDirection();

	x = fabs(norm.x);
	y = fabs(norm.y);
	z = fabs(norm.z);

	Vector3 smallest(0, 0, 1);
	if (x < y && x < z)
	{
		smallest = Vector3(1, 0, 0);
	}
	else if (y < z)
	{
		smallest = Vector3(0, 1, 0);
	}

	Vector3 tmp = CrossProduct(smallest, norm);
	smallest = CrossProduct(tmp, norm);
	smallest.Normalize();

	// we have a normal perpendicular to the cone normal
	// now find the rest 2 vertices of the triangle
	Vector3 b = center + radius * smallest;
	Vector3 c = center - radius * smallest;

	return GetOutSphere(Triangle(vert, b, c));
}

// smart box - a bounding box used for skinned meshes.
// it is calculated as a bounding box of the bones only
// and then is scaled by 25% . The scaling is not uniform
// in order to produce a more cube-like box.
Box SmartBox(const std::vector<Vector3> &points)
{
	Box ret;

	ret = Box(points);

	// smart scale
	Vector3 diag = ret.GetVertex(7) - ret.GetVertex(1);
	Vector3 scale = Vector3(1, 1, 1) - diag.Normalized();
	scale.Normalize();

	// average of scale in x, y and z will be 1.25
	scale *= 3 / (scale.x + scale.y + scale.z);
	scale *= 0.25f;
	scale += Vector3(1, 1, 1);

	// centered scaling matrix
	Vector3 center = ret.GetCenter();
	Matrix4x4 scm;
	scm.SetTranslation(center);
	scm.Scale(scale);
	scm.Translate(-center);
	ret.Transform(scm);

	return ret;
}












