/*
 * usage reminder:
 * ---------------
 * Sphere *sph = new Sphere(Vector3(0, 0, 0), 10);
 * Box *bb = new Box(front_quad, back_quad);
 * Box *frustum = new Box(front_clip_plane, back_clip_plane, left .........);
 * sph->SetTransform(matrix);
 * if (((BoundingVolume *) sph)->FrustumCheck(*frustum))
 * {
 *		Render();
 * }
 * 
 * Hierarchy:
 * ----------
 * sph->AddVolume((BoundingVolume*) bb);
 * BoundingVolume *bvol = new BoundingVolume();
 * bvol->AddVolume(sph);
 * (bvol->sph->bb)
 * 
 * A bounding volume may be empty. In that case FrustumCheck returns true.
 *
 * How to add bounding volumes:
 * ----------------------------
 * 1) Derive your bvol class from BoundingVolume.
 * 2) DO overload FrustumCheck(Box& frustum, Matrix4x4& mat);
 * 3) DO NOT overload FrustumCheck(Box& frustum);
 * 
 * Memory management:
 * ------------------
 * When you use spheres, tetrahedra and boxes individually, use "new" to
 * create them and "delete" to free them:
 * Sphere *sph = new Sphere();
 * delete sph;
 *
 * But when you use bounding volume hierarchies, you must delete the parent
 * volume *only*!!!
 * BoundingVolume *bvol = new BoundingVolume();
 * Sphere *sph = new Sphere();
 * bvol->AddVolume(sph);		// From now on you shouldn't "delete" sph !!!
 * delete bvol;					// .. because "delete" frees the entire hierarchy !!!
 */
#ifndef _GEOM_H_
#define _GEOM_H_

#include <vector>
#include "vector.h"
#include "matrix.h"

enum Classifications
{
	CP_FRONT,
	CP_BACK,
	CP_ONPLANE,
	CP_SPLIT
};

// fwd declaration
class Box;

class Plane
{
protected:
	Vector3 pos, normal;
public:
	Plane();
	Plane(float a, float b, float c, float d);
	Plane(const Vector4 &v);
	Plane(const Vector3& normal, const Vector3& pos);

	Vector3 GetPos() const;
	void SetPos(const Vector3& pos);

	// this was inlined
	Vector3 GetNormal() const
	{
		return normal;
	}
	void SetNormal(const Vector3& normal);

	unsigned int ClassifyPoint (const Vector3 &point) const;

}; // end class Plane


// fwd declaration
class Ray;

class Line
{
protected:
	Vector3 verts[2];
public:
	Line();
	Line(const Vector3& a, const Vector3& b);

	// this was made inline
	Vector3 GetVertex(unsigned int vert) const
	{
		return verts[vert];
	}
	void SetVertex(unsigned int vert, const Vector3& vertex);

	Ray GetRay() const;
	float GetLength() const;

	bool Inside(const Vector3 &v) const;

	Vector3 GetMidpoint() const;

	Line GetReverse() const;
}; // end class Line


class Ray
{
protected:
	Vector3 origin, direction;
public:
	Ray();
	Ray(const Vector3& origin, const Vector3& direction);

	Vector3 GetOrigin() const;
	Vector3 GetDirection() const;
	void SetOrigin(const Vector3& origin);
	void SetDirection(const Vector3& direction);

	Vector3 GetPos(float dist) const;
	float GetDist(const Vector3 &pos) const;

	Vector3 Project(const Vector3 &v) const;
}; // end class Ray

class Triangle
{
protected:
	Vector3 verts[3];
public:
	Triangle();
	Triangle(const Vector3& a, const Vector3& b, const Vector3& c);

	Vector3 GetNormal() const;
	Vector3 GetVertex(unsigned int vert) const;
	void SetVertex(unsigned int vert, const Vector3& vertex);

	Plane GetPlane() const;

	// returnd a plane that intersects the edge given
	// the new plane has a right angle with the triangle's plane
	// The new plane's normal points inside the triangle
	Plane GetEdgePlane(unsigned int i) const;

	Line GetLine(unsigned int lin) const;

	Line GetSmallLine() const;
	Line GetMediumLine() const;
	Line GetLargeLine() const;

	Vector3 GetCenter() const;

	void Transform(const Matrix4x4 &xform);
}; // end class Triangle

class Quad
{
protected:
	Vector3 verts[4];
public:
	Quad();
	Quad(const Vector3& a, const Vector3& b, 
		const Vector3& c, const Vector3& d);

	Vector3 GetVertex(unsigned int vert) const;
	void SetVertex(unsigned int vert, const Vector3& vertex);

	// This was made inline
	Line GetLine(unsigned int lin) const
	{
		return Line(verts[lin], verts[(lin + 1) % 4]);
	}

	Triangle GetTriangle(unsigned int tri) const;
	Vector3 GetNormal() const;

	Plane GetPlane() const;

	// returnd a plane that intersects the edge given
	// the new plane has a right angle with the triangle's plane
	// The new plane's normal points inside the triangle
	Plane GetEdgePlane(unsigned int i) const;

	bool FrustumCheck(const Box &frustum) const;

	void Flip();
}; // end class Quad


class BoundingVolume
{
protected:
	std::vector<BoundingVolume*> children;
	Matrix4x4 transform;
public:
	BoundingVolume();
	virtual ~BoundingVolume();
	virtual bool FrustumCheck(const Box& frustum, const Matrix4x4& transform) const = 0;
	
	bool FrustumCheck(const Box& frustum) const;
	
	void SetTransform(const Matrix4x4& mat);
	void AddVolume(BoundingVolume* vol);

};

/* definition:
 *-------------
 *        0
 *        /\
 *       /  \
 *      / X  \
 *    2/______\1
 * 
 * 0-1-2 = trianglethat lies on the plane of the paper (or monitor :) )
 * X = vertex "behind the triangle" as seen from the viewer
 * X = 3
 * triangles are numbered:
 * 0 = 0-1-2
 * 1 = 0-X-1
 * 2 = 1-X-2
 * 3 = 2-X-0
 */
class Tetrahedron : public BoundingVolume
{
protected:
	Vector3 verts[4];
public:
	Tetrahedron();
	Tetrahedron(const Vector3& a, const Vector3& b, 
		const Vector3& c, const Vector3& d);

	Vector3 GetVertex(unsigned int vert) const;
	void SetVertex(unsigned int vert, const Vector3& vertex);

	Triangle GetTriangle(unsigned int tri) const;
	
	virtual bool FrustumCheck(const Box& frustum, const Matrix4x4& transform) const;
	void Transform(const Matrix4x4& mat);
}; // end class Tetrahedron

enum FrustumPlane
{
	FP_FRONT,
	FP_BACK,
	FP_LEFT,
	FP_RIGHT,
	FP_TOP,
	FP_BOTTOM
}; // end enum FrustumPlane

/* Box edge numbering:
 * 
 *       4+-------7+
 *       /        /|
 *     0+-------3+ |
 *      |        | |
 *      | 5      | |6
 *     1|_______2|/
 *
 * Quad numbering:
 * quad 0 = 0-1-2-3
 * quad 1 = 4-5-6-7
 */
class Box : public BoundingVolume
{
protected:
	Plane planes[6];
	Vector3 verts[8];

	void CalculatePlanes();
	void CalculateVertices();

public:
	Box();
	Box(const Plane& front, const Plane& back,
		const Plane& left, const Plane& right,
		const Plane& top, const Plane& bottom);
	Box(const Quad& front, const Quad& back);
	
	// Axis aligned box
	Box(const Vector3 &from, const Vector3 &to);

	// creates a bounding box for a set of vertices
	Box(const std::vector<Vector3> &verts);

	// 

	Quad GetQuad(unsigned int i) const;

	// OPT: this is used frequently, so I made it inline
	// and skipped the if (pln<6) step.
	Plane GetPlane(unsigned int pln) const
	{
		return planes[pln];
	}
	void SetPlane(unsigned int pln, const Plane& plane);

	// OPT: this is used frequently, so I made it inline
	// and skipped the if (vert<8) step.
	Vector3 GetVertex(unsigned int vert) const
	{
		return verts[vert];
	}
	void SetVertex(unsigned int vert, const Vector3& vertex);
	
	virtual bool FrustumCheck(const Box& frustum, const Matrix4x4& transform) const;
	void Transform(const Matrix4x4& mat);

	unsigned int ClassifyWithPlane (const Plane &plane) const;

	Vector3 GetViewer() const;

	void ConvertToPyramid(const Vector3 &v);

	Vector3 GetCenter() const;

}; // end class Box

class Sphere : public BoundingVolume
{
protected:
	float radius;
	Vector3 center;
public:
	Sphere();
	Sphere(const Vector3 center, float radius);

	float GetRadius() const;
	Vector3 GetCenter() const;

	void SetRadius(float radius);
	void SetCenter(const Vector3& center);

	virtual bool FrustumCheck(const Box& frustum, const Matrix4x4& transform) const;
	void Transform(const Matrix4x4& mat);
}; // end class Sphere

bool UnderPlane(const Vector3& point, const Plane& plane);
bool UnderPlane(const Triangle& triangle, const Plane& plane);
bool UnderPlane(const Quad& quad, const Plane& plane);
bool UnderPlane(const Tetrahedron& tetrahedron, const Plane& plane);
bool UnderPlane(const Box& box, const Plane& plane);
bool UnderPlane(const Sphere& sphere, const Plane& plane);

bool OverPlane(const Sphere &sphere, const Plane &plane);

bool Intersect(const Ray &ray, const Plane &plane);
bool Intersect(const Ray &ray, const Triangle &triangle);
bool Intersect(const Ray &ray, const Quad &quad);
bool Intersect(const Line& line, const Triangle& triangle);
bool Intersect(const Triangle& a, const Triangle& b);
bool Intersect(const Triangle& triangle, const Tetrahedron& tetrahedron);
bool Intersect(const Line& line, const Quad& quad);
bool Intersect(const Triangle& triangle, const Quad& quad);
bool Intersect(const Quad& a, const Quad& b);
bool Intersect(const Line& l, const Plane& p);
bool Intersect(const Sphere& s, const Plane& p);
bool Intersect(const Sphere &s, const Triangle &t);
bool Intersect(const Ray &ray, const Box &box);
bool Intersect(const Ray &ray, const Sphere &sphere);
bool Intersect(const Line &line, const Sphere &sphere);
bool Intersect(const Box &box1, const Box &box2);
bool Intersect(const Sphere &sphere, const Box &box);
bool Intersect(const Vector3 &point, const Box &box);
bool Intersect(const Quad &quad, const Box &box);

Ray RayOfIntersection(const Plane& a, const Plane& b);

Vector3 PointOfIntersection(const Ray& ray, const Plane& plane);
Vector3 PointOfIntersection(const Ray &ray, const Triangle &triangle);
Vector3 PointOfIntersection(const Plane& a, const Plane& b, const Plane& c);
Vector3 PointOfIntersection(const Ray &ray, const Sphere &sphere);

float DistanceFromPlane(const Vector3& point, const Plane& plane);

Vector3 ProjectToPlane(const Vector3& point, const Plane& plane);

Vector3 Project(const Vector3 &point, const Ray &ray);
float Distance(const Vector3 &point, const Ray &ray);

// Viewing frustum calculation
Box CalculateViewFrustum(const Matrix4x4 &view, const Matrix4x4 &proj);

Box ReduceFrustum(const Quad &portal, const Box &frustum);

bool IsPortalVisible(const Quad &portal, const Box &frustum, 
					 const Vector3 &viewer, Box &reduced_fr);

bool CompareRay(const Ray &r1, const Ray &r2);

Sphere Concatenate(const Sphere &a, const Sphere &b);

Sphere GetOutSphere(const Triangle &tri);

Sphere GetOutSphereOfCone(const Vector3 &vert, const Vector3 &center, float radius);

// smart box - a bounding box used for skinned meshes.
// it is calculated as a bounding box of the bones only
// and then is scaled by 25% . The scaling is not uniform
// in order to produce a more cube-like box.
Box SmartBox(const std::vector<Vector3> &points);

#endif // ndef _GEOM_H_
