#include <math.h>
#include <cstring>
#include "matrix.h"
#include "math3d.h"

using namespace std;

// ----------- Matrix3x3 --------------

Matrix3x3 Matrix3x3::identity = Matrix3x3(1, 0, 0, 0, 1, 0, 0, 0, 1);

Matrix3x3::Matrix3x3() {
	*this = identity;
}

Matrix3x3::Matrix3x3(	float m11, float m12, float m13,
						float m21, float m22, float m23,
						float m31, float m32, float m33) {
	m[0][0] = m11; m[0][1] = m12; m[0][2] = m13;
	m[1][0] = m21; m[1][1] = m22; m[1][2] = m23;
	m[2][0] = m31; m[2][1] = m32; m[2][2] = m33;
}

Matrix3x3::Matrix3x3(const Matrix4x4 &mat4x4) {
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			m[i][j] = mat4x4[i][j];
		}
	}
}

Matrix3x3 operator +(const Matrix3x3 &m1, const Matrix3x3 &m2) {
	Matrix3x3 res;
	const float *op1 = m1.m[0], *op2 = m2.m[0];
	float *dest = res.m[0];
	
	for(int i=0; i<9; i++) {
		*dest++ = *op1++ + *op2++;
	}
	return res;
}

Matrix3x3 operator -(const Matrix3x3 &m1, const Matrix3x3 &m2) {
	Matrix3x3 res;
	const float *op1 = m1.m[0], *op2 = m2.m[0];
	float *dest = res.m[0];
	
	for(int i=0; i<9; i++) {
		*dest++ = *op1++ - *op2++;
	}
	return res;
}

Matrix3x3 operator *(const Matrix3x3 &m1, const Matrix3x3 &m2) {
	Matrix3x3 res;
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			res.m[i][j] = m1.m[i][0] * m2.m[0][j] + m1.m[i][1] * m2.m[1][j] + m1.m[i][2] * m2.m[2][j];
		}
	}
	return res;
}

void operator +=(Matrix3x3 &m1, const Matrix3x3 &m2) {
	float *op1 = m1.m[0];
	const float *op2 = m2.m[0];
	
	for(int i=0; i<9; i++) {
		*op1++ += *op2++;
	}
}

void operator -=(Matrix3x3 &m1, const Matrix3x3 &m2) {
	float *op1 = m1.m[0];
	const float *op2 = m2.m[0];
	
	for(int i=0; i<9; i++) {
		*op1++ -= *op2++;
	}
}

void operator *=(Matrix3x3 &m1, const Matrix3x3 &m2) {
	Matrix3x3 res;
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			res.m[i][j] = m1.m[i][0] * m2.m[0][j] + m1.m[i][1] * m2.m[1][j] + m1.m[i][2] * m2.m[2][j];
		}
	}
	memcpy(m1.m, res.m, 9 * sizeof(float));
}

Matrix3x3 operator *(const Matrix3x3 &mat, float scalar) {
	Matrix3x3 res;
	const float *mptr = mat.m[0];
	float *dptr = res.m[0];
	
	for(int i=0; i<9; i++) {
		*dptr++ = *mptr++ * scalar;
	}
	return res;
}

Matrix3x3 operator *(float scalar, const Matrix3x3 &mat) {
	Matrix3x3 res;
	const float *mptr = mat.m[0];
	float *dptr = res.m[0];
	
	for(int i=0; i<9; i++) {
		*dptr++ = *mptr++ * scalar;
	}
	return res;
}

void operator *=(Matrix3x3 &mat, float scalar) {
	float *mptr = mat.m[0];
	
	for(int i=0; i<9; i++) {
		*mptr++ *= scalar;
	}
}

void Matrix3x3::Translate(const Vector2 &trans) {
	Matrix3x3 tmat(1, 0, trans.x, 0, 1, trans.y, 0, 0, 1);
	*this *= tmat;
}

void Matrix3x3::SetTranslation(const Vector2 &trans) {
	*this = Matrix3x3(1, 0, trans.x, 0, 1, trans.y, 0, 0, 1);
}

void Matrix3x3::Rotate(float angle) {
	float cos_a = cos(angle);
	float sin_a = sin(angle);
	Matrix3x3 rmat(	cos_a, 	-sin_a, 	0,
					sin_a, 	cos_a, 		0,
					0,		0,			1);
	*this *= rmat;
}

void Matrix3x3::SetRotation(float angle) {
	float cos_a = cos(angle);
	float sin_a = sin(angle);
	*this = Matrix3x3(cos_a, -sin_a, 0, sin_a, cos_a, 0, 0, 0, 1);
}

void Matrix3x3::Rotate(const Vector3 &euler_angles) {
	Matrix3x3 xrot, yrot, zrot;
	
	xrot = Matrix3x3(	1,			0,					0,
						0,	cos(euler_angles.x),	-sin(euler_angles.x),
						0,	sin(euler_angles.x),	cos(euler_angles.x));
	
	yrot = Matrix3x3(	cos(euler_angles.y),	0,	sin(euler_angles.y),
								0,				1,				0,
						-sin(euler_angles.y),	0,	cos(euler_angles.y));
	
	zrot = Matrix3x3(	cos(euler_angles.z),	-sin(euler_angles.z),	0,
						sin(euler_angles.z),	cos(euler_angles.z),	0,
								0,						0,				1);
	
	*this *= xrot * yrot * zrot;
}

void Matrix3x3::SetRotation(const Vector3 &euler_angles) {
	Matrix3x3 xrot, yrot, zrot;
	
	xrot = Matrix3x3(	1,			0,					0,
						0,	cos(euler_angles.x),	-sin(euler_angles.x),
						0,	sin(euler_angles.x),	cos(euler_angles.x));
	
	yrot = Matrix3x3(	cos(euler_angles.y),	0,	sin(euler_angles.y),
								0,				1,				0,
						-sin(euler_angles.y),	0,	cos(euler_angles.y));
	
	zrot = Matrix3x3(	cos(euler_angles.z),	-sin(euler_angles.z),	0,
						sin(euler_angles.z),	cos(euler_angles.z),	0,
								0,						0,				1);
	
	*this = xrot * yrot * zrot;
}

void Matrix3x3::Rotate(const Vector3 &axis, float angle) {
	float sina = (float)sin(angle);
	float cosa = (float)cos(angle);
	float invcosa = 1-cosa;
	float nxsq = axis.x * axis.x;
	float nysq = axis.y * axis.y;
	float nzsq = axis.z * axis.z;

	Matrix3x3 xform;
	xform.m[0][0] = nxsq + (1-nxsq) * cosa;
	xform.m[0][1] = axis.x * axis.y * invcosa - axis.z * sina;
	xform.m[0][2] = axis.x * axis.z * invcosa + axis.y * sina;
	xform.m[1][0] = axis.x * axis.y * invcosa + axis.z * sina;
	xform.m[1][1] = nysq + (1-nysq) * cosa;
	xform.m[1][2] = axis.y * axis.z * invcosa - axis.x * sina;
	xform.m[2][0] = axis.x * axis.z * invcosa - axis.y * sina;
	xform.m[2][1] = axis.y * axis.z * invcosa + axis.x * sina;
	xform.m[2][2] = nzsq + (1-nzsq) * cosa;

	*this *= xform;
}

void Matrix3x3::SetRotation(const Vector3 &axis, float angle) {
	float sina = (float)sin(angle);
	float cosa = (float)cos(angle);
	float invcosa = 1-cosa;
	float nxsq = axis.x * axis.x;
	float nysq = axis.y * axis.y;
	float nzsq = axis.z * axis.z;

	ResetIdentity();
	m[0][0] = nxsq + (1-nxsq) * cosa;
	m[0][1] = axis.x * axis.y * invcosa - axis.z * sina;
	m[0][2] = axis.x * axis.z * invcosa + axis.y * sina;
	m[1][0] = axis.x * axis.y * invcosa + axis.z * sina;
	m[1][1] = nysq + (1-nysq) * cosa;
	m[1][2] = axis.y * axis.z * invcosa - axis.x * sina;
	m[2][0] = axis.x * axis.z * invcosa - axis.y * sina;
	m[2][1] = axis.y * axis.z * invcosa + axis.x * sina;
	m[2][2] = nzsq + (1-nzsq) * cosa;
}

void Matrix3x3::Scale(const Vector3 &scale_vec) {
	Matrix3x3 smat(	scale_vec.x, 0, 0,
					0, scale_vec.y, 0,
					0, 0, scale_vec.z);
	*this *= smat;
}

void Matrix3x3::SetScaling(const Vector3 &scale_vec) {
	*this = Matrix3x3(	scale_vec.x, 0, 0,
						0, scale_vec.y, 0,
						0, 0, scale_vec.z);
}

void Matrix3x3::SetColumnVector(const Vector3 &vec, unsigned int col_index) {
	m[0][col_index] = vec.x;
	m[1][col_index] = vec.y;
	m[2][col_index] = vec.z;
}

void Matrix3x3::SetRowVector(const Vector3 &vec, unsigned int row_index) {
	m[row_index][0] = vec.x;
	m[row_index][1] = vec.y;
	m[row_index][2] = vec.z;
}

Vector3 Matrix3x3::GetColumnVector(unsigned int col_index) const {
	return Vector3(m[0][col_index], m[1][col_index], m[2][col_index]);
}

Vector3 Matrix3x3::GetRowVector(unsigned int row_index) const {
	return Vector3(m[row_index][0], m[row_index][1], m[row_index][2]);
}

void Matrix3x3::Transpose() {
	Matrix3x3 tmp = *this;
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			m[i][j] = tmp[j][i];
		}
	}
}

Matrix3x3 Matrix3x3::Transposed() const {
	Matrix3x3 res;
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			res[i][j] = m[j][i];
		}
	}
	return res;
}

float Matrix3x3::Determinant() const {
	return 	m[0][0] * (m[1][1]*m[2][2] - m[1][2]*m[2][1]) -
			m[0][1] * (m[1][0]*m[2][2] - m[1][2]*m[2][0]) +
			m[0][2] * (m[1][0]*m[2][1] - m[1][1]*m[2][0]);
}

Matrix3x3 Matrix3x3::Inverse() const {
	// TODO: implement 3x3 inverse
	return *this;
}

ostream &operator <<(ostream &out, const Matrix3x3 &mat) {
	for(int i=0; i<3; i++) {
		char str[100];
		sprintf(str, "[ %12.5f %12.5f %12.5f ]\n", (float)mat.m[i][0], (float)mat.m[i][1], (float)mat.m[i][2]);
		out << str;
	}
	return out;
}



////////////////////// Matrix4x4 implementation ///////////////////////


Matrix4x4 Matrix4x4::identity = Matrix4x4(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

Matrix4x4::Matrix4x4() {
	*this = identity;
}

Matrix4x4::Matrix4x4(	float m11, float m12, float m13, float m14,
						float m21, float m22, float m23, float m24,
						float m31, float m32, float m33, float m34,
						float m41, float m42, float m43, float m44) {
	m[0][0] = m11; m[0][1] = m12; m[0][2] = m13; m[0][3] = m14;
	m[1][0] = m21; m[1][1] = m22; m[1][2] = m23; m[1][3] = m24;
	m[2][0] = m31; m[2][1] = m32; m[2][2] = m33; m[2][3] = m34;
	m[3][0] = m41; m[3][1] = m42; m[3][2] = m43; m[3][3] = m44;
}

Matrix4x4::Matrix4x4(const Matrix3x3 &mat3x3) {
	ResetIdentity();
	for(int i=0; i<3; i++) {
		for(int j=0; j<3; j++) {
			m[i][j] = mat3x3[i][j];
		}
	}
}

Matrix4x4 operator +(const Matrix4x4 &m1, const Matrix4x4 &m2) {
	Matrix4x4 res;
	const float *op1 = m1.m[0], *op2 = m2.m[0];
	float *dest = res.m[0];
	
	for(int i=0; i<16; i++) {
		*dest++ = *op1++ + *op2++;
	}
	return res;
}

Matrix4x4 operator -(const Matrix4x4 &m1, const Matrix4x4 &m2) {
	Matrix4x4 res;
	const float *op1 = m1.m[0], *op2 = m2.m[0];
	float *dest = res.m[0];
	
	for(int i=0; i<16; i++) {
		*dest++ = *op1++ - *op2++;
	}
	return res;
}

Matrix4x4 operator *(const Matrix4x4 &m1, const Matrix4x4 &m2) {
	Matrix4x4 res;
	
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			res.m[i][j] = m1.m[i][0] * m2.m[0][j] + m1.m[i][1] * m2.m[1][j] + m1.m[i][2] * m2.m[2][j] + m1.m[i][3] * m2.m[3][j];
		}
	}

	return res;
}

void operator +=(Matrix4x4 &m1, const Matrix4x4 &m2) {
	float *op1 = m1.m[0];
	const float *op2 = m2.m[0];
	
	for(int i=0; i<16; i++) {
		*op1++ += *op2++;
	}
}

void operator -=(Matrix4x4 &m1, const Matrix4x4 &m2) {
	float *op1 = m1.m[0];
	const float *op2 = m2.m[0];
	
	for(int i=0; i<16; i++) {
		*op1++ -= *op2++;
	}
}

void operator *=(Matrix4x4 &m1, const Matrix4x4 &m2) {
	Matrix4x4 res;
	
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			res.m[i][j] = m1.m[i][0] * m2.m[0][j] + m1.m[i][1] * m2.m[1][j] + m1.m[i][2] * m2.m[2][j] + m1.m[i][3] * m2.m[3][j];
		}
	}

	memcpy(m1.m, res.m, 16 * sizeof(float));
}

Matrix4x4 operator *(const Matrix4x4 &mat, float scalar) {
	Matrix4x4 res;
	const float *mptr = mat.m[0];
	float *dptr = res.m[0];
	
	for(int i=0; i<16; i++) {
		*dptr++ = *mptr++ * scalar;
	}
	return res;
}

Matrix4x4 operator *(float scalar, const Matrix4x4 &mat) {
	Matrix4x4 res;
	const float *mptr = mat.m[0];
	float *dptr = res.m[0];
	
	for(int i=0; i<16; i++) {
		*dptr++ = *mptr++ * scalar;
	}
	return res;
}

void operator *=(Matrix4x4 &mat, float scalar) {
	float *mptr = mat.m[0];
	
	for(int i=0; i<16; i++) {
		*mptr++ *= scalar;
	}
}

void Matrix4x4::Translate(const Vector3 &t) {
	Matrix4x4 tmat(1, 0, 0, t.x, 0, 1, 0, t.y, 0, 0, 1, t.z, 0, 0, 0, 1);
	*this *= tmat;
}

void Matrix4x4::SetTranslation(const Vector3 &t) {
	*this = Matrix4x4(1, 0, 0, t.x, 0, 1, 0, t.y, 0, 0, 1, t.z, 0, 0, 0, 1);
}

void Matrix4x4::Rotate(const Vector3 &euler) {
	Matrix4x4 xrot, yrot, zrot;
	
	xrot = Matrix4x4(	1,		0,				0, 			0,
						0,	cos(euler.x),	-sin(euler.x),	0,
						0,	sin(euler.x),	cos(euler.x),	0,
		 				0,		0,				0,			1);
	
	yrot = Matrix4x4(	cos(euler.y),	0,	sin(euler.y),	0,
								0,		1,		0,			0,
						-sin(euler.y),	0,	cos(euler.y),	0,
								0,		0,		0,			1);
	
	zrot = Matrix4x4(	cos(euler.z),	-sin(euler.z),	0,	0,
						sin(euler.z),	cos(euler.z),	0,	0,
								0,			0,			1,	0,
								0,			0,			0,	1);
	
	*this *= xrot * yrot * zrot;
}

void Matrix4x4::SetRotation(const Vector3 &euler) {
	Matrix4x4 xrot, yrot, zrot;

	xrot = Matrix4x4(	1,		0,				0, 			0,
						0,	cos(euler.x),	-sin(euler.x),	0,
						0,	sin(euler.x),	cos(euler.x),	0,
		 				0,		0,				0,			1);
	
	yrot = Matrix4x4(	cos(euler.y),	0,	sin(euler.y),	0,
								0,		1,		0,			0,
						-sin(euler.y),	0,	cos(euler.y),	0,
								0,		0,		0,			1);
	
	zrot = Matrix4x4(	cos(euler.z),	-sin(euler.z),	0,	0,
						sin(euler.z),	cos(euler.z),	0,	0,
								0,			0,			1,	0,
								0,			0,			0,	1);
	
	*this = xrot * yrot * zrot;
}

void Matrix4x4::Rotate(const Vector3 &axis, float angle) {
	float sina = sin(angle);
	float cosa = cos(angle);
	float invcosa = 1-cosa;
	float nxsq = axis.x * axis.x;
	float nysq = axis.y * axis.y;
	float nzsq = axis.z * axis.z;

	Matrix4x4 xform;
	xform[0][0] = nxsq + (1-nxsq) * cosa;
	xform[0][1] = axis.x * axis.y * invcosa - axis.z * sina;
	xform[0][2] = axis.x * axis.z * invcosa + axis.y * sina;
	xform[1][0] = axis.x * axis.y * invcosa + axis.z * sina;
	xform[1][1] = nysq + (1-nysq) * cosa;
	xform[1][2] = axis.y * axis.z * invcosa - axis.x * sina;
	xform[2][0] = axis.x * axis.z * invcosa - axis.y * sina;
	xform[2][1] = axis.y * axis.z * invcosa + axis.x * sina;
	xform[2][2] = nzsq + (1-nzsq) * cosa;

	*this *= xform;
}

void Matrix4x4::SetRotation(const Vector3 &axis, float angle) {
	float sina = sin(angle);
	float cosa = cos(angle);
	float invcosa = 1.0f-cosa;
	float nxsq = axis.x * axis.x;
	float nysq = axis.y * axis.y;
	float nzsq = axis.z * axis.z;

	ResetIdentity();
	m[0][0] = nxsq + (1-nxsq) * cosa;
	m[0][1] = axis.x * axis.y * invcosa - axis.z * sina;
	m[0][2] = axis.x * axis.z * invcosa + axis.y * sina;
	m[1][0] = axis.x * axis.y * invcosa + axis.z * sina;
	m[1][1] = nysq + (1-nysq) * cosa;
	m[1][2] = axis.y * axis.z * invcosa - axis.x * sina;
	m[2][0] = axis.x * axis.z * invcosa - axis.y * sina;
	m[2][1] = axis.y * axis.z * invcosa + axis.x * sina;
	m[2][2] = nzsq + (1-nzsq) * cosa;
}

void Matrix4x4::Scale(const Vector4 &s) {
	Matrix4x4 smat(	s.x, 0, 0, 0,
					0, s.y, 0, 0,
					0, 0, s.z, 0,
					0, 0, 0, s.w);
	*this *= smat;
}

void Matrix4x4::SetScaling(const Vector4 &s) {
	*this = Matrix4x4(	s.x, 0, 0, 0,
						0, s.y, 0, 0,
						0, 0, s.z, 0,
						0, 0, 0, s.w);
}

void Matrix4x4::SetColumnVector(const Vector4 &vec, unsigned int col) {
	m[0][col] = vec.x;
	m[1][col] = vec.y;
	m[2][col] = vec.z;
	m[3][col] = vec.w;
}

void Matrix4x4::SetRowVector(const Vector4 &vec, unsigned int row) {
	m[row][0] = vec.x;
	m[row][1] = vec.y;
	m[row][2] = vec.z;
	m[row][3] = vec.w;
}

Vector4 Matrix4x4::GetColumnVector(unsigned int col) const {
	return Vector4(m[0][col], m[1][col], m[2][col], m[3][col]);
}

Vector4 Matrix4x4::GetRowVector(unsigned int row) const {
	return Vector4(m[row][0], m[row][1], m[row][2], m[row][3]);
}

void Matrix4x4::Transpose() {
	Matrix4x4 tmp = *this;
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			m[i][j] = tmp[j][i];
		}
	}
}

Matrix4x4 Matrix4x4::Transposed() const {
	Matrix4x4 res;
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			res[i][j] = m[j][i];
		}
	}
	return res;
}

float Matrix4x4::Determinant() const {

	float det11 =	(m[1][1] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
						(m[1][2] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) +
						(m[1][3] * (m[2][1] * m[3][2] - m[3][1] * m[2][2]));

	float det12 =	(m[1][0] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
						(m[1][2] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
						(m[1][3] * (m[2][0] * m[3][2] - m[3][0] * m[2][2]));

	float det13 =	(m[1][0] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) -
						(m[1][1] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
						(m[1][3] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));

	float det14 =	(m[1][0] * (m[2][1] * m[3][2] - m[3][1] * m[2][2])) -
						(m[1][1] * (m[2][0] * m[3][2] - m[3][0] * m[2][2])) +
						(m[1][2] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));

	return m[0][0] * det11 - m[0][1] * det12 + m[0][2] * det13 - m[0][3] * det14;
}


Matrix4x4 Matrix4x4::Adjoint() const {

	Matrix4x4 coef;

	coef.m[0][0] =	(m[1][1] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
					(m[1][2] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) +
					(m[1][3] * (m[2][1] * m[3][2] - m[3][1] * m[2][2]));
	coef.m[0][1] =	(m[1][0] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
					(m[1][2] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
					(m[1][3] * (m[2][0] * m[3][2] - m[3][0] * m[2][2]));
	coef.m[0][2] =	(m[1][0] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) -
					(m[1][1] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
					(m[1][3] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));
	coef.m[0][3] =	(m[1][0] * (m[2][1] * m[3][2] - m[3][1] * m[2][2])) -
					(m[1][1] * (m[2][0] * m[3][2] - m[3][0] * m[2][2])) +
					(m[1][2] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));

	coef.m[1][0] =	(m[0][1] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
					(m[0][2] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) +
					(m[0][3] * (m[2][1] * m[3][2] - m[3][1] * m[2][2]));
	coef.m[1][1] =	(m[0][0] * (m[2][2] * m[3][3] - m[3][2] * m[2][3])) -
					(m[0][2] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
					(m[0][3] * (m[2][0] * m[3][2] - m[3][0] * m[2][2]));
	coef.m[1][2] =	(m[0][0] * (m[2][1] * m[3][3] - m[3][1] * m[2][3])) -
					(m[0][1] * (m[2][0] * m[3][3] - m[3][0] * m[2][3])) +
					(m[0][3] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));
	coef.m[1][3] =	(m[0][0] * (m[2][1] * m[3][2] - m[3][1] * m[2][2])) -
					(m[0][1] * (m[2][0] * m[3][2] - m[3][0] * m[2][2])) +
					(m[0][2] * (m[2][0] * m[3][1] - m[3][0] * m[2][1]));

	coef.m[2][0] =	(m[0][1] * (m[1][2] * m[3][3] - m[3][2] * m[1][3])) -
					(m[0][2] * (m[1][1] * m[3][3] - m[3][1] * m[1][3])) +
					(m[0][3] * (m[1][1] * m[3][2] - m[3][1] * m[1][2]));
	coef.m[2][1] =	(m[0][0] * (m[1][2] * m[3][3] - m[3][2] * m[1][3])) -
					(m[0][2] * (m[1][0] * m[3][3] - m[3][0] * m[1][3])) +
					(m[0][3] * (m[1][0] * m[3][2] - m[3][0] * m[1][2]));
	coef.m[2][2] =	(m[0][0] * (m[1][1] * m[3][3] - m[3][1] * m[1][3])) -
					(m[0][1] * (m[1][0] * m[3][3] - m[3][0] * m[1][3])) +
					(m[0][3] * (m[1][0] * m[3][1] - m[3][0] * m[1][1]));
	coef.m[2][3] =	(m[0][0] * (m[1][1] * m[3][2] - m[3][1] * m[1][2])) -
					(m[0][1] * (m[1][0] * m[3][2] - m[3][0] * m[1][2])) +
					(m[0][2] * (m[1][0] * m[3][1] - m[3][0] * m[1][1]));

	coef.m[3][0] =	(m[0][1] * (m[1][2] * m[2][3] - m[2][2] * m[1][3])) -
					(m[0][2] * (m[1][1] * m[2][3] - m[2][1] * m[1][3])) +
					(m[0][3] * (m[1][1] * m[2][2] - m[2][1] * m[1][2]));
	coef.m[3][1] =	(m[0][0] * (m[1][2] * m[2][3] - m[2][2] * m[1][3])) -
					(m[0][2] * (m[1][0] * m[2][3] - m[2][0] * m[1][3])) +
					(m[0][3] * (m[1][0] * m[2][2] - m[2][0] * m[1][2]));
	coef.m[3][2] =	(m[0][0] * (m[1][1] * m[2][3] - m[2][1] * m[1][3])) -
					(m[0][1] * (m[1][0] * m[2][3] - m[2][0] * m[1][3])) +
					(m[0][3] * (m[1][0] * m[2][1] - m[2][0] * m[1][1]));
	coef.m[3][3] =	(m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2])) -
					(m[0][1] * (m[1][0] * m[2][2] - m[2][0] * m[1][2])) +
					(m[0][2] * (m[1][0] * m[2][1] - m[2][0] * m[1][1]));

	coef.Transpose();

	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			coef.m[i][j] = j%2 ? -coef.m[i][j] : coef.m[i][j];
			if(i%2) coef.m[i][j] = -coef.m[i][j];
		}
	}

	return coef;
}

Matrix4x4 Matrix4x4::Inverse() const {

	Matrix4x4 AdjMat = Adjoint();

	return AdjMat * (1.0f / Determinant());
}

const float *Matrix4x4::GetGLMatrix() const {
	return (const float*)m;
}

std::ostream &operator <<(std::ostream &out, const Matrix4x4 &mat) {
	for(int i=0; i<4; i++) {
		//char str[100];
		//sprintf(str, "[ %12.5f %12.5f %12.5f %12.5f ]\n", (float)mat.m[i][0], (float)mat.m[i][1], (float)mat.m[i][2], (float)mat.m[i][3]);
		//out << str;
		Vector4 row = mat.GetRowVector(i);
		cout << row << std::endl;
	}
	return out;
}

Matrix4x4 PerspectiveProj(float fov, float aspect, float znear, float zfar)
{
    Matrix4x4 m;

    float h = 1.0f / tan (DEG_TO_RAD(fov)/2.0f);
    float w = h / aspect;

    m[0][0] = w;
    m[1][1] = h;
    m[2][2] = zfar / (zfar - znear);
    m[3][2] = 1.0f;
    m[2][3] = -znear * zfar / (zfar - znear);
    m[3][3] = 0.0f;

    return m;
}

Matrix4x4 OrthoProj(float xmin, float xmax, float ymin, float ymax, float zmin, float zmax)
{
	float w = xmax - xmin;
	float h = ymax - ymin;
	float d = zmax - zmin;
	Matrix4x4 ret(
		2/w,		0,			0,				-1.0f,
		0,			-2/h,		0,				1.0f,
		0,			0,			1/d,			0,
		0,			0,			0/*-zmin/d*/,	1);

	return ret;
}
#include <stdio.h>
void PrintMatrix(const Matrix4x4 &mat)
{
	printf("[MATRIX]\n");
	for (unsigned int i=0; i<4; i++)
	{
		Vector4 v = mat.GetRowVector(i);
		printf("[%f, %f, %f, %f]\n", v.x, v.y, v.z, v.w);
	}
}

// prs decomposition
Vector3 Matrix4x4::ExtractTranslation() const
{
	Vector3 sc = ExtractScaling();
	Vector3 ret = GetColumnVector(3);

	if (fabs(sc.x) > 0.00000001f) ret.x /= sc.x;
	if (fabs(sc.y) > 0.00000001f) ret.y /= sc.y;
	if (fabs(sc.z) > 0.00000001f) ret.z /= sc.z;

	return ret;
}
	
Vector3 Matrix4x4::ExtractScaling() const
{
	return Vector3(GetColumnVector(0).Length(), 
		GetColumnVector(1).Length(), GetColumnVector(2).Length());
}

Matrix3x3 Matrix4x4::ExtractRotation() const
{
	Vector3 col[3];
	for (int i=0; i<3; i++)
	{
		col[i] = GetColumnVector(i);
	}

	Vector3 sc = ExtractScaling();
	if (fabsf(sc.x) > 0.00000001f)	col[0] /= sc.x;
	if (fabsf(sc.y) > 0.00000001f)	col[1] /= sc.y;
	if (fabsf(sc.z) > 0.00000001f)	col[2] /= sc.z;

	Matrix3x3 ret;
	for (int i=0; i<3; i++)
	{
		ret.SetColumnVector(col[i], i);
	}

	return ret;
}
