#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <iostream>
#include <cstring>
#include "vector.h"

class Matrix3x3 {
private:
	float m[3][3];
public:
	
	static Matrix3x3 identity;

	Matrix3x3();
	Matrix3x3(	float m11, float m12, float m13,
				float m21, float m22, float m23,
				float m31, float m32, float m33);
	
	Matrix3x3(const Matrix4x4 &mat4x4);
	
	// binary operations matrix (op) matrix
	friend Matrix3x3 operator +(const Matrix3x3 &m1, const Matrix3x3 &m2);
	friend Matrix3x3 operator -(const Matrix3x3 &m1, const Matrix3x3 &m2);
	friend Matrix3x3 operator *(const Matrix3x3 &m1, const Matrix3x3 &m2);
	
	friend void operator +=(Matrix3x3 &m1, const Matrix3x3 &m2);
	friend void operator -=(Matrix3x3 &m1, const Matrix3x3 &m2);
	friend void operator *=(Matrix3x3 &m1, const Matrix3x3 &m2);
	
	// binary operations matrix (op) scalar and scalar (op) matrix
	friend Matrix3x3 operator *(const Matrix3x3 &mat, float scalar);
	friend Matrix3x3 operator *(float scalar, const Matrix3x3 &mat);
	
	friend void operator *=(Matrix3x3 &mat, float scalar);
	
	inline float *operator [](int index);
	inline const float *operator [](int index) const;
	
	inline void ResetIdentity();
	
	void Translate(const Vector2 &trans);
	void SetTranslation(const Vector2 &trans);
	
	void Rotate(float angle);
	void Rotate(const Vector3 &euler_angles);
	void Rotate(const Vector3 &axis, float angle);
	void SetRotation(float angle);
	void SetRotation(const Vector3 &euler_angles);
	void SetRotation(const Vector3 &axis, float angle);
	
	void Scale(const Vector3 &scale_vec);
	void SetScaling(const Vector3 &scale_vec);
	
	void SetColumnVector(const Vector3 &vec, unsigned int col_index);
	void SetRowVector(const Vector3 &vec, unsigned int row_index);
	Vector3 GetColumnVector(unsigned int col_index) const;
	Vector3 GetRowVector(unsigned int row_index) const;

	void Transpose();
	Matrix3x3 Transposed() const;	
	float Determinant() const;
	Matrix3x3 Inverse() const;
	
	friend std::ostream &operator <<(std::ostream &out, const Matrix3x3 &mat);
};


class Matrix4x4 {
private:
	float m[4][4];

public:
	static Matrix4x4 identity;

	Matrix4x4();
	Matrix4x4(	float m11, float m12, float m13, float m14,
				float m21, float m22, float m23, float m24,
				float m31, float m32, float m33, float m34,
				float m41, float m42, float m43, float m44);

	Matrix4x4(const Matrix3x3 &mat3x3);

	// binary operations matrix (op) matrix
	friend Matrix4x4 operator +(const Matrix4x4 &m1, const Matrix4x4 &m2);
	friend Matrix4x4 operator -(const Matrix4x4 &m1, const Matrix4x4 &m2);
	friend Matrix4x4 operator *(const Matrix4x4 &m1, const Matrix4x4 &m2);
	
	friend void operator +=(Matrix4x4 &m1, const Matrix4x4 &m2);
	friend void operator -=(Matrix4x4 &m1, const Matrix4x4 &m2);
	friend void operator *=(Matrix4x4 &m1, const Matrix4x4 &m2);
	
	// binary operations matrix (op) scalar and scalar (op) matrix
	friend Matrix4x4 operator *(const Matrix4x4 &mat, float scalar);
	friend Matrix4x4 operator *(float scalar, const Matrix4x4 &mat);
	
	friend void operator *=(Matrix4x4 &mat, float scalar);
	
	inline float *operator [](int index);
	inline const float *operator [](int index) const;
	
	inline void ResetIdentity();
	
	void Translate(const Vector3 &t);
	void SetTranslation(const Vector3 &t);
	
	void Rotate(const Vector3 &euler);		// 3d rotation with euler angles
	void Rotate(const Vector3 &axis, float angle);	// 3d axis/angle rotation
	void SetRotation(const Vector3 &euler);
	void SetRotation(const Vector3 &axis, float angle);
	
	void Scale(const Vector4 &s);
	void SetScaling(const Vector4 &s);
	
	void SetColumnVector(const Vector4 &vec, unsigned int col);
	void SetRowVector(const Vector4 &vec, unsigned int row);
	Vector4 GetColumnVector(unsigned int col) const;
	Vector4 GetRowVector(unsigned int row) const;
	
	void Transpose();
	Matrix4x4 Transposed() const;
	float Determinant() const;
	Matrix4x4 Adjoint() const;
	Matrix4x4 Inverse() const;

	// prs decomposition
	Vector3 ExtractTranslation() const;
	Vector3 ExtractScaling() const;
	Matrix3x3 ExtractRotation() const;
	
	const float *GetGLMatrix() const;
		
	friend std::ostream &operator <<(std::ostream &out, const Matrix4x4 &mat);
};

Matrix4x4 PerspectiveProj(float fov, float aspect, float znear, float zfar);
Matrix4x4 OrthoProj(float xmin, float xmax, float ymin, float ymax, float zmin, float zmax);
void PrintMatrix(const Matrix4x4 &mat);

#include "matrix.inl"

#endif	// _MATRIX_H_
