#include "quaternion.h"
#include "matrix.h"
#include "math3d.h"

using namespace std; // for C++Builder compatibility

Quaternion::Quaternion() {
	s = 1.0;
	v.x = v.y = v.z = 0.0;
}

Quaternion::Quaternion(float s, float x, float y, float z) {
	v.x = x;
	v.y = y;
	v.z = z;
	this->s = s;
}

Quaternion::Quaternion(float s, const Vector3 &v) {
	this->s = s;
	this->v = v;
}

Quaternion::Quaternion(const Vector3 &axis, float angle) {
	SetRotation(axis, angle);
}

// quaternion that represents the angle between two vectors
Quaternion::Quaternion(const Vector3 &a, const Vector3 &b)
{
	Vector3 v1 = a.Normalized();
	Vector3 v2 = b.Normalized();
	float dot = DotProduct(v1, v2);
	if (fabs(dot) > 0.99999999)
	{
		SetRotation(Vector3(1, 0, 0), 0);
		return;
	}
	Vector3 axis = CrossProduct(v1, v2).Normalized();
	float angle = acos(dot);
	SetRotation(axis, angle);
}

Quaternion::Quaternion(const Matrix3x3 &rot_mat)
{
	float	trace = rot_mat[0][0] + rot_mat[1][1] + rot_mat[2][2] + 1.0f;
	float	xs;

	if (trace > 0.0000001f)
	{
		xs = 0.5f/sqrtf(trace);
		s = 0.25f/xs;
		v.x = (rot_mat[2][1] - rot_mat[1][2])*xs;
		v.y = (rot_mat[0][2] - rot_mat[2][0])*xs;
		v.z = (rot_mat[1][0] - rot_mat[0][1])*xs;
	}
	else
	{
		if (rot_mat[0][0] > rot_mat[1][1] && rot_mat[0][0] > rot_mat[2][2])
		{
			xs = 2.0f*sqrtf(1.0f + rot_mat[0][0] - rot_mat[1][1] - rot_mat[2][2]);
			s = (rot_mat[1][2] - rot_mat[2][1])/xs;
			v.x = 0.25f*xs;
			v.y = (rot_mat[0][1] + rot_mat[1][0])/xs;
			v.z = (rot_mat[0][2] + rot_mat[2][0])/xs;
		}
		else if (rot_mat[1][1] > rot_mat[2][2]) {
			xs = 2.0f*sqrtf(1.0f + rot_mat[1][1] - rot_mat[0][0] - rot_mat[2][2]);
			s = (rot_mat[0][2] - rot_mat[2][0])/xs;
			v.x = (rot_mat[0][1] + rot_mat[1][0])/xs;
			v.y = 0.25f*xs;
			v.z = (rot_mat[1][2] + rot_mat[2][1])/xs;
		} else {
			xs = 2.0f*sqrtf(1.0f + rot_mat[2][2] - rot_mat[0][0] - rot_mat[1][1]);
			s = (rot_mat[0][1] - rot_mat[1][0])/xs;
			v.x = (rot_mat[0][2] + rot_mat[2][0])/xs;
			v.y = (rot_mat[1][2] + rot_mat[2][1])/xs;
			v.z = 0.25f*xs;
		}
	}
}

Quaternion Quaternion::operator +(const Quaternion &quat) const {
	return Quaternion(s + quat.s, v + quat.v);
}

Quaternion Quaternion::operator -(const Quaternion &quat) const {
	return Quaternion(s - quat.s, v - quat.v);
}

Quaternion Quaternion::operator -() const {
	return Quaternion(-s, -v);
}

/** Quaternion Multiplication:
 * Q1*Q2 = [s1*s2 - v1.v2,  s1*v2 + s2*v1 + v1(x)v2]
 */
Quaternion Quaternion::operator *(const Quaternion &quat) const {
	Quaternion newq;	
	newq.s = s * quat.s - DotProduct(v, quat.v);
	newq.v = quat.v * s + v * quat.s + CrossProduct(v, quat.v);	
	return newq;
}

void Quaternion::operator +=(const Quaternion &quat) {
	*this = Quaternion(s + quat.s, v + quat.v);
}

void Quaternion::operator -=(const Quaternion &quat) {
	*this = Quaternion(s - quat.s, v - quat.v);
}

void Quaternion::operator *=(const Quaternion &quat) {
	*this = *this * quat;
}

void Quaternion::ResetIdentity() {
	s = 1.0;
	v.x = v.y = v.z = 0.0;
}

Quaternion Quaternion::Conjugate() const {
	return Quaternion(s, -v);
}

float Quaternion::Length() const {
	return (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
}

/** Q * ~Q = ||Q||^2 */
float Quaternion::LengthSq() const {
	return v.x*v.x + v.y*v.y + v.z*v.z + s*s;
}

void Quaternion::Normalize() {
	float len = (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
	v.x /= len;
	v.y /= len;
	v.z /= len;
	s /= len;
}

Quaternion Quaternion::Normalized() const {
	Quaternion nq = *this;
	float len = (float)sqrt(v.x*v.x + v.y*v.y + v.z*v.z + s*s);
	nq.v.x /= len;
	nq.v.y /= len;
	nq.v.z /= len;
	nq.s /= len;
	return nq;
}

/** Quaternion Inversion: Q^-1 = ~Q / ||Q||^2 */
Quaternion Quaternion::Inverse() const {
	Quaternion inv = Conjugate();
	float lensq = LengthSq();
	inv.v /= lensq;
	inv.s /= lensq;

	return inv;
}


void Quaternion::SetRotation(const Vector3 &axis, float angle) {
	float half_angle = angle / 2.0f;
	s = cos(half_angle);
	v = axis * sin(half_angle);
}

void Quaternion::Rotate(const Vector3 &axis, float angle) {
	Quaternion q;
	float half_angle = angle / 2.0f;
	q.s = cos(half_angle);
	q.v = axis * sin(half_angle);

	*this *= q;
}

void Quaternion::Rotate(const Quaternion &q) {
	*this = q * *this * q.Conjugate();
}

Matrix3x3 Quaternion::GetRotationMatrix() const {
	return Matrix3x3(	1.0f - 2.0f * v.y*v.y - 2.0f * v.z*v.z,	2.0f * v.x  * v.y + 2.0f * s * v.z,		2.0f * v.z * v.x - 2.0f * s * v.y,
						2.0f * v.x  * v.y - 2.0f * s * v.z,		1.0f - 2.0f * v.x*v.x - 2.0f * v.z*v.z,	2.0f * v.y * v.z + 2.0f * s * v.x,
						2.0f * v.z  * v.x + 2.0f * s * v.y,		2.0f * v.y  * v.z - 2.0f * s * v.x,		1.0f - 2.0f * v.x*v.x - 2.0f * v.y*v.y);
}


/** Spherical linear interpolation (slerp) */
Quaternion Slerp(const Quaternion &qua1, const Quaternion &q2, float t)
{
	Quaternion q1 = qua1;
	float s1, s2;
	float dot = DotProduct(Vector4(q1.s, q1.v.x, q1.v.y, q1.v.z), Vector4(q2.s, q2.v.x, q2.v.y, q2.v.z));

	if(fabs(dot) < XSmallNumber) {
		return q1;
	}
	if(dot < 0.0) {
		q1.v = -q1.v;
		q1.s = -q1.s;
		dot = -dot;
	}

	if((1.0 - dot) >= 0.05) {
		float angle = acos(dot);
		s1 = sin(angle * (1.0f - t)) / sin(angle);
		s2 = sin(angle * t) / sin(angle);
	} else {
		s1 = 1.0f - t;
		s2 = t;
	}

	Quaternion res;
	res.v.x = q1.v.x * s1 + q2.v.x * s2;
	res.v.y = q1.v.y * s1 + q2.v.y * s2;
	res.v.z = q1.v.z * s1 + q2.v.z * s2;
	res.s = q1.s * s1 + q2.s * s2;
	return res;
}
	

float DotProduct(const Quaternion &a, const Quaternion &b)
{
	return a.s * b.s + a.v.x * b.v.x + a.v.y * b.v.y + a.v.z * b.v.z;
}


std::ostream &operator <<(std::ostream &out, const Quaternion &q) {
	out << "(" << q.s << ", " << q.v << ")";
	return out;
}
