#include "vector.h"
#include "quaternion.h"
#include "matrix.h"
#include "sphvec.h"
#include "math3d.h"

using namespace std; // for C++ builder compatibility

// ---------- Vector2 -----------

Vector2::Vector2(float x, float y) {
	this->x = x;
	this->y = y;
}

Vector2::Vector2(const Vector3 &vec) {
	x = vec.x;
	y = vec.y;
}

Vector2::Vector2(const Vector4 &vec) {
	x = vec.x;
	y = vec.y;
}

void Vector2::Normalize() {
	float len = Length();
	x /= len;
	y /= len;
}

Vector2 Vector2::Normalized() const {	
	float len = Length();
	return Vector2(x / len, y / len);
}

void Vector2::Transform(const Matrix3x3 &mat) {
	float nx = mat[0][0] * x + mat[0][1] * y + mat[0][2];
	y = mat[1][0] * x + mat[1][1] * y + mat[1][2];
	x = nx;
}

Vector2 Vector2::Transformed(const Matrix3x3 &mat) const {
	Vector2 vec;
	vec.x = mat[0][0] * x + mat[0][1] * y + mat[0][2];
	vec.y = mat[1][0] * x + mat[1][1] * y + mat[1][2];
	return vec;
}

void Vector2::Rotate(float angle) {
	*this = Vector2(cos(angle) * x - sin(angle) * y, sin(angle) * x + cos(angle) * y);
}

Vector2 Vector2::Rotated(float angle) const {
	return Vector2(cos(angle) * x - sin(angle) * y, sin(angle) * x + cos(angle) * y);
}
	
Vector2 Vector2::Reflection(const Vector2 &normal) const {
	return 2.0f * DotProduct(*this, normal) * normal - *this;
}

Vector2 Vector2::Refraction(const Vector2 &normal, float src_ior, float dst_ior) const {
	// quick and dirty implementation :)
	Vector3 v3refr = Vector3(this->x, this->y, 1.0f).Refraction(Vector3(this->x, this->y, 1), src_ior, dst_ior);
	return Vector2(v3refr.x, v3refr.y);
}

std::ostream &operator <<(std::ostream &out, const Vector2 &vec) {
	out << "[" << vec.x << " " << vec.y << "]";
	return out;
}



// --------- Vector3 ----------

Vector3::Vector3(float x, float y, float z) {
	this->x = x;
	this->y = y;
	this->z = z;
}

Vector3::Vector3(const Vector2 &vec) {
	x = vec.x;
	y = vec.y;
	z = 1;
}

Vector3::Vector3(const Vector4 &vec) {
	x = vec.x;
	y = vec.y;
	z = vec.z;
}

Vector3::Vector3(const SphVector &sph) {
	*this = sph;
}

Vector3 &Vector3::operator =(const SphVector &sph) {
	x = sph.r * cos(sph.theta) * sin(sph.phi);
	z = sph.r * sin(sph.theta) * sin(sph.phi);
	y = sph.r * cos(sph.phi);
	return *this;
}

void Vector3::Normalize() {
	float len = Length();
	x /= len;
	y /= len;
	z /= len;	
}

Vector3 Vector3::Normalized() const {
	float len = Length();
	return Vector3(x / len, y / len, z / len);
}

Vector3 Vector3::Reflection(const Vector3 &normal) const {
	return -(2.0f * DotProduct(*this, normal) * normal - *this);
}

Vector3 Vector3::Refraction(const Vector3 &normal, float src_ior, float dst_ior) const {
	float cos_inc = DotProduct(*this, -normal);
	float ior = src_ior / dst_ior;

	float radical = 1.0f + SQ(ior) * (SQ(cos_inc) - 1.0f);

	if(radical < 0.0f) {		// total internal Reflection
		return Reflection(normal);
	}

	float beta = ior * cos_inc - sqrt(radical);

	return *this * ior + normal * beta;
}

void Vector3::Transform(const Matrix3x3 &mat) {
	float nx = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z;
	float ny = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z;
	z = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z;
	x = nx;
	y = ny;
}

Vector3 Vector3::Transformed(const Matrix3x3 &mat) const {
	Vector3 vec;
	vec.x = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z;
	vec.y = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z;
	vec.z = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z;
	return vec;
}

void Vector3::Transform(const Matrix4x4 &mat) {
	float nx = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z + mat[0][3];
	float ny = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z + mat[1][3];
	z = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z + mat[2][3];
	x = nx;
	y = ny;
}

Vector3 Vector3::Transformed(const Matrix4x4 &mat) const {
	Vector3 vec;
	vec.x = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z + mat[0][3];
	vec.y = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z + mat[1][3];
	vec.z = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z + mat[2][3];
	return vec;
}

void Vector3::Transform(const Quaternion &quat) {
	Quaternion vq(0.0f, *this);
	vq = quat * vq * quat.Inverse();
	*this = vq.v;
}

Vector3 Vector3::Transformed(const Quaternion &quat) const {
	Quaternion vq(0.0f, *this);
	vq = quat * vq * quat.Inverse();
	return vq.v;
}

void Vector3::Rotate(const Vector3 &euler) {
	Matrix4x4 rot;
	rot.SetRotation(euler);
	Transform(rot);
}

Vector3 Vector3::Rotated(const Vector3 &euler) const {
	Matrix4x4 rot;
	rot.SetRotation(euler);
	return Transformed(rot);
}


std::ostream &operator <<(std::ostream &out, const Vector3 &vec) {
	out << "[" << vec.x << " " << vec.y << " " << vec.z << "]";
	return out;
}



// -------------- Vector4 --------------
Vector4::Vector4(float x, float y, float z, float w) {
	this->x = x;
	this->y = y;
	this->z = z;
	this->w = w;
}

Vector4::Vector4(const Vector2 &vec) {
	x = vec.x;
	y = vec.y;
	z = 1;
	w = 1;
}

Vector4::Vector4(const Vector3 &vec) {
	x = vec.x;
	y = vec.y;
	z = vec.z;
	w = 1;
}

void Vector4::Normalize() {
	float len = (float)sqrt(x*x + y*y + z*z + w*w);
	x /= len;
	y /= len;
	z /= len;
	w /= len;
}

Vector4 Vector4::Normalized() const {
	float len = (float)sqrt(x*x + y*y + z*z + w*w);
	return Vector4(x / len, y / len, z / len, w / len);
}

void Vector4::Transform(const Matrix4x4 &mat) {
	float nx = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z + mat[0][3] * w;
	float ny = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z + mat[1][3] * w;
	float nz = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z + mat[2][3] * w;
	w = mat[3][0] * x + mat[3][1] * y + mat[3][2] * z + mat[3][3] * w;
	x = nx;
	y = ny;
	z = nz;
}

Vector4 Vector4::Transformed(const Matrix4x4 &mat) const {
	Vector4 vec;
	vec.x = mat[0][0] * x + mat[0][1] * y + mat[0][2] * z + mat[0][3] * w;
	vec.y = mat[1][0] * x + mat[1][1] * y + mat[1][2] * z + mat[1][3] * w;
	vec.z = mat[2][0] * x + mat[2][1] * y + mat[2][2] * z + mat[2][3] * w;
	vec.w = mat[3][0] * x + mat[3][1] * y + mat[3][2] * z + mat[3][3] * w;
	return vec;
}

// TODO: implement 4D vector Reflection
Vector4 Vector4::Reflection(const Vector4 &normal) const {
	return *this;
}

// TODO: implement 4D vector Refraction
Vector4 Vector4::Refraction(const Vector4 &normal, float src_ior, float dst_ior) const {
	return *this;
}

std::ostream &operator <<(std::ostream &out, const Vector4 &vec) {
	out << "[" << vec.x << " " << vec.y << " " << vec.z << " " << vec.w << "]";
	return out;
}
