#ifndef VECTOR_H_
#define VECTOR_H_

#include <iostream>

class Vector3;
class Vector4;
class Matrix3x3;
class Matrix4x4;
class Quaternion;
class SphVector;

class Vector2 {
public:
	float x, y;

	Vector2(float x = 0.0, float y = 0.0);
	Vector2(const Vector3 &vec);
	Vector2(const Vector4 &vec);
	
	inline float &operator [](int elem);	
	
	// unary operations
	friend inline Vector2 operator -(const Vector2 &vec);
	
	// binary vector (op) vector operations
	friend inline float DotProduct(const Vector2 &v1, const Vector2 &v2);
	
	friend inline Vector2 operator +(const Vector2 &v1, const Vector2 &v2);
	friend inline Vector2 operator -(const Vector2 &v1, const Vector2 &v2);	
	friend inline Vector2 operator *(const Vector2 &v1, const Vector2 &v2);
	friend inline Vector2 operator /(const Vector2 &v1, const Vector2 &v2);
	friend inline bool operator ==(const Vector2 &v1, const Vector2 &v2);
	
	friend inline void operator +=(Vector2 &v1, const Vector2 &v2);
	friend inline void operator -=(Vector2 &v1, const Vector2 &v2);
	friend inline void operator *=(Vector2 &v1, const Vector2 &v2);
	friend inline void operator /=(Vector2 &v1, const Vector2 &v2);
	
	// binary vector (op) scalar and scalar (op) vector operations
	friend inline Vector2 operator +(const Vector2 &vec, float scalar);
	friend inline Vector2 operator +(float scalar, const Vector2 &vec);
	friend inline Vector2 operator -(const Vector2 &vec, float scalar);
	friend inline Vector2 operator -(float scalar, const Vector2 &vec);
	friend inline Vector2 operator *(const Vector2 &vec, float scalar);
	friend inline Vector2 operator *(float scalar, const Vector2 &vec);
	friend inline Vector2 operator /(const Vector2 &vec, float scalar);
	friend inline Vector2 operator /(float scalar, const Vector2 &vec);
	
	friend inline void operator +=(Vector2 &vec, float scalar);
	friend inline void operator -=(Vector2 &vec, float scalar);
	friend inline void operator *=(Vector2 &vec, float scalar);
	friend inline void operator /=(Vector2 &vec, float scalar);
	
	inline float Length() const;
	inline float LengthSq() const;
	void Normalize();
	Vector2 Normalized() const;

	void Transform(const Matrix3x3 &mat);
	Vector2 Transformed(const Matrix3x3 &mat) const;

	void Rotate(float angle);
	Vector2 Rotated(float angle) const;
			
	Vector2 Reflection(const Vector2 &normal) const;
	Vector2 Refraction(const Vector2 &normal, float src_ior, float dst_ior) const;
	
	friend std::ostream &operator <<(std::ostream &out, const Vector2 &vec);
};

class Vector3 {
public:
	float x, y, z;

	Vector3(float x = 0.0, float y = 0.0, float z = 0.0);
	Vector3(const Vector2 &vec);
	Vector3(const Vector4 &vec);
	Vector3(const SphVector &sph);

	Vector3 &operator =(const SphVector &sph);
	
	inline float &operator [](int elem);	
	
	// unary operations
	friend inline Vector3 operator -(const Vector3 &vec);
	
	// binary vector (op) vector operations
	friend inline float DotProduct(const Vector3 &v1, const Vector3 &v2);
	friend inline Vector3 CrossProduct(const Vector3 &v1, const Vector3 &v2);
	
	friend inline Vector3 operator +(const Vector3 &v1, const Vector3 &v2);
	friend inline Vector3 operator -(const Vector3 &v1, const Vector3 &v2);	
	friend inline Vector3 operator *(const Vector3 &v1, const Vector3 &v2);	
	friend inline Vector3 operator /(const Vector3 &v1, const Vector3 &v2);	
	friend inline bool operator ==(const Vector3 &v1, const Vector3 &v2);
	
	friend inline void operator +=(Vector3 &v1, const Vector3 &v2);
	friend inline void operator -=(Vector3 &v1, const Vector3 &v2);
	friend inline void operator *=(Vector3 &v1, const Vector3 &v2);
	friend inline void operator /=(Vector3 &v1, const Vector3 &v2);
	
	// binary vector (op) scalar and scalar (op) vector operations
	friend inline Vector3 operator +(const Vector3 &vec, float scalar);
	friend inline Vector3 operator +(float scalar, const Vector3 &vec);
	friend inline Vector3 operator -(const Vector3 &vec, float scalar);
	friend inline Vector3 operator -(float scalar, const Vector3 &vec);
	friend inline Vector3 operator *(const Vector3 &vec, float scalar);
	friend inline Vector3 operator *(float scalar, const Vector3 &vec);
	friend inline Vector3 operator /(const Vector3 &vec, float scalar);
	friend inline Vector3 operator /(float scalar, const Vector3 &vec);
	
	friend inline void operator +=(Vector3 &vec, float scalar);
	friend inline void operator -=(Vector3 &vec, float scalar);
	friend inline void operator *=(Vector3 &vec, float scalar);
	friend inline void operator /=(Vector3 &vec, float scalar);
	
	inline float Length() const;
	inline float LengthSq() const;
	void Normalize();
	Vector3 Normalized() const;
	
	void Transform(const Matrix3x3 &mat);
	Vector3 Transformed(const Matrix3x3 &mat) const;
	void Transform(const Matrix4x4 &mat);
	Vector3 Transformed(const Matrix4x4 &mat) const;
	void Transform(const Quaternion &quat);
	Vector3 Transformed(const Quaternion &quat) const;

	void Rotate(const Vector3 &euler);
	Vector3 Rotated(const Vector3 &euler) const;
	
	Vector3 Reflection(const Vector3 &normal) const;
	Vector3 Refraction(const Vector3 &normal, float src_ior, float dst_ior) const;
	
	friend std::ostream &operator <<(std::ostream &out, const Vector3 &vec);
};

class Vector4 {
public:
	float x, y, z, w;

	Vector4(float x = 0.0, float y = 0.0, float z = 0.0, float w = 0.0);
	Vector4(const Vector2 &vec);
	Vector4(const Vector3 &vec);
	
	inline float &operator [](int elem);	
	
	// unary operations
	friend inline Vector4 operator -(const Vector4 &vec);
		
	// binary vector (op) vector operations
	friend inline float DotProduct(const Vector4 &v1, const Vector4 &v2);
	friend inline Vector4 CrossProduct(const Vector4 &v1, const Vector4 &v2, const Vector4 &v3);
	
	friend inline Vector4 operator +(const Vector4 &v1, const Vector4 &v2);
	friend inline Vector4 operator -(const Vector4 &v1, const Vector4 &v2);	
	friend inline Vector4 operator *(const Vector4 &v1, const Vector4 &v2);	
	friend inline Vector4 operator /(const Vector4 &v1, const Vector4 &v2);	
	friend inline bool operator ==(const Vector4 &v1, const Vector4 &v2);
	
	friend inline void operator +=(Vector4 &v1, const Vector4 &v2);
	friend inline void operator -=(Vector4 &v1, const Vector4 &v2);
	friend inline void operator *=(Vector4 &v1, const Vector4 &v2);
	friend inline void operator /=(Vector4 &v1, const Vector4 &v2);
	
	// binary vector (op) scalar and scalar (op) vector operations
	friend inline Vector4 operator +(const Vector4 &vec, float scalar);
	friend inline Vector4 operator +(float scalar, const Vector4 &vec);
	friend inline Vector4 operator -(const Vector4 &vec, float scalar);
	friend inline Vector4 operator -(float scalar, const Vector4 &vec);
	friend inline Vector4 operator *(const Vector4 &vec, float scalar);
	friend inline Vector4 operator *(float scalar, const Vector4 &vec);
	friend inline Vector4 operator /(const Vector4 &vec, float scalar);
	friend inline Vector4 operator /(float scalar, const Vector4 &vec);
	
	friend inline void operator +=(Vector4 &vec, float scalar);
	friend inline void operator -=(Vector4 &vec, float scalar);
	friend inline void operator *=(Vector4 &vec, float scalar);
	friend inline void operator /=(Vector4 &vec, float scalar);
	
	inline float Length() const;
	inline float LengthSq() const;
	void Normalize();
	Vector4 Normalized() const;

	void Transform(const Matrix4x4 &mat);
	Vector4 Transformed(const Matrix4x4 &mat) const;
		
	Vector4 Reflection(const Vector4 &normal) const;
	Vector4 Refraction(const Vector4 &normal, float src_ior, float dst_ior) const;
	
	friend std::ostream &operator <<(std::ostream &out, const Vector4 &vec);
};

#include "vector.inl"

#endif	// VECTOR_H_
