// exporter.cpp
#include "plugin.h"
#include <errlog.h>
#include "vertex.h"
#include <math3d.h>
#include <direct.h>
#include <map>
#include <stdlib.h>

#define FRAME_INT 1

#define Info	__MyInfo

static void __MyInfo(char *lala, ...)
{
}

static unsigned int curr_add_node = 0;
static void add_node_to_map(std::map<unsigned int, unsigned int> &map, IGameNode *node)
{
	map[node->GetNodeID()] = curr_add_node;
	curr_add_node ++;

	for (unsigned int i=0; i<node->GetChildCount(); i++)
	{
		add_node_to_map(map, node->GetNodeChild(i));
	}
}

// utility to create a map for nodes
static std::map<unsigned int, unsigned int> create_node_map(IGameScene *scene)
{
	curr_add_node = 0;

	std::map<unsigned int, unsigned int> ret;
	for (unsigned int i=0; i<scene->GetTopLevelNodeCount(); i++)
	{
		add_node_to_map(ret, scene->GetTopLevelNode(i));
	}

	return ret;
}

bool ExportPlugin::InitExport()
{
	// begin logging
	SetLogFile("C:\\plugin_log.txt", LOG_ALL);

	scene = GetIGameInterface();
	if (!scene)
	{
		Info("GetIGameInterface() returned zero pointer!\n");
		return false;
	}

	// Set D3D coord system
	GetConversionManager()->SetCoordSystem(IGameConversionManager::IGAME_D3D); 
	scene->InitialiseIGame(false);
	// select frame 0
	scene->SetStaticFrame(0);

	return true;
}

void ExportPlugin::CloseExport()
{
	if (scene)
	{
		scene->ReleaseIGame();
	}
}

static std::string GetFilename(std::string full_path)
{
	char *buf = new char[full_path.size() + 1];
	strcpy(buf, full_path.c_str());

	char *last_slash = strrchr(buf, '/');
	if(!last_slash) {
		last_slash = strrchr(buf, '\\');
	}

	if(last_slash) {
		full_path = std::string(last_slash + 1);
	}
	delete [] buf;
	return full_path;
}

static std::string GetPath(std::string full_path)
{
	char *buf = new char[full_path.size() + 1];
	strcpy(buf, full_path.c_str());

	char *last_slash = strrchr(buf, '/');
	if(!last_slash) {
		last_slash = strrchr(buf, '\\');
	}

	if(last_slash) {
		*last_slash = 0;
	}
	else {
		buf[0] = 0;
	}
	full_path = std::string(buf);
	
	delete [] buf;
	return full_path;
}

bool Prefix(std::string test, std::string prefix)
{
	if (prefix.size() > test.size()) return false;
	for (unsigned int i=0; i<prefix.size(); i++)
	{
		if (prefix[i] != test[i]) return false;
	}
	return true;
}

static const char *GetToolsDir()
{
	HKEY		key;
	static char	buff[16384];
	DWORD		type, len;
	RegOpenKey(HKEY_LOCAL_MACHINE, "SOFTWARE\\Track7Games\\EngineTools", &key);
	len = 16384;
	type = REG_SZ;
	RegQueryValueEx(key, "InstallPath", NULL, &type, (LPBYTE)buff, &len);
	RegCloseKey(key);

	return buff;
}

static DWORD WINAPI SceviewThread(LPVOID param)
{
	ShellExecute(0, "open", "sceview", 
		(std::string("\"") + (char*) param + '\"').c_str(),
		"C:\mypath", 0);
	return 0;
}

// Export file
int ExportPlugin::DoExport(const TCHAR *name, ExpInterface *ei,
		Interface *i, BOOL suppressPrompts,
		DWORD options)
{
	// Launch plugin dialog
	if (!suppressPrompts)
	{
		if (!DisplayDialog()) return TRUE;
		dialog_data = GetDialogData();
	}
	else
	{
		dialog_data.normals = true;
		dialog_data.tangents = true;
		dialog_data.tc0 = true;
		dialog_data.tc1 = true;
		dialog_data.tex_func = 0;
	}

	if (!InitExport()) return FALSE;

	//show progress bar
	StartProgressDialog();

	// Init chunk hierarchy
	SceneFile* scene_file = new SceneFile();
	scene_file->SetVersion(10); // 3 digit - decimal
	scene_file->SetMessage("Track7 scene format v0.1");
	curr_file_node = scene_file;

	// Process materials
	Info("-----------\nNumber of root materials in scene: %d\n", scene->GetRootMaterialCount());
	for (unsigned long i=0; i<scene->GetRootMaterialCount(); i++)
	{
		IGameMaterial* mat = scene->GetRootMaterial(i);
		ProcessMaterial(mat);
	}
	Info("Total materials exported: %d\n", mat_count);

	Info("Adding default material\n");
	// add a new material chunk
	MaterialChunk* mc = new MaterialChunk();
	scene_file->Mount(mc);
	mc->SetName("Dummy plugin material");
	// insert a new material to the material map
	mat_map["default plugin material"] = mat_count;
	mat_count ++;

	// process scene data
	int root_nodes = scene->GetTopLevelNodeCount();
	int nodes = scene->GetTotalNodeCount();

	Info("Root node count = %d\nTotal node count = %d\n", root_nodes, nodes);

	// process nodes
	for (unsigned int i=0; i<root_nodes; i++)
	{	
		ProcessNode(scene->GetTopLevelNode(i));
		SetProgress((float) (i+1) / (float) root_nodes);
		SetProgressMsg(scene->GetTopLevelNode(i)->GetName());
	}

	// get file path
	std::string file_path = GetPath(name) + "\\";
	std::string data_path = file_path + "data\\";

	// perform requested texture action
	std::vector<std::string> texture_list;
	ChunkList tcl;
	if (dialog_data.tex_func)
	{
		std::map<std::string, int> texlist_map;
		tcl = scene_file->GetIDChildrenRec(SCE_TEXTURE);
		
		// fill the map with textures
		for (int i=0; i<tcl.size(); i++)
		{
			texlist_map[((NamedChunk*)tcl[i])->GetName()]++;
		}

		// collect unique filenames
		for (std::map<std::string, int>::iterator i=texlist_map.begin(); i!=texlist_map.end(); i++)
		{
			texture_list.push_back(i->first);
		}
	}
	if (dialog_data.tex_func == 1)
	{
		// texture copy
		//std::string mkdir_cmd = std::string("md \"") + file_path + "data\"";
		//system(mkdir_cmd.c_str());

		std::string dir_name = file_path + "data";

		CreateDirectory(dir_name.c_str(), 0);
		//HANDLE dir_handle = CreateFile(dir_name.c_str(), 0, 0, 0, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, 0);
		//if (dir_handle == INVALID_HANDLE_VALUE)
		//{
		//	MessageBox(0, "Couldn't create \"data\" directory", "Error", 0);
		//}
		 
		//else
		//{
		for (int i=0; i<texture_list.size(); i++)
		{
			//std::string cp_command = std::string("copy \"") + texture_list[i] 
			//	+ "\" \"" + data_path + GetFilename(texture_list[i]) + "\"";

			//system(cp_command.c_str());
			SetProgress((float) (i+1) / (float) texture_list.size());
			SetProgressMsg(texture_list[i].c_str());

			std::string dst = data_path + GetFilename(texture_list[i]);
			CopyFile(texture_list[i].c_str(), dst.c_str(), FALSE);
		}

		// fix texture chunks
		for (int i=0; i<tcl.size(); i++)
		{
			TextureChunk* tex_chunk = (TextureChunk*) tcl[i];
			std::string fname = GetFilename(tex_chunk->GetName());
			tex_chunk->SetName(std::string("data\\") + fname);
		}
		//}
	}
	if (dialog_data.tex_func == 2)
	{
		// texture pack
		for (int i=0; i<texture_list.size(); i++)
		{
			const char *skata = texture_list[i].c_str();
			FILE* fp_tex = fopen(texture_list[i].c_str(), "rb");
			if (!fp_tex) continue;
			unsigned int filesize;
			fseek(fp_tex, 0, SEEK_END);
			filesize = ftell(fp_tex);
			fseek(fp_tex, 0, SEEK_SET);
			FileChunk *fc = new FileChunk();
			fc->SetName(texture_list[i]);
			char *tmp = new char[filesize];
			fread(tmp, 1, filesize, fp_tex);
			fc->SetData(filesize, tmp);
			delete [] tmp;
			scene_file->Mount(fc);
			fclose(fp_tex);

			SetProgress((float) (i+1) / (float) texture_list.size());
			SetProgressMsg(texture_list[i].c_str());
		}
	}

	// fix node ids to be sequential
	std::map<unsigned int, unsigned int> node_map = create_node_map(scene);
	ChunkList all_nodes = scene_file->GetIDChildrenRec(SCE_NODE);
	for (unsigned int i=0; i<all_nodes.size(); i++)
	{
		NodeChunk * curr_node = (NodeChunk*) all_nodes[i];
		Info("orig. id = %d, new_id = %d\n", curr_node->GetNodeID(), node_map[curr_node->GetNodeID()]);
		curr_node->SetNodeID(node_map[curr_node->GetNodeID()]);
	}

	// fix bone maps with new node id's
	ChunkList all_meshes = scene_file->GetIDChildrenRec(SCE_MESH);
	for (unsigned int i=0; i<all_meshes.size(); i++)
	{
		MeshChunk *mesh = (MeshChunk*)all_meshes[i];
		GeometryArray *bone_map = mesh->GetArray(BONE_MAP_ARRAY);
		if (!bone_map) continue;
		unsigned int *bm = (unsigned int*)bone_map->GetArrayData();
		for (unsigned int j=0; j<bone_map->GetElementCount(); j++)
		{
			bm[j] = node_map[bm[j]];
		}
	}

	// convert nodes by name
	// prefix proxy_, pvs_ and link_ are functional meshes
	Info("Processing functional meshes (PVS / proxy)\n");
	all_nodes = scene_file->GetIDChildrenRec(SCE_NODE);
	char tb[1024];
	for (unsigned int i=0; i<all_nodes.size(); i++)
	{
		bool mesh=false, pvs=false, proxy=false, link=false;
		NodeChunk *n = (NodeChunk*)all_nodes[i];
		if (Prefix(n->GetName(), std::string("proxy_"))) proxy = true;
		if (Prefix(n->GetName(), std::string("pvs_"))) pvs = true;
		if (Prefix(n->GetName(), std::string("link_"))) link = true;
		if (!proxy && !pvs && !link) mesh = true;
		if (proxy) Info("Node %s is proxy\n", n->GetName().c_str());
		if (pvs) Info("Node %s is PVS\n", n->GetName().c_str());
		if (link) Info("Node %s is a Link\n", n->GetName().c_str());
		ChunkList meshes = n->GetIDChildren(SCE_MESH);
		for (unsigned int j=0; j<meshes.size(); j++)
		{
			MeshChunk *m = (MeshChunk*) meshes[j];
			m->SetMesh(mesh);
			m->SetProxy(proxy);
			m->SetPVS(pvs);
			m->SetLink(link);
		}

		// batch all meshes of the same material and all
		// pvs/proxy/link of the same node
		n->BatchMeshes();

		SetProgress((float) (i+1) / (float) all_nodes.size());
		sprintf(tb, "Batching geometry: %s", n->GetName().c_str());
		SetProgressMsg(tb);
	}

	// transform all pvs and link meshes to their permanent position
	MeshList pvsl = scene_file->GetPVSMeshes();
	for (unsigned int i=0; i<pvsl.size(); i++)
	{
		NodeChunk *n = (NodeChunk*)pvsl[i]->GetParent();
		pvsl[i]->Transform(n->GetTransform());
	}
	MeshList linkl = scene_file->GetLinkMeshes();
	for (unsigned int i=0; i<linkl.size(); i++)
	{
		NodeChunk *n = (NodeChunk*)linkl[i]->GetParent();
		linkl[i]->Transform(n->GetTransform());
	}

	// give each mesh a unique name, based on the parent node
	all_meshes = scene_file->GetIDChildrenRec(SCE_MESH);
	std::map<std::string, int> node_name_usage;
	for (unsigned int i=0; i<all_meshes.size(); i++)
	{
		MeshChunk *m = (MeshChunk*) all_meshes[i];
		NodeChunk *n = (NodeChunk*) m->GetParent();

		std::string new_name = n->GetName();
		if (node_name_usage[n->GetName()])
		{
			sprintf(tb, "%d", node_name_usage[n->GetName()]);
			new_name += tb;
		}
	
		((NamedChunk*)m)->SetName(new_name);

		node_name_usage[n->GetName()]++;
	}

	// write data to file
	Info("Trying to save to: %s\n", name);
	VFS_FILE *fp = vfs_open(name, "wb");
	if (!fp)
	{
		Info("Failed opening file %s for writing\n");
		return FALSE;
	}
	scene_file->Save(fp);
	vfs_close(fp);

	std::string mat_filename = name;
	mat_filename += ".mat";
	FILE *mat_fp = fopen(mat_filename.c_str(), "wb");
	if (mat_fp)
	{
		ChunkList all_materials = scene_file->GetIDChildrenRec(SCE_MATERIAL);

		fprintf(mat_fp, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
		fprintf(mat_fp, "<materials version=\"1.0\">\n");

		for (unsigned int i=0; i<all_materials.size(); i++)
		{
			MaterialChunk *mat_chunk = (MaterialChunk*) all_materials[i];
			std::string material_name = mat_chunk->GetName();
			
			fprintf(mat_fp, "<material name=\"%s\" parent=\"mesh_base\">\n",
				material_name.c_str());
			
			TextureChunk *tex;
			std::string tex_filename;
			tex = mat_chunk->GetTexture(STT_DIFFUSE);
			if (tex)
			{
				tex_filename = std::string("textures/") + tex->GetName();
				fprintf(mat_fp, "\t<texture name=\"diffuse\">%s</texture>\n", 
					tex_filename.c_str());
			}
			tex = mat_chunk->GetTexture(STT_NORMAL);
			if (tex)
			{
				tex_filename = std::string("textures/") + tex->GetName();
				fprintf(mat_fp, "\t<texture name=\"normal\">%s</texture>\n", 
					tex_filename.c_str());
			}
		
			fprintf(mat_fp, "</material>\n");
		}

		fprintf(mat_fp, "</materials>\n");

		fclose(mat_fp);
	}

	// cleanup
	delete scene_file;
	scene_file = 0;
	curr_file_node = 0;
	//vfs_umount("/");
    CloseExport();

	CloseProgressDialog();

	if (dialog_data.preview)
	{
		// show preview window
		std::string prev_cmd = GetToolsDir();
		chdir(prev_cmd.c_str());
		prev_cmd += "\\sceview \"";
		prev_cmd += name;
		prev_cmd += "\"";

		WinExec(prev_cmd.c_str(), SW_SHOW);
		//CreateThread(0, 0, SceviewThread, (LPVOID) name, 0, 0);
		//SystemCmdThreadProc(prev_cmd.c_str());
	}

	return TRUE;
}

static PRS GMatrixToPRS(const GMatrix &mat) {
	Vector3 pos(mat.Translation().x, mat.Translation().y, mat.Translation().z);
	Quaternion rot(mat.Rotation().w, mat.Rotation().x, mat.Rotation().y, mat.Rotation().z);
	Vector3 scale(mat.Scaling().x, mat.Scaling().y, mat.Scaling().z);
	return PRS(pos, rot, scale);
}

static void DumpPRS(const PRS &prs) {
	Info("position: %.3f, %.3f, %.3f\n", prs.GetPosition().x, prs.GetPosition().y, prs.GetPosition().z);
	Info("rotation: <%.3f, %.3f, %.3f>, %.3f\n", prs.GetRotation().v.x, prs.GetRotation().v.y, prs.GetRotation().v.z, prs.GetRotation().s);
	Info(" scaling: %.3f, %.3f, %.3f\n", prs.GetScaling().x, prs.GetScaling().y, prs.GetScaling().z);
}

void ExportPlugin::ProcessSingleNode(IGameNode* node)
{
	Info("Processing node: %s\n", node->GetName());
	Info("Node ID: %d\n", node->GetNodeID());

	// save node info
	((NodeChunk*)curr_file_node)->SetName(node->GetName());
	((NodeChunk*)curr_file_node)->SetNodeID(node->GetNodeID());
	((NodeChunk*)curr_file_node)->SetDynamic(!node->GetLocalTM().IsIdentity());

	GMatrix lmat = node->GetLocalTM();
	GMatrix wmat = node->GetWorldTM();
	GMatrix omat = node->GetObjectTM();

	NodeChunk *cur_node = (NodeChunk*)curr_file_node;
	cur_node->SetWorldPRS(GMatrixToPRS(wmat));
	cur_node->SetLocalPRS(GMatrixToPRS(lmat));
	cur_node->SetObjectPRS(GMatrixToPRS(omat));

	// get pivot point
	INode *max_node = node->GetMaxNode();
	Point3 pivot_pos = max_node->GetObjOffsetPos();
	Quat pivot_rot = max_node->GetObjOffsetRot();
	Point3 pivot_scale = max_node->GetObjOffsetScale().s;
	Quat pivot_quat = max_node->GetObjOffsetScale().q;

	Vector3 ppos(pivot_pos.x, pivot_pos.z, pivot_pos.y);
	Quaternion prot(pivot_rot.w, pivot_rot.x, pivot_rot.z, pivot_rot.y);
	Vector3 pscale(pivot_scale.x, pivot_scale.z, pivot_scale.y);
	Quaternion pqscale(pivot_quat.w, pivot_quat.x, pivot_quat.z, pivot_quat.y);
	pscale.Transform(pqscale.GetRotationMatrix());

	cur_node->SetPivotPRS(PRS(ppos, prot, pscale));

	// get the corresponding object
	IGameObject* obj = node->GetIGameObject();
	if (!obj)
	{
		Info("Invalid object!\n---\n");
		return;
	}

	// get the object type
	IGameObject::ObjectTypes type = obj->GetIGameType();

	switch (type)
	{
		case IGameObject::IGAME_UNKNOWN:
			Info("Ignoring unknown object\n");
			break;
		case IGameObject::IGAME_LIGHT:
			ProcessLight((IGameLight*) obj, node);
			break;
		case IGameObject::IGAME_MESH:
			ProcessMesh((IGameMesh*) obj, node);
			break;
		case IGameObject::IGAME_SPLINE:
			Info("Ignoring spline object\n");
			break;
		case IGameObject::IGAME_CAMERA:
			ProcessCamera((IGameCamera*) obj, node);
			break;
		case IGameObject::IGAME_HELPER:
			Info("Ignoring helper object\n");
			break;
		case IGameObject::IGAME_BONE:
			Info("Ignoring bone object\n");
			break;
		case IGameObject::IGAME_IKCHAIN:
			Info("Ignoring IKChain object\n");
			break;
	}

	Info("----------------------------\n");

	node->ReleaseIGameObject();
	obj = 0;
}

void ExportPlugin::ProcessNode(IGameNode* node)
{
	// add a new node chunk to the current chunk
	NodeChunk* new_node = new NodeChunk();
	curr_file_node->Mount(new_node);
	curr_file_node = new_node;

	// process current node
	ProcessSingleNode(node);

	// process children
	for (unsigned int i=0; i<node->GetChildCount(); i++)
	{
		ProcessNode(node->GetNodeChild(i));
	}

	// go back to parent -- this is safe since we added a child eralier
	curr_file_node = curr_file_node->GetParent();
}

void ExportPlugin::ProcessCamera(IGameCamera* camera, IGameNode* node)
{
	Info("Processing camera:\n");

	// add a new camera chunk
	CameraChunk* cc = new CameraChunk();
	curr_file_node->Mount(cc);

	float f;
	f = 80.0;
	if (camera->GetCameraFOV())
		camera->GetCameraFOV()->GetPropertyValue(f);
	cc->SetFOV(f);

	Info("FOV = %f\n", f);

	f = 10000.0f;
	if (camera->GetCameraFarClip())
		camera->GetCameraFarClip()->GetPropertyValue(f);
	cc->SetFarClip(f);

	Info("Far clip = %f\n", f);

	f = 1.0f;
	if (camera->GetCameraNearClip())
		camera->GetCameraNearClip()->GetPropertyValue(f);
	cc->SetNearClip(f);

	Info("Near clip = %f\n", f);

	IGameNode* n = camera->GetCameraTarget();
	if (n)
	{
		Info("This is a target camera\n");
		Info("Target node ID: %d\n", n->GetNodeID());
		cc->SetTargetNodeID(n->GetNodeID());
		cc->SetTarget(true);
	}
	else
	{
		Info("This is a free camera\n");
		cc->SetTargetNodeID(0);
		cc->SetTarget(false);
	}
}

void ExportPlugin::ProcessLight(IGameLight* light, IGameNode* node)
{
	if (!light->IsEntitySupported()) return;

	// Get max lightobject and lightsettings, because some
	// IGame functions like GetAttenStart etc won't work
	INode *max_node = node->GetMaxNode();
	Object *lightobj = max_node->GetObjectRef();
	ObjectState os = max_node->EvalWorldState(0);
	GenLight* max_light = (GenLight*)os.obj;
	struct LightState ls;
	max_light->EvalLightState(0, FOREVER, &ls);

	IGameLight::LightType lt = light->GetLightType();
	LightChunkType type = LT_OMNI;
	switch (lt)
	{
		case IGameLight::IGAME_OMNI:
			Info("Omni light\n");
			type = LT_OMNI;
			break;
		case IGameLight::IGAME_TSPOT:
			Info("Target spot light\n");
			type = LT_SPOT;
			break;
		case IGameLight::IGAME_FSPOT:
			Info("Spot light\n");
			type = LT_SPOT;
			break;

		case IGameLight::IGAME_DIR:
			Info("Directional light\n");
			return; // unsupported light
			break;
		case IGameLight::IGAME_TDIR:
			Info("Target directional light\n");
			return; // unsupported light
			break;
		default:
			Info("Unknown light type!\n");
			return; // unsupported light
			break;
	}

	// add a new light chunk
	LightChunk* lc = new LightChunk();
	curr_file_node->Mount(lc);

	lc->SetType(type);

	Point3 clr;
	clr.x = clr.y = clr.z = 1.0f;
	if (light->GetLightColor())
	{
		light->GetLightColor()->GetPropertyValue(clr);
	}
	else
	{
		clr.x = ls.color.r;
		clr.y = ls.color.g;
		clr.z = ls.color.b;
	}
	Vector4 color(clr.x, clr.y, clr.z, 1);
	lc->SetColor(color);

	float f = 1.0f;
	if (light->GetLightMultiplier())
	{
		light->GetLightMultiplier()->GetPropertyValue(f);
	}
	lc->SetMultiplier(f);

	// save attenuation start / end
	float fl;
	fl =501.0f;
	if (light->GetLightAttenStart())
	{
		light->GetLightAttenStart()->GetPropertyValue(fl);
	}
	else
	{
		fl = ls.attenStart;
	}
	lc->SetAttenStart(fl);
	Info("AttenStart = %f\n", fl);
	
	fl = 100.0f;
	if (light->GetLightAttenEnd())
	{
		light->GetLightAttenEnd()->GetPropertyValue(fl);
	}
	else
	{
		fl = ls.attenEnd;
	}
	Info("AttenEEnd = %f\n", fl);
	lc->SetAttenEnd(fl);

	unsigned long target_node_id = 0;
	if (lt == IGameLight::IGAME_TSPOT)
	{
		target_node_id = light->GetLightTarget()->GetNodeID();
		if (light->GetLightAspectRatio())
		{
			light->GetLightAspectRatio()->GetPropertyValue(f);
		}
		else
		{
			f = ls.aspect;
		}
		lc->SetAspectRatio(f);
	}
	if (lt == IGameLight::IGAME_TSPOT || lt == IGameLight::IGAME_FSPOT)
	{
		if (light->GetLightFallOff())
		{
			light->GetLightFallOff()->GetPropertyValue(f);
		}
		else
		{
			f = ls.fallsize;
		}
		lc->SetFallOff(f);

		if (light->GetLightHotSpot())
		{
			light->GetLightHotSpot()->GetPropertyValue(f);
		}
		else
		{
			f = ls.hotsize;
		}
		lc->SetHotSpot(f);
	}
	lc->SetTargetNodeID(target_node_id);
}

void ExportPlugin::ProcessMesh(IGameMesh* mesh, IGameNode* node)
{
	// initialize data
	mesh->SetCreateOptimizedNormalList(); // make normals (and tangents/binormals) per vertex
	mesh->SetUseWeightedNormals(); // is this better than the default?
	mesh->InitializeData();
	
	Info("Processing mesh data\n");

	// get modifiers
	int num_mod = mesh->GetNumModifiers();
	Info("Number of modifiers: %d\n", num_mod);

	IGameSkin *skin = 0;

	for (int i=0; i<num_mod; i++)
	{
		IGameModifier *mod = mesh->GetIGameModifier(i);
		IGameModifier::ModType type = mod->GetModifierType();
		if (type == IGameModifier::IGAME_SKINNING)
		{
			Info("Found skinning modifier\n");
			skin = (IGameSkin*) mod;
		}
	}

	std::vector<int> bone_map;
	std::map<int, int> rev_bone_map;

	if (skin)
	{
		Info("Processing skin modifier ... \n");

		unsigned int num_skin_verts = skin->GetNumOfSkinnedVerts();
		Info("\tskineed verts: %d\n", num_skin_verts);

		int curr_bone = 1;

		for (unsigned int i=0; i<num_skin_verts; i++)
		{
			Info("\tVertex %d\n", i);
			unsigned int bone_count = skin->GetNumberOfBones(i);
			for (unsigned int j=0; j<bone_count; j++)
			{
				Info("\t\t Bone %d, BoneID: %d,weight : %f\n", j, skin->GetBoneID(i, j), skin->GetWeight(i, j));
				
				// fill bone map
				int bone_id = skin->GetBoneID(i, j);
				if (rev_bone_map[bone_id] == 0)
				{
					rev_bone_map[bone_id] = curr_bone;
					curr_bone ++;
				}
			}
		}

		// unroll bone map
		for (int i=1; i<curr_bone; i++)
		{
			// find the bone in the reverse map
			std::map<int, int>::iterator it;
			for (it = rev_bone_map.begin(); it != rev_bone_map.end(); ++it)
			{
				if (it->second == i)
				{
					bone_map.push_back(it->first);
				}
			}
		}

		// fix reverse map (start at 0 not 1)
		std::map<int, int>::iterator it;
		for (it = rev_bone_map.begin(); it != rev_bone_map.end(); ++it)
		{
			it->second --;
		}

		Info("Bone map:\n-----------\n");
		for (int i=0; i<bone_map.size(); i++)
		{
			Info("%d\t---> %d\n", i, bone_map[i]);
		}
	}

	// get number of active materials (this is also the number of
	// submeshes we are going to export)
	unsigned long num_materials = mesh->GetActiveMatIDs().Count();
	Info("Number of materials in the mesh:%d\n", num_materials);
	for (unsigned long i=0; i<num_materials; i++)
	{
		// add a new mesh to the current node
		MeshChunk* new_mesh = new MeshChunk();
		curr_file_node->Mount(new_mesh);

		// for now export only actual meshes
		new_mesh->SetProxy(false);

		// get a face list
		Tab<FaceEx*> faces = mesh->GetFacesFromMatID(mesh->GetActiveMatIDs()[i]);

		IGameMaterial* mat = mesh->GetMaterialFromFace(faces[0]);
		unsigned long mat_id;
		if (mat)
		{
			mat_id = mat_map[mat->GetMaterialName()];
			Info("Submesh material id: %d, name: %s\n", mat_id, mat->GetMaterialName());
		}
		else
		{
			mat_id = mat_map["default plugin material"];
			Info("Submesh material id: %d, name: %s\n", mat_id, "default plugin material");
		}
		new_mesh->SetMaterialID(mat_id);

		// process index array
		GeometryArray* index_array = new GeometryArray(INDEX_ARRAY, SCE_BT_INT32, 1, 
			3 * faces.Count());
		new_mesh->Mount(index_array);
		VertexBuffer vb;

		// See what mapping channels are used by the mesh
		Tab<int> map_channels = mesh->GetActiveMapChannelNum();
		// support only the first four mapping channels
		bool map_channel_used[4];
		for (unsigned long i=0; i<4; i++)
		{
			map_channel_used[i] = false;
		}
		for (unsigned long i=0; i<map_channels.Count(); i++)
		{
			if (map_channels[i] >= 1 && map_channels[i] <= 4)
			{
				map_channel_used[map_channels[i] - 1] = true;
			}
		}

		unsigned long* index_data = (unsigned long*)index_array->GetArrayData();

		/* not anymore
		// add a smoothing group array (used for normal calculation)
		GeometryArray *smgrp_array = new GeometryArray(SMOOTHING_GROUP_ARRAY,
			SCE_BT_INT32, 1, faces.Count());
		new_mesh->Mount(smgrp_array);
		unsigned int *smgrparr = (unsigned int*) smgrp_array->GetArrayData();
		*/

		for (unsigned long f=0; f<faces.Count(); f++)
		{
			//smgrparr[f] = faces[i]->smGrp;

			for (unsigned long v=0; v<3; v++)
			{
				unsigned long pos_index = faces[f]->vert[v];
				unsigned long nor_index = faces[f]->norm[v];

				unsigned long tc0_index = 0;
				unsigned long tc1_index = 0;
				unsigned long tc2_index = 0;
				unsigned long tc3_index = 0;

				unsigned long tmp[3];

				if (dialog_data.tc0 && map_channel_used[0])
				{
					mesh->GetMapFaceIndex(1, faces[f]->meshFaceIndex, tmp);
					tc0_index = tmp[v];
				}
				if (dialog_data.tc1 && map_channel_used[1])
				{
					mesh->GetMapFaceIndex(2, faces[f]->meshFaceIndex, tmp);
					tc1_index = tmp[v];
				}
				if (dialog_data.tc2 && map_channel_used[2])
				{
					mesh->GetMapFaceIndex(3, faces[f]->meshFaceIndex, tmp);
					tc2_index = tmp[v];
				}
				if (dialog_data.tc3 && map_channel_used[3])
				{
					mesh->GetMapFaceIndex(4, faces[f]->meshFaceIndex, tmp);
					tc3_index = tmp[v];
				}

				// produce an index and store it to the index buffer
				unsigned long index = vb.Insert(Vertex(pos_index, 
					tc0_index, tc1_index, tc2_index, tc3_index, nor_index));

				// flip triangle
				unsigned int flip;
				if (v == 0) flip = 2;
				if (v == 1) flip = 1;
				if (v == 2) flip = 0;

				index_data[3*f + flip] = index;
			}

			
		}

		Info("Submesh %d:\n", i);
		Info("Number of vertices (after breaks): %d\n", vb.GetNumVerts());

		// produce vertex arrays
		unsigned long num_verts = vb.GetNumVerts();

		// fill position buffer
		if (mesh->GetNumberOfVerts())
		{
			GeometryArray* position_array = new GeometryArray(POSITION_ARRAY, SCE_BT_FLOAT, 3, num_verts);
			new_mesh->Mount(position_array);
			Vector3* pos = (Vector3*)position_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				// get worldspace vertex for now
				// false = worldspace, true = modelspace
				Point3 mv = mesh->GetVertex(vb.GetArxidia(v).position, true);
				pos[v] = Vector3(mv.x, mv.y, mv.z);
			}
		}

		// mapping channels	
		Tab<int> map_ch = mesh->GetActiveMapChannelNum();
		Info("Number of mapping channels: %d\n", map_ch.Count());
		//for (unsigned long i=0; i<map_ch.Count(); i++)
		//{
		//	Info("%d\t", map_ch[i]);
		//}
		

		// fill mapping channels
		if (map_channel_used[0] && dialog_data.tc0)
		{
			
			GeometryArray* texcoord0_array = new GeometryArray(TEXCOORD0_ARRAY, SCE_BT_FLOAT, 2, num_verts);
			new_mesh->Mount(texcoord0_array);
			Vector2* tc0 = (Vector2*)texcoord0_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetMapVertex(1, vb.GetArxidia(v).tc0);
				tc0[v] = Vector2(mv.x, 1.0f - mv.y);
			}
		}

		if (map_channel_used[1] && dialog_data.tc1)
		{
			
			GeometryArray* texcoord1_array = new GeometryArray(TEXCOORD1_ARRAY, SCE_BT_FLOAT, 2, num_verts);			
			new_mesh->Mount(texcoord1_array);
			Vector2* tc1 = (Vector2*)texcoord1_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetMapVertex(2, vb.GetArxidia(v).tc1);
				tc1[v] = Vector2(mv.x, 1.0f - mv.y);
			}
		}
		if (map_channel_used[2] && dialog_data.tc2)
		{
			GeometryArray* texcoord2_array = new GeometryArray(TEXCOORD2_ARRAY, SCE_BT_FLOAT, 2, num_verts);
			new_mesh->Mount(texcoord2_array);
			Vector2* tc2 = (Vector2*)texcoord2_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetMapVertex(3, vb.GetArxidia(v).tc2);
				tc2[v] = Vector2(mv.x, 1.0f - mv.y);
			}
		}
		if (map_channel_used[3] && dialog_data.tc3)
		{
			GeometryArray* texcoord3_array = new GeometryArray(TEXCOORD3_ARRAY, SCE_BT_FLOAT, 2, num_verts);
			new_mesh->Mount(texcoord3_array);
			Vector2* tc3 = (Vector2*)texcoord3_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetMapVertex(4, vb.GetArxidia(v).tc3);
				tc3[v] = Vector2(mv.x, 1.0f - mv.y);
			}
		}

		Quaternion q = ((NodeChunk*)curr_file_node)->GetWorldPRS().GetRotation();
		q = q.Inverse();
		Matrix3x3 inv_rot_mat = q.GetRotationMatrix();

		// fill normal buffer
		Matrix3x3 world_rot = (Matrix3x3) ((NodeChunk*)curr_file_node)->GetWorldPRS().GetTransform();
		if (mesh->GetNumberOfNormals() && dialog_data.normals)
		{
			GeometryArray* normal_array = new GeometryArray(NORMAL_ARRAY, SCE_BT_FLOAT, 3, num_verts);
			new_mesh->Mount(normal_array);
			Vector3* nor = (Vector3*)normal_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetNormal(vb.GetArxidia(v).normal);
				Vector3 normal = Vector3(mv.x, mv.y, mv.z);
				normal.Transform(inv_rot_mat);
				nor[v] = normal;
			}
		}

		/*
		// fill tangent buffer
		if (mesh->GetNumberOfTangents() && dialog_data.tangents)
		{
			GeometryArray* tangent_array = new GeometryArray(TANGENT_ARRAY, SCE_BT_FLOAT, 3, num_verts);
			new_mesh->Mount(tangent_array);
			Vector3* tan = (Vector3*)tangent_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetTangent(vb.GetArxidia(v).normal);
				Vector3 tangent = Vector3(mv.x, mv.y, mv.z);
				tangent.Transform(inv_rot_mat);
				tan[v] = tangent;
			}
		}

		// fill binormal buffer
		Info("**********CHECK1**********\n");
		if (mesh->GetNumberOfBinormals() && dialog_data.binormals)
		{
			GeometryArray* binormal_array = new GeometryArray(BINORMAL_ARRAY, SCE_BT_FLOAT, 3, num_verts);
			new_mesh->Mount(binormal_array);
			Vector3* bnor = (Vector3*)binormal_array->GetArrayData();
			for (unsigned long v=0; v<num_verts; v++)
			{
				Point3 mv = mesh->GetBinormal(vb.GetArxidia(v).normal);
				Vector3 binormal = Vector3(mv.x, mv.y, mv.z);
				binormal.Transform(inv_rot_mat);
				bnor[v] = binormal;
			}
		}
		Info("**********CHECK2**********\n");
		*/

		// fill bone weight/index buffers
		if (mesh->GetNumberOfVerts() && skin)
		{
			GeometryArray* bone_weight_array = new GeometryArray(BONE_WEIGHT_ARRAY, SCE_BT_FLOAT, 3, num_verts);
			new_mesh->Mount(bone_weight_array);
			float* weights = (float*)bone_weight_array->GetArrayData();

			GeometryArray* bone_index_array = new GeometryArray(BONE_INDEX_ARRAY, SCE_BT_INT32, 1, num_verts);
			new_mesh->Mount(bone_index_array);
			unsigned int* indices = (unsigned int*)bone_index_array->GetArrayData();

			for (unsigned long v=0; v<num_verts; v++)
			{
				int vert = vb.GetArxidia(v).position;
				weights[3*v + 0] = 0.0f;
				weights[3*v + 1] = 0.0f;
				weights[3*v + 2] = 0.0f;

				// NOTE: little endian only !!!
				unsigned int packed_bone_indices = 0;
				unsigned char* cp = (unsigned char*) &packed_bone_indices;

				for (int i=0; i<skin->GetNumberOfBones(vert); i++)
				{
					if (i == 3)
					{
						Error("Vertex %d has more than 3 bones. Fix it!!!\n");
						break;
					}

					// write weight
					weights[3*v + i] = skin->GetWeight(vert, i);

					// write index
					cp[i] = rev_bone_map[skin->GetBoneID(vert, i)];
				}

				indices[v] = packed_bone_indices;

				Info("Vertex: %d\n\t[PACKED BONE IND] 0x%x\n\t[BONE_WEIGHTS] %f, %f, %f\n", v, packed_bone_indices, 
					weights[3*v +0],
					weights[3*v +1],
					weights[3*v +2]);
			}

			// add bone map array
			GeometryArray *bone_map_array = new GeometryArray(BONE_MAP_ARRAY, SCE_BT_INT32, 1, bone_map.size());
			new_mesh->Mount(bone_map_array);
			unsigned int *bm = (unsigned int *) bone_map_array->GetArrayData();
			for (int i=0; i<bone_map.size(); i++)
			{
				bm[i] = bone_map[i];
			}
		}

		// calculate tangents, if needed
		if (dialog_data.tangents)
		{
			if (!new_mesh->CalculateTangents())
			{
				Error("Could not calculate tangents for this mesh!!!!!!!!!\n");
				Error("Please supply at least one texture coordinate stream!!!\n");
			}
		}


		//new_mesh->CalculateNormals();
		//
		//{
		//	MessageBox(0, "Could not calculate tangents", "Error", 0);
		//}

		// transform mesh to model space
		NodeChunk* curr_node = (NodeChunk*)curr_file_node;
		Matrix4x4 world = curr_node->GetWorldPRS().GetTransform();
		//world = world * curr_node->GetPivotPRS().GetTransform();
		//world = curr_node->GetPivotPRS().GetTransform() * world;
		
		//new_mesh->Transform(world.Inverse());
	} // end material submesh loop

	// log stuff
	Info("Max mesh info:\n-----\n");
	Info("Vertex count = %d\n", mesh->GetNumberOfVerts());
	Info("Normal count: %d\n", mesh->GetNumberOfNormals());
	Info("Tangent count: %d\n", mesh->GetNumberOfTangents());
	Info("Binormal count: %d\n", mesh->GetNumberOfBinormals());
	Info("Face count = %d\n", mesh->GetNumberOfFaces());
	Info("Smoothing group count: %d\n", mesh->GetActiveSmgrps().Count());
	Info("Active materials: %d\n", mesh->GetActiveMatIDs().Count());
}

static SceTextureType MaxSlotToTT(int slot)
{
	switch (slot)
	{
		case ID_DI:
			return STT_DIFFUSE;
			break;
		case ID_SI:
			return STT_LIGHT;
			break;
		case ID_OP:
			return STT_OPACITY;
			break;
		case ID_BU:
			return STT_NORMAL;
			break;
	}

	return STT_UNSUPPORTED;
}

void ExportPlugin::ProcessMaterial(IGameMaterial* mat)
{
	Info("Material Name = %s\n", mat->GetMaterialName());

	if (!mat->GetSubMaterialCount()) // don't save multimaterials
	{
		// add a new material chunk
		MaterialChunk* mc = new MaterialChunk();
		curr_file_node->Mount(mc);

		// insert a new material to the material map
		mat_map[mat->GetMaterialName()] = mat_count;
		mat_count ++;

		// save material name
		mc->SetName(mat->GetMaterialName());

		Point3 c;
		c.x = c.y = c.z = 1.0f;
		if (mat->GetAmbientData())
			mat->GetAmbientData()->GetPropertyValue(c);	
		mc->SetAmbient(Vector4(c.x, c.y, c.z));
		Info("\tAmbient: r = %f g = %f b = %f\n", c.x, c.y, c.z);

		c.x = c.y = c.z = 1.0f;
		if (mat->GetDiffuseData())
			mat->GetDiffuseData()->GetPropertyValue(c);	
		mc->SetDiffuse(Vector4(c.x, c.y, c.z));
		Info("\tDiffuse: r = %f g = %f b = %f\n", c.x, c.y, c.z);

		c.x = c.y = c.z = 1.0f;
		if (mat->GetSpecularData())
			mat->GetSpecularData()->GetPropertyValue(c);	
		mc->SetSpecular(Vector4(c.x, c.y, c.z));
		Info("\tSpecular: r = %f g = %f b = %f\n", c.x, c.y, c.z);

		c.x = c.y = c.z = 1.0f;
		if (mat->GetEmissiveData())
			mat->GetEmissiveData()->GetPropertyValue(c);	
		mc->SetEmissive(Vector4(c.x, c.y, c.z));
		Info("\tEmissive: r = %f g = %f b = %f\n", c.x, c.y, c.z);

		// process texture maps
		unsigned long tex_count = mat->GetNumberOfTextureMaps();
		for (unsigned long i=0; i<tex_count; i++)
		{
			IGameTextureMap* map = mat->GetIGameTextureMap(i);
			// get texture slot
			int slot = map->GetStdMapSlot();
			if (slot == -1) continue;
			SceTextureType tt = MaxSlotToTT(slot);
			if (tt == STT_UNSUPPORTED) continue;
			TextureChunk* tc = mc->GetTexture(tt);

			if (map->IsEntitySupported())
			{
				tc->SetName(map->GetBitmapFileName());
			}
			else
			{
				// support only bump of non-supported
				if (slot == ID_BU)
				{
					Texmap* max_texture = map->GetMaxTexmap();
					int sub_count = max_texture->NumSubTexmaps();
					
					for (int i=0; i<sub_count; i++)
					{
						Texmap *sub_map = max_texture->GetSubTexmap(i);
						if (!sub_map) continue;
						//Class_ID id = sub_map->ClassID();
						//if (Class_ID(TEXMAP_CLASS_ID, BMTEX_CLASS_ID) == id)
						
						//if (dynamic_cast<BitmapTex*>(sub_map))
						//{
							BitmapTex *sub_map_bmp = (BitmapTex*) sub_map;
							tc->SetName(sub_map_bmp->GetMapName());
						//}
					}
				}
			}
		} // end texture loop

		// log texture maps
		TextureChunk *tc = mc->GetTexture(STT_DIFFUSE);
		std::string name;
		if (tc->GetName() != "") name = tc->GetName(); else name = "NO_TEXTURE";
		Info("\tDiffuse tex: %s\n", name.c_str());
		tc = mc->GetTexture(STT_NORMAL);
		if (tc->GetName() != "") name = tc->GetName(); else name = "NO_TEXTURE";
		Info("\tNormal tex: %s\n", name.c_str());
	}

	// process sub-materials
	for (unsigned long i=0; i<mat->GetSubMaterialCount(); i++)
	{
		ProcessMaterial(mat->GetSubMaterial(i));
	}
}

// ----------------------------------------------------------------------
// animation plugin
// ----------------------------------------------------------------------

void AnimationPlugin::ProcessSingleNode(IGameNode* node)
{
	Info("------------------\n");
	Info("Processing node: %s\n", node->GetName());
	Info("Node ID: %d\n", node->GetNodeID());

	if (node->GetNodeID() == 498)
	{
		printf("skata");
	}

	// get the corresponding object
	IGameObject* obj = node->GetIGameObject();
	if (!obj)
	{
		Info("Invalid object!\n");
		return;
	}

	
	IGameKeyTab pos_keys, rot_keys, scale_keys;
	IGameControl* control = node->GetIGameControl(); 

	IGameKeyTab tm_keys;

	// create a new KeyframeChunk
	KeyframeChunk* keys = new KeyframeChunk;
	animation->Mount(keys);	
	keys->SetNodeID(node->GetNodeID());
	keys->SetKeyframeCount(0);
	if (control->GetFullSampledKeys(tm_keys, FRAME_INT, IGameControlType(IGAME_TM), true))
	{
		keys->SetKeyframeCount(tm_keys.Count());
		PRS* prs = keys->GetKeyframes();

		for (unsigned int i=0; i<tm_keys.Count(); i++)
		{
			prs[i] = GMatrixToPRS(tm_keys[i].sampleKey.gval);
		}
	}

	/*
	unsigned int pos_key_count, rot_key_count, scale_key_count;


	bool result;
	result = control->GetFullSampledKeys(pos_keys, FRAME_INT, IGAME_POS, false);
	if (!result) pos_key_count = 0;
	else pos_key_count = pos_keys.Count();
	result = control->GetFullSampledKeys(rot_keys, FRAME_INT, IGAME_ROT, false);
	if (!result) rot_key_count = 0;
	else rot_key_count = rot_keys.Count();
	result = control->GetFullSampledKeys(scale_keys, FRAME_INT, IGAME_SCALE, false);
	if (!result) scale_key_count = 0;
	else scale_key_count = scale_keys.Count();
	
	unsigned int max_keys = pos_key_count;
	if (max_keys < rot_key_count) max_keys = rot_key_count;
	if (max_keys < scale_key_count) max_keys = scale_key_count;
*/


	// add default local matrix
	GMatrix lmat = node->GetLocalTM();
	keys->SetDefaultPRS(GMatrixToPRS(lmat));

	/*
	bool key_changed = false;

	// initialize keys
	for (unsigned int i=0; i<pos_keys.Count(); i++)
	{
		prs[i] = PRS();
	}

	for (unsigned long i=0; i<pos_key_count; i++)
	{
		Point3 k = pos_keys[i].sampleKey.pval;

		prs[i].SetPosition(Vector3(k.x, k.y, k.z));

		if (k != pos_keys[0].sampleKey.pval)
		{
			key_changed = true;
		}
		Info("Position keyframe: %f %f %f\n", k.x, k.y, k.z);
	}
	for (unsigned long i=0; i<rot_key_count; i++)
	{
		Quat k = rot_keys[i].sampleKey.qval;

		Quaternion q = Quaternion(k.w, k.x, k.y, k.z).Inverse();
			
		prs[i].SetRotation(q);

		Quat z = rot_keys[0].sampleKey.qval;

		if (k.x != z.x || k.y != z.y || k.z != z.z || k.w != z.w)
		{
			key_changed = true;
		}
		Info("Rotation keyframe: <%f %f %f> %f\n", k.x, k.y, k.z, k.w);
	}
	for (unsigned long i=0; i<scale_key_count; i++)
	{
		Point3 k = scale_keys[i].sampleKey.sval.s;

		prs[i].SetScaling(Vector3(k.x, k.y, k.z));

		Point3 z = scale_keys[0].sampleKey.sval.s;

		if (k != z)
		{
			key_changed = true;
		}
		Info("Scaling keyframe: %f %f %f\n", k.x, k.y, k.z);
	}

	if (key_changed)
	{
		Info("this object is animated\n");
	}
	else
	{
		keys->SetKeyframeCount(0);
		Info("This object is static\n");
	}
	*/

	node->ReleaseIGameObject();
	obj = 0;
}

void AnimationPlugin::ProcessNode(IGameNode* node)
{
	// process current node
	ProcessSingleNode(node);

	// process children
	for (unsigned int i=0; i<node->GetChildCount(); i++)
	{
		ProcessNode(node->GetNodeChild(i));
	}
}

int AnimationPlugin::DoExport(const TCHAR *name, ExpInterface *ei,
		Interface *i, BOOL suppressPrompts,
		DWORD options)
{
	scene = GetIGameInterface();
	if (!scene)
	{
		Info("GetIGameInterface() returned zero pointer!\n");
		Cleanup();
		return false;
	}

	// Set D3D coord system
	GetConversionManager()->SetCoordSystem(IGameConversionManager::IGAME_D3D); 
	scene->InitialiseIGame(false);
	// select frame 0
	//scene->SetStaticFrame(0);


	// create chunk file root
	animation = new AnimationFile();
	animation->SetVersion(100);
	animation->SetMessage("Track7 animation file format version 1.0");
	animation->SetFPS(GetFrameRate());

	// process nodes
	for (unsigned int i=0; i<scene->GetTopLevelNodeCount(); i++)
	{
		ProcessNode(scene->GetTopLevelNode(i));
	}

	// fix node ids to be sequential
	std::map<unsigned int, unsigned int> node_map = create_node_map(scene);
	ChunkList all_nodes = animation->GetIDChildren(ANM_KEYFRAME_CHUNK);
	for (unsigned int i=0; i<all_nodes.size(); i++)
	{
		KeyframeChunk * curr_node = (KeyframeChunk*) all_nodes[i];
		curr_node->SetNodeID(node_map[curr_node->GetNodeID()]);
	}

	// save to disk
	std::string file_path = GetPath(name);
	//vfs_mount(file_path.c_str(), "/");
	//std::string vfs_fname = std::string("/") + GetFilename(name);
	//VFS_FILE* fp = vfs_open(vfs_fname.c_str(), "wb");
	
	VFS_FILE *fp = vfs_open(name, "wb");
	
	if (!fp)
	{
		Info("Could not open %s for writing\n", name);
		Cleanup();
		return false;
	}

	animation->Save(fp);


	vfs_close(fp);
	//vfs_umount("/");
	Cleanup();

	return TRUE;
}

void AnimationPlugin::Cleanup()
{
	if (scene)
	{
		scene->ReleaseIGame();
		scene = 0;
	}

	if (animation)
	{
		delete animation;
		animation = 0;
	}
}









